#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    tkScenarist - screen writing made simpler

    Copyright (c) 2014+ Raphaël Seban <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see http://www.gnu.org/licenses/
"""

from tkRAD.widgets.rad_canvas import RADCanvas


class CharactersCanvas (RADCanvas):

    CONFIG = {
        "bg": "grey80",
        "bd": 0,
        "highlightbackground": "grey20",
        "highlightthickness": 1,
    }

    CONFIG_LINK = {
        "font": "helvetica 8 italic",
        "anchor": "center",
        "color": "grey90",
        "background": "grey30",
        "outline": "grey10",
        "box": (-5, -5, +5, +5),
        "label_width": 150,
        "width": 0,
    }

    CONFIG_NAME = {
        "font": "helvetica 10 bold",
        "anchor": "center",
        "color": "royal blue",
        "background": "grey90",
        "outline": "grey10",
        "box": (-10, -10, +10, +10),
        "label_width": 200,
        "width": 1,
    }

    ITEM_BOX = (-10, -10, +10, +10)

    TAG_NAME = "name"
    TAG_LINK = "link"

    def bbox_add (self, *bboxes):
        return tuple(map(sum, zip(*bboxes)))
    # end def

    def bind_events (self, *args, **kw):
        self.bind("<Configure>", self.update_canvas)
        self.bind("<Button-1>", self.on_start_drag)
        self.bind("<Shift-Button-1>", self.on_start_drag_link)
        self.bind("<Control-Button-1>", self.on_start_drag_none)
        self.bind("<Double-1>", self.on_double_click)
        for _key in ("Button-4", "Button-5", "MouseWheel"):
            self.bind("<{}>".format(_key), self.on_mouse_wheel_y)
            self.bind("<Shift-{}>".format(_key), self.on_mouse_wheel_x)
        # end for
    # end def

    def center_x (self):
        return (self.winfo_width() or self.winfo_reqwidth()) // 2
    # end def

    def center_xy (self):
        return (self.center_x(), self.center_y())
    # end def

    def center_y (self):
        return (self.winfo_height() or self.winfo_reqheight()) // 2
    # end def

    def character_name_add (self, name, **kw):
        xy = kw.get("xy")
        if not xy:
            _x, _y = self.viewport_center_xy()
            _offset = 20 * self.instance_counter
            xy = (_x + _offset, _y + _offset)
        # end if
        self.character_labels[name] = self.create_label(
            self.TAG_NAME, xy, text=name, **self.CONFIG_NAME
        )
        self.update_canvas()
        self.modified = True
    # end def

    def character_name_remove (self, name):
        try:
            _tag = self.get_tag_from_name(name)
        except:
            pass
        else:
            self.delete(_tag)
            self.remove_links(_tag)
            self.canvas_groups.pop(_tag, None)
            self.character_labels.pop(name, None)
            self.update_canvas()
            self.modified = True
        # end try
    # end def

    def character_name_rename (self, old_name, new_name):
        _group = self.character_labels.get(old_name)
        if _group:
            self.update_label(_group, text=new_name)
            self.character_labels[new_name] = _group
            self.character_labels.pop(old_name, None)
            self.update_canvas()
            self.modified = True
        # end if
    # end def

    def clear_canvas (self, *args, **kw):
        self.delete("all")
        self.configure(scrollregion=(0, 0, 0, 0))
        self.xview_moveto(0)
        self.yview_moveto(0)
        self.modified = True
    # end def

    def create_label (self, tag_prefix, xy, **kw):
        _tag = self.get_new_tag(tag_prefix)
        _item_box = kw.get("box") or self.ITEM_BOX
        _x, _y = xy
        _text_id = self.create_text(
            _x, _y,
            text=kw.get("text") or "label",
            anchor=kw.get("anchor"),
            font=kw.get("font"),
            fill=kw.get("color"),
            width=kw.get("label_width"),
            tags=_tag,
        )
        _frame_id = self.create_rectangle(
            self.bbox_add(self.bbox(_text_id), _item_box),
            outline=kw.get("outline"),
            fill=kw.get("background"),
            width=kw.get("width"),
            tags=_tag,
        )
        self.tag_raise(_text_id, _frame_id)
        self.modified = True
        return self.group_add(
            _tag,
            tag=_tag, text=_text_id, frame=_frame_id, box=_item_box
        )
    # end def

    def create_relationship (self, tag0, tag1, text=None):
        if tag0 != tag1 and not self.tags_linked(tag0, tag1):
            _center0 = self.get_bbox_center(tag0)
            _center1 = self.get_bbox_center(tag1)
            _line = self.create_line(
                _center0 + _center1,
                fill=self.CONFIG_LINK.get("outline"),
                width=1,
            )
            self.tag_lower(_line, "all")
            _group = self.create_label(
                self.TAG_LINK,
                self.get_segment_center(_center0, _center1),
                text=text or _("Relationship"),
                **self.CONFIG_LINK
            )
            _group.update(line=_line, tag0=tag0, tag1=tag1)
            self.register_link(tag0, tag1, _group)
            self.modified = True
            return True
        # end if
        return False
    # end def

    def dispose_label (self, name, coords):
        _group = self.character_labels.get(name)
        if _group:
            self.coords(_group["text"], *coords)
            self.update_label(_group)
            self.update_links(_group["tag"])
            self.update_canvas()
            self.modified = True
        # end if
    # end def

    def dnd_reset (self, *args, **kw):
        self.drag_tag = ""
        self.drag_start_xy = (0, 0)
        self.drag_last_pos = (0, 0)
        self.unbind("<B1-Motion>")
        self.unbind("<ButtonRelease-1>")
        self.unbind("<Control-ButtonRelease-1>")
    # end def

    def get_actual_position (self, x, y):
        return (self.canvasx(x), self.canvasy(y))
    # end def

    def get_bbox_center (self, tag):
        _x1, _y1, _x2, _y2 = self.bbox(tag)
        return ((_x1 + _x2) // 2, (_y1 + _y2) // 2)
    # end def

    def get_file_contents (self):
        import json
        _pos = {
            _name: self.coords(_group["text"])
            for _name, _group in self.character_labels.items()
        }
        return json.dumps(
            dict(
                label_positions=_pos,
                links=self.get_named_relationships()
            )
        )
    # end def

    def get_group_tag (self, list_ids):
        try:
            return self.gettags(list_ids[-1])[0]
        except:
            return ""
        # end try
    # end def

    def get_mouse_wheel_step (self, event):
        import os
        _platform = os.name.lower()
        if _platform == "nt":
            return int(-event.delta // 120)
        elif _platform == "mac":
            return int(-event.delta)
        else:
            return int((event.num == 5) - (event.num == 4))
        # end if
    # end def

    def get_name_from_tag (self, tag):
        for _name, _group in self.character_labels.items():
            if _group and _group["tag"] == tag:
                return _name
            # end if
        # end for
        return ""
    # end def

    def get_named_relationships (self):
        _tag_names = {
            _group["tag"]: _name
            for _name, _group in self.character_labels.items()
        }
        return tuple(
            dict(
                name0=_tag_names[_group["tag0"]],
                name1=_tag_names[_group["tag1"]],
                text=self.itemcget(_group["text"], "text"),
            )
            for _tag, _group in self.canvas_groups.items()
            if self.TAG_LINK in _tag
        )
    # end def

    def get_new_tag (self, tag_prefix=None):
        tag_prefix = tag_prefix or "group"
        self.instance_counter += 1
        return "{}#{}".format(tag_prefix, self.instance_counter)
    # end def

    def get_segment_center (self, start_xy, end_xy):
        _x0, _y0 = start_xy
        _x1, _y1 = end_xy
        return ((_x0 + _x1) // 2, (_y0 + _y1) // 2)
    # end def

    def get_tag_from_name (self, name):
        return self.character_labels[name]["tag"]
    # end def

    def get_tag_xy (self, event):
        _xy = self.get_actual_position(event.x, event.y)
        _tag = self.get_group_tag(self.find_overlapping(*(_xy * 2)))
        return (_tag,) + _xy
    # end def

    def group_add (self, name, **kw):
        if name:
            if name in self.canvas_groups:
                raise KeyError(
                    _("canvas group name '{gname}' already exists.")
                    .format(gname=name)
                )
            else:
                self.canvas_groups[name] = kw
                self.modified = True
                return kw
            # end if
        # end if
        return None
    # end def

    def initialize (self, *args, **kw):
        self.reset(*args, **kw)
        self.bind_events()
    # end def

    def is_viewport_out (self, x, y):
        _w, _h = self.viewport_size()
        return not(0 <= x <= _w and 0 <= y <= _h)
    # end def

    def link_create_relationship (self, from_tag, to_tag):
        if self.TAG_NAME in from_tag and self.TAG_NAME in to_tag:
            if from_tag == to_tag:
                pass
                #~ self.user_show_warning(
                    #~ _("Linking a name to itself makes *NO* sense.")
                #~ )
            elif not self.create_relationship(from_tag, to_tag):
                self.user_show_warning(
                    _(
                        "These two character names "
                        "are *ALREADY LINKED* together."
                    )
                )
            # end if
        # end if
    # end def

    def link_edit_relationship (self, tag):
        if self.TAG_LINK in tag:
            _group = self.canvas_groups[tag]
            _text_id = _group["text"]
            from tkinter.simpledialog import askstring
            _new_text = askstring(
                title=_("Characters relationship"),
                prompt=_(
                    "Relationship '{from_name}' <---> '{to_name}'"
                ).format(
                    from_name=self.get_name_from_tag(_group["tag0"]),
                    to_name=self.get_name_from_tag(_group["tag1"])
                ),
                initialvalue=self.itemcget(_text_id, "text"),
                parent=self
            )
            if _new_text:
                self.update_label(_group, text=_new_text)
                self.update_canvas()
                self.modified = True
            # end if
        # end if
    # end def

    def link_remove_relationship (self, tag):
        if self.TAG_LINK in tag:
            _group = self.canvas_groups[tag]
            _tag0 = _group["tag0"]
            _tag1 = _group["tag1"]
            _confirm = self.user_ask(
                _(
                    "Relationship: '{from_name}' <---> '{to_name}'.\n"
                    "Do you really want to remove this "
                    "relationship link?"
                ).format(
                    from_name=self.get_name_from_tag(_tag0),
                    to_name=self.get_name_from_tag(_tag1)
                )
            )
            if _confirm:
                _tags = self.relationship_links.get(_tag0) or dict()
                _tags.pop(_tag1, None)
                _tags = self.relationship_links.get(_tag1) or dict()
                _tags.pop(_tag0, None)
                self.remove_group_link(_group)
                self.modified = True
            # end if
        # end if
    # end def

    def on_double_click (self, event=None, *args, **kw):
        _tag, _x, _y = self.get_tag_xy(event)
        if self.TAG_LINK in _tag:
            self.link_edit_relationship(_tag)
        elif self.TAG_NAME in _tag:
            self.events.raise_event(
                "Characters:List:Rename", xy=(_x, _y)
            )
        elif not _tag:
            self.events.raise_event(
                "Characters:List:Add", xy=(_x, _y)
            )
        # end if
    # end def

    def on_drag_canvas (self, event=None, *args, **kw):
        self.scan_dragto(event.x, event.y, gain=-5)
    # end def

    def on_drag_label (self, event=None, *args, **kw):
        _x, _y = self.get_actual_position(event.x, event.y)
        _x0, _y0 = self.drag_last_pos
        _dx, _dy = (_x - _x0), (_y - _y0)
        self.drag_last_pos = (_x, _y)
        self.move(self.drag_tag, _dx, _dy)
        self.update_links(self.drag_tag)
        self.update_canvas()
        self.modified = True
        self.scan_dragto(event.x, event.y, gain=-1)
    # end def

    def on_drag_link (self, event=None, *args, **kw):
        self.drag_last_pos = self.get_actual_position(event.x, event.y)
        self.coords(
            self.drag_link_id, self.drag_start_xy + self.drag_last_pos
        )
        self.update_canvas()
        self.modified = True
        self.scan_dragto(event.x, event.y, gain=-1)
    # end def

    def on_drag_none (self, *args, **kw):
        self.dnd_reset()
    # end def

    def on_drop_label (self, event=None, *args, **kw):
        _name = self.get_name_from_tag(self.drag_tag)
        if self.is_viewport_out(event.x, event.y):
            self.show_label(_name)
        # end if
        self.events.raise_event("Characters:Name:Selected", name=_name)
        self.modified = True
        self.dnd_reset()
        self.update_canvas()
    # end def

    def on_drop_link (self, event=None, *args, **kw):
        self.delete(self.drag_link_id)
        self.link_create_relationship(
            self.drag_tag, self.get_tag_xy(event)[0]
        )
        self.modified = True
        self.dnd_reset()
        self.update_canvas()
    # end def

    def on_drop_remove (self, event=None, *args, **kw):
        _tag = self.get_tag_xy(event)[0]
        if self.TAG_LINK in _tag:
            self.link_remove_relationship(_tag)
        elif self.TAG_NAME in _tag:
            self.events.raise_event(
                "Characters:Name:Selected",
                name=self.get_name_from_tag(_tag)
            )
            self.events.raise_event("Characters:List:Delete")
        # end if
        self.dnd_reset()
        self.update_canvas()
    # end def

    def on_mouse_wheel_x (self, event=None, *args, **kw):
        self.xview_scroll(self.get_mouse_wheel_step(event), "units")
    # end def

    def on_mouse_wheel_y (self, event=None, *args, **kw):
        self.yview_scroll(self.get_mouse_wheel_step(event), "units")
    # end def

    def on_start_drag (self, event=None, *args, **kw):
        self.dnd_reset()
        self.scan_mark(event.x, event.y)
        _tag, _x, _y = self.get_tag_xy(event)
        if self.TAG_NAME in _tag:
            self.drag_start_xy = (_x, _y)
            self.drag_last_pos = self.drag_start_xy
            self.drag_tag = _tag
            self.tag_raise(_tag, "all")
            self.bind("<B1-Motion>", self.on_drag_label)
            self.bind("<ButtonRelease-1>", self.on_drop_label)
        else:
            self.bind("<B1-Motion>", self.on_drag_canvas)
        # end if
    # end def

    def on_start_drag_link (self, event=None, *args, **kw):
        self.dnd_reset()
        _tag = self.get_tag_xy(event)[0]
        if self.TAG_NAME in _tag:
            self.scan_mark(event.x, event.y)
            self.drag_start_xy = self.get_bbox_center(_tag)
            self.drag_last_pos = self.drag_start_xy
            self.drag_tag = _tag
            self.drag_link_id = self.create_line(
                self.drag_start_xy * 2
            )
            self.tag_lower(self.drag_link_id, "all")
            self.bind("<B1-Motion>", self.on_drag_link)
            self.bind("<ButtonRelease-1>", self.on_drop_link)
        # end if
    # end def

    def on_start_drag_none (self, *args, **kw):
        self.dnd_reset()
        self.bind("<B1-Motion>", self.on_drag_none)
        self.bind("<Control-ButtonRelease-1>", self.on_drop_remove)
    # end def

    def register_link (self, tag1, tag2, group):
        _tags = self.relationship_links.setdefault(tag1, dict())
        _tags[tag2] = group
        _tags = self.relationship_links.setdefault(tag2, dict())
        _tags[tag1] = group
        self.modified = True
    # end def

    def relationship_add (self, name0, name1, text):
        self.create_relationship(
            self.get_tag_from_name(name0),
            self.get_tag_from_name(name1),
            text
        )
        self.modified = True
    # end def

    def remove_group_link (self, group):
        if group:
            self.delete(group["line"])
            self.delete(group["tag"])
            self.canvas_groups.pop(group["tag"], None)
            self.modified = True
        # end if
    # end def

    def remove_links (self, tag):
        _tags = self.relationship_links.setdefault(tag, dict())
        for _group in _tags.values():
            self.remove_group_link(_group)
        # end for
        for _tags in self.relationship_links.values():
            _tags.pop(tag, None)
        # end for
        self.relationship_links.pop(tag, None)
        self.modified = True
    # end def

    def reset (self, *args, **kw):
        self.clear_canvas()
        self.dnd_reset()
        self.instance_counter = 0
        self.character_labels = dict()
        self.relationship_links = dict()
        self.canvas_groups = dict()
        self.modified = False
    # end def

    def show_label (self, name):
        try:
            _x0, _y0, _x1, _y1 = map(int, self["scrollregion"].split())
            _x, _y = self.get_bbox_center(self.get_tag_from_name(name))
        except:
            pass
        else:
            _cw, _ch = abs(_x1 - _x0), abs(_y1 - _y0)
            _cx, _cy = self.center_xy()
            self.xview_moveto((_x - _x0 - _cx)/_cw)
            self.yview_moveto((_y - _y0 - _cy)/_ch)
        # end try
    # end def

    def tags_linked (self, tag1, tag2):
        _tags = self.relationship_links.get(tag1) or dict()
        return bool(tag2 in _tags)
    # end def

    def update_canvas (self, *args, **kw):
        _modified = self.modified
        try:
            _x0, _y0, _x1, _y1 = self.bbox("all")
            _w, _h = self.viewport_size()
            _x0, _y0 = (min(0, _x0), min(0, _y0))
            _x1, _y1 = (max(_x1, _w - 2 - _x0), max(_y1, _h - 2 - _y0))
            self.configure(scrollregion=(_x0, _y0, _x1, _y1))
        except:
            self.reset()
        # end try
        self.modified = _modified
    # end def

    def update_label (self, group, **kw):
        """Update name label or relationship label."""
        self.itemconfigure(group["text"], **kw)
        self.coords(
            group["frame"],
            self.bbox_add(self.bbox(group["text"]), group["box"])
        )
        self.modified = True
    # end def

    def update_links (self, tag):
        _tags = self.relationship_links.get(tag)
        if _tags:
            _start_xy = self.get_bbox_center(tag)
            for _tag, _group in _tags.items():
                _end_xy = self.get_bbox_center(_tag)
                self.coords(_group["line"], _start_xy + _end_xy)
                self.tag_lower(_group["line"], "all")
                self.coords(
                    _group["text"],
                    self.get_segment_center(_start_xy, _end_xy)
                )
                self.update_label(_group)
            # end for
        # end if
    # end def

    def user_ask (self, question):
        from tkinter import messagebox
        return messagebox.askyesno(
            title=_("Question"), message=str(question), parent=self
        )
    # end def

    def user_show_warning (self, msg):
        from tkinter import messagebox
        messagebox.showwarning(
            title=_("Attention"), message=str(msg), parent=self
        )
    # end def

    def viewport_center_xy (self):
        _x, _y = self.center_xy()
        return (self.canvasx(_x), self.canvasy(_y))
    # end def

    def viewport_size (self):
        return (self.winfo_width(), self.winfo_height())
    # end def

# end class
