#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    tkScenarist - screen writing made simpler

    Copyright (c) 2014+ Raphaël Seban <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see http://www.gnu.org/licenses/
"""

import tkRAD
from . import text_helpers as th


class ComponentCharacters (tkRAD.RADXMLFrame):

    def auto_save (self, *args, **kw):
        _name = self.get_current_name()
        if _name:
            self.character_logs[_name] = (
                th.text_get_contents(self.w_text)
            )
        # end if
    # end def

    def bind_events (self, **kw):
        self.events.connect_dict(
            {
                "Characters:List:Add": self.on_list_add,
                "Characters:List:Delete": self.on_list_delete,
                "Characters:List:Purge": self.on_list_purge,
                "Characters:List:Rename": self.on_list_rename,
                "Characters:Name:Selected": self.on_name_selected,

                "Component:Reset": self.on_component_reset,

                "Reset:Modified": self.on_reset_modified,

                "Scenario:Character:Name:Detected":
                    self.on_character_name_detected,
            }
        )
        self.bind("<FocusIn>", self.on_focus_in)
        self.w_listbox.bind(
            "<<ListboxSelect>>", self.on_listbox_item_selected
        )
        self.w_text.bind("<KeyRelease>", self.on_text_keypress)
    # end def

    def canvas_add_name (self, name, **kw):
        self.w_canvas.character_name_add(name, **kw)
    # end def

    def canvas_add_relationship (self, **kw):
        self.w_canvas.relationship_add(**kw)
    # end def

    def canvas_delete_name (self, name):
        self.w_canvas.character_name_remove(name)
    # end def

    def canvas_dispose_label (self, name, coords):
        self.w_canvas.dispose_label(name, coords)
    # end def

    def canvas_rename (self, old_name, new_name):
        self.w_canvas.character_name_rename(old_name, new_name)
    # end def

    def canvas_show_name (self, name):
        self.w_canvas.show_label(name)
    # end def

    def do_add_character_name (self, name, show_error=False, **kw):
        name = self.format_name(name)
        if name:
            if name in self.character_logs:
                if show_error:
                    self.error_already_exists(name)
                # end if
            else:
                self.listbox_add_name(name, **kw)
                self.canvas_add_name(name, **kw)
            # end if
        # end if
    # end def

    def do_delete_character_name (self, name, show_error=False, **kw):
        name = self.format_name(name)
        if name:
            if name in self.character_logs:
                self.listbox_delete_name(name)
                self.canvas_delete_name(name)
            elif show_error:
                self.error_not_found(name)
            # end if
        # end if
    # end def

    def do_purge_character_names (self, show_error=False, **kw):
        _text = self.get_scenario_text()
        for _name in self.get_character_names():
            if not self.is_mentioned(_name, _text):
                self.do_delete_character_name(_name)
            # end if
        # end for
    # end def

    def do_rename_character_name (self, old_name, new_name,
                                                show_error=False, **kw):
        old_name = self.format_name(old_name)
        new_name = self.format_name(new_name)
        if old_name and new_name and new_name != old_name:
            if new_name in self.character_logs:
                if show_error:
                    self.error_already_exists(new_name)
                # end if
            else:
                self.listbox_rename(old_name, new_name)
                self.canvas_rename(old_name, new_name)
            # end if
        # end if
    # end def

    def do_select_character_name (self, name=None):
        name = name or self.current_name
        try:
            _index = self.get_character_names().index(name)
        except:
            pass
        else:
            self.w_listbox.selection_clear(0, "end")
            self.w_listbox.selection_set(_index)
            self.w_listbox.see(_index)
            self.update_character_log()
        # end try
    # end def

    def error_already_exists (self, name):
        self.user_show_warning(
            _(
                "Character name '{name}' already exists."
            ).format(name=name)
        )
    # end def

    def error_not_found (self, name):
        self.user_show_warning(
            _(
                "Character name '{name}' has *NOT* been found in list."
            ).format(name=name)
        )
    # end def

    def find_nearest_name (self, contents, index):
        contents = self.format_name(contents)
        _names = self.get_character_names()
        if contents in _names:
            return (contents, 0)
        # end if
        import re
        for _name in reversed(_names):
            _name = self.format_name(_name)
            _len = len(_name)
            _start = max(0, index - _len)
            _sub = contents[_start:(index + _len)]
            _found = re.search(r"\b{}\b".format(_name), _sub)
            if _found:
                return (_name, _start + _found.start())
            # end if
        # end for
        return ("", 0)
    # end def

    def format_name (self, name):
        return str(name).strip().upper()
    # end def

    def get_character_log (self, character_name):
        try:
            return str(
                self.character_logs[self.format_name(character_name)]
                or _("(no history log)")
            ).strip()
        except:
            return ""
        # end try
    # end def

    def get_character_names (self):
        return sorted(self.character_logs)
    # end def

    def get_current_name (self):
        if not self.character_logs:
            self.current_name = ""
        # end if
        return self.current_name
    # end def

    def get_file_contents (self, files):
        import json
        _dict = {
            files["names"]: "\n".join(self.get_character_names()),
            files["logs"]: json.dumps(self.character_logs),
            files["relationships"]: self.w_canvas.get_file_contents(),
        }
        return _dict
    # end def

    def get_matching_names (self, word):
        _matchups = []
        _word = self.format_name(word)
        if _word:
            for _name in self.get_character_names():
                if _name.startswith(_word):
                    _matchups.append(_name)
                # end if
            # end for
            if len(_matchups) == 1 and _matchups[0] == _word:
                _matchups = []
            # end if
        # end if
        return _matchups
    # end def

    def get_scenario_text (self):
        return th.text_get_contents(
            self.app.mainwindow.get_component("scenario").w_text
        )
    # end def

    def initialize (self, *args, **kw):
        from tkRAD.core.async import get_async_manager
        self.async = get_async_manager()
        self.character_logs = dict()
        self.current_name = ""
        self.xml_build("component_characters")
        self.bind_events(**kw)
        self.on_component_reset()
    # end def

    def is_mentioned (self, name, text=None):
        import re
        return bool(
            re.search(
                r"(?i)\b{}\b".format(name),
                text or self.get_scenario_text()
            )
        )
    # end def

    def is_registered (self, name):
        return bool(self.format_name(name) in self.character_logs)
    # end def

    def listbox_add_name (self, name, **kw):
        if name:
            self.character_logs[name] = ""
            self.update_listbox(new_name=name)
        # end if
    # end def

    def listbox_delete_name (self, name, *args, **kw):
        if name:
            self.character_logs.pop(name, None)
            self.current_name = ""
            kw.update(delete=True)
            self.update_listbox(*args, **kw)
        # end if
    # end def

    def listbox_get_selected_name (self):
        try:
            return self.w_listbox.get(self.w_listbox.curselection()[0])
        except:
            return self.get_current_name()
        # end try
    # end def

    def listbox_rename (self, old_name, new_name):
        if old_name and new_name and new_name != old_name:
            self.character_logs[new_name] = (
                self.character_logs[old_name]
            )
            self.listbox_delete_name(old_name, new_name=new_name)
        # end if
    # end def

    @property
    def modified (self):
        return bool(
            self.w_canvas.modified or self.w_text.edit_modified()
        )
    # end def

    def on_character_name_detected (self, *args, name=None, **kw):
        self.do_add_character_name(name, show_error=False, **kw)
    # end def

    def on_component_reset (self, *args, **kw):
        self.character_logs.clear()
        self.update_listbox()
        self.w_canvas.reset()
    # end def

    def on_focus_in (self, *args, **kw):
        self.on_name_selected()
    # end def

    def on_list_add (self, *args, **kw):
        _name = self.user_ask_string(
            _("Character name"), xy=kw.get("xy")
        )
        if _name:
            self.do_add_character_name(_name, show_error=True, **kw)
        # end if
    # end def

    def on_list_delete (self, *args, **kw):
        _name = self.listbox_get_selected_name()
        if _name:
            if self.is_mentioned(_name):
                self.user_show_warning(
                    _(
                        "Cannot delete name '{name}'.\n"
                        "This character name is still "
                        "mentioned into scenario's text."
                    ).format(name=_name)
                )
            elif self.user_confirm_deletion(_name):
                self.do_delete_character_name(_name)
            # end if
        # end if
    # end def

    def on_list_purge (self, *args, **kw):
        if self.character_logs and self.user_confirm_purge():
            self.save_now()
            self.do_purge_character_names()
        # end if
    # end def

    def on_list_rename (self, *args, **kw):
        _old_name = self.listbox_get_selected_name()
        if _old_name:
            if self.is_mentioned(_old_name):
                self.user_show_warning(
                    _(
                        "Cannot rename '{name}'.\n"
                        "This character name is still "
                        "mentioned into scenario's text."
                    ).format(name=_old_name)
                )
            else:
                _new_name = self.user_ask_string(
                    _("Renaming"), _old_name, kw.get("xy")
                )
                if _new_name:
                    self.save_now()
                    self.do_rename_character_name(
                        _old_name, _new_name, show_error=True
                    )
                # end if
            # end if
        # end if
    # end def

    def on_listbox_item_selected (self, event=None, *args, **kw):
        _name = self.listbox_get_selected_name()
        self.canvas_show_name(_name)
        if _name != self.current_name:
            self.update_character_log()
        # end if
    # end def

    def on_name_selected (self, *args, name=None, **kw):
        self.do_select_character_name(name)
    # end def

    def on_reset_modified (self, *args, **kw):
        self.w_canvas.modified = False
        self.w_text.edit_modified(False)
    # end def

    def on_text_keypress (self, event=None, *args, **kw):
        th.text_edit_autoseparator(self.w_text, event)
        self.async.run_after(3000, self.auto_save)
    # end def

    def save_now (self):
        self.async.stop(self.auto_save)
        self.auto_save()
    # end def

    def setup_component (self, files, archive):
        import json
        _read = lambda f: json.loads(archive.read(f).decode(ENCODING))
        self.character_logs = _read(files["logs"])
        _relationships = _read(files["relationships"])
        self.update_listbox()
        # backward compatibility
        _label_positions = "label_positions"
        if _label_positions not in _relationships:
            _label_positions = "label_pos"
        # end if
        for _name in self.character_logs:
            self.canvas_add_name(_name)
            self.canvas_dispose_label(
                _name, _relationships[_label_positions][_name]
            )
        # end for
        for _group in _relationships["links"]:
            self.canvas_add_relationship(**_group)
        # end for
    # end def

    def update_character_log (self, *args, new_name=None, **kw):
        self.save_now()
        _name = new_name or self.listbox_get_selected_name()
        _cvar = self.get_stringvar("log_character_name")
        self.enable_widget(self.w_text, True)
        if _name:
            _cvar.set(_name)
            self.current_name = _name
            th.text_reset_contents(
                self.w_text, self.character_logs[_name]
            )
            if new_name:
                self.w_text.focus_set()
            # end if
        else:
            _cvar.set("")
            self.current_name = ""
            th.text_reset_contents(self.w_text)
            self.enable_widget(self.w_text, False)
        # end if
    # end def

    def update_listbox (self, *args, **kw):
        _names = self.get_character_names()
        _flag = bool(_names)
        _new_name = kw.get("new_name")
        _index = 0
        if kw.get("delete"):
            try:
                _index = min(
                    self.w_listbox.curselection()[0], len(_names) - 1
                )
            except:
                pass
            # end try
        # end if
        self.w_listbox.delete(0, "end")
        if _names:
            self.w_listbox.insert("end", *_names)
            if _new_name:
                _index = _names.index(_new_name)
            # end if
            self.w_listbox.selection_set(_index)
            self.w_listbox.see(_index)
            self.events.raise_event(
                "Characters:List:Changed",
                contents=_names,
                current_selected=_index,
            )
        # end if
        self.enable_widget(self.btn_delete, _flag)
        self.enable_widget(self.btn_rename, _flag)
        self.enable_widget(self.btn_purge, _flag)
        self.update_character_log(*args, **kw)
    # end def

    def user_ask_string (self, prompt, default=None, xy=None):
        from tkinter import simpledialog
        return simpledialog.askstring(
            _("Please, insert"), str(prompt), initialvalue=default,
            parent=self.w_canvas if xy else self,
        )
    # end def

    def user_ask_yes_no (self, question):
        from tkinter import messagebox
        return messagebox.askyesno(
            title=_("Question"), message=str(question), parent=self,
        )
    # end def

    def user_confirm_deletion (self, name):
        return self.user_ask_yes_no(
            _(
                "Character '{name}' has *NOT* "
                "been found into scenario's text.\n"
                "Do you really want to delete "
                "this character?"
            ).format(name=name)
        )
    # end def

    def user_confirm_purge (self):
        return self.user_ask_yes_no(
            _(
                "Do you really want to delete *ALL* names "
                "that are *NOT* mentioned into scenario's text?"
            )
        )
    # end def

    def user_show_warning (self, msg):
        from tkinter import messagebox
        messagebox.showwarning(
            title=_("Attention"), message=str(msg), parent=self,
        )
    # end def

# end class
