#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    tkScenarist - screen writing made simpler

    Copyright (c) 2014+ Raphaël Seban <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see http://www.gnu.org/licenses/
"""

import tkRAD
from . import text_helpers as th


class ComponentDraftNotes (tkRAD.RADXMLFrame):

    def auto_save (self, *args, **kw):
        _index = self.w_listbox.last_selected
        if _index >= 0:
            self.draft_notes[self.w_listbox.get(_index)] = (
                th.text_get_contents(self.w_text)
            )
        # end if
    # end def

    def bind_events (self, **kw):
        self.events.connect_dict(
            {
                "Component:Reset": self.on_component_reset,

                "Notes:List:Add": self.on_listbox_add,
                "Notes:List:Delete": self.on_listbox_delete,
                "Notes:List:Rename": self.on_listbox_rename,

                "Pitch:Template:Insert": self.on_template_insert,

                "Reset:Modified": self.on_reset_modified,
            }
        )
        self.bind("<Expose>", self.on_expose)
        self.bind("<FocusIn>", self.on_focus_in)
        self.w_listbox.bind(
            "<<ListboxSelect>>", self.on_listbox_select
        )
        self.w_text.bind("<KeyRelease>", self.on_text_keypress)
    # end def

    def clear_listbox (self, *args, **kw):
        self.w_listbox.delete(0, "end")
        self.w_listbox.selection_clear(0, "end")
        self.w_listbox.modified = False
        self.w_listbox.last_selected = -1
    # end def

    def get_current_selected (self):
        try:
            return self.w_listbox.curselection()[0]
        except:
            return -1
        # end try
    # end def

    def get_file_contents (self, files):
        import json
        self.save_now()
        return {
            files["text"]: "", # kept for backward compatibility
            files["data"]: json.dumps(self.draft_notes),
        }
    # end def

    def initialize (self, *args, **kw):
        from tkRAD.core.async import get_async_manager
        self.async = get_async_manager()
        self.xml_build("component_draft_notes")
        self.cvar_title = self.get_stringvar("lbl_note_title")
        self.draft_notes = dict()
        self.on_component_reset()
        self.bind_events(**kw)
    # end def

    @property
    def modified (self):
        return bool(
            self.w_text.modified or self.w_listbox.modified
        )
    # end def

    def on_component_reset (self, *args, **kw):
        self.clear_listbox()
        self.draft_notes.clear()
        self.w_text.modified = False
        self.update_component()
    # end def

    def on_expose (self, *args, **kw):
        self.on_focus_in()
    # end def

    def on_focus_in (self, *args, **kw):
        _index = self.w_listbox.last_selected
        if _index < 0 and self.w_listbox.size():
            _index = 0
        # end if
        if _index >= 0:
            self.w_listbox.selection_clear(0, "end")
            self.w_listbox.selection_set(_index)
            self.w_listbox.see(_index)
            self.update_component()
        # end if
    # end def

    def on_listbox_add (self, *args, **kw):
        self.save_now()
        _title = self.user_ask(_("New note"))
        if _title:
            if _title in self.draft_notes:
                self.user_warn(_("This note title already exists."))
            else:
                self.draft_notes[_title] = ""
                self.update_listbox(_title)
            # end if
        # end if
    # end def

    def on_listbox_delete (self, *args, **kw):
        self.save_now()
        _index = self.get_current_selected()
        if _index >= 0:
            _title = self.w_listbox.get(_index)
            if self.user_confirm_deletion(_title):
                self.draft_notes.pop(_title, None)
                self.update_listbox()
        # end if
    # end def

    def on_listbox_rename (self, *args, **kw):
        self.save_now()
        _index = self.get_current_selected()
        if _index >= 0:
            _old = self.w_listbox.get(_index)
            _new = self.user_ask(_("Rename"), _old)
            if _new and _new != _old:
                if _new in self.draft_notes:
                    self.user_warn(
                        _("This note title already exists.")
                    )
                else:
                    self.draft_notes[_new] = self.draft_notes.pop(_old)
                    self.update_listbox(_new)
                # end if
            # end if
        # end if
    # end def

    def on_listbox_select (self, *args, **kw):
        if self.get_current_selected() != self.w_listbox.last_selected:
            self.update_component()
        # end if
    # end def

    def on_reset_modified (self, *args, **kw):
        self.w_text.edit_modified(False)
        self.w_text.modified = False
        self.w_listbox.modified = False
    # end def

    def on_template_insert (self, *args, notes=None, **kw):
        if notes:
            try:
                th.text_set_contents(self.w_text, notes)
                self.save_now()
                self.w_text.modified = True
            except:
                pass
            # end try
        # end if
    # end def

    def on_text_keypress (self, event=None, *args, **kw):
        self.w_text.modified = bool(
            self.w_text.modified or self.w_text.edit_modified()
        )
        th.text_edit_autoseparator(self.w_text, event)
        self.async.run_after(3000, self.auto_save)
    # end def

    def save_now (self, *args, **kw):
        self.async.stop(self.auto_save)
        self.auto_save()
    # end def

    def setup_component (self, files, archive):
        import json
        _read = lambda f: archive.read(f).decode(ENCODING)
        _text = _read(files["text"])
        _data = json.loads(_read(files["data"])) or dict()
        # backward compatibility
        if _text:
            _data[_("New note")] = str(_text)
        # end if
        self.draft_notes = _data
        self.update_listbox()
    # end def

    def update_component (self, *args, **kw):
        self.save_now()
        self.enable_widget(self.w_text, True)
        _index = self.get_current_selected()
        self.w_listbox.last_selected = _index
        _sel = bool(_index >= 0)
        if _sel:
            _title = self.w_listbox.get(_index)
            self.cvar_title.set(_title)
            th.text_reset_contents(
                self.w_text, self.draft_notes[_title]
            )
            self.after_idle(self.w_text.focus_set)
        else:
            self.cvar_title.set("")
            th.text_reset_contents(self.w_text)
            self.enable_widget(self.w_text, False)
        # end if
        self.enable_widget(self.btn_delete, _sel)
        self.enable_widget(self.btn_rename, _sel)
    # end def

    def update_listbox (self, title=None):
        _lsel = self.w_listbox.last_selected
        _titles = sorted(self.draft_notes)
        self.clear_listbox()
        self.w_listbox.insert("end", *_titles)
        try:
            _index = _titles.index(title)
        except:
            _index = min(_lsel, self.w_listbox.size() - 1)
        # end try
        if _index >= 0:
            self.w_listbox.selection_set(_index)
            self.w_listbox.see(_index)
        # end if
        self.w_listbox.modified = True
        self.update_component()
    # end def

    def user_ask (self, prompt, default_value=None):
        from tkinter.simpledialog import askstring
        return askstring(
            title=_("Please, insert"),
            prompt=prompt,
            initialvalue=default_value,
            parent=self
        )
    # end def

    def user_confirm_deletion (self, title):
        from tkinter.messagebox import askyesno
        return askyesno(
            title=_("Question"),
            message=_("Do you really want to delete '{note}'?")
                .format(note=title),
            parent=self
        )
    # end def

    def user_warn (self, message):
        from tkinter.messagebox import showwarning
        showwarning(
            title=_("Attention"), message=str(message), parent=self
        )
    # end def

# end class
