#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    tkScenarist - screen writing made simpler

    Copyright (c) 2014+ Raphaël Seban <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see http://www.gnu.org/licenses/
"""

import tkRAD
from . import text_helpers as th


class ComponentResources (tkRAD.RADXMLFrame):

    def auto_save (self, *args, **kw):
        if self.listbox_item.last_selected >= 0:
            _dict = {
                _key: _entry.get()
                for _key, _entry in self.entries.items()
            }
            _dict["notes"] = th.text_get_contents(self.text_notes)
            if any(_dict.values()):
                _dict["fk_type"] = self.get_fk_type()
                self.database.res_update_item(**_dict)
            # end if
        # end if
    # end def

    def bind_events (self, **kw):
        self.events.connect_dict(
            {
                "Component:Reset": self.on_component_reset,

                "Reset:Modified": self.on_reset_modified,

                "Resources:Item:Add": self.on_res_item_add,
                "Resources:Item:Delete": self.on_res_item_delete,
                "Resources:Item:Rename": self.on_res_item_rename,

                "Resources:Section:Add": self.on_res_section_add,
                "Resources:Section:Delete": self.on_res_section_delete,
                "Resources:Section:Rename": self.on_res_section_rename,

                "Resources:Type:Add": self.on_res_type_add,
                "Resources:Type:Delete": self.on_res_type_delete,
                "Resources:Type:Rename": self.on_res_type_rename,
            }
        )
        self.bind("<Expose>", self.on_expose)
        self.bind("<FocusIn>", self.on_expose)
        self.combobox_type.bind(
            "<<ComboboxSelected>>", self.on_type_combobox_selected
        )
        self.combobox_section.bind(
            "<<ComboboxSelected>>", self.on_section_combobox_selected
        )
        self.listbox_item.bind(
            "<<ListboxSelect>>", self.on_item_listbox_select
        )
        self.text_notes.bind("<KeyRelease>", self.on_text_keypress)
        for _entry in self.entries.values():
            _entry.bind("<KeyRelease>", self.on_entry_keypress)
            _entry.bind(
                "<Control-KeyRelease>", self.on_entry_ctrl_keypress
            )
        # end for
    # end def

    def clear_comboboxes (self, *comboboxes):
        for _combobox in comboboxes:
            _combobox.set("")
            _combobox.configure(values=[])
            _combobox.selection_clear()
            _combobox.items = dict()
        # end for
    # end def

    def clear_listboxes (self, *listboxes):
        for _listbox in listboxes:
            _listbox.delete(0, "end")
            _listbox.selection_clear(0, "end")
            _listbox.items = dict()
            _listbox.last_selected = -1
        # end for
    # end def

    def clear_listbox_item (self, *args, **kw):
        self.clear_listboxes(self.listbox_item)
        self.w_canvas.reset(except_scale=True)
    # end def

    def combobox_add_item (self, combobox, fk_parent=None):
        self.save_now()
        _new_name = self.user_ask_item_name()
        if _new_name:
            if _new_name in combobox.items:
                self.error_already_exists(_new_name)
            else:
                _rowid = self.database.res_add_type(
                    fk_parent, _new_name
                )
                combobox.items[_new_name] = _rowid
                _items = list(combobox.cget("values"))
                _items.append(_new_name)
                combobox.configure(values=sorted(_items))
                combobox.set(_new_name)
                combobox.event_generate("<<ComboboxSelected>>")
                self.combobox_modified = True
            # end if
        # end if
    # end def

    def combobox_delete_item (self, combobox):
        _index = combobox.current()
        if _index >= 0:
            _label = combobox.get()
            if self.user_confirm_deletion(_label):
                _rowid = combobox.items.pop(_label)
                self.database.res_del_type(_rowid)
                self.clear_listbox_item()
                _items = list(combobox.cget("values"))
                _items.pop(_index)
                combobox.configure(values=_items)
                _index = max(-1, min(len(_items) - 1, _index))
                if _index >= 0:
                    combobox.current(_index)
                    combobox.event_generate("<<ComboboxSelected>>")
                else:
                    self.clear_comboboxes(combobox)
                # end if
                self.on_update_inputs()
                self.combobox_modified = True
            # end if
        # end if
        return _index
    # end def

    def combobox_rename_item (self, combobox):
        _index = combobox.current()
        if _index >= 0:
            _old_name = combobox.get()
            _new_name = self.user_ask_rename(_old_name)
            if _new_name and _new_name != _old_name:
                if _new_name in combobox.items:
                    self.error_already_exists(_new_name)
                else:
                    _rowid = combobox.items.pop(_old_name)
                    combobox.items[_new_name] = _rowid
                    self.database.res_rename_type(_rowid, _new_name)
                    _items = list(combobox.cget("values"))
                    _items.pop(_index)
                    _items.insert(_index, _new_name)
                    combobox.configure(values=sorted(_items))
                    combobox.set(_new_name)
                    self.combobox_modified = True
                # end if
            # end if
        # end if
    # end def

    def error_already_exists (self, item_name):
        from tkinter.messagebox import showwarning
        showwarning(
            title=_("Attention"),
            message=_("Item '{item}' already exists.")
                .format(item=item_name),
            parent=self
        )
    # end def

    def get_current_selected_item (self):
        try:
            self.listbox_item.last_selected = (
                self.listbox_item.curselection()[0]
            )
        except:
            if not self.listbox_item.size():
                self.clear_listbox_item()
            # end if
        # end try
        return self.listbox_item.last_selected
    # end def

    def get_file_contents (self, files):
        import json
        self.save_now()
        _db = self.database
        _dict = {
            "types": [dict(i) for i in _db.res_get_all_types()],
            "items": [dict(i) for i in _db.res_get_all_items()],
            "datebars": [dict(i) for i in _db.res_get_all_datebars()],
            "config": {
                "date_scale": self.w_canvas.get_date_scale(),
            },
        }
        _fcontents = json.dumps(_dict)
        return {files: _fcontents}
    # end def

    def get_fk_type (self):
        return self.listbox_item.items[
            self.listbox_item.get(self.listbox_item.last_selected)
        ]
    # end def

    def get_res_items (self):
        return self.database.res_get_types(
            fk_parent=self.combobox_section
                            .items[self.combobox_section.get()]
        )
    # end def

    def get_res_section (self):
        return self.database.res_get_types(
            fk_parent=self.combobox_type
                            .items[self.combobox_type.get()]
        )
    # end def

    def initialize (self, *args, **kw):
        from tkRAD.core.async import get_async_manager
        from .database import get_database
        self.async = get_async_manager()
        self.database = get_database()
        self.xml_build("component_resources")
        _readonly = ["readonly"]
        self.combobox_type.state(_readonly)
        self.combobox_section.state(_readonly)
        self.entries = {
            _key: getattr(self, "entry_res_{}".format(_key))
            for _key in ("name", "role", "contact", "phone", "email")
        }
        self.on_component_reset()
        self.on_reset_modified()
        self.bind_events(**kw)
    # end def

    @property
    def modified (self):
        return bool(
            self.w_canvas.modified
            or self.combobox_modified
            or self.listbox_modified
            or self.entry_modified
            or self.text_notes.edit_modified()
        )
    # end def

    def on_component_reset (self, *args, **kw):
        self.database.res_reset()
        self.reset_resources()
    # end def

    def on_entry_ctrl_keypress (self, *args, **kw):
        self.async.run_after(3000, self.auto_save)
    # end def

    def on_entry_keypress (self, *args, **kw):
        self.entry_modified = True
        self.async.run_after(3000, self.auto_save)
    # end def

    def on_expose (self, *args, **kw):
        _index = self.listbox_item.last_selected
        if _index >= 0:
            self.listbox_item.see(_index)
            self.listbox_item.selection_set(_index)
        # end if
    # end def

    def on_item_listbox_select (self, *args, **kw):
        self.save_now()
        self.on_update_inputs()
        _index = self.get_current_selected_item()
        if _index >= 0:
            _row = self.database.res_get_item(self.get_fk_type())
            for _key, _entry in self.entries.items():
                _entry.insert("end", _row.get(_key) or "")
            # end for
            th.text_reset_contents(
                self.text_notes, _row.get("notes") or ""
            )
            self.after_idle(self.entries["name"].focus_set)
        # end if
    # end def

    def on_res_item_add (self, *args, **kw):
        self.save_now()
        if self.combobox_section.current() >= 0:
            _fk_parent = (
                self.combobox_section
                    .items[self.combobox_section.get()]
            )
            _new_name = self.user_ask_item_name()
            if _new_name:
                if _new_name in self.listbox_item.items:
                    self.error_already_exists(_new_name)
                else:
                    _rowid = self.database.res_add_type(
                        _fk_parent, _new_name
                    )
                    self.listbox_item.items[_new_name] = _rowid
                    _items = list(self.listbox_item.get(0, "end"))
                    _items.append(_new_name)
                    _items = sorted(_items)
                    _index = _items.index(_new_name)
                    self.listbox_item.delete(0, "end")
                    self.listbox_item.insert("end", *_items)
                    self.listbox_item.see(_index)
                    self.listbox_item.selection_set(_index)
                    self.on_item_listbox_select()
                    self.w_canvas.item_list_update(
                        self.listbox_item.items
                    )
                    self.listbox_modified = True
                # end if
            # end if
        # end if
    # end def

    def on_res_item_delete (self, *args, **kw):
        _index = self.get_current_selected_item()
        if _index >= 0:
            _label = self.listbox_item.get(_index)
            if self.user_confirm_deletion(_label):
                _rowid = self.listbox_item.items.pop(_label)
                self.database.res_del_type(_rowid)
                self.listbox_item.delete(_index)
                self.on_update_inputs()
                _index = (
                    max(-1, min(self.listbox_item.size() - 1, _index))
                )
                if _index >= 0:
                    self.listbox_item.see(_index)
                    self.listbox_item.selection_set(_index)
                    self.on_item_listbox_select()
                # end if
                self.w_canvas.item_list_update(self.listbox_item.items)
                self.listbox_modified = True
            # end if
        # end if
    # end def

    def on_res_item_rename (self, *args, **kw):
        _index = self.get_current_selected_item()
        if _index >= 0:
            _old_name = self.listbox_item.get(_index)
            _new_name = self.user_ask_rename(_old_name)
            if _new_name and _new_name != _old_name:
                if _new_name in self.listbox_item.items:
                    self.error_already_exists(_new_name)
                else:
                    self.listbox_item.delete(_index)
                    _items = list(self.listbox_item.get(0, "end"))
                    _items.append(_new_name)
                    _items = sorted(_items)
                    _index = _items.index(_new_name)
                    self.listbox_item.delete(0, "end")
                    self.listbox_item.insert("end", *_items)
                    self.listbox_item.see(_index)
                    self.listbox_item.selection_set(_index)
                    _rowid = self.listbox_item.items.pop(_old_name)
                    self.listbox_item.items[_new_name] = _rowid
                    self.database.res_rename_type(_rowid, _new_name)
                    self.w_canvas.item_list_update(
                        self.listbox_item.items
                    )
                    self.on_item_listbox_select()
                    self.listbox_modified = True
                # end if
            # end if
        # end if
    # end def

    def on_res_section_add (self, *args, **kw):
        if self.combobox_type.current() >= 0:
            _fk_parent = (
                self.combobox_type.items[self.combobox_type.get()]
            )
            self.combobox_add_item(self.combobox_section, _fk_parent)
        # end if
    # end def

    def on_res_section_delete (self, *args, **kw):
        self.combobox_delete_item(self.combobox_section)
    # end def

    def on_res_section_rename (self, *args, **kw):
        self.combobox_rename_item(self.combobox_section)
    # end def

    def on_res_type_add (self, *args, **kw):
        self.combobox_add_item(self.combobox_type)
    # end def

    def on_res_type_delete (self, *args, **kw):
        if self.combobox_delete_item(self.combobox_type) < 0:
            self.clear_comboboxes(self.combobox_section)
            self.on_update_inputs()
        # end if
    # end def

    def on_res_type_rename (self, *args, **kw):
        self.combobox_rename_item(self.combobox_type)
    # end def

    def on_reset_modified (self, *args, **kw):
        self.w_canvas.modified = False
        self.combobox_modified = False
        self.listbox_modified = False
        self.entry_modified = False
        self.text_notes.edit_modified(False)
    # end def

    def on_section_combobox_selected (self, *args, **kw):
        self.save_now()
        _dict = self.get_res_items()
        self.clear_listbox_item()
        self.listbox_item.insert("end", *sorted(_dict))
        self.listbox_item.items = _dict
        self.on_update_inputs()
        self.w_canvas.item_list_update(_dict)
    # end def

    def on_text_keypress (self, event=None, *args, **kw):
        th.text_edit_autoseparator(self.text_notes, event)
        self.async.run_after(3000, self.auto_save)
    # end def

    def on_type_combobox_selected (self, *args, **kw):
        self.save_now()
        _dict = self.get_res_section()
        self.clear_listbox_item()
        self.clear_comboboxes(self.combobox_section)
        self.combobox_section.configure(values=sorted(_dict))
        self.combobox_section.items = _dict
        if _dict:
            self.combobox_section.current(0)
            self.on_section_combobox_selected()
        else:
            self.on_update_inputs()
        # end if
    # end def

    def on_update_inputs (self, *args, **kw):
        _seltype = bool(self.combobox_type.current() + 1)
        _selsection = bool(self.combobox_section.current() + 1)
        _selitem = bool(self.get_current_selected_item() + 1)
        self.enable_widget(self.btn_delete_type, _seltype)
        self.enable_widget(self.btn_rename_type, _seltype)
        self.enable_widget(self.btn_add_section, _seltype)
        self.enable_widget(self.btn_delete_section, _selsection)
        self.enable_widget(self.btn_rename_section, _selsection)
        self.enable_widget(self.btn_add_item, _selsection)
        self.enable_widget(self.btn_delete_item, _selitem)
        self.enable_widget(self.btn_rename_item, _selitem)
        for _entry in self.entries.values():
            self.enable_widget(_entry, True)
            _entry.delete(0, "end")
            self.enable_widget(_entry, _selitem)
        # end for
        self.enable_widget(self.text_notes, True)
        th.text_reset_contents(self.text_notes)
        self.enable_widget(self.text_notes, _selitem)
    # end def

    def reset_resources (self):
        self.clear_comboboxes(
            self.combobox_type, self.combobox_section
        )
        self.clear_listbox_item()
        _dict = self.database.res_get_types()
        self.combobox_type.configure(values=sorted(_dict))
        self.combobox_type.items = _dict
        if _dict:
            self.combobox_type.current(0)
            self.on_type_combobox_selected()
        # end if
    # end def

    def save_now (self):
        self.async.stop(self.auto_save)
        self.auto_save()
    # end def

    def setup_component (self, files, archive):
        import json
        _dict = json.loads(files or "{}")
        _type_rows = _dict.get("types")
        if _type_rows:
            self.database.res_import_types(_type_rows)
            self.database.res_import_items(_dict.get("items"))
            self.database.res_import_datebars(_dict.get("datebars"))
            self.reset_resources()
        else:
            self.on_component_reset()
        # end if
        _cfg = _dict.get("config") or dict()
        self.w_canvas.set_date_scale(_cfg.get("date_scale"))
    # end def

    def user_ask_item_name (self):
        from tkinter.simpledialog import askstring
        return str(
            askstring(_("Please, insert"), _("New name"), parent=self)
            or ""
        ).strip()
    # end def

    def user_ask_rename (self, initial_value=""):
        from tkinter.simpledialog import askstring
        return str(
            askstring(
                title=_("Please, insert"),
                prompt=_("Renaming"),
                initialvalue=initial_value,
                parent=self
            ) or ""
        ).strip()
    # end def

    def user_confirm_deletion (self, label):
        from tkinter.messagebox import askyesno
        return askyesno(
            title=_("Attention"),
            message=_("Do you really want to delete\n'{item}'?")
                .format(item=label),
            parent=self
        )
    # end def

# end class
