#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    tkScenarist - screen writing made simpler

    Copyright (c) 2014+ Raphaël Seban <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see http://www.gnu.org/licenses/
"""

import tkRAD


class ComponentScenario (tkRAD.RADXMLFrame):

    DEFAULT_HINT = [
        _("ONE PAGE of script is roughly equal to ONE MINUTE "
        "of movie."),
    ]

    INFO_HINTS_FPATH = _("^/data/json/info_hints.en.json")

    LINES_PER_PAGE = 42

    def bind_events (self, **kw):
        self.events.connect_dict(
            {
                "Component:Reset": self.on_component_reset,

                "Reset:Modified": self.on_reset_modified,

                "Scenario:Current:Element:Update":
                    self.on_current_element_update,
                "Scenario:Elements:Init": self.on_elements_init,
                "Scenario:Text:Clicked": self.on_text_click,
                "Scenario:Text:FocusedIn": self.on_text_focus_in,
            }
        )
        self.bind("<Expose>", self.on_expose)
        self.w_combobox.bind(
            "<<ComboboxSelected>>", self.on_combobox_selected
        )
        self.w_listbox.bind(
            "<<ListboxSelect>>", self.on_listbox_select
        )
        self.w_popup.bind("<Button-1>", self.on_popup_click)
        self.w_popup.bind("<FocusOut>", self.on_text_focus_out)
        self.w_text.bind("<FocusOut>", self.on_text_focus_out)
    # end def

    @property
    def component_characters (self):
        return self.app.mainwindow.get_component("characters")
    # end def

    def get_file_contents (self, files):
        return self.w_text.get_file_contents(files)
    # end def

    def hide_popup_list (self, *args, **kw):
        self.async.stop(self.on_autocomplete)
        self.w_popup.hide()
    # end def

    def initialize (self, *args, **kw):
        from tkRAD.core.async import get_async_manager
        from .text_completion_popup import TextCompletionPopup
        self.async = get_async_manager()
        self.xml_build("component_scenario")
        self.w_combobox.state(["readonly"])
        self.w_combobox.current_selected = ""
        self.w_combobox.elements = dict()
        _cvar = self.get_stringvar
        self.cvar_combobox = _cvar("combobox_current_element")
        self.cvar_on_tab = _cvar("lbl_on_tab")
        self.cvar_on_return = _cvar("lbl_on_return")
        self.cvar_on_ctrl_return = _cvar("lbl_on_ctrl_return")
        self.cvar_info_hint = _cvar("lbl_info_hint")
        self.cvar_character_name = _cvar("lbl_character_name")
        self.cvar_page_count = _cvar("lbl_page_count")
        self.cvar_movie_duration = _cvar("lbl_movie_duration")
        self.w_popup = TextCompletionPopup(
            self.w_text, on_insert=self.on_popup_insert
        )
        self.on_component_reset()
        self.bind_events(**kw)
    # end def

    @property
    def modified (self):
        return self.w_text.edit_modified()
    # end def

    def on_autocomplete (self, *args, **kw):
        _word = self.w_text.get_word()
        _si = _word["start_index"]
        _names = self.component_characters.get_matching_names(
            _word["word"]
        )
        if not _names:
            _names = self.component_characters.get_matching_names(
                self.w_text.get_line_contents()
            )
            _si = "insert linestart"
        # end if
        if _names:
            self.show_popup_list(choices=_names, start_index=_si)
        else:
            self.hide_popup_list()
        # end if
    # end def

    def on_combobox_selected (self, *args, **kw):
        _text = self.cvar_combobox.get()
        if _text != self.w_combobox.current_selected:
            self.update_selected(_text)
            self.events.raise_event(
                "Scenario:Switch:Line",
                new_tag=self.w_combobox.elements[_text]
            )
            self.after_idle(self.w_text.focus_set)
        # end if
    # end def

    def on_component_reset (self, *args, **kw):
        self.w_text.reset()
        self.reset_scene_browser()
        self.reset_hints()
    # end def

    def on_current_element_update (self, *args, **kw):
        _map = self.w_text.get_element_mappings()
        _label = lambda n: self.w_text.get_label(
            _map["on_{}_create".format(n)] or
            _map["on_{}_switch".format(n)]
        )
        self.set_combobox_text(self.w_text.get_label(_map["tag"]))
        self.cvar_on_tab.set(_label("tab"))
        self.cvar_on_return.set(_label("return"))
        self.cvar_on_ctrl_return.set(_label("ctrl_return"))
        self.async.run_after(1000, self.update_hints)
        self.async.run_after(500, self.update_stats)
        self.async.run_after(200, self.update_scene_browser)
        self.async.run_after(100, self.update_character_log)
    # end def

    def on_elements_init (self, *args, elements=None, **kw):
        if elements:
            self.w_combobox.current_selected = ""
            self.w_combobox.elements = {
                self.w_text.get_label(_element): _element
                for _element in elements
            }
            self.w_combobox.configure(
                values=sorted(self.w_combobox.elements)
            )
        # end if
    # end def

    def on_expose (self, *args, **kw):
        self.w_text.focus_set()
    # end def

    def on_focus_out (self, *args, **kw):
        self.async.run_after(100, self.hide_popup_list)
    # end def

    def on_listbox_select (self, *args, **kw):
        try:
            _index = float(
                self.w_listbox.current_lines[
                    self.w_listbox.curselection()[0]
                ]
            )
        except:
            pass
        else:
            self.w_text.see(_index)
            self.w_text.move_cursor("{} lineend".format(_index))
            self.after_idle(self.w_text.focus_set)
        # end try
    # end def

    def on_popup_click (self, *args, **kw):
        self.after_idle(
            self.async.stop, self.hide_popup_list, self.on_autocomplete
        )
    # end def

    def on_popup_insert (self, text, start_index):
        self.w_text.replace_text(text, start_index, smart_delete=True)
    # end def

    def on_reset_modified (self, *args, **kw):
        self.w_text.edit_modified(False)
    # end def

    def on_text_click (self, *args, **kw):
        self.hide_popup_list()
    # end def

    def on_text_focus_in (self, *args, **kw):
        self.hide_popup_list()
    # end def

    def on_text_focus_out (self, *args, **kw):
        self.async.run_after(100, self.hide_popup_list)
    # end def

    def reset_hints (self, filepath=None):
        import json
        from tkRAD.core.path import normalize
        filepath = normalize(filepath or self.INFO_HINTS_FPATH)
        with open(filepath, encoding=ENCODING) as _file:
            self.INFO_HINTS = json.load(_file)
        # end with
    # end def

    def reset_scene_browser (self, *args, **kw):
        self.w_listbox.current_lines = []
        self.w_listbox.delete(0, "end")
    # end def

    def set_combobox_text (self, text):
        if text in self.w_combobox.elements and \
                            text != self.w_combobox.current_selected:
            self.update_selected(text)
            self.cvar_combobox.set(text)
        # end if
    # end def

    def setup_component (self, files, archive):
        self.w_text.file_setup(files, archive)
    # end def

    def show_popup_list (self, *args, **kw):
        self.async.stop(self.hide_popup_list, self.on_autocomplete)
        self.w_popup.show(*args, **kw)
    # end def

    def update_character_log (self, *args, **kw):
        from . import text_helpers as th
        _line = self.w_text.get_line_contents()
        _name, _start_index = (
            self.component_characters.find_nearest_name(
                _line, self.w_text.get_column_index()
            )
        )
        self.enable_widget(self.text_characters_log, True)
        if self.component_characters.is_registered(_name):
            self.cvar_character_name.set(_name)
            th.text_set_contents(
                self.text_characters_log,
                self.component_characters.get_character_log(_name)
            )
            if _name not in _line:
                _index = "insert linestart+{}c"
                self.w_text.replace_text(
                    _name,
                    _index.format(_start_index),
                    _index.format(_start_index + len(_name)),
                    keep_cursor=True,
                )
            # end if
            self.hide_popup_list()
        else:
            self.cvar_character_name.set("")
            th.text_reset_contents(self.text_characters_log)
            self.async.run_after_idle(self.on_autocomplete)
        # end if
        self.enable_widget(self.text_characters_log, False)
    # end def

    def update_hints (self):
        import random
        _tag = self.w_text.get_line_tag()
        _hints = (
            self.INFO_HINTS.get(_tag)
            or self.INFO_HINTS.get("default")
            or self.DEFAULT_HINT
        )
        if random.randint(1, 10) == 7:
            _hints = self.DEFAULT_HINT
        # end if
        self.cvar_info_hint.set(str(random.choice(_hints)))
    # end def

    def update_scene_browser (self):
        _dict = self.w_text.get_lines("scene")
        _cursor = _dict["current"]
        _lines = self.w_listbox.current_lines = _dict["lines"]
        _index = -1
        self.w_listbox.delete(0, "end")
        self.w_listbox.insert("end", *_dict["texts"])
        self.w_listbox.selection_clear(0, "end")
        try:
            _index = _lines.index(_cursor)
        except:
            pass
        else:
            if not (self.w_text.line_selected(float(_cursor)) or
                                        self.w_popup.winfo_viewable()):
                self.w_listbox.selection_set(_index)
                self.w_listbox.see(_index)
            # end if
        # end try
        self.events.raise_event(
            "Scenario:Scene:Browser:Changed",
            contents=_dict["texts"],
            lines=_dict["lines"],
            line_ids=_dict["line_ids"],
            current_line=_cursor,
            current_selected=_index,
        )
    # end def

    def update_selected (self, text):
        self.w_combobox.current_selected = text
        self.w_combobox.selection_clear()
    # end def

    def update_stats (self, *args, **kw):
        _lpp = self.LINES_PER_PAGE
        _lines = self.w_text.get_nb_of_lines()
        _nb_pages = _lines // _lpp + 1
        _cur_page = self.w_text.get_line_number() // _lpp + 1
        _mduration = int(0.9 * (_lines / _lpp + 1))
        self.cvar_page_count.set(
            _("{page} of {total}")
            .format(page=_cur_page, total=_nb_pages)
        )
        if _mduration < 1:
            _time_fmt = _("< 1 minute")
        else:
            _time_fmt = _(
                "{hours:02d} h {minutes:02d} min"
            ).format(
                hours=(_mduration // 60), minutes=(_mduration % 60)
            )
        # end if
        self.cvar_movie_duration.set(_time_fmt)
        self.events.raise_event(
            "Scenario:Stats:Updated",
            nb_of_lines=_lines,
            total_pages=_nb_pages,
            current_page=_cur_page,
            movie_duration=_mduration,
            movie_duration_label=_time_fmt,
        )
    # end def

# end class
