#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    tkScenarist - screen writing made simpler

    Copyright (c) 2014+ Raphaël Seban <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see http://www.gnu.org/licenses/
"""

from tkRAD.widgets.rad_canvas import RADCanvas


class DBViewCanvas (RADCanvas):

    CONFIG = {
        "bg": "white",
        "highlightbackground": "grey80",
        "highlightthickness": 0,
    }

    CONFIG_FIELD = {
        "body": {
            "text": {"align": "left", "font": "sans 10"},
            "box": {},
        },
        "header": {
            "text": {"align": "center", "font": "sans 11 bold"},
            "box": {},
        },
    }

    def bind_events (self, *args, **kw):
        for _sequence in ("<Button-4>", "<Button-5>", "<MouseWheel>"):
            self.bind(_sequence, self.on_mouse_wheel)
        # end for
    # end def

    def clear_canvas (self, *args, **kw):
        self.delete("all")
        self.configure(scrollregion=(0, 0, 0, 0))
        self.xview_moveto(0)
        self.yview_moveto(0)
    # end def

    def do_set_field_options (self, group_tag, fieldname, **options):
        try:
            _options = (
                self.field_options[group_tag].get(fieldname)
                or self.CONFIG_FIELD[group_tag].copy()
            )
            _options.update(options)
            self.field_options[group_tag][fieldname] = _options
        except:
            pass
        # end try
    # end def

    def do_update_all (self):
        self.do_update_dimension(self.columns, (1, 0))
        self.do_update_dimension(self.rows, (0, 1))
        self.update_canvas()
    # end def

    def do_update_canvas (self):
        try:
            _x0, _y0, _x1, _y1 = self.bbox("all")
            _x1 = max(_x1, self.winfo_width() - 1)
            _y1 = max(_y1, self.winfo_height() - 1)
            self.configure(scrollregion=(_x0, _y0, _x1, _y1))
        except:
            self.reset()
        # end try
    # end def

    def do_update_dimension (self, managers, move_xy):
        _dmove = 0
        _previous = None
        _sx, _sy = move_xy
        for _manager in managers:
            _manager.resize_all()
            if _previous:
                _dmove += _previous.get_move_delta()
                self.move(_manager.tag, _sx * _dmove, _sy * _dmove)
                _previous.reset_move()
            # end if
            _previous = _manager
        # end for
    # end def

    def ensure_visible_header (self, *args, **kw):
        try:
            _x0, _y0, _x1, _y1 = self.bbox("header")
            y = self.canvasy(-0.5)
            self.move("header", 0, y - _y0)
            self.tag_raise("header", "all")
        except:
            pass
        # end try
    # end def

    def finalize (self, *args, **kw):
        self.reset()
    # end def

    def get_column_manager (self, index):
        try:
            _manager = self.columns[index]
        except:
            _manager = None
        # end try
        if not _manager:
            _manager = DBViewColumnManager()
            self.columns.insert(index, _manager)
        # end if
        return _manager
    # end def

    def get_field_options (self, group_tag, fieldname):
        try:
            return self.field_options[group_tag].get(fieldname)
        except:
            return dict()
        # end try
    # end def

    def get_insertion_xy (self, row_manager, column_manager):
        try:
            _x = self.coords(column_manager.tag)[0]
        except:
            _x = 0
        # end try
        try:
            _y = self.coords(row_manager.tag)[1]
        except:
            _y = 0
        # end try
        return (_x, _y)
    # end def

    def get_row_manager (self, index):
        try:
            _manager = self.rows[index]
        except:
            _manager = None
        # end try
        if not _manager:
            _manager = DBViewRowManager()
            self.rows.insert(index, _manager)
        # end if
        return _manager
    # end def

    def initialize (self, *args, **kw):
        from tkRAD.core.async import get_async_manager
        self.async = get_async_manager()
        self.reset(*args, **kw)
        self.bind_events(*args, **kw)
    # end def

    def insert_label (self, group_tag, text, **kw):
        _row = kw.get("row") or self.row_index
        _column = kw.get("column") or self.column_index
        _field_options = kw.get("field_options") or dict()
        _text_options = _field_options.get("text")
        _box_options = _field_options.get("box")
        _row_manager = self.get_row_manager(_row)
        _column_manager = self.get_column_manager(_column)
        _tags = (group_tag, _row_manager.tag, _column_manager.tag)
        _text_options.update(tags=_tags)
        _box_options.update(tags=_tags)
        _label = DBViewLabel(
            self,
            text_options=_text_options,
            box_options=_box_options,
            on_width_changed=_column_manager.on_dimension_changed,
            on_height_changed=_row_manager.on_dimension_changed,
        )
        _x, _y = self.get_insertion_xy(_row_manager, _column_manager)
        _label.create_label(_x, _y, text)
        _row_manager.insert_item(_row, _label)
        _column_manager.insert_item(_column, _label)
        if kw.get("column") is None:
            self.column_index += 1
        # end if
    # end def

    def insert_row (self, row_dict, row_index=None):
        if row_index is None:
            row_index = self.row_index
            self.next_row()
        # end if
        for _column, _fieldname in enumerate(self.field_sequence):
            _text = str(row_dict.get(_fieldname) or "")
            _field_options = self.get_field_options("body", _fieldname)
            self.insert_label(
                "body", row=row_index, column=_column, text=_text,
                field_options=_field_options,
            )
        # end for
        self.update_all()
    # end def

    def next_row (self, *args, **kw):
        self.row_index += 1
        self.column_index = 0
    # end def

    def on_mouse_wheel (self, event=None, *args, **kw):
        import os
        _platform = os.name.lower()
        if _platform == "nt":
            _step = -event.delta // 120
        elif _platform == "mac":
            _step = -event.delta
        else:
            _step = (event.num == 5) - (event.num == 4)
        # end if
        self.yview_scroll(_step, "units")
    # end def

    def reset (self, *args, **kw):
        self.async.stop(self.do_update_all, self.do_update_canvas)
        self.clear_canvas(*args, **kw)
        self.field_sequence = None
        self.field_options = dict(body=dict(), header=dict())
        self.column_index = 0
        self.row_index = 0
        self.rows = []
        self.columns = []
    # end def

    def set_field_options (self, group_tag, fieldname, **options):
        if fieldname in self.field_sequence:
            group_tag = str(group_tag).lower()
            if group_tag in ("body", "both", "all"):
                self.do_set_field_options("body", fieldname, **options)
            # end if
            if group_tag in ("header", "both", "all"):
                self.do_set_field_options(
                    "header", fieldname, **options
                )
            # end if
            return self.field_options
        # end if
    # end def

    def set_header (self, *fieldnames, header_options=None,
                                                    body_options=None):
        self.reset()
        self.field_sequence = fieldnames
        header_options = header_options or dict()
        body_options = body_options or dict()
        for _fieldname in fieldnames:
            self.set_field_options(
                "header", _fieldname,
                **(header_options.get(_fieldname) or header_options)
            )
            _field_options = self.set_field_options(
                "body", _fieldname,
                **(body_options.get(_fieldname) or body_options)
            )
            self.insert_label(
                "header",
                text=_(_fieldname), # i18n support
                field_options=_field_options["header"][_fieldname]
            )
        # end for
        self.next_row()
        self.update_all()
    # end def

    def update_all (self, *args, **kw):
        self.async.run_after_idle(self.do_update_all)
    # end def

    def update_canvas (self, *args, **kw):
        self.async.run_after_idle(self.do_update_canvas)
    # end def

    def yview (self, *args):
        super().yview(*args)
        self.ensure_visible_header()
    # end def

    def yview_moveto (self, *args):
        super().yview_moveto(*args)
        self.ensure_visible_header()
    # end def

    def yview_scroll (self, *args):
        super().yview_scroll(*args)
        self.ensure_visible_header()
    # end def

# end class


class DBViewDimensionManager:

    DIMENSION_NAME = "dimension"

    def __init__ (self, tag=None):
        self.tag = tag or "group#{}".format(id(self))
        self.dimension = 0
        self.move_delta = 0
        self.collection = list()
    # end def

    def add_item (self, item):
        self.collection.append(item)
    # end def

    def get_move_delta (self):
        return self.move_delta
    # end def

    def insert_item (self, index, item):
        self.collection.insert(index, item)
    # end def

    def on_dimension_changed (self, new_dimension):
        _dmove = new_dimension - self.dimension
        if _dmove > 0:
            self.dimension = new_dimension
            self.move_delta += _dmove
        # end if
    # end def

    def reset_move (self, *args, **kw):
        self.move_delta = 0
    # end def

    def resize_all (self, *args, **kw):
        for _item in self.collection:
            _item.resize(**{self.DIMENSION_NAME: self.dimension})
        # end for
    # end def

# end class


class DBViewColumnManager (DBViewDimensionManager):

    DIMENSION_NAME = "width"

# end class


class DBViewRowManager (DBViewDimensionManager):

    DIMENSION_NAME = "height"

# end class


class DBViewLabel:

    BOX_OPTIONS = {
        "fill": "white",
        "outline": "black",
        "width": 1,
    }

    LABEL_BOX = (-5, -5, +5, +5)

    TEXT_OPTIONS = {
        "align": "left",
        "fill": "black",
        "font": "helvetica 10",
        "width": None,
    }

    def __init__ (self, dbviewcanvas, **kw):
        self.__box_options = self.BOX_OPTIONS.copy()
        self.__text_options = self.TEXT_OPTIONS.copy()
        self.canvas = dbviewcanvas
        self.id_box = 0
        self.id_text = 0
        self.on_width_changed = kw.get("on_width_changed")
        self.on_height_changed = kw.get("on_height_changed")
        self.text_options = kw.get("text_options")
        self.box_options = kw.get("box_options")
    # end def

    def box_add (self, *boxes):
        return tuple(map(sum, zip(*boxes)))
    # end def

    def box_center (self, _x0, _y0, _x1, _y1):
        return ((_x0 + _x1) // 2, (_y0 + _y1) // 2)
    # end def

    def box_coords (self, new_coords=None):
        if self.id_box:
            if new_coords:
                self.canvas.coords(self.id_box, *new_coords)
            else:
                return self.canvas.coords(self.id_box)
            # end if
        # end if
    # end def

    @property
    def box_options (self):
        return self.__box_options
    # end def

    @box_options.setter
    def box_options (self, options):
        if options:
            self.__box_options.update(options)
        # end if
    # end def

    def box_resize (self, width=0, height=0, box=None):
        _box = box or self.box_coords()
        if width:
            _box[2] = _box[0] + width
        # end if
        if height:
            _box[3] = _box[1] + height
        # end if
        self.box_coords(_box)
    # end def

    def box_size (self, box):
        try:
            _x0, _y0, _x1, _y1 = box
            return (abs(_x1 - _x0), abs(_y1 - _y0))
        except:
            return (0, 0)
        # end try
    # end def

    def configure_box (self, **box_options):
        self.box_options = box_options
        if self.id_box:
            self.canvas.itemconfigure(self.id_box, **self.box_options)
        # end if
    # end def

    def configure_text (self, text, **text_options):
        self.text_options = text_options
        if self.id_text:
            _tk_opts = self.text_options.copy()
            _tk_opts.pop("align", None)
            self.canvas.itemconfigure(
                self.id_text, text=text, **_tk_opts
            )
            self.update_box()
        # end if
    # end def

    def create_label (self, left_x, top_y, text, **text_options):
        if self.id_text:
            self.configure_text(text, **text_options)
        else:
            _xb0, _yb0, _xb1, _yb1 = self.LABEL_BOX
            self.id_text = self.canvas.create_text(
                left_x - _xb0 + 1, top_y - _yb0,
                anchor="nw",
            )
            _box = self.get_surrounding_box()
            self.id_box = self.canvas.create_rectangle(
                _box, **self.box_options
            )
            self.configure_text(text, **text_options)
            self.canvas.tag_raise(self.id_text, self.id_box)
        # end if
    # end def

    def get_surrounding_box (self):
        if self.id_text:
            return self.box_add(
                self.canvas.bbox(self.id_text), self.LABEL_BOX
            )
        # end if
    # end def

    def label_size_changed (self, width, height):
        try:
            self.on_width_changed(width)
        except:
            pass
        # end try
        try:
            self.on_height_changed(height)
        except:
            pass
        # end try
    # end def

    def reset (self, *args, **kw):
        self.canvas.delete(self.id_box, self.id_text)
        self.id_box = self.id_text = 0
    # end def

    def resize (self, width=0, height=0):
        self.box_resize(width, height)
        if width:
            self.update_text()
        # end if
    # end def

    @property
    def text_options (self):
        return self.__text_options
    # end def

    @text_options.setter
    def text_options (self, options):
        if options:
            self.__text_options.update(options)
            self.__text_options.pop("anchor", None)
        # end if
    # end def

    def update_box (self, *args, **kw):
        if self.id_box and self.id_text:
            _box = self.box_coords()
            _w0, _h0 = self.box_size(_box)
            _w1, _h1 = self.box_size(self.get_surrounding_box())
            _dw = _w1 - _w0
            _dh = _h1 - _h0
            if _dw or _dh:
                self.box_resize(box=_box, width=_w1, height=_h1)
                self.label_size_changed(width=_w1, height=_h1)
            # end if
        # end if
    # end def

    def update_text (self, *args, **kw):
        if self.id_box and self.id_text:
            _align = self.text_options.get("align") or "left"
            _x0, _y0, _x1, _y1 = kw.get("box") or self.box_coords()
            _xb0, _yb0, _xb1, _yb1 = self.LABEL_BOX
            if _align == "center":
                _x, _y = self.box_center(_x0, _y0, _x1, _y1)
                _anchor = "center"
            elif _align == "right":
                _x, _y = (_x1 - _xb1, _y1 - _yb1)
                _anchor = "se"
            else:
                _x, _y = (_x0 - _xb0 + 1, _y0 - _yb0)
                _anchor = "nw"
            # end if
            self.canvas.coords(self.id_text, _x, _y)
            self.canvas.itemconfigure(self.id_text, anchor=_anchor)
        # end if
    # end def

# end class
