#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    tkScenarist - screen writing made simpler

    Copyright (c) 2014+ Raphaël Seban <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see http://www.gnu.org/licenses/
"""

from tkRAD.widgets.rad_dialog import RADDialog


class DateBarDialog (RADDialog):

    def bind_events (self, **kw):
        self.bind("<Escape>", self.cancel)
        for _combobox in self.begin_comboboxes[1:]:
            _combobox.bind(
                "<<ComboboxSelected>>", self.on_begin_combobox_selected
            )
        # end for
        for _combobox in self.end_comboboxes[1:]:
            _combobox.bind(
                "<<ComboboxSelected>>", self.on_end_combobox_selected
            )
        # end for
    # end def

    def get_date (self, group):
        from datetime import date
        _day, _month, _year = group
        _y = int(_year.get())
        _m = _month.current() + 1
        try:
            return date(_y, _m, int(_day.get()))
        except:
            return date(_y, _m, 1)
        # end try
    # end def

    def get_days (self, cdate):
        from calendar import monthrange
        return tuple(
            "{:02d}".format(i + 1)
            for i in range(monthrange(cdate.year, cdate.month)[1])
        )
    # end def

    def init_comboboxes (self, *groups):
        from calendar import month_name
        from datetime import date
        _days = self.get_days(date.today())
        _months = tuple(month_name)[1:]
        _year = date.today().year
        _years = tuple(range(_year - 1, _year + 2))
        _readonly = ["readonly"]
        for _group in groups:
            _day, _month, _year = _group
            _day.configure(values=_days, state=_readonly)
            _day.current(0)
            _month.configure(values=_months, state=_readonly)
            _month.current(0)
            _year.configure(values=_years, state=_readonly)
            _year.current(0)
        # end for
    # end def

    def on_begin_combobox_selected (self, *args, **kw):
        self.update_days_range(self.begin_comboboxes)
    # end def

    def on_end_combobox_selected (self, *args, **kw):
        self.update_days_range(self.end_comboboxes)
    # end def

    def reset_comboboxes (self, **kw):
        from datetime import date
        if self.datebar:
            _begin = self.datebar.date_begin
            _end = self.datebar.date_end
        else:
            _begin = _end = date.today()
        # end if
        self.reset_date(_begin, self.begin_comboboxes)
        self.reset_date(_end, self.end_comboboxes)
    # end def

    def reset_date (self, cdate, group):
        from datetime import date
        _year = date.today().year
        _ymin = min(_year - 1, cdate.year)
        _ymax = max(_year + 5, cdate.year)
        _day, _month, _year = group
        _day.configure(values=self.get_days(cdate))
        _day.current(cdate.day - 1)
        _month.current(cdate.month - 1)
        _year.configure(values=tuple(range(_ymin, _ymax + 1)))
        _year.set(cdate.year)
    # end def

    def setup_gui (self, *args, **kw):
        from tkinter import ttk
        from tkRAD import RADXMLFrame
        _frame = ttk.Frame(self, padding=5)
        self.body = RADXMLFrame(_frame)
        self.body.xml_build("dlg_date_bar")
        self.body.pack(expand=1, fill="both")
        ttk.Separator(_frame, orient="horizontal").pack(
            expand=1, fill="x", pady=5
        )
        _frame2 = ttk.Frame(_frame)
        self.btn_ok = ttk.Button(
            _frame2, text=_("OK"), command=self.validate
        )
        self.btn_ok.pack(side="left", padx=5)
        self.btn_cancel = ttk.Button(
            _frame2, text=_("Cancel"), command=self.cancel
        )
        self.btn_cancel.pack(side="right", padx=5)
        _frame2.pack()
        _frame.pack()
        self.datebar = kw.get("datebar")
        _cvar = self.body.get_stringvar
        self.cvar_name = _cvar("item_name")
        self.cvar_name.set(kw.get("item_name") or "sample demo")
        self.cvar_status = _cvar("opt_status")
        self.cvar_status.set(
            self.datebar and self.datebar.status == "N/A" and "N/A"
            or "OK"
        )
        self.begin_comboboxes = (
            self.body.combo_begin_day,
            self.body.combo_begin_month,
            self.body.combo_begin_year,
        )
        self.end_comboboxes = (
            self.body.combo_end_day,
            self.body.combo_end_month,
            self.body.combo_end_year,
        )
        self.init_comboboxes(
            self.begin_comboboxes, self.end_comboboxes
        )
        self.reset_comboboxes(**kw)
        self.bind_events(**kw)
    # end def

    def update_days_range (self, group):
        _days = self.get_days(self.get_date(group))
        _day = group[0]
        _index = _day.current()
        _day.configure(values=_days)
        _day.current(min(_index, len(_days) - 1))
    # end def

    def validate (self, *args, **kw):
        _begin = self.get_date(self.begin_comboboxes)
        _end = self.get_date(self.end_comboboxes)
        if _begin > _end:
            _begin, _end = _end, _begin
        # end if
        self.events.raise_event(
            "Dialog:DateBar:Validate",
            datebar=self.datebar,
            item_name=self.cvar_name.get(),
            status=self.cvar_status.get(),
            date_begin=_begin,
            date_end=_end,
        )
        self.destroy()
    # end def

# end class
