#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    tkScenarist - screen writing made simpler

    Copyright (c) 2014+ Raphaël Seban <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see http://www.gnu.org/licenses/
"""

from tkRAD.widgets.rad_dialog import RADDialog


class ExportPDFDialog (RADDialog):

    ITEM_NAMES = (
        "characters", "draft_notes", "pitch_concept", "resources",
        "scenario", "storyboard",
    )

    OPT_NAMES = (
        "print_scene_left", "print_scene_right", "print_shot_left",
        "print_shot_right", "show_in_viewer",
    )

    ALL_NAMES = ITEM_NAMES + OPT_NAMES

    def bind_events (self, **kw):
        self.events.connect_dict(
            {
                "Dialog:ExportPDF:Checkbutton:Click":
                    self.on_checkbutton_click,
                "Dialog:ExportPDF:Export": self.on_export_pdf,
            }
        )
        self.bind("<Escape>", self.cancel)
    # end def

    def cancel (self, *args, **kw):
        if self.running_task:
            self.user_show_warning(
                _(
                    "Cannot quit dialog. "
                    "Exportation procedure is yet running."
                )
            )
        else:
            self.destroy()
        # end if
    # end def

    def check_cvar (self, cvarname, state=True):
        from tkRAD.core.tools import ensure_int
        if state is not None:
            self.body.get_stringvar(cvarname).set(
                "1" if ensure_int(state) else ""
            )
        # end if
    # end def

    def cvar_checked (self, cvarname):
        return bool(self.body.get_stringvar(cvarname).get() == "1")
    # end def

    def export_loop (self, kw):
        if self.running_task and not self.user_cancelled:
            if not kw.get("export_list"):
                self.show_status(_("Nothing to export. Aborted."))
            else:
                try:
                    exec(
                        "self.export_step_{}(kw)"
                        .format(kw.get("step") or 0)
                    )
                    self.async.run_after(50, self.export_loop, kw)
                    return
                except:
                    self.show_status(
                        _(
                            "Got an error while trying to export. "
                            "Aborting."
                        )
                    )
                    raise
                # end try
            # end if
        # end if
        self.set_progressbar(100)
        self.enable_widget(self.btn_close, True)
        self.body.btn_export.configure(
            text=_("Export"), command=self.on_export_pdf
        )
        self.async.run_after(5000, self.reset)
        self.running_task = False
    # end def

    def export_step_0 (self, kw):
        _export_list = kw.get("export_list")
        _doc_index = kw.setdefault("doc_index", 0)
        if _doc_index < len(_export_list):
            if kw.get("doc"):
                self.show_in_viewer(kw)
            # end if
            _doc_name = _export_list[_doc_index]
            self.show_status(
                _("Trying to export '{}'...")
                .format(_(self.get_fancy_name(_doc_name)))
            )
            from .pdf_export import get_pdf_document
            kw["doc"] = get_pdf_document(_doc_name, **kw)
            kw["doc_index"] += 1
            kw["progress"] = 0
            kw["step"] = 1
        else:
            self.show_status(
                _("All selected items exported. Done.")
            )
            self.show_in_viewer(kw)
            self.user_cancelled = True
        # end if
    # end def

    def export_step_1 (self, kw):
        from .pdf_export import build_elements
        self.generic_step(
            _("Building elements..."), kw, build_elements
        )
    # end def

    def export_step_2 (self, kw):
        from .pdf_export import build_document
        self.generic_step(_("Exporting to PDF..."), kw, build_document)
        kw["step"] %= 3
    # end def

    def finalize (self, *args, **kw):
        self.stop_async()
    # end def

    def generic_step (self, message, kw, callback):
        self.show_status(message)
        self.set_progressbar(
            25.0 * (kw["step"] + (kw["progress"] / 100.0))
        )
        kw["progress"] = callback(kw["doc"])
        if kw["progress"] >= 100:
            kw["progress"] = 0
            kw["step"] += 1
        # end if
    # end def

    def get_export_list (self):
        return [
            _name
            for _name in self.ITEM_NAMES
            if self.cvar_checked("chk_" + _name)
        ]
    # end def

    def get_fancy_name (self, doc_name):
        return str(doc_name).title().replace("_", "/")
    # end def

    def get_options (self):
        return {
            _name: self.cvar_checked("chk_" + _name)
            for _name in self.OPT_NAMES
        }
    # end def

    def init_rc_options (self, **kw):
        _section = self.get_rc_section()
        for _cvarname in self.ALL_NAMES:
            _chk = "chk_" + _cvarname
            self.check_cvar(
                _chk, self.options.get(_section, _chk, fallback=None)
            )
        # end for
    # end def

    def on_checkbutton_click (self, *args, **kw):
        self.save_rc_options()
    # end def

    def on_export_pdf (self, *args, **kw):
        self.stop_async()
        self.enable_widget(self.btn_close, False)
        self.body.btn_export.configure(
            text=_("Stop"), command=self.on_stop_export,
        )
        self.show_status(
            _("Trying to export selected items, please wait.")
        )
        self.progressbar_wait()
        self.running_task = True
        self.user_cancelled = False
        self.viewer_failure = False
        self.async.run_after_idle(
            self.export_loop,
            dict(
                export_list=self.get_export_list(),
                options=self.get_options(),
                data=self.component_title_data.get_data(),
            )
        )
    # end def

    def on_stop_export (self, *args, **kw):
        self.user_cancelled = True
        self.show_status(
            _("User asked for cancellation.")
        )
    # end def

    def progressbar_wait (self, *args, **kw):
        self.body.w_progressbar.stop()
        self.body.w_progressbar.configure(mode="indeterminate")
        self.body.w_progressbar.start()
    # end def

    def reset (self, *args, **kw):
        self.show_status("")
        self.set_progressbar(0)
    # end def

    def save_rc_options (self, *args, **kw):
        _section = self.get_rc_section()
        for _cvarname in self.ALL_NAMES:
            _chk = "chk_" + _cvarname
            self.options[_section][_chk] = str(
                int(self.cvar_checked(_chk))
            )
        # end for
    # end def

    def set_progressbar (self, value):
        self.body.w_progressbar.configure(mode="determinate")
        self.body.w_progressbar.stop()
        self.cvar_progressbar_value.set(str(int(value)))
    # end def

    def setup_gui (self, **kw):
        from tkinter import ttk
        from tkRAD import RADXMLFrame
        from tkRAD.core.async import get_async_manager
        self.async = get_async_manager()
        self.component_title_data = (
            self.app.mainwindow.get_component("title_data")
        )
        self.running_task = False
        self.user_cancelled = False
        self.viewer_failure = False
        _frame = ttk.Frame(self, padding=5)
        self.body = RADXMLFrame(_frame)
        self.body.xml_build("dlg_export_pdf")
        self.body.pack(expand=1, fill="both")
        ttk.Separator(_frame, orient="horizontal").pack(
            expand=1, fill="x", pady=5
        )
        self.btn_close = ttk.Button(
            _frame, text=_("Close"), command=self.cancel
        )
        self.btn_close.pack()
        _frame.pack()
        _cvar = self.body.get_stringvar
        self.cvar_status = _cvar("lbl_export_status")
        self.cvar_progressbar_value = _cvar("progressbar_value")
        self.init_rc_options(**kw)
        self.bind_events(**kw)
    # end def

    def show_in_viewer (self, kw):
        if kw["options"]["show_in_viewer"] and not \
                                                self.viewer_failure:
            import os
            _filepath = kw["doc"].filepath
            try:
                if hasattr(os, "startfile"):
                    os.startfile(_filepath)
                else:
                    _ret = os.system('xdg-open "{}"'.format(_filepath))
                    if _ret != os.EX_OK:
                        raise RuntimeError
                    # end if
                # end if
            except:
                self.viewer_failure = True
                self.user_show_warning(
                    _("Could not launch any PDF viewer.")
                )
            # end try
        # end if
    # end def

    def show_status (self, message):
        self.cvar_status.set(str(message))
    # end def

    def stop_async (self, *args, **kw):
        self.async.stop(self.export_loop, self.reset)
    # end def

    def user_show_warning (self, msg):
        from tkinter.messagebox import showwarning
        showwarning(
            title=_("Attention"), message=str(msg), parent=self
        )
    # end def

# end class
