#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    tkScenarist - screen writing made simpler

    Copyright (c) 2014+ Raphaël Seban <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see http://www.gnu.org/licenses/
"""

from tkRAD.widgets.rad_dialog import RADDialog


class NameDatabaseDialog (RADDialog):

    ROW_LIMIT = 50

    def bind_events (self, **kw):
        self.events.connect_dict(
            {
                "Dialog:NameDB:Search:Criteria:Changed":
                    self.on_search_criteria_changed,

                "Dialog:NameDB:Search:Filter:All":
                    self.on_search_filter_all,
                "Dialog:NameDB:Search:Filter:Clicked":
                    self.on_search_filter_clicked,

                "Dialog:NameDB:Import:File": self.on_import_file,

                "Dialog:NameDB:Show:First": self.on_show_first,
                "Dialog:NameDB:Show:Next": self.on_show_next,
                "Dialog:NameDB:Show:Previous": self.on_show_previous,
            }
        )
        self.bind("<Escape>", self.cancel)
    # end def

    def do_search_criteria (self, *args, **kw):
        self.switch_buttons(False)
        _cvar = lambda n: self.body.get_stringvar(n).get()
        _criteria = {
            "query": _cvar("search_mention"),
            "matchup": _cvar("search_rdo_query"),
            "name": _cvar("search_chk_name"),
            "origin": _cvar("search_chk_origin"),
            "description": _cvar("search_chk_description"),
            "all": _cvar("search_chk_all"),
            "male": _cvar("search_chk_male"),
            "female": _cvar("search_chk_female"),
        }
        _rows = self.database.get_character_names(
            limit=self.ROW_LIMIT, offset=self.current_offset,
            **_criteria
        )
        self.body.w_dbview.set_header(
            *self.database.get_column_names(),
            body_options={
                "Gender": {"text": {"align": "center"}},
                "Description": {"text": {"width": 400}},
            }
        )
        if _rows:
            self.offset_max = max(self.current_offset, self.offset_max)
            for _row in _rows:
                self.body.w_dbview.insert_row(dict(_row))
            # end for
        elif self.current_offset > self.offset_max:
            self.current_offset = self.offset_max
            self.do_search_criteria()
        # end if
        self.switch_buttons(True)
    # end def

    def finalize (self, *args, **kw):
        self.async.stop(self.do_search_criteria)
    # end def

    def on_import_file (self, *args, **kw):
        from .dlg_namedb_import import NameDBImportDialog
        NameDBImportDialog(self).show()
        self.on_search_criteria_changed(delay=100)
    # end def

    def on_search_criteria_changed (self, *args, **kw):
        self.reset_offset()
        self.async.run_after(
            kw.get("delay") or 1000, self.do_search_criteria
        )
    # end def

    def on_search_filter_all (self, *args, **kw):
        self.body.get_stringvar("search_chk_male").set("")
        self.body.get_stringvar("search_chk_female").set("")
        self.on_search_criteria_changed(delay=500)
    # end def

    def on_search_filter_clicked (self, *args, **kw):
        self.body.get_stringvar("search_chk_all").set("")
        self.on_search_criteria_changed(delay=500)
    # end def

    def on_show_first (self, *args, **kw):
        if self.current_offset:
            self.current_offset = 0
            self.async.run_after(500, self.do_search_criteria)
        # end if
    # end def

    def on_show_next (self, *args, **kw):
        self.current_offset += self.ROW_LIMIT
        self.async.run_after(500, self.do_search_criteria)
    # end def

    def on_show_previous (self, *args, **kw):
        if self.current_offset > 0:
            self.current_offset = max(
                0, self.current_offset - self.ROW_LIMIT
            )
            self.async.run_after(500, self.do_search_criteria)
        # end if
    # end def

    def reset_offset (self, *args, **kw):
        self.current_offset = 0
        self.offset_max = 0
    # end def

    def setup_gui (self, *args, **kw):
        from tkinter import ttk
        from tkRAD import RADXMLFrame
        from tkRAD.core.async import get_async_manager
        from .database import get_database
        self.async = get_async_manager()
        self.database = get_database()
        _frame = ttk.Frame(self, padding=5)
        self.body = RADXMLFrame(_frame)
        self.body.xml_build("dlg_name_database")
        self.body.pack(expand=1, fill="both")
        ttk.Separator(_frame, orient="horizontal").pack(
            expand=1, fill="x", pady=5
        )
        ttk.Button(_frame, text=_("OK"), command=self.cancel).pack()
        _frame.pack()
        self.reset_offset()
        self.bind_events(**kw)
        self.on_search_criteria_changed(delay=200)
    # end def

    def switch_buttons (self, flag=True):
        self.enable_widget(self.body.btn_show_first, flag)
        self.enable_widget(self.body.btn_show_next, flag)
        self.enable_widget(self.body.btn_show_previous, flag)
        self.enable_widget(self.body.btn_import_file, flag)
    # end def

# end class
