#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    tkScenarist - screen writing made simpler

    Copyright (c) 2014+ Raphaël Seban <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see http://www.gnu.org/licenses/
"""

from tkRAD.widgets.rad_dialog import RADDialog


class NameDBImportDialog (RADDialog):

    DEFAULT_DIR = "^/data/csv"

    FIELD_NAMES = ("name", "gender", "origin", "description")

    FILE_TYPES = (
        (_("CSV files"), "*.csv"),
        (_("Text files"), "*.txt"),
        (_("All files"), "*.*"),
    )

    def bind_events (self, **kw):
        self.events.connect_dict(
            {
                "Dialog:Import:File:Browse": self.on_file_browse,
                "Dialog:Import:File:Import": self.on_file_import,
            }
        )
        self.bind("<Escape>", self.cancel)
    # end def

    def cancel (self, *args, **kw):
        if self.running_task:
            self.user_show_warning(
                _("Cannot quit. Import procedure is yet active.")
            )
        else:
            self.async.stop(self.do_import_loop)
            self.destroy()
        # end if
    # end def

    def do_import_file (self, filepath, findices):
        import csv
        import os
        if self.running_task:
            self.user_show_warning(
                _("Import procedure is already running.")
            )
        else:
            self.running_task = True
            self.user_cancelled = False
            self.database.auto_commit = False
            self.reset_progressbar()
            self.show_status(_("importing CSV file, please wait..."))
            self.body.btn_import.configure(
                text=_("Stop"), command=self.on_stop_import
            )
            self.enable_widget(self.body.btn_browse_file, False)
            self.enable_widget(self.btn_ok, False)
            _csvfile = open(filepath, encoding=ENCODING, newline='')
            _csvreader = csv.reader(_csvfile)
            next(_csvreader, None)
            kw = dict(
                findices=findices.copy(),
                csvreader=_csvreader,
                csvfile=_csvfile,
                fsize=os.path.getsize(filepath),
                consumed=0,
            )
            self.async.run_after_idle(self.do_import_loop, kw)
        # end if
    # end def

    def do_import_loop (self, kw):
        _row = next(kw["csvreader"], None)
        if _row and not self.user_cancelled:
            kw["consumed"] += len(",".join(_row))
            self.set_progressbar(100 * kw["consumed"] / kw["fsize"])
            _fields = {
                _fieldname: _row[_index]
                for _fieldname, _index in kw["findices"].items()
            }
            self.show_status(
                _("importing name '{name}'").format(**_fields)
            )
            try:
                self.database.import_character_name(**_fields)
            except:
                from warnings import warn
                warn("Failed to import row: {}".format(_fields))
            # end try
            if kw["consumed"] % 1024 == 0:
                self.database.commit()
            # end if
            self.async.run_after(0, self.do_import_loop, kw)
        else:
            kw["csvfile"].close()
            kw.clear()
            self.database.commit()
            self.database.auto_commit = True
            if self.user_cancelled:
                self.show_status(_("importation stopped by user."))
            else:
                self.show_status(_("importation succeeded. OK."))
                self.set_progressbar(100)
            # end if
            self.enable_widget(self.btn_ok, True)
            self.enable_widget(self.body.btn_browse_file, True)
            self.body.btn_import.configure(
                text=_("Import"), command=self.on_file_import
            )
            self.running_task = False
        # end if
    # end def

    def do_open_file (self, filepath):
        import os
        from tkRAD.core.path import shorten_path
        if self.is_csv(filepath):
            self.options["dirs"]["namedb_import_dir"] = (
                os.path.dirname(filepath)
            )
            self.current_path = filepath
            self.body.get_stringvar("lbl_file_path").set(
                shorten_path(filepath, width=50)
            )
            self.fill_preview(filepath)
            self.fill_fields(filepath)
            self.show_status("")
            self.set_progressbar(0)
            self.body.btn_import.configure(
                text=_("Import"), command=self.on_file_import
            )
        else:
            self.user_show_warning(
                _("Invalid CSV file format. Not supported.")
            )
        # end if
    # end def

    def fill_combos (self, choices=[], matches=None):
        _choices = [_("--- not found ---")]
        _choices.extend(choices)
        matches = matches or (0,) * len(self.FIELD_NAMES)
        for _index, _fieldname in enumerate(self.FIELD_NAMES):
            _combo = self.get_combobox(_fieldname)
            _combo.configure(values=_choices)
            _combo.state(["readonly"])
            _combo.current(matches[_index])
        # end for
    # end def

    def fill_fields (self, filepath):
        import csv
        with open(filepath, encoding=ENCODING, newline='') as _csvfile:
            _row = next(csv.reader(_csvfile), [])
            _choices = [
                _("Column#{}").format(i + 1) for i in range(len(_row))
            ]
            _row = tuple(map(lambda s: str(s).lower(), _row))
            _matches = []
            for _field in self.FIELD_NAMES:
                try:
                    _index = _row.index(_field.lower()) + 1
                except:
                    _index = 0
                # end try
                _matches.append(_index)
            # end for
        # end with
        self.fill_combos(_choices, _matches)
    # end def

    def fill_preview (self, filepath):
        _line = 1
        self.enable_widget(self.body.w_text, True)
        self.body.w_text.delete(1.0, "end")
        with open(filepath, encoding=ENCODING) as _txtfile:
            for _row in _txtfile:
                self.body.w_text.insert("end", _row)
                _line += 1
                if _line > 10: break
            # end for
        # end with
        self.body.w_text.insert("end", "...")
        self.enable_widget(self.body.w_text, False)
    # end def

    def get_combobox (self, field_name):
        return getattr(self.body, "column_{}".format(field_name))
    # end def

    def get_field_indices (self):
        _indices = dict()
        for _fieldname in self.FIELD_NAMES:
            _index = self.get_combobox(_fieldname).current()
            if _index:
                _indices[_fieldname] = _index - 1
            # end if
        # end for
        return _indices
    # end def

    def is_csv (self, filepath):
        import csv
        with open(filepath, encoding=ENCODING, newline='') as _csvfile:
            try:
                csv.Sniffer().sniff(_csvfile.read(1024))
                return True
            except:
                return False
            # end try
        # end with
    # end def

    def on_file_browse (self, *args, **kw):
        from tkinter import filedialog
        _path = filedialog.askopenfilename(
            title=_("Please, select a file to import"),
            initialdir=self.options.get(
                "dirs", "namedb_import_dir", fallback=self.DEFAULT_DIR
            ),
            filetypes=self.FILE_TYPES,
            parent=self,
        )
        if _path:
            self.do_open_file(_path)
        # end if
    # end def

    def on_file_import (self, *args, **kw):
        if not self.current_path:
            self.user_show_warning(
                _(
                    "Please, select a comma-separated "
                    "values (CSV) file to import."
                )
            )
        else:
            _indices = self.get_field_indices()
            if not _indices:
                self.user_show_warning(_("Nothing to import."))
            elif "name" not in _indices:
                self.user_show_warning(
                    _(
                        "Cannot import names without "
                        "a valid 'name' field."
                    )
                )
            else:
                self.do_import_file(self.current_path, _indices)
            # end if
        # end if
    # end def

    def on_stop_import (self, *args, **kw):
        self.user_cancelled = True
    # end def

    def reset_progressbar (self, *args, **kw):
        self.set_progressbar(0)
        self.body.w_progressbar.configure(mode="indeterminate")
        self.body.w_progressbar.start()
    # end def

    def set_progressbar (self, value):
        self.body.w_progressbar.configure(mode="determinate")
        self.body.w_progressbar.stop()
        self.body.get_stringvar("pgbar_value").set(str(int(value)))
    # end def

    def setup_gui (self, *args, **kw):
        from tkinter import ttk
        from tkRAD import RADXMLFrame
        from tkRAD.core.async import get_async_manager
        from tkRAD.core.path import normalize
        from .database import get_database
        self.async = get_async_manager()
        self.database = get_database()
        self.DEFAULT_DIR = normalize(self.DEFAULT_DIR)
        self.current_path = ""
        self.user_cancelled = False
        self.running_task = False
        _frame = ttk.Frame(self, padding=5)
        self.body = RADXMLFrame(_frame)
        self.body.xml_build("dlg_namedb_import")
        self.body.pack(expand=1, fill="both")
        ttk.Separator(_frame, orient="horizontal").pack(
            expand=1, fill="x", pady=5
        )
        self.btn_ok = ttk.Button(
            _frame, text=_("OK"), command=self.cancel
        )
        self.btn_ok.pack()
        _frame.pack()
        self.fill_combos()
        self.bind_events(**kw)
    # end def

    def show_status (self, message):
        self.body.get_stringvar("lbl_status").set(str(message))
    # end def

    def user_show_warning (self, msg):
        from tkinter import messagebox
        messagebox.showwarning(
            title=_("Attention"), message=msg, parent=self
        )
    # end def

# end class
