#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    tkScenarist - screen writing made simpler

    Copyright (c) 2014+ Raphaël Seban <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see http://www.gnu.org/licenses/
"""

import os
from tkRAD.core import path
from tkRAD.widgets.rad_dialog import RADDialog
from . import text_helpers as th


class PitchTemplatesDialog (RADDialog):

    DEFAULT_DIR = "^/data/templates/pitch/"

    FILE_EXT = ".txt"

    FILE_PATTERN = "*" + FILE_EXT

    FILE_TYPES = [
        (_("Text files"), FILE_PATTERN),
        (_("All files"), "*.*"),
    ]

    def auto_save (self, *args, **kw):
        if self.current_path and self.body.w_text.edit_modified():
            try:
                with open(self.current_path, "w", encoding=ENCODING) \
                                                               as _file:
                    _file.write(self.get_preview_text())
                # end with
            except Exception as e:
                self.user_show_error(
                    _(
                        "An error has occurred:\n"
                        "[{ecn}] {err}\n"
                        "Could *NOT* save text contents."
                    ).format(ecn=e.__class__.__name__, err=e)
                )
            finally:
                self.body.w_text.edit_modified(False)
            # end try
        # end if
    # end def

    def bind_events (self, **kw):
        self.events.connect_dict(
            {
                "Dialog:Template:Browse": self.on_browse,
                "Dialog:Template:Delete": self.on_delete,
                "Dialog:Template:New": self.on_new,
                "Dialog:Template:ResetDir": self.on_reset_dir,
            }
        )
        self.bind("<Escape>", self.cancel)
        self.body.w_listbox.bind(
            "<<ListboxSelect>>", self.on_listbox_select
        )
        self.body.w_text.bind("<Control-z>", self.on_edit_undo)
        self.body.w_text.bind("<Control-Z>", self.on_edit_redo)
        self.body.w_text.bind("<Control-a>", self.on_edit_select_all)
        self.body.w_text.bind("<KeyRelease>", self.on_text_keypress)
    # end def

    def clean_name (self, filepath):
        return os.path.splitext(os.path.basename(filepath))[0]
    # end def

    def clear_preview (self, lock=False):
        _flag = self.widget_enabled(self.body.w_text) and not lock
        self.enable_widget(self.body.w_text, True)
        th.text_reset_contents(self.body.w_text)
        self.enable_widget(self.body.w_text, _flag)
        self.set_template_name("")
    # end def

    @property
    def current_dir (self):
        return self.__current_dir
    # end def

    @current_dir.setter
    def current_dir (self, value):
        value = path.normalize(value)
        if os.path.isdir(value):
            self.__current_dir = value
            self.options["dirs"]["pitch_template_dir"] = value
            self.cvar_current_dir.set(
                path.shorten_path(value, width=70)
            )
        else:
            raise NotADirectoryError
        # end if
    # end def

    @property
    def current_path (self):
        return self.__current_path
    # end def

    @current_path.setter
    def current_path (self, value):
        value = path.normalize(value)
        if not os.path.isfile(value):
            value = ""
        # end if
        self.__current_path = value
    # end def

    def finalize (self, *args, **kw):
        self.save_now()
    # end def

    def get_chk_option (self, cvarname):
        return bool(self.body.get_stringvar(cvarname).get() == "1")
    # end def

    def get_filepath_from_listbox (self):
        try:
            _index = self.body.w_listbox.curselection()[0]
            _filepath = os.path.join(
                self.current_dir,
                self.body.w_listbox.get(_index) + self.FILE_EXT
            )
            if os.path.isfile(_filepath):
                return _filepath
            # end if
        except:
            pass
        # end try
        return ""
    # end def

    def get_preview_text (self):
        return th.text_get_contents(self.body.w_text)
    # end def

    def get_template_name (self):
        return self.cvar_template_name.get()
    # end def

    def on_browse (self, *args, **kw):
        from tkinter import filedialog
        _dirpath = filedialog.askdirectory(
            title=_("Please, select a directory"),
            initialdir=self.current_dir,
            mustexist=True,
            parent=self,
        )
        if _dirpath:
            self.update_current_dir(_dirpath)
        # end if
    # end def

    def on_delete (self, *args, **kw):
        _filepath = self.get_filepath_from_listbox()
        if _filepath and self.user_confirm():
            self.update_preview("")
            os.remove(_filepath)
            self.update_listbox()
        # end if
    # end def

    def on_edit_redo (self, *args, **kw):
        self.body.w_text.event_generate("<<Redo>>")
        return "break"
    # end def

    def on_edit_select_all (self, *args, **kw):
        self.body.w_text.event_generate("<<SelectAll>>")
        return "break"
    # end def

    def on_edit_undo (self, *args, **kw):
        self.body.w_text.event_generate("<<Undo>>")
        return "break"
    # end def

    def on_listbox_select (self, *args, **kw):
        _filepath = self.get_filepath_from_listbox()
        if _filepath != self.current_path:
            self.update_preview(_filepath)
        # end if
    # end def

    def on_new (self, *args, **kw):
        from tkinter import filedialog
        _filepath = filedialog.asksaveasfilename(
            title=_("New template"),
            defaultextension=self.FILE_EXT,
            filetypes=self.FILE_TYPES,
            initialdir=self.current_dir,
            confirmoverwrite=True,
            parent=self,
        )
        if _filepath:
            open(_filepath, "w").close()
            self.update_current_dir(os.path.dirname(_filepath))
            self.update_preview(_filepath)
            self.body.w_text.focus_set()
        # end if
    # end def

    def on_reset_dir (self, *args, **kw):
        if self.current_dir != self.DEFAULT_DIR:
            self.update_current_dir(self.DEFAULT_DIR)
        # end if
    # end def

    def on_text_keypress (self, event=None, *args, **kw):
        th.text_edit_autoseparator(self.body.w_text, event)
        self.async.run_after(3000, self.auto_save)
    # end def

    def save_now (self):
        self.async.stop(self.auto_save)
        self.auto_save()
    # end def

    def set_template_name (self, value):
        self.cvar_template_name.set(self.clean_name(value))
    # end def

    def setup_gui (self, *args, **kw):
        from tkinter import ttk
        from tkRAD import RADXMLFrame
        from tkRAD.core.async import get_async_manager
        self.async = get_async_manager()
        self.DEFAULT_DIR = path.normalize(_(self.DEFAULT_DIR))
        self.current_path = ""
        _frame = ttk.Frame(self, padding=5)
        self.body = RADXMLFrame(_frame)
        self.body.xml_build("dlg_pitch_templates")
        self.body.pack(expand=1, fill="both")
        ttk.Separator(_frame, orient="horizontal").pack(
            expand=1, fill="x", pady=5
        )
        _frame2 = ttk.Frame(_frame)
        self.btn_ok = ttk.Button(
            _frame2, text=_("OK"), command=self.validate
        )
        self.btn_ok.pack(side="left", padx=5)
        self.btn_cancel = ttk.Button(
            _frame2, text=_("Cancel"), command=self.cancel
        )
        self.btn_cancel.pack(side="right", padx=5)
        _frame2.pack()
        _frame.pack()
        _cvar = self.body.get_stringvar
        self.cvar_template_name = _cvar("template_name")
        self.cvar_current_dir = _cvar("current_dir")
        self.update_current_dir(
            self.options.get(
                "dirs", "pitch_templates_dir",
                fallback=self.DEFAULT_DIR
            )
        )
        self.bind_events(**kw)
    # end def

    def update_current_dir (self, new_dir):
        self.update_preview("")
        try:
            self.current_dir = new_dir
        except:
            self.current_dir = "~"
        # end try
        self.update_listbox(new_dir)
    # end def

    def update_listbox (self, *args, new_dir=None, **kw):
        from glob import glob
        new_dir = path.normalize(new_dir or self.current_dir)
        if os.path.isdir(new_dir):
            self.body.w_listbox.delete(0, "end")
            _flist = sorted(
                map(
                    self.clean_name,
                    glob(os.path.join(new_dir, self.FILE_PATTERN))
                )
            )
            self.body.w_listbox.insert("end", *_flist)
            if not _flist:
                self.clear_preview(lock=True)
            # end if
            self.enable_widget(self.body.btn_delete, False)
        # end if
    # end def

    def update_preview (self, filepath):
        self.save_now()
        filepath = path.normalize(filepath)
        self.current_path = ""
        self.clear_preview(lock=True)
        self.enable_widget(self.body.btn_delete, False)
        if os.path.isfile(filepath):
            self.current_path = filepath
            self.set_template_name(filepath)
            self.enable_widget(self.body.w_text, True)
            try:
                with open(filepath, encoding=ENCODING) as _file:
                    th.text_reset_contents(
                        self.body.w_text, _file.read()
                    )
                # end with
            except Exception as e:
                self.user_show_error(
                    _(
                        "An error has occurred:\n"
                        "[{ecn}] {err}\n"
                        "Could *NOT* load text contents."
                    ).format(ecn=e.__class__.__name__, err=e)
                )
            # end try
            self.enable_widget(self.body.btn_delete, True)
        # end if
    # end def

    def user_confirm (self):
        from tkinter import messagebox
        return messagebox.askyesno(
            title=_("Question"),
            message=_("Are you sure?"),
            parent=self
        )
    # end def

    def user_show_error (self, msg):
        from tkinter import messagebox
        messagebox.showerror(
            title=_("Error"), message=str(msg), parent=self
        )
    # end def

    def validate (self, *args, **kw):
        if self.current_path:
            self.save_now()
            _pitch = _notes = ""
            _text = self.get_preview_text()
            _chk_pitch = self.get_chk_option("chk_copy_to_pitch")
            _chk_notes = self.get_chk_option("chk_copy_to_notes")
            if _chk_pitch:
                _pitch = _text
            # end if
            if _chk_notes:
                _notes = _text
            # end if
            if _pitch or _notes:
                self.events.raise_event(
                    "Pitch:Template:Insert", pitch=_pitch, notes=_notes
                )
            # end if
        # end if
        self.destroy()
    # end def

# end class
