#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    tkScenarist - screen writing made simpler

    Copyright (c) 2014+ Raphaël Seban <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see http://www.gnu.org/licenses/
"""

from tkRAD.widgets.rad_dialog import RADDialog


class ScenarioElementsEditorDialog (RADDialog):

    def bind_events (self, **kw):
        self.events.connect_dict(
            {
                "Dialog:Color:Choose": self.on_choose_color,
                "Dialog:Text:Align": self.on_store_looknfeel,
            }
        )
        self.bind("<Escape>", self.cancel)
        self.body.w_notebook.bind(
            "<<NotebookTabChanged>>", self.on_notebook_tab_changed
        )
        self.body.combo_current_element.bind(
            "<<ComboboxSelected>>", self.on_current_element_changed
        )
        self.body.text_preview.bind(
            "<ButtonRelease-1>", self.on_preview_clicked
        )
    # end def

    def get_current_element (self):
        return (
            self.current_settings["element"]
                                    [self.get_current_element_tag()]
        )
    # end def

    def get_current_element_tag (self):
        return (
            self.element_names[self.body.combo_current_element.get()]
        )
    # end def

    def get_current_tab_index (self):
        return self.body.w_notebook.index("current")
    # end def

    def get_font_attrs (self, font_str):
        from tkinter.font import Font
        _font = Font(font=font_str or "courier 12")
        _family = _font.cget("family")
        _size = _font.cget("size")
        _weight = _font.cget("weight").replace("normal", "")
        _slant = _font.cget("slant").replace("roman", "")
        _style = "{} {}".format(_weight, _slant).strip()
        _style = _style or "normal"
        return (_family, _size, _style)
    # end def

    def get_formatted_margin (self, cvar_entry, combobox):
        _value = int(self.body.get_stringvar(cvar_entry).get() or 0)
        _units = combobox.get()[0] if combobox.current() else ""
        return "{}{}".format(_value, _units).strip().lower()
    # end def

    def get_label (self, element_tag):
        _element = (
            self.current_settings["element"].get(element_tag) or dict()
        )
        return _element.get("label") or ""
    # end def

    def key_filter_digits (self, input_char):
        self.async.run_after(500, self.on_store_looknfeel)
        return bool(input_char in "0123456789")
    # end def

    def on_choose_color (self, *args, widget=None, **kw):
        if widget:
            from tkinter import colorchooser
            _bg = widget.cget("background")
            _color = colorchooser.askcolor(
                color=_bg,
                title=_("Please, choose a color"),
                parent=self,
            )[-1] or _bg
            widget.configure(background=_color)
            self.set_color_label(widget)
            self.on_store_looknfeel()
        # end if
    # end def

    def on_current_element_changed (self, *args, **kw):
        self.current_settings["current_selected"] = (
            self.body.combo_current_element.current()
        )
        self.on_update_linked_items()
    # end def

    def on_notebook_tab_changed (self, *args, **kw):
        _index = self.get_current_tab_index()
        self.current_settings = self.settings[_index]
        self.on_update_current_selected()
    # end def

    def on_preview_clicked (self, *args, **kw):
        _tags = self.body.text_preview.tag_names("current")
        if _tags:
            _label = self.get_label(_tags[0])
            if _label in self.element_names:
                self.body.combo_current_element.set(_label)
                self.on_current_element_changed()
            # end if
        # end if
    # end def

    def on_reset_to_defaults (self, *args, **kw):
        from copy import deepcopy
        _index = self.get_current_tab_index()
        if _index == 0:
            self.settings[_index] = {
                "element": deepcopy(self.w_text.ELEMENT_DEFAULTS),
                "current_selected": 0,
            }
        elif _index == 1:
            self.settings[_index] = {
                "element": deepcopy(self.settings[0]["element"]),
                "current_selected": 0,
            }
        # end if
        self.on_notebook_tab_changed()
    # end def

    def on_store_chainings (self, *args, **kw):
        _element = self.get_current_element()
        for _widget, _name in self.chainings:
            _element["on_{}".format(_name)] = (
                self.element_names.get(_widget.get()) or ""
            )
        # end for
    # end def

    def on_store_looknfeel (self, *args, **kw):
        _element = self.get_current_element()
        _cfg = _element.setdefault("config", dict())
        _cfg["background"] = self.body.btn_choose_bg.cget("background")
        _cfg["foreground"] = self.body.btn_choose_fg.cget("background")
        _cfg["font"] = (
            "{{{family}}} {size} {style}"
            .format(
                family=self.body.combo_font_family.get(),
                size=self.body.combo_font_size.get(),
                style=self.body.combo_font_style.get(),
            ).strip().lower()
        )
        _cfg["justify"] = (
            self.body.get_stringvar("options_text_align").get()
        )
        self.reset_margin(
            "entry_margin_left", self.body.combo_lmargin_units
        )
        _cfg["lmargin1"] = _cfg["lmargin2"] = (
            self.get_formatted_margin(
                "entry_margin_left", self.body.combo_lmargin_units
            )
        )
        self.reset_margin(
            "entry_margin_right", self.body.combo_rmargin_units
        )
        _cfg["rmargin"] = (
            self.get_formatted_margin(
                "entry_margin_right", self.body.combo_rmargin_units
            )
        )
        self.on_update_preview()
    # end def

    def on_update_current_selected (self, *args, **kw):
        self.body.combo_current_element.current(
            self.current_settings["current_selected"]
        )
        self.on_update_linked_items()
    # end def

    def on_update_linked_items (self, *args, **kw):
        _element = self.get_current_element()
        for _widget, _name in self.chainings:
            _widget.set(
                self.get_label(_element.get("on_{}".format(_name)))
            )
        # end for
        _config = _element.setdefault("config", dict())
        _face, _size, _style = self.get_font_attrs(_config.get("font"))
        self.body.combo_font_family.set(_face)
        self.body.combo_font_size.set(_size)
        self.body.combo_font_style.set(_style)
        _bg = _config.get("background") or "white"
        _fg = _config.get("foreground") or "black"
        self.body.btn_choose_bg.configure(background=_bg)
        self.body.btn_choose_fg.configure(background=_fg)
        self.set_color_label(
            self.body.btn_choose_bg, self.body.btn_choose_fg
        )
        self.body.get_stringvar("options_text_align").set(
            _config.get("justify") or "left"
        )
        self.reset_margin(
            "entry_margin_left",
            self.body.combo_lmargin_units,
            _config.get("lmargin1") or _config.get("lmargin2") or 0
        )
        self.reset_margin(
            "entry_margin_right",
            self.body.combo_rmargin_units,
            _config.get("rmargin") or 0
        )
        self.on_update_preview()
    # end def

    def on_update_preview (self, *args, **kw):
        for _tag, _element in self.current_settings["element"].items():
            _config = _element.get("config")
            if _config:
                self.body.text_preview.tag_delete(_tag)
                self.body.text_preview.tag_configure(_tag, **_config)
            # end if
        # end for
        self.body.text_preview.tag_raise("sel")
        self.set_preview_contents()
        self.body.text_preview.see(
            self.body.text_preview.tag_ranges(
                self.get_current_element_tag()
            )[0]
        )
    # end def

    def reformat_line (self, contents, tag):
        _text, _adjust = self.w_text.switch_to_method(
            "reformat_line_{}".format(tag), contents
        )
        return (_text + "\n", tag)
    # end def

    def reset_margin (self, cvar_entry, combobox, value=None):
        import re
        if value is None:
            value = self.get_formatted_margin(cvar_entry, combobox)
        # end if
        value = str(value or 0).strip()
        _units = "px"
        _match = re.search(r"(\d+)\s*(\w*)", value)
        if _match:
            value = _match.group(1)
            _unit = _match.group(2)
            if _unit:
                _units = {
                    "c": "cm", "i": "in", "m": "mm", "p": "pt",
                }.get(_unit[0]) or "px"
            # end if
        # end if
        self.body.get_stringvar(cvar_entry).set(value)
        combobox.set(_units)
    # end def

    def set_color_label (self, *widgets):
        for _w in widgets:
            _color = "#{:02X}{:02X}{:02X}".format(
                *map(
                    lambda x: x//256,
                    _w.winfo_rgb(_w.cget("background"))
                )
            )
            _w.configure(text=_color)
        # end for
    # end def

    def set_preview_contents (self, *args, **kw):
        self.enable_widget(self.body.text_preview, True)
        self.body.text_preview.delete(1.0, "end")
        _action = self.reformat_line(
            _("This is an 'ACTION' line."), "action"
        )
        _character = self.reformat_line(
            _("'CHARACTER' name"), "character"
        )
        _dialogue = self.reformat_line(
            _("This is a 'DIALOGUE' line."), "dialogue"
        )
        _parenthetical = self.reformat_line(
            _("this is a 'PARENTHETICAL' line"), "parenthetical"
        )
        _scene = self.reformat_line(
            _("ext - day - this is a 'SCENE' line"), "scene"
        )
        _transition = self.reformat_line(
            _("'TRANSITION' line."), "transition"
        )
        self.body.text_preview.insert(
            "end",
            *(
                _scene + _action + _character + _parenthetical +
                _dialogue + _character + _dialogue + _action +
                _transition + _scene + _transition + _action
            )
        )
        self.enable_widget(self.body.text_preview, False)
    # end def

    def setup_gui (self, *args, **kw):
        from copy import deepcopy
        from tkinter import font, ttk
        from tkRAD import RADXMLFrame
        from tkRAD.core.async import get_async_manager
        self.async = get_async_manager()
        self.w_text = kw.get("w_text")
        self.settings = [
            {
                "element": deepcopy(self.w_text.get_options_element()),
                "current_selected": 0,
            },
            {
                "element": deepcopy(self.w_text.ELEMENT),
                "current_selected": 0,
            },
        ]
        self.current_settings = self.settings[0]
        self.element_names = {
            _value["label"]: _key
            for _key, _value in
                self.current_settings["element"].items()
        }
        _frame = ttk.Frame(self, padding=5)
        self.body = RADXMLFrame(_frame)
        self.body.xml_build("dlg_scenario_elements_editor")
        self.body.pack(expand=1, fill="both")
        ttk.Separator(_frame, orient="horizontal").pack(
            expand=1, fill="x", pady=5
        )
        _frame2 = ttk.Frame(_frame)
        self.btn_ok = ttk.Button(
            _frame2, text=_("OK"), command=self.validate
        )
        self.btn_ok.pack(side="left", padx=5)
        self.btn_reset = ttk.Button(
            _frame2,
            text=_("Reset to defaults"),
            command=self.on_reset_to_defaults
        )
        self.btn_reset.pack(side="left", padx=5)
        self.btn_cancel = ttk.Button(
            _frame2, text=_("Cancel"), command=self.cancel
        )
        self.btn_cancel.pack(side="left", padx=5)
        _frame2.pack()
        _frame.pack()
        _names = sorted(self.element_names)
        _readonly = ["readonly"]
        self.body.combo_current_element.configure(
            values=_names, state=_readonly
        )
        self.body.combo_current_element.current(0)
        _names.insert(0, "")
        _chaining_names = tuple(
            "{}_{}".format(_key, _t)
            for _key in ("tab", "return", "ctrl_return")
            for _t in ("switch", "create")
        )
        _chaining_combos = tuple(
            getattr(self.body, "combo_{}".format(_name))
            for _name in _chaining_names
        )
        self.chainings = tuple(zip(_chaining_combos, _chaining_names))
        for _w in _chaining_combos:
            _w.configure(values=_names, state=_readonly)
            _w.current(0)
            _w.bind("<<ComboboxSelected>>", self.on_store_chainings)
        # end for
        self.body.combo_font_family.configure(
            values=['courier', 'helvetica', 'times', 'tkdefaultfont'] +
            sorted(font.families())
        )
        _filter = self.register(self.key_filter_digits)
        self.body.entry_lmargin.configure(
            validate="key", validatecommand=(_filter, "%S")
        )
        self.body.entry_rmargin.configure(
            validate="key", validatecommand=(_filter, "%S")
        )
        _combos = (
            self.body.combo_font_family,
            self.body.combo_font_size,
            self.body.combo_font_style,
            self.body.combo_lmargin_units,
            self.body.combo_rmargin_units,
        )
        for _w in _combos:
            _w.state(_readonly)
            _w.current(0)
            _w.bind("<<ComboboxSelected>>", self.on_store_looknfeel)
        # end if
        self.body.text_preview.configure(
            font=self.w_text.cget("font"),
            background=self.w_text.cget("background"),
            foreground=self.w_text.cget("foreground"),
            height=10, width=1, wrap="word",
        )
        self.bind_events(**kw)
        self.body.w_notebook.select(
            self.options.get(
                self.get_rc_section(), "startup_tab_index", fallback=1
            )
        )
    # end def

    def user_confirm (self):
        from tkinter import messagebox
        return messagebox.askyesno(
            title=_("Question"),
            message=_("Are you sure?"),
            parent=self
        )
    # end def

    def validate (self, *args, **kw):
        _global, _project = self.settings
        self.events.raise_event(
            "Scenario:Settings:Update",
            settings={
                "global": _global["element"],
                "project": _project["element"],
            }
        )
        self.destroy()
    # end def

# end class
