#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    tkScenarist - screen writing made simpler

    Copyright (c) 2014+ Raphaël Seban <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see http://www.gnu.org/licenses/
"""

import tkRAD

__builtins__["ENCODING"] = "UTF-8"


class MainWindow (tkRAD.RADXMLWindow):

    MAILING_LIST_URL = _(
        "https://mail.python.org/mailman/listinfo/tkscenarist-help"
    )
    OFFLINE_DOC_URL = _("^/html/en/index.html")
    ONLINE_DOC_URL = _("https://github.com/tarball69/tkScenarist/wiki")

    def bind_events (self, **kw):
        self.events.connect_dict(
            {
                "Edit:Redo": self.on_edit_redo,
                "Edit:Select:All": self.on_edit_select_all,
                "Edit:Undo": self.on_edit_undo,

                "Help:About": self.on_help_about,
                "Help:Mailing:List": self.on_help_mailing_list,
                "Help:Offline:Documentation":
                    self.on_help_offline_documentation,
                "Help:Online:Documentation":
                    self.on_help_online_documentation,

                "Project:Export:PDF": self.project_fm.on_export_pdf,
                "Project:New": self.project_fm.on_new,
                "Project:Open": self.project_fm.on_open,
                "Project:Save:As": self.project_fm.on_save_as,
                "Project:Save": self.project_fm.on_save,

                "Tools:NameDatabase": self.on_tools_name_db,
                "Tools:Pitch:Templates": self.on_tools_pitch_templates,
                "Tools:Scenario:Elements:Editor":
                    self.on_tools_scenario_elements_editor,
            }
        )
        _handler = lambda e: None
        for _class in ("TEntry", "Text"):
            for _key in "aknopqsyzAKNOPQSYZ":
                self.bind_class(
                    _class, "<Control-{}>".format(_key), _handler
                )
            # end for
        # end for
        self.bind_class(
            "TCombobox", "<<ComboboxSelected>>",
            lambda e: e.widget.selection_clear()
        )
    # end def

    def center_on_screen (self):
        self.update_idletasks()
        _w, _h = self.winfo_reqwidth(), self.winfo_reqheight()
        self.geometry(
            "{w}x{h}+{x}+{y}".format(
                w=_w, h=_h,
                x=(self.winfo_screenwidth() - _w)//2,
                y=(self.winfo_screenheight() - _h)//2,
            )
        )
    # end def

    def confirm_quit (self, *args, **kw):
        return bool(
            self.project_fm.ensure_saved() != self.project_fm.CANCEL
        )
    # end def

    def edit_raise_event (self, event_name):
        try:
            self.focus_lastfor().event_generate(
                "<<{}>>".format(event_name)
            )
        except:
            pass
        # end try
    # end def

    def finalize (self, *args, **kw):
        try:
            self.async.clear_all()
            self.database.finalize()
            self.options.save()
        except:
            pass
        # end try
    # end def

    def get_component (self, component_name):
        return self.mainframe.get_object_by_id(
            "component_{}".format(str(component_name).lower())
        )
    # end def

    def initialize (self, *args, **kw):
        self.window_state = "hidden"
        super().initialize(*args, **kw)
        from tkRAD.core.async import get_async_manager
        from .database import get_database
        from .project_file_management import ProjectFileManagement
        self.async = get_async_manager()
        self.database = get_database()
        self.project_fm = ProjectFileManagement(self)
        self.services.register_service("PFM", self.project_fm)
        self.topmenu.xml_build("topmenu")
        self.connect_statusbar("show_statusbar")
        self.mainframe.xml_build("mainwindow")
        self.bind_events()
        self.update_caption()
    # end def

    def launch_web_browser (self, url):
        if url:
            import webbrowser
            from tkinter import messagebox
            webbrowser.open(url)
            messagebox.showinfo(
                title=_("Attention"),
                message=_("Launching web browser, please wait."),
                parent=self,
            )
        # end if
    # end def

    def on_edit_redo (self, event=None, *args, **kw):
        self.edit_raise_event("Redo")
    # end def

    def on_edit_select_all (self, event=None, *args, **kw):
        self.edit_raise_event("SelectAll")
    # end def

    def on_edit_undo (self, event=None, *args, **kw):
        self.edit_raise_event("Undo")
    # end def

    def on_help_about (self, *args, **kw):
        from tkinter import messagebox
        messagebox.showinfo(
            title=_("About..."),
            message=_(
                "{name} v{version}\n"
                "{description}\n"
                "{copyright}\n"
                "\nCurrent maintainer: {maintainer}\n"
                "PDF toolkit: {pdflib}\n"
                "{pdflib_author}"
            ).format(**self.app.APP),
            parent=self,
        )
    # end def

    def on_help_mailing_list (self, *args, **kw):
        self.launch_web_browser(self.MAILING_LIST_URL)
    # end def

    def on_help_offline_documentation (self, *args, **kw):
        from tkRAD.core.path import normalize
        self.launch_web_browser(
            "file://{}".format(normalize(self.OFFLINE_DOC_URL))
        )
    # end def

    def on_help_online_documentation (self, *args, **kw):
        self.launch_web_browser(self.ONLINE_DOC_URL)
    # end def

    def on_tools_name_db (self, *args, **kw):
        from .dlg_name_database import NameDatabaseDialog
        NameDatabaseDialog(self).show()
    # end def

    def on_tools_pitch_templates (self, *args, **kw):
        from .dlg_pitch_templates import PitchTemplatesDialog
        PitchTemplatesDialog(self).show()
    # end def

    def on_tools_scenario_elements_editor (self, *args, **kw):
        from . import dlg_scenario_elements_editor as DSEE
        DSEE.ScenarioElementsEditorDialog(
            self, w_text=self.get_component("scenario").w_text
        ).show()
    # end def

    def show (self, *args, **kw):
        self.center_on_screen()
        super().show(*args, **kw)
    # end def

    def update_caption (self, *args, **kw):
        import os
        _title = "{title}".format(**self.app.APP)
        _directory, _filename = os.path.split(
            self.project_fm.project_path
        )
        if _directory and _filename:
            self.options["dirs"]["pfm_preferred_dir"] = _directory
            _title = "{} - {}".format(_title, _filename)
        # end if
        if self.project_fm.project_modified:
            _title += "*"
        # end if
        self.title(_title)
        self.async.run_after(500, self.update_caption)
    # end def

# end class
