#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    tkScenarist - screen writing made simpler

    Copyright (c) 2014+ Raphaël Seban <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see http://www.gnu.org/licenses/
"""

from tkRAD.core import tools
from tkRAD.core import path


class ProjectFileManagement:

    ARCHIVE_FILES = {
        "title_data": "title_data.txt",
        "draft_notes": {
            "text": "draft_notes.txt",
            "data": "draft_notes_data.txt",
        },
        "pitch_concept": {
            "text": "pitch_concept.txt",
            "data": "pitch_concept_data.txt",
        },
        "characters": {
            "names": "characters_names.txt",
            "logs": "characters_logs.txt",
            "relationships": "characters_relationships.txt",
        },
        "scenario": {
            "text": "scenario_text.txt",
            "data": "scenario_data.txt",
            "elements": "scenario_elements.txt",
        },
        "storyboard": "storyboard.txt",
        "resources": "resources.txt",
    }

    CANCEL, NO, YES = (None, False, True)

    FILE_EXT = "scn"

    def __init__ (self, tk_owner):
        self.mainwindow = tk_owner
        self.notify = tk_owner.statusbar.notify
        self.current_dir = tk_owner.options.get(
            "dirs", "pfm_preferred_dir", fallback="~"
        )
        self.ALL_FILES = self.get_files_from_dict(self.ARCHIVE_FILES)
        self.FILE_EXT = self.normalize_file_ext(self.FILE_EXT)
        self.FILE_TYPES = [
            (_("tkScenarist files"), "*{}".format(self.FILE_EXT)),
            (_("zip files"), "*.zip"),
        ]
        self.reset_members()
    # end def

    def check_integrity (self, filepath):
        filepath = path.normalize(filepath)
        try:
            import zipfile
            with zipfile.ZipFile(filepath, "r") as _archive:
                _zipfiles = _archive.namelist()
            # end with
            return bool(set(self.ALL_FILES) == set(_zipfiles))
        except:
            return False
        # end if
    # end def

    @property
    def current_dir (self):
        return self.__current_dir
    # end def

    @current_dir.setter
    def current_dir (self, value):
        self.__current_dir = path.normalize(value)
    # end def

    @current_dir.deleter
    def current_dir (self):
        del self.__current_dir
    # end def

    def do_open_project (self, filepath):
        if self.check_integrity(filepath):
            self.notify(_("Opening project, please wait."))
            self.do_reset_project()
            import zipfile
            with zipfile.ZipFile(filepath, 'r') as _archive:
                for _id, _files in self.ARCHIVE_FILES.items():
                    if tools.is_pstr(_files):
                        _files = str(
                            _archive.read(_files).decode(ENCODING)
                        ).strip()
                    # end if
                    try:
                        getattr(
                            self.mainwindow.get_component(_id),
                            "setup_component"
                        )(_files, _archive)
                    except:
                        pass
                    # end try
                # end for
            # end with
            self.project_path = filepath
            self.notify(_("Project opened OK."))
            self.reset_modified()
            return self.YES
        else:
            self.user_show_error(
                _(
                    "Could not open project. "
                    "Incorrect file path or file format."
                )
            )
            return self.NO
        # end if
    # end def

    def do_reset_project (self):
        self.reset_members()
        self.mainwindow.events.raise_event("Component:Reset")
        self.reset_modified()
    # end def

    def do_save_project (self, filepath):
        filepath = path.normalize(filepath)
        if tools.is_pstr(filepath):
            self.notify(_("Saving project, please wait."))
            import zipfile
            with zipfile.ZipFile(filepath, 'w') as _archive:
                for _id, _files in self.ARCHIVE_FILES.items():
                    _contents = self.get_file_contents(_id, _files)
                    for _filename, _fcontents in _contents.items():
                        _archive.writestr(
                            _filename, bytes(_fcontents, ENCODING)
                        )
                    # end for
                # end for
            # end with
            self.project_path = filepath
            self.notify(_("Project saved OK."))
            self.reset_modified()
            return self.YES
        else:
            self.user_show_error(
                _("Could not save project. Incorrect file path.")
            )
            return self.NO
        # end if
    # end def

    def ensure_saved (self):
        _response = self.YES
        if self.project_modified:
            _response = self.user_ask_yesnocancel(
                _("Project has been modified. Save it?")
            )
            if _response is self.YES:
                _response = self.on_save()
            # end if
        # end if
        return _response and self.check_integrity(self.project_path)
    # end def

    def get_file_contents (self, component_name, files):
        try:
            return getattr(
                self.mainwindow.get_component(component_name),
                "get_file_contents"
            )(files)
        except:
            return dict()
        # end try
    # end def

    def get_files_from_dict (self, dict_object):
        _flist = list()
        for _object in dict_object.values():
            if tools.is_pdict(_object):
                _flist.extend(self.get_files_from_dict(_object))
            elif tools.is_pstr(_object):
                _flist.append(_object)
            # end if
        # end for
        return _flist
    # end def

    def normalize_file_ext (self, file_ext):
        return (
            ".{}".format(tools.normalize_id(file_ext) or "zip").lower()
        )
    # end def

    def on_export_pdf (self, *args, **kw):
        from .dlg_export_pdf import ExportPDFDialog
        _response = (
            self.check_integrity(self.project_path)
            or self.ensure_saved()
        )
        if _response is self.YES:
            ExportPDFDialog(self.mainwindow).show()
        elif _response is self.NO:
            self.user_show_warning(
                _(
                    "Cannot export data: "
                    "need a saved project file path."
                )
            )
        # end if
    # end def

    def on_new (self, *args, **kw):
        if self.ensure_saved() is not self.CANCEL:
            self.do_reset_project()
        else:
            self.notify(_("Project > New: cancelled."))
        # end if
    # end def

    def on_open (self, *args, **kw):
        _filepath = self.CANCEL
        if self.ensure_saved() is not self.CANCEL:
            self.prepare_filedialog()
            from tkinter import filedialog
            _filepath = filedialog.askopenfilename(
                title=_("Open project file..."),
                defaultextension=self.FILE_EXT,
                filetypes=self.FILE_TYPES,
                initialdir=self.current_dir,
                multiple=False,
                parent=self.mainwindow
            )
        # end if
        if _filepath:
            return self.do_open_project(_filepath)
        else:
            self.notify(_("Project > Open: cancelled."))
            return self.CANCEL
        # end if
    # end def

    def on_save (self, *args, **kw):
        if not self.check_integrity(self.project_path):
            return self.on_save_as()
        else:
            return self.do_save_project(self.project_path)
        # end if
    # end def

    def on_save_as (self, *args, **kw):
        self.prepare_filedialog()
        from tkinter import filedialog
        _filepath = filedialog.asksaveasfilename(
            title=_("Save as..."),
            defaultextension=self.FILE_EXT,
            filetypes=self.FILE_TYPES,
            initialdir=self.current_dir,
            confirmoverwrite=True,
            parent=self.mainwindow
        )
        if _filepath:
            return self.do_save_project(_filepath)
        else:
            self.notify(_("Project > Save as...: cancelled."))
            return self.CANCEL
        # end if
    # end def

    def prepare_filedialog (self, *args, **kw):
        from tkinter import TclError
        _tk = self.mainwindow.tk
        try:
            _tk.call("tk_getOpenFile", "-dummy")
        except TclError:
            pass
        # end try
        try:
            _tk.call(
                "set", "::tk::dialog::file::showHiddenBtn", "1"
            )
            _tk.call(
                "set", "::tk::dialog::file::showHiddenVar", "0"
            )
        except:
            pass
        # end try
    # end def

    @property
    def project_modified (self):
        for _id in self.ARCHIVE_FILES:
            try:
                if self.mainwindow.get_component(_id).modified:
                    return True
                # end if
            except:
                pass
            # end try
        # end for
        return False
    # end def

    @property
    def project_path (self):
        return self.__project_path
    # end def

    @project_path.setter
    def project_path (self, value):
        import os
        _filepath = path.normalize(value)
        self.__project_path = _filepath
        _directory, _filename = os.path.split(_filepath)
        if _directory:
            self.current_dir = _directory
        # end if
        self.mainwindow.events.raise_event(
            "Project:Path:Update",
            new_path=_filepath,
            filename=_filename,
            directory=_directory
        )
    # end def

    @project_path.deleter
    def project_path (self):
        del self.__project_path
    # end def

    def reset_members (self):
        self.project_path = ""
    # end def

    def reset_modified (self, *args, **kw):
        self.mainwindow.events.raise_event("Reset:Modified")
    # end def

    def user_ask_yesnocancel (self, question):
        from tkinter import messagebox
        return messagebox.askyesnocancel(
            title=_("Question"),
            message=str(question),
            parent=self.mainwindow
        )
    # end def

    def user_show_error (self, msg):
        from tkinter import messagebox
        messagebox.showerror(
            title=_("Error"),
            message=str(msg),
            parent=self.mainwindow
        )
    # end def

    def user_show_warning (self, msg):
        from tkinter import messagebox
        messagebox.showwarning(
            title=_("Attention"),
            message=str(msg),
            parent=self.mainwindow
        )
    # end def

# end class
