#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    tkScenarist - screen writing made simpler

    Copyright (c) 2014+ Raphaël Seban <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see http://www.gnu.org/licenses/
"""

from tkRAD.widgets.rad_canvas import RADCanvas


class ResourcesCanvas (RADCanvas):

    CONFIG = {
        "background": "grey80",
        "highlightbackground": "grey20",
        "highlightthickness": 1,
        "takefocus": 0,
    }

    def bbox_size (self, tag_or_id):
        try:
            _x0, _y0, _x1, _y1 = self.bbox(tag_or_id)
            return (abs(_x1 - _x0), abs(_y1 - _y0))
        except:
            return (0, 0)
        # end try
    # end def

    def bind_events (self, **kw):
        self.events.connect_dict(
            {
                "Dialog:DateBar:Validate": self.on_datebar_validate,
            }
        )
        self.bind("<Configure>", self.update_canvas)
        self.bind("<Control-ButtonRelease-1>", self.on_remove_item)
        self.bind("<Double-1>", self.on_double_click)
        for _seq in ("Button-4", "Button-5", "MouseWheel"):
            self.bind(
                "<{}>".format(_seq), self.on_mouse_wheel
            )
            self.bind(
                "<Control-{}>".format(_seq), self.on_change_date_scale
            )
        # end for
    # end def

    def box_rel (self, xy0, width, height):
        _x0, _y0 = xy0
        return (_x0, _y0, _x0 + width, _y0 + height)
    # end def

    def can_scroll_x (self):
        _w = self.bbox_size("all")[0]
        _cw = self.winfo_width()
        return _w and _w >= _cw - 1
    # end def

    def can_scroll_y (self):
        _h = self.bbox_size("all")[1]
        _ch = self.winfo_height()
        return _h and _h >= _ch - 1
    # end def

    def clear_canvas (self, *args, **kw):
        self.delete("all")
        self.configure(scrollregion=(0, 0, 0, 0))
        self.xview_moveto(0, override=True)
        self.yview_moveto(0, override=True)
    # end def

    def get_date_scale (self):
        return self.date_ruler.scale
    # end def

    def get_group_tag (self, x, y):
        try:
            return (
                self.gettags(self.find_overlapping(x, y, x, y)[-1])[0]
            )
        except:
            return ""
        # end try
    # end def

    def get_real_pos (self, x, y):
        return (self.canvasx(x), self.canvasy(y))
    # end def

    def get_xy_pos (self, cdate, item_name):
        _x = (
            self.date_ruler.XY_ORIGIN[0]
            + self.date_ruler.tick_offset
            + self.date_ruler.get_width(
                self.date_ruler.date_min, cdate, offset=True
            )
        )
        _y = self.item_list.get_y_pos(item_name)
        return (_x, _y)
    # end def

    def initialize (self, *args, **kw):
        from tkRAD.core.async import get_async_manager
        from .database import get_database
        self.async = get_async_manager()
        self.database = get_database()
        self.date_ruler = RCDateRuler(self)
        self.item_list = RCItemList(self)
        self.date_bars = dict()
        self.modified = False
        self.reset(**kw)
        self.bind_events()
    # end def

    def item_list_update (self, item_dict):
        self.clear_canvas()
        if item_dict:
            self.item_list.fill_list(item_dict)
            self.reset_datebars(item_dict)
            _w, _h = self.item_list.size
            self.date_ruler.tick_offset = _w - 2
            self.update_datebars()
        else:
            self.reset(except_scale=True)
        # end if
    # end def

    def on_change_date_scale (self, event=None, *args, **kw):
        if self.bbox("all"):
            _scale = self.date_ruler.get_scale_value(
                self.date_ruler.scale
                + int(event.num == 5 or event.delta < 0)
                - int(event.num == 4 or event.delta > 0)
            )
            if _scale != self.date_ruler.scale:
                self.set_date_scale(_scale)
            # end if
        # end if
    # end def

    def on_datebar_validate (self, *args, **kw):
        _datebar = kw.get("datebar")
        _item_name = kw.get("item_name")
        if not _datebar:
            _datebar = RCDateBar(
                self,
                rowid=self.item_list.items[_item_name],
            )
            self.date_bars[_datebar.tag] = _datebar
        # end if
        _datebar.status = kw.get("status")
        _datebar.date_begin, _datebar.date_end = (
            self.date_ruler.get_correct_interval(
                kw.get("date_begin"), kw.get("date_end")
            )
        )
        self.database.res_update_datebar(
            fk_type=_datebar.rowid,
            tag=_datebar.tag,
            status=_datebar.status,
            date_begin=str(_datebar.date_begin),
            date_end=str(_datebar.date_end),
        )
        self.update_datebars()
        self.modified = True
    # end def

    def on_double_click (self, event=None, *args, **kw):
        if self.bbox("all"):
            from .dlg_date_bar import DateBarDialog
            _x, _y = self.get_real_pos(event.x, event.y)
            _item = self.item_list.get_item_from(_x, _y)
            _tag = self.get_group_tag(_x, _y)
            _datebar = self.date_bars.get(_tag)
            DateBarDialog(
                self, item_name=_item["name"], datebar=_datebar
            ).show()
        # end if
    # end def

    def on_mouse_wheel (self, event=None, *args, **kw):
        import os
        _os = os.name.lower()
        if _os == "nt":
            _step = -event.delta // 120
        elif _os == "mac":
            _step = -event.delta
        else:
            _step = (event.num == 5) - (event.num == 4)
        # end if
        if event.state & 0x01:
            self.xview_scroll(_step, "units")
        else:
            self.yview_scroll(_step, "units")
        # end if
    # end def

    def on_remove_item (self, event=None, *args, **kw):
        if self.bbox("all"):
            _tag = self.get_group_tag(
                *self.get_real_pos(event.x, event.y)
            )
            if _tag in self.date_bars and self.user_confirm_deletion():
                _datebar = self.date_bars.pop(_tag)
                _datebar.clear()
                self.database.res_del_datebar(_datebar.tag)
                self.update_datebars()
                self.modified = True
            # end if
        # end if
    # end def

    def reset (self, *args, **kw):
        self.clear_canvas()
        self.date_bars.clear()
        self.date_ruler.reset(**kw)
        self.item_list.reset(**kw)
    # end def

    def reset_datebars (self, item_dict):
        self.date_bars.clear()
        if item_dict:
            _rows = self.database.res_get_datebars(item_dict.values())
            for _row in _rows:
                _datebar = RCDateBar(
                    self,
                    rowid=_row["fk_type"],
                    tag=_row["tag"],
                    status=_row["status"],
                    date_begin=_row["date_begin"],
                    date_end=_row["date_end"],
                )
                self.date_bars[_datebar.tag] = _datebar
            # end for
        # end if
    # end def

    def set_date_scale (self, value):
        self.date_ruler.scale = value
        self.update_datebars()
    # end def

    def update_canvas (self, *args, **kw):
        self.update_pos()
        _bbox = self.bbox("all")
        if _bbox:
            x0, y0, x1, y1 = _bbox
            x0, y0 = (min(0, x0 + 1), min(0, y0 + 1))
            self.configure(scrollregion=(x0, y0, x1, y1))
        else:
            self.reset()
        # end if
    # end def

    def update_datebars (self, *args, **kw):
        def deferred ():
            if self.date_bars:
                _dmin = _dmax = None
                for _datebar in self.date_bars.values():
                    if _dmin:
                        _dmin = min(_dmin, _datebar.date_begin)
                        _dmax = max(_dmax, _datebar.date_end)
                    else:
                        _dmin = _datebar.date_begin
                        _dmax = _datebar.date_end
                    # end if
                # end for
                self.xview_moveto(0, override=True)
                self.date_ruler.update(date_min=_dmin, date_max=_dmax)
                _names = self.item_list.swapped_items
                for _datebar in self.date_bars.values():
                    _datebar.draw(_names[_datebar.rowid])
                # end for
            else:
                self.date_ruler.update()
            # end if
            self.update_canvas()
        # end def
        self.async.run_after_idle(deferred)
    # end def

    def update_pos (self, *args, **kw):
        self.item_list.update_pos()
        self.date_ruler.update_pos()
    # end def

    def user_confirm_deletion (self):
        from tkinter.messagebox import askyesno
        return askyesno(
            title=_("Attention"),
            message=_("Do you really want to delete this date bar?"),
            parent=self
        )
    # end def

    def xview (self, *args):
        if self.can_scroll_x():
            super().xview(*args)
            self.update_pos()
        # end if
    # end def

    def xview_moveto (self, *args, override=False):
        if self.can_scroll_x() or override:
            super().xview_moveto(*args)
            self.update_pos()
        # end if
    # end def

    def xview_scroll (self, *args):
        if self.can_scroll_x():
            super().xview_scroll(*args)
            self.update_pos()
        # end if
    # end def

    def yview (self, *args):
        if self.can_scroll_y():
            super().yview(*args)
            self.update_pos()
        # end if
    # end def

    def yview_moveto (self, *args, override=False):
        if self.can_scroll_y() or override:
            super().yview_moveto(*args)
            self.update_pos()
        # end if
    # end def

    def yview_scroll (self, *args):
        if self.can_scroll_y():
            super().yview_scroll(*args)
            self.update_pos()
        # end if
    # end def

# end class


class RCCanvasItem:

    def __init__ (self, canvas, **kw):
        self.canvas = canvas
        self.tag = kw.get("tag")
        self.tag_labels = "{}_labels".format(self.tag)
        self.init_members(**kw)
    # end def

    def clear (self, *args, **kw):
        self.canvas.delete(self.tag, self.tag_labels)
    # end def

    def init_members (self, *args, **kw):
        pass
    # end def

    def reset (self, *args, **kw):
        self.clear(*args, **kw)
        self.init_members(*args, **kw)
    # end def

    @property
    def tag (self):
        return self.__tag
    # end def

    @tag.setter
    def tag (self, value):
        if value and isinstance(value, str):
            self.__tag = value
        else:
            self.__tag = (
                "{}#{}".format(self.__class__.__name__, id(self))
            )
        # end if
    # end def

    @tag.deleter
    def tag (self):
        del self.__tag
    # end def

# end class


class RCDateBar (RCCanvasItem):

    COLORS = {
        "OK": "royal blue",
        "N/A": "firebrick",
    }

    @property
    def date_begin (self):
        return self.__date_begin
    # end def

    @date_begin.setter
    def date_begin (self, value):
        self.__date_begin = self.get_date_value(value)
    # end def

    @property
    def date_end (self):
        return self.__date_end
    # end def

    @date_end.setter
    def date_end (self, value):
        self.__date_end = self.get_date_value(value)
    # end def

    def draw (self, item_name):
        self.clear()
        _height = int(self.canvas.item_list.LINE_HEIGHT * 0.8)
        _width = self.canvas.date_ruler.get_width(
            self.date_begin, self.date_end,
            overlap=1 # last day is *INCLUDED*
        )
        _x, _y = self.canvas.get_xy_pos(self.date_begin, item_name)
        _y -= _height // 2 - 2
        self.canvas.create_rectangle(
            self.canvas.box_rel((_x, _y), _width, _height),
            outline="black",
            fill=self.COLORS[self.status],
            width=1,
            tags=self.tag,
        )
    # end def

    def get_date_value (self, value):
        from datetime import date
        if isinstance(value, date):
            return value
        elif value and isinstance(value, str):
            return self.get_iso_date(value)
        elif value is not None:
            raise TypeError(
                "Date attribute must be of datetime.date type or "
                "at least an ISO 8601 'YYYY-MM-DD' string of chars."
            )
        # end if
        return None
    # end def

    def get_iso_date (self, date_str):
        from datetime import date, datetime
        _dt = datetime.strptime(date_str, "%Y-%m-%d")
        return date(_dt.year, _dt.month, _dt.day)
    # end def

    def init_members (self, *args, **kw):
        self.rowid = kw.get("rowid")
        self.status = kw.get("status")
        self.date_begin = kw.get("date_begin")
        self.date_end = kw.get("date_end")
    # end def

# end class


class RCDateRuler (RCCanvasItem):

    FONT = "tkdefaultfont 8"
    PAD_X = 5
    RULER_HEIGHT = 24
    SCALES = ("days", "weeks", "months")
    XY_ORIGIN = (0, 0)

    @property
    def date_max (self):
        return self.__date_max
    # end def

    @date_max.setter
    def date_max (self, value):
        from datetime import date, timedelta
        if isinstance(value, date):
            self.__date_max = value
        else:
            self.__date_max = date.today() + timedelta(days=7)
        # end if
    # end def

    @property
    def date_min (self):
        return self.__date_min
    # end def

    @date_min.setter
    def date_min (self, value):
        from datetime import date
        if isinstance(value, date):
            self.__date_min = value
        else:
            self.__date_min = date.today()
        # end if
    # end def

    def day_ratio (self, cdate):
        from calendar import monthrange
        return float(
            (cdate.day - 1) / monthrange(cdate.year, cdate.month)[1]
        )
    # end def

    def draw_ruler (self, *args, **kw):
        self.tick_width = 0
        _cb_label = kw.get("get_date_label")
        _cb_next = kw.get("next_date")
        _labels = list()
        _x0, _y0 = self.XY_ORIGIN
        _y0 += self.RULER_HEIGHT
        _y = _y0 - 5
        _cur_date, _end_date = self.get_correct_interval(
            self.date_min, self.date_max
        )
        _end_date = _cb_next(_end_date)
        self.clear()
        while _cur_date <= _end_date:
            _id = self.canvas.create_text(
                _x0, _y,
                anchor="s",
                fill="black",
                font=self.FONT,
                text=_cb_label(_cur_date),
                tags=(self.tag, self.tag_labels),
            )
            _labels.append(_id)
            _w, _h = self.canvas.bbox_size(_id)
            self.tick_width = max(_w, self.tick_width)
            _cur_date = _cb_next(_cur_date)
        # end while
        self.tick_width += self.PAD_X
        self.tick_offset = max(
            self.tick_width // 2 + self.PAD_X, self.tick_offset
        )
        _x1 = _x0 + self.tick_offset
        for _index, _id in enumerate(_labels):
            _x = _x1 + _index * self.tick_width
            self.canvas.coords(_id, _x, _y)
            self.canvas.create_line(
                _x, _y0, _x, _y0 - 3,
                fill="black",
                tags=(self.tag, self.tag_labels),
                width=1,
            )
        # end for
        self.frame_id = self.canvas.create_rectangle(
            _x0, _y0,
            self.canvas.bbox(self.tag_labels)[2],
            _y0 - self.RULER_HEIGHT,
            outline="black",
            fill="grey90",
            width=1,
            tags=self.tag,
        )
        self.canvas.tag_raise(self.tag, "all")
        self.canvas.tag_raise(self.tag_labels, self.frame_id)
    # end def

    def fill_with_days (self, *args, **kw):
        from datetime import timedelta
        self.draw_ruler(
            *args,
            get_date_label=lambda d: d.strftime("%a %x"),
            next_date=lambda d: d + timedelta(days=1),
            **kw
        )
    # end def

    def fill_with_months (self, *args, **kw):
        def next_month (cdate):
            from datetime import date
            _y = cdate.year
            _m = cdate.month + 1
            if _m > 12:
                _y += 1
                _m = 1
            # end if
            return date(_y, _m, 1)
        # end def
        self.draw_ruler(
            *args,
            get_date_label=lambda d: d.strftime("%b %Y"),
            next_date=next_month,
            **kw
        )
    # end def

    def fill_with_weeks (self, *args, **kw):
        from datetime import timedelta
        self.draw_ruler(
            *args,
            get_date_label=lambda d: d.strftime("%a %x"),
            next_date=lambda d: d + timedelta(days=7),
            **kw
        )
    # end def

    def get_correct_interval (self, date_min, date_max):
        if date_min > date_max:
            return (date_max, date_min)
        else:
            return (date_min, date_max)
        # end if
    # end def

    def get_scale_name (self, scale=None):
        return self.SCALES[
            self.get_scale_value(
                self.scale if scale is None else scale
            )
        ]
    # end def

    def get_scale_value (self, value):
        try:
            return max(0, min(len(self.SCALES) - 1, int(value)))
        except:
            return 0
        # end try
    # end def

    def get_width (self, date_min, date_max, overlap=0, offset=False):
        _dmin, _dmax = self.get_correct_interval(date_min, date_max)
        _delta = _dmax - _dmin
        _days = _delta.days + overlap
        if self.scale == 0:
            return int(self.tick_width * _days)
        elif self.scale == 1:
            return int(self.tick_width * _days / 7.0)
        elif self.scale == 2:
            _rmin = self.day_ratio(_dmin)
            _w0 = _rmin if offset else 0
            _months = int(
                (_dmax.year - _dmin.year) * 12
                + _dmax.month - _dmin.month
            )
            _days = self.day_ratio(_dmax) - _rmin
            # end if
            return int(self.tick_width * (_w0 + _months + _days))
        # end if
    # end def

    def init_members (self, *args, **kw):
        self.date_min = kw.get("date_min")
        self.date_max = kw.get("date_max")
        if not kw.get("except_scale"):
            self.scale = kw.get("scale")
        # end if
        self.tick_offset = 0
        self.tick_width = 0
    # end def

    @property
    def scale (self):
        return self.__scale
    # end def

    @scale.setter
    def scale (self, value):
        self.__scale = self.get_scale_value(value)
    # end def

    def update (self, *args, **kw):
        self.date_min = kw.pop("date_min", None)
        self.date_max = kw.pop("date_max", None)
        getattr(
            self, "fill_with_{}".format(self.get_scale_name())
        )(*args, **kw)
    # end def

    def update_pos (self, *args, **kw):
        try:
            self.canvas.move(
                self.tag,
                0, self.canvas.canvasy(0)-self.canvas.bbox(self.tag)[1]
            )
            self.canvas.tag_raise(self.tag, "all")
        except:
            pass
        # end try
    # end def

# end class


class RCItemList (RCCanvasItem):

    LINE_HEIGHT = 20

    XY_ORIGIN = (0, RCDateRuler.RULER_HEIGHT)

    def fill_list (self, item_dict):
        self.reset()
        if item_dict:
            self.items = item_dict
            self.sorted_items = sorted(item_dict)
            self.swapped_items = {
                _value: _key for _key, _value in item_dict.items()
            }
            _x0, _y0 = self.XY_ORIGIN
            _x0 += 5
            _y0 += 5
            for _index, _item in enumerate(self.sorted_items):
                self.canvas.create_text(
                    _x0, _y0 + _index * self.LINE_HEIGHT,
                    anchor="nw",
                    fill="black",
                    text=_item,
                    tags=(self.tag, self.tag_labels),
                )
            # end for
            _w, _h = self.canvas.bbox_size(self.tag_labels)
            self.frame_id = self.canvas.create_rectangle(
                self.canvas.box_rel(self.XY_ORIGIN, _w + 10, _h + 10),
                outline="black",
                fill="grey90",
                width=1,
                tags=self.tag,
            )
            self.canvas.tag_raise(self.tag, "all")
            self.canvas.tag_raise(self.tag_labels, self.frame_id)
        # end if
    # end def

    def get_item_from (self, x, y):
        if self.items:
            _x0, _y0 = self.XY_ORIGIN
            _index = min(
                len(self.items) - 1,
                max(0, y - _y0) // self.LINE_HEIGHT
            )
            _name = self.sorted_items[int(_index)]
            return dict(name=_name, rowid=self.items[_name])
        # end if
    # end def

    def get_y_pos (self, item_name):
        return int(
            self.XY_ORIGIN[1] +
            (self.sorted_items.index(item_name) + 0.5) *
            self.LINE_HEIGHT
        )
    # end def

    def init_members (self, *args, **kw):
        self.items = None
        self.sorted_items = None
        self.swapped_items = None
    # end def

    @property
    def size (self):
        return self.canvas.bbox_size(self.tag)
    # end def

    def update_pos (self, *args, **kw):
        try:
            self.canvas.move(
                self.tag,
                self.canvas.canvasx(0)-self.canvas.bbox(self.tag)[0], 0
            )
            self.canvas.tag_raise(self.tag, "all")
        except:
            pass
        # end try
    # end def

# end class
