#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    tkScenarist - screen writing made simpler

    Copyright (c) 2014+ Raphaël Seban <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see http://www.gnu.org/licenses/
"""

from tkinter import Text
from tkRAD.widgets.rad_widget_base import RADWidgetBase
from tkRAD.core import tools


class ScenarioText (RADWidgetBase, Text):

    CONFIG = {
        "autoseparators": False,
        "background": "white",
        "font": "courier 12",
        "foreground": "black",
        "highlightthickness": 1,
        "undo": True,
        "wrap": "word",
    }

    CONFIG_DEFAULTS = {
        "background": "white",
        "foreground": "black",
        "font": "courier 12 normal",
        "justify": "left",
        "lmargin1": "0",
        "lmargin2": "0",
        "rmargin": "0",
        "spacing1": "0",
        "spacing2": "0",
        "spacing3": "0",
    }

    DEAD_KEYS = (
        "Shift_L", "Shift_R", "Control_L", "Control_R", "Alt_L",
        "Alt_R", "Caps_Lock",
    )

    DEFAULT_TAG = "scene"

    ELEMENT_DEFAULTS = {
        "action": {
            "label": _("Action"),
            "config": {
                "spacing1": "5",
                "spacing3": "5",
            },
            "on_tab_create": "character",
            "on_tab_switch": "character",
            "on_return_create": "action",
            "on_return_switch": "",
            "on_ctrl_return_create": "transition",
            "on_ctrl_return_switch": "scene",
        },
        "character": {
            "label": _("Character"),
            "config": {
                "font": "courier 12 bold",
                "foreground": "grey40",
                "lmargin1": "5c",
                "lmargin2": "5c",
                "rmargin": "1c",
                "spacing1": "5",
            },
            "on_tab_create": "parenthetical",
            "on_tab_switch": "action",
            "on_return_create": "dialogue",
            "on_return_switch": "",
            "on_ctrl_return_create": "action",
            "on_ctrl_return_switch": "",
        },
        "dialogue": {
            "label": _("Dialogue"),
            "config": {
                "foreground": "grey30",
                "lmargin1": "3c",
                "lmargin2": "3c",
                "rmargin": "1c",
                "spacing1": "5",
                "spacing3": "10",
            },
            "on_tab_create": "action",
            "on_tab_switch": "parenthetical",
            "on_return_create": "character",
            "on_return_switch": "",
            "on_ctrl_return_create": "scene",
            "on_ctrl_return_switch": "",
        },
        "parenthetical": {
            "label": _("Parenthetical"),
            "config": {
                "foreground": "grey50",
                "font": "courier 12 italic",
                "lmargin1": "4c",
                "lmargin2": "4c",
                "rmargin": "1c",
            },
            "on_tab_create": "dialogue",
            "on_tab_switch": "dialogue",
            "on_return_create": "dialogue",
            "on_return_switch": "",
            "on_ctrl_return_create": "action",
            "on_ctrl_return_switch": "",
        },
        "scene": {
            "label": _("Scene"),
            "config": {
                "background": "grey90",
                "font": "courier 12 bold",
                "spacing1": "5",
                "spacing3": "5",
            },
            "on_tab_create": "character",
            "on_tab_switch": "action",
            "on_return_create": "action",
            "on_return_switch": "",
            "on_ctrl_return_create": "transition",
            "on_ctrl_return_switch": "",
        },
        "transition": {
            "label": _("Transition"),
            "config": {
                "justify": "right",
                "rmargin": "10",
            },
            "on_tab_create": "transition",
            "on_tab_switch": "scene",
            "on_return_create": "scene",
            "on_return_switch": "",
            "on_ctrl_return_create": "action",
            "on_ctrl_return_switch": "",
        },
    }

    INS_LINE = ("insert linestart", "insert linestart+1l")

    INS_LINE_END = ("insert linestart", "insert lineend")

    UNDO_LIMIT = 2000

    def __init__ (self, master=None, **kw):
        Text.__init__(self, master)
        self.CONFIG = self.CONFIG.copy()
        self.CONFIG.update(kw)
        self.configure(**self._only_tk(self.CONFIG))
        RADWidgetBase.__init__(self, **kw)
        self.initialize(**kw)
    # end def

    def _do_delete (self, index1, index2=None):
        self.delete_line_ids(index1, index2)
        super().delete(index1, index2)
        self.update_line()
        self.update_modified()
    # end def

    def _do_insert (self, index, chars, *args):
        super().insert(index, chars, *args)
        self.insert_line_ids(index, chars, *args)
        self.update_line()
        self.update_modified()
    # end def

    def bind_events (self, **kw):
        self.events.connect_dict(
            {
                "Scenario:Switch:Line": self.switch_line,
                "Scenario:Settings:Update": self.on_update_settings,
            }
        )
        self.bind("<Key>", self.on_keypress)
        self.bind("<ButtonRelease>", self.on_click)
        self.bind("<MouseWheel>", self.on_click)
        self.bind("<FocusIn>", self.on_focus_in)
        self.bind("<Tab>", self.on_key_tab)
        self.bind("<Return>", self.on_key_return)
        self.bind("<Control-Return>", self.on_key_ctrl_return)
        self.bind("<Control-a>", self.on_select_all)
        self.bind("<Control-A>", self.on_select_all)
        self.bind("<Control-z>", self.on_edit_undo)
        self.bind("<Control-Z>", self.on_edit_redo)
        self.bind("<Delete>", self.on_key_delete)
        self.bind("<Control-Delete>", self.on_key_ctrl_delete)
        self.bind("<BackSpace>", self.on_key_backspace)
        self.bind("<Control-BackSpace>", self.on_key_ctrl_backspace)
    # end def

    def clear_text (self, *args, **kw):
        self.delete(1.0, "end")
        self.edit_reset()
    # end def

    def compat_filter (self, e_dict):
        from copy import deepcopy
        _elements = deepcopy(self.get_options_element())
        self.dict_update_strict(_elements, e_dict)
        return _elements
    # end def

    def copy_styles_into (self, text_widget):
        for _tag in self.tag_names():
            _config = {
                _option: self.tag_cget(_tag, _option)
                for _option in self.CONFIG_DEFAULTS
            }
            text_widget.tag_configure(_tag, **_config)
        # end for
    # end def

    def create_element_line (self, element_tag, index=None):
        index = self.index(
            "{} linestart+1l".format(index or "insert")
        )
        if element_tag in self.ELEMENT:
            try:
                self.switch_to_method(
                    "manage_line_{}".format(self.get_line_tag())
                )
            except AttributeError:
                pass
            # end try
            try:
                return self.switch_to_method(
                    "create_element_line_{}".format(element_tag),
                    element_tag, index
                )
            except AttributeError:
                return self.insert_new_line(element_tag, index)
            # end try
        # end if
    # end def

    def create_element_line_parenthetical (self, tag, index):
        self.insert_new_line(tag, index)
        self.insert(index, "()", (tag,))
        self.move_cursor("{}+1c".format(index))
    # end def

    def delete (self, index1, index2=None):
        if self.undo_enabled():
            self.undo_stack.push_delete(
                self.index(index1),
                *self.get_tagged_text(index1, index2)
            )
            try:
                self._do_delete(index1, index2)
            except:
                self.undo_stack.pop()
            # end try
        else:
            self._do_delete(index1, index2)
        # end if
    # end def

    def delete_line_ids (self, index1, index2):
        _text = self.get(index1, index2)
        _c = _text.count("\n")
        _index = self.get_line_number(index1)
        del self.line_ids[_index:_index + _c]
    # end def

    def delete_selection (self, *args, **kw):
        try:
            self.delete("sel.first", "sel.last")
            return True
        except:
            return False
        # end try
    # end def

    def dict_update_strict (self, s_dict, u_dict):
        if tools.is_pdict(s_dict) and tools.is_pdict(u_dict):
            _keys = set(s_dict).intersection(set(u_dict))
            # strict at first level only
            for _k in _keys:
                s_dict[_k] = u_dict[_k]
            # end for
        # end if
    # end def

    def edit_redo (self):
        if self.undo_enabled():
            _sequence = self.undo_stack.get_redo_elements()
            _index = None
            for _element in _sequence:
                if _element.mode == "+":
                    self._do_insert(
                        _element.start_index, *_element.args
                    )
                    _index = _element.end_index
                else:
                    self._do_delete(
                        _element.start_index, _element.end_index
                    )
                    _index = _element.start_index
                # end if
            # end for
            if _index:
                self.move_cursor(_index)
                self.update_line()
                self.after_idle(self.see, "insert")
            # end if
        # end if
    # end def

    def edit_reset (self):
        super().edit_reset()
        self.undo_stack.reset()
    # end def

    def edit_separator (self):
        if self.undo_enabled():
            self.undo_stack.add_separator()
        # end if
    # end def

    def edit_undo (self):
        if self.undo_enabled():
            _sequence = self.undo_stack.get_undo_elements()
            _index = None
            for _element in _sequence:
                if _element.mode == "+":
                    self._do_delete(
                        _element.start_index, _element.end_index
                    )
                    _index = _element.start_index
                else:
                    self._do_insert(
                        _element.start_index, *_element.args
                    )
                    _index = _element.end_index
                # end if
            # end for
            if _index:
                self.move_cursor(_index)
                self.after_idle(self.see, "insert")
            # end if
        # end if
    # end def

    def file_setup (self, fname, archive):
        import json
        _get_fc = lambda k: archive.read(fname[k]).decode(ENCODING)
        self.reset()
        self.insert("end", _get_fc("text").rstrip())
        self.reset_elements(
            self.compat_filter(json.loads(_get_fc("elements")))
        )
        _data = json.loads(_get_fc("data"))
        self.line_ids = _data.get("line_ids") or self.line_ids
        _tags = _data.get("tags") or []
        for _index, _tag in enumerate(_tags):
            self.tag_add(
                _tag, float(_index + 1), float(_index + 2)
            )
        # end for
        self.edit_reset()
        self.move_cursor(1.0)
        self.update_line()
        self.update_modified(flag=False)
    # end def

    def get_all_lines_range (self):
        return range(1, 1 + self.get_nb_of_lines())
    # end def

    def get_column_index (self, index=None):
        index = self.index(index or "insert")
        return tools.ensure_int(index.split(".")[-1])
    # end def

    def get_element_mappings (self, index=None):
        index = index or "insert"
        _tag = self.get_line_tag(index)
        if _tag in self.ELEMENT:
            _element = self.ELEMENT[_tag]
            if self.inserted_chars(index):
                _map = {
                    "tag": _tag,
                    "on_tab_create":
                        _element.get("on_tab_create") or "",
                    "on_return_create":
                        _element.get("on_return_create") or "",
                    "on_ctrl_return_create":
                        _element.get("on_ctrl_return_create") or "",
                    "on_tab_switch": "",
                    "on_return_switch": "",
                    "on_ctrl_return_switch": "",
                }
            else:
                _map = {
                    "tag": _tag,
                    "on_tab_create": "",
                    "on_return_create": "",
                    "on_ctrl_return_create": "",
                    "on_tab_switch":
                        _element.get("on_tab_switch") or "",
                    "on_return_switch":
                        _element.get("on_return_switch") or "",
                    "on_ctrl_return_switch":
                        _element.get("on_ctrl_return_switch") or "",
                }
            # end if
            return _map
        # end if
    # end def

    def get_fc_tags (self):
        _tags = list()
        for _line in self.get_all_lines_range():
            _tn = self.tag_names(float(_line))
            if _tn:
                _tags.append(_tn[0])
            # end if
        # end for
        return _tags
    # end def

    def get_file_contents (self, fname):
        import json
        return {
            fname["text"]: self.get(1.0, "end"),
            fname["data"]: json.dumps(
                {
                    "tags": self.get_fc_tags(),
                    "line_ids": self.line_ids,
                }
            ),
            fname["elements"]: json.dumps(self.ELEMENT),
        }
    # end def

    def get_label (self, element_tag):
        _element = self.ELEMENT.get(element_tag) or dict()
        return _element.get("label") or ""
    # end def

    def get_line_contents (self, index=None):
        index = index or "insert"
        return self.get(
            "{} linestart".format(index), "{} lineend".format(index)
        )
    # end def

    def get_line_number (self, index=None):
        index = index or "insert"
        return tools.ensure_int(
            self.index("{} linestart".format(index))
        )
    # end def

    def get_line_tag (self, index=None, strict=False):
        index = "{} linestart".format(index or "insert")
        _tags = self.tag_names(index)
        if _tags and _tags[0] in self.ELEMENT:
            return _tags[0]
        # end if
        if not strict:
            return self.current_tag
        # end if
    # end def

    def get_lines (self, element_tag):
        _lines = []
        _line_ids = []
        _texts = []
        _count = 1
        for _line in self.get_all_lines_range():
            _index = float(_line)
            _tag = self.get_line_tag(_index, strict=True)
            if _tag == element_tag:
                _lines.append(_line)
                _line_ids.append(self.line_ids[_line - 1])
                _texts.append(
                    "#{} {}"
                    .format(_count, self.get_line_contents(_index))
                )
                _count += 1
            # end if
        # end for
        return {
            "current": self.get_line_number(),
            "lines": _lines,
            "line_ids": _line_ids,
            "texts": _texts,
        }
    # end def

    def get_nb_of_lines (self):
        return self.get_line_number("end") - 1
    # end def

    def get_new_id (self):
        self.instance_counter += 1
        return int(self.instance_counter)
    # end def

    def get_options_element (self):
        import json
        return json.loads(
            self.options.get(
                self.get_rc_section(), "text_settings", fallback="[]"
            )
        ) or self.ELEMENT_DEFAULTS
    # end def

    def get_tagged_text (self, index1, index2=None):
        index1 = index1 or "insert"
        index2 = index2 or "insert"
        if self.get_line_number(index1) == self.get_line_number(index2):
            return (self.get(index1, index2), self.tag_names(index1))
        else:
            _index = index1
            _tags = self.tag_names(_index)
            _args = []
            while self.compare(_index, "<", index2):
                _tn = self.tag_names(_index)
                if _tn != _tags:
                    _args.append(self.get(index1, _index))
                    _args.append(_tags)
                    index1 = _index
                    _tags = _tn
                # end if
                _index = self.index("{}+1c".format(_index))
            # end while
            _args.append(self.get(index1, index2))
            _args.append(self.tag_names(index1))
            return tuple(_args)
        # end if
    # end def

    def get_word (self, index=None):
        index = index or "insert"
        _start = "{} linestart".format(index)
        _end = "{} lineend".format(index)
        _word = ""
        _text = self.get(_start, index)
        _pos = _text.rfind(" ")
        if _pos >= 0:
            _word += _text[_pos + 1:]
            _start = "{}+{}c".format(_start, _pos + 1)
        else:
            _word += _text
        # end if
        _text = self.get(index, _end)
        _pos = _text.find(" ")
        if _pos >= 0:
            _word += _text[:_pos]
            _end = "{}+{}c".format(_start, _pos)
        else:
            _word += _text
        # end if
        return {
            "word": _word.rstrip(" .:,;?!\"']})"),
            "start_index": _start,
            "end_index": _end,
        }
    # end def

    def init_members (self, **kw):
        self.instance_counter = 0
        self.current_tag = self.DEFAULT_TAG
        self.line_ids = [self.get_new_id()]
        self.reset_elements(self.get_options_element())
    # end def

    def init_styles (self, **kw):
        for _tag, _element in self.ELEMENT.items():
            _config = _element.get("config")
            try:
                _cfg = self.CONFIG_DEFAULTS.copy()
                _cfg.update(_config)
                self.tag_configure(_tag, **_cfg)
            except:
                pass
            # end try
        # end for
        self.tag_configure(
            "sel", background="grey30", foreground="white"
        )
        self.tag_raise("sel")
    # end def

    def initialize (self, *args, **kw):
        import os
        import string
        self.STATE_MASK = (
            {"nt": 0x20084, "posix": 0x8c}.get(os.name, 0x00)
        )
        self.SEPARATORS = string.whitespace + string.punctuation
        self.undo_stack = TextUndoStack(limit=self.UNDO_LIMIT)
        self.line_ids = list()
        def deferred ():
            self.reset(**kw)
            self.bind_events(**kw)
        # end def
        self.after_idle(deferred)
    # end def

    def insert (self, index, chars, *args):
        if self.undo_enabled():
            self.undo_stack.push_insert(
                self.index(index), chars, *args
            )
        # end if
        self._do_insert(index, chars, *args)
    # end def

    def insert_line_ids (self, index, chars, *args):
        _text = chars + "".join(args[1::2])
        _c = _text.count("\n")
        _i = self.get_line_number(index)
        self.line_ids[_i:_i] = [self.get_new_id() for n in range(_c)]
    # end def

    def insert_new_line (self, new_tag, index):
        if not self.delete_selection():
            self.edit_separator()
            self.insert("insert lineend", "\n")
            self.move_cursor(index)
            self.tag_remove(self.tag_names("insert"), *self.INS_LINE)
            self.tag_add(new_tag, *self.INS_LINE)
        # end if
    # end def

    def inserted_chars (self, index=None):
        index = index or "insert"
        _chars = self.get_line_contents(index).strip("\n\t")
        _tag = self.get_line_tag(index)
        if _tag == "parenthetical":
            return bool(len(_chars) > 2)
        else:
            return bool(_chars)
        # end if
    # end def

    def key_filter (self, event, use_filter=None):
        _char = event.char
        _modifiers = (event.state & self.STATE_MASK)
        if _char >= " " and not _modifiers:
            self.delete_selection()
            if use_filter:
                if callable(use_filter):
                    _char = use_filter(event)
                elif use_filter.startswith("up"):
                    _char = _char.upper()
                elif use_filter.startswith("low"):
                    _char = _char.lower()
                elif use_filter.startswith("cap"):
                    _ins = "insert"
                    if self.compare(_ins, "==", self.INS_LINE[0]) or \
                            self.get("{}-2c".format(_ins), _ins) in \
                                                    (". ", "! ", "? "):
                        _char = _char.upper()
                    # end if
                # end if
            # end if
            self.insert(
                "insert", _char, self.get_line_tag()
            )
            return "break"
        # end if
        return None
    # end def

    def line_selected (self, index=None):
        _line = self.get_line_number(index)
        try:
            _first = self.get_line_number("sel.first")
            _last = self.get_line_number("sel.last")
        except:
            return False
        else:
            return bool(_line in (_first, _last))
        # end try
    # end def

    def manage_line (self, create_key, switch_key):
        _map = self.get_element_mappings()
        if _map:
            self.after_idle(self.see, "insert")
            if _map.get(create_key):
                self.create_element_line(_map.get(create_key))
            elif _map.get(switch_key):
                self.switch_line(new_tag=_map.get(switch_key))
            # end if
        # end if
        return "break"
    # end def

    def manage_line_character (self):
        _name = self.get_line_contents()
        if _name:
            self.events.raise_event(
                "Scenario:Character:Name:Detected", name=_name,
            )
        # end if
    # end def

    def move_cursor (self, index):
        self.mark_set("insert", index)
    # end def

    def on_click (self, event=None, *args, **kw):
        self.update_line()
        self.events.raise_event("Scenario:Text:Clicked", event=event)
    # end def

    def on_edit_redo (self, *args, **kw):
        self.edit_redo()
        return "break"
    # end def

    def on_edit_undo (self, *args, **kw):
        self.edit_undo()
        return "break"
    # end def

    def on_focus_in (self, event=None, *args, **kw):
        self.update_line()
        self.events.raise_event("Scenario:Text:FocusedIn", event=event)
    # end def

    def on_key_backspace (self, *args, **kw):
        self.reformat_line(keep_cursor=True)
        if not self.delete_selection():
            self.delete("insert-1c")
        # end if
        return "break"
    # end def

    def on_key_ctrl_backspace (self, *args, **kw):
        if not self.delete_selection():
            _start = self.index("insert linestart")
            _text = self.get(_start, "insert")
            _pos = _text.rfind(" ")
            if _pos >= 0:
                _start = self.index("{}+{}c".format(_start, _pos))
            # end if
            self.delete(_start, "insert")
        # end if
        return "break"
    # end def

    def on_key_ctrl_delete (self, *args, **kw):
        if not self.delete_selection():
            _end = self.index("insert lineend")
            _text = self.get("insert", _end)
            _pos = _text.find(" ")
            if _pos >= 0:
                _end = self.index("insert+{}c".format(_pos + 1))
            # end if
            self.delete("insert", _end)
        # end if
        return "break"
    # end def

    def on_key_ctrl_return (self, *args, **kw):
        return self.manage_line(
            "on_ctrl_return_create", "on_ctrl_return_switch"
        )
    # end def

    def on_key_delete (self, *args, **kw):
        self.reformat_line(keep_cursor=True)
        if not self.delete_selection():
            self.delete("insert")
        # end if
        return "break"
    # end def

    def on_key_return (self, *args, **kw):
        return self.manage_line("on_return_create", "on_return_switch")
    # end def

    def on_key_tab (self, *args, **kw):
        return self.manage_line("on_tab_create", "on_tab_switch")
    # end def

    def on_keypress (self, event=None, *args, **kw):
        if event.keysym not in self.DEAD_KEYS:
            self.update_line()
            self.see("insert")
            if event.char and event.char in self.SEPARATORS:
                self.edit_separator()
            # end if
        # end if
        return self.switch_to_method(
            "on_keypress_{}".format(self.get_line_tag()), event
        )
    # end def

    def on_keypress_action (self, event=None, *args, **kw):
        return self.key_filter(event, use_filter="capitalize")
    # end def

    def on_keypress_character (self, event=None, *args, **kw):
        return self.on_keypress_scene(event, *args, **kw)
    # end def

    def on_keypress_dialogue (self, event=None, *args, **kw):
        return self.on_keypress_action(event)
    # end def

    def on_keypress_parenthetical (self, event=None, *args, **kw):
        return self.on_keypress_action(event)
    # end def

    def on_keypress_scene (self, event=None, *args, **kw):
        return self.key_filter(event, use_filter="upper")
    # end def

    def on_keypress_transition (self, event=None, *args, **kw):
        return self.on_keypress_scene(event, *args, **kw)
    # end def

    def on_select_all (self, *args, **kw):
        self.tag_add("sel", *self.INS_LINE)
        return "break"
    # end def

    def on_update_settings (self, *args, settings=None, **kw):
        if tools.is_pdict(settings):
            self.reset_elements(settings["project"])
            self.set_options_element(settings["global"])
            self.edit_modified(True)
        # end if
    # end def

    def reformat_line (self, *args, **kw):
        def deferred ():
            _tag = self.update_current_tag("insert")
            _text = self.get(*self.INS_LINE_END)
            _cursor = self.index("insert")
            _text, _adjust = self.switch_to_method(
                "reformat_line_{}".format(_tag), _text
            )
            self._do_delete(*self.INS_LINE_END)
            self._do_insert(self.INS_LINE_END[0], _text, _tag)
            if kw.get("keep_cursor"):
                _adjust = ""
            # end if
            self.move_cursor("{} {}".format(_cursor, _adjust))
        # end def
        self.after_idle(deferred)
    # end def

    def reformat_line_action (self, text):
        return (text.strip("()\t\n"), "")
    # end def

    def reformat_line_character (self, text):
        return self.reformat_line_scene(text)
    # end def

    def reformat_line_dialogue (self, text):
        return self.reformat_line_action(text)
    # end def

    def reformat_line_parenthetical (self, text):
        return ("({})".format(text.strip("()\t\n")), "+1c")
    # end def

    def reformat_line_scene (self, text):
        return (text.strip("()\t\n").upper(), "")
    # end def

    def reformat_line_transition (self, text):
        return self.reformat_line_scene(text)
    # end def

    def replace_text (self, text, start=None, end=None,
                                smart_delete=False, keep_cursor=False):
        start = start or "insert"
        end = end or "insert"
        _cursor = self.index("insert")
        _tags = tuple(set(self.tag_names(start) + self.tag_names(end)))
        if smart_delete:
            import re
            _endl = "{} lineend".format(end)
            _text = self.get(end, _endl)
            _found = re.search(r"[^\w\-]", _text)
            if _found:
                end = "{} +{}c".format(end, _found.start())
            else:
                end = _endl
            # end if
        # end if
        self.delete(start, end)
        self.insert(start, text, _tags)
        if keep_cursor:
            self.move_cursor(_cursor)
        # end if
    # end def

    def reset (self, *args, **kw):
        self.clear_text(**kw)
        self.init_members(**kw)
        self.update_line(**kw)
        self.update_modified(flag=False)
    # end def

    def reset_elements (self, new_dict):
        from copy import deepcopy
        if tools.is_pdict(new_dict):
            self.ELEMENT = deepcopy(new_dict)
            self.init_styles()
            self.events.raise_event(
                "Scenario:Elements:Init",
                elements=tuple(sorted(self.ELEMENT))
            )
        # end if
    # end def

    def set_options_element (self, e_dict):
        import json
        if tools.is_pdict(e_dict):
            self.options[self.get_rc_section()]["text_settings"] = (
                json.dumps(e_dict)
            )
        # end if
    # end def

    def switch_line (self, *args, new_tag=None, **kw):
        self.update_line(force_tag=new_tag)
        self.reformat_line()
    # end def

    def switch_to_method (self, method_name, *args, **kw):
        if method_name:
            _method = getattr(self, method_name, None)
            if callable(_method):
                return _method(*args, **kw)
            else:
                raise AttributeError(
                    _("unable to call method '{method}()': not found.")
                    .format(method=method_name)
                )
            # end if
        # end if
        return None
    # end def

    def undo_enabled (self):
        return bool(self.cget("undo"))
    # end def

    def update_current_tag (self, *args, index=None, **kw):
        _index = index or "insert"
        _tag = kw.get("force_tag") or self.get_line_tag(_index)
        if _tag in self.ELEMENT:
            if self.compare(_index, "==", "insert"):
                self.current_tag = _tag
            else:
                return _tag
            # end if
        # end if
        return self.current_tag
    # end def

    def update_line (self, *args, **kw):
        def deferred ():
            _index = kw.pop("index", "insert")
            _tag = self.update_current_tag(index=_index, **kw)
            if _tag in self.ELEMENT:
                _index1 = "{} linestart".format(_index)
                _index2 = "{}+1l".format(_index1)
                self.tag_remove(
                    self.tag_names(_index), _index1, _index2
                )
                self.tag_add(_tag, _index1, _index2)
                if self.compare(_index, "==", "insert"):
                    self.events.raise_event(
                        "Scenario:Current:Element:Update",
                        element_tag=_tag
                    )
                # end if
            # end if
        # end def
        self.after_idle(deferred)
    # end def

    def update_modified (self, *args, flag=True, **kw):
        self.edit_modified(flag)
    # end def

# end class


class TextUndoStack (list):

    SEPARATOR = object()

    class Element:

        def __init__ (self, mode, index, *args):
            self.mode = mode
            self.start_index = index
            self.args = args
        # end def

        def __repr__ (self):
            return str(self)
        # end def

        def __str__ (self):
            return "<Element '{}' ({}, {}) {}>".format(
                self.mode, self.start_index, self.end_index, self.args
            )
        # end def

        @property
        def end_index (self):
            return "{}+{}c".format(self.start_index, self.text_length())
        # end def

        def text_length (self):
            return sum(map(len, self.args[::2]))
        # end def

    # end class

    def __init__ (self, limit=200):
        self.limit = limit
        self.reset()
    # end def

    def add_separator (self):
        if not self or self[-1] is not self.SEPARATOR:
            self.append(self.SEPARATOR)
        # end if
    # end def

    def append (self, element):
        if self.current_index >= 0:
            del self[self.current_index + 1:]
        else:
            self.clear()
        # end if
        if self.limit > 1:
            del self[:1 - self.limit]
        elif self.limit == 1:
            self.clear()
        # end if
        if len(self) and hasattr(element, "mode"):
            _last = self[-1]
            if hasattr(_last, "mode") and _last.mode != element.mode:
                super().append(self.SEPARATOR)
            # end if
        # end if
        super().append(element)
        self.update_index()
    # end def

    def get_redo_elements (self):
        _sequence = []
        self.current_index += 1
        _ci = self.trap(self.SEPARATOR, +1)
        if _ci < len(self):
            if self.SEPARATOR not in self[_ci + 1:]:
                _sequence = self[_ci:]
                self.current_index = len(self) - 1
            else:
                _sep = self[_ci + 1:].index(self.SEPARATOR)
                _sequence = self[_ci:_ci + _sep + 1]
                self.current_index = _ci + _sep + 1
            # end if
        # end if
        self.rebind_index()
        return _sequence
    # end def

    def get_undo_elements (self):
        _sequence = []
        _ci = self.trap(self.SEPARATOR, -1)
        if _ci >= 0:
            if self.SEPARATOR not in self[:_ci]:
                _sequence = self[:_ci + 1]
                self.current_index = -1
            else:
                _sep = self[_ci::-1].index(self.SEPARATOR)
                _sequence = self[_ci - _sep + 1:_ci + 1]
                self.current_index = _ci - _sep
            # end if
        # end if
        return reversed(_sequence)
    # end def

    def pop (self):
        super().pop()
        self.update_index()
    # end def

    def push_delete (self, index, chars, *args):
        if chars or args:
            self.append(self.Element("-", index, chars, *args))
        # end if
    # end def

    def push_insert (self, index, chars, *args):
        if chars or args:
            self.append(self.Element("+", index, chars, *args))
        # end if
    # end def

    def rebind_index (self):
        self.current_index = (
            max(0, min(len(self) - 1, self.current_index))
        )
    # end def

    def reset (self):
        self.clear()
        self.update_index()
    # end def

    def trap (self, element, sx=1):
        _len = len(self)
        _ci = self.current_index
        if sx and 0 <= _ci < _len:
            while self[_ci] is element:
                _ci += sx
                if not (0 <= _ci < _len):
                    break
                # end if
            # end while
        # end if
        self.current_index = max(-1, min(_len, _ci))
        return self.current_index
    # end def

    def update_index (self):
        self.current_index = len(self) - 1
        return self.current_index
    # end def

# end class
