#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    tkScenarist - screen writing made simpler

    Copyright (c) 2014+ Raphaël Seban <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see http://www.gnu.org/licenses/
"""

from tkinter import Toplevel


class TextCompletionPopup (Toplevel):

    def __init__ (self, w_text, on_insert=None):
        super().__init__()
        self.w_text = w_text
        self.insert_callback = on_insert
        self.withdraw()
        self.transient(w_text)
        self.overrideredirect(True)
        self.initialize()
    # end def

    def bind_events (self, *args, **kw):
        self.bind("<Key>", self.on_keypress)
        self.w_listbox.bind(
            "<<ListboxSelect>>", self.on_listbox_select
        )
        self.w_listbox.bind("<Double-1>", self.on_listbox_double_click)
    # end def

    def fill_listbox (self, choices):
        choices = tuple(choices or ())
        if choices and choices != self.w_listbox.get(0, "end"):
            self.current_index = 0
            self.w_listbox.delete(0, "end")
            self.w_listbox.insert("end", *choices)
            self.w_listbox.configure(
                height=min(5, len(choices)),
                width=min(40, max(map(len, choices)))
            )
        # end if
        self.w_listbox.selection_clear(0, "end")
        try:
            self.w_listbox.selection_set(self.current_index)
            self.w_listbox.see(self.current_index)
        except:
            self.w_listbox.selection_set(0)
            self.w_listbox.see(0)
            self.current_index = 0
        # end try
    # end def

    def hide (self, *args, **kw):
        self.unroute_events()
        self.withdraw()
    # end def

    def initialize (self, *args, **kw):
        import os
        from tkinter import Listbox, Scrollbar
        self.w_listbox = Listbox(
            self,
            font="courier 12",
            background="lemon chiffon",
            foreground="grey30",
            selectbackground="grey30",
            selectforeground="lemon chiffon",
            highlightthickness=0,
            borderwidth=1,
            relief="solid",
        )
        self.w_scrollbar = Scrollbar(
            self, orient="vertical", relief="flat", width=5,
        )
        self.w_listbox.configure(yscrollcommand=self.w_scrollbar.set)
        self.w_scrollbar.configure(command=self.w_listbox.yview)
        self.w_listbox.pack(side="left", expand=1, fill="both")
        self.w_scrollbar.pack(side="right", expand=1, fill="y")
        self.win32 = bool(os.name.lower() == "nt")
        self.start_index = None
        self.current_index = 0
        self.bind_events(*args, **kw)
    # end def

    def on_insert_selection (self, *args, **kw):
        try:
            self.insert_callback(
                self.w_listbox.get(self.w_listbox.curselection()[0]),
                self.start_index
            )
        except:
            pass
        # end try
        self.hide()
        self.w_text.focus_set()
    # end def

    def on_key_arrows (self, key, *args, **kw):
        try:
            _index = (
                self.w_listbox.curselection()[0]
                + (key == "Down") - (key == "Up")
            )
        except:
            _index = 0
        # end try
        _index = max(0, min(_index, self.w_listbox.size() - 1))
        self.current_index = _index
        self.w_listbox.selection_clear(0, "end")
        self.w_listbox.selection_set(_index)
        self.w_listbox.see(_index)
    # end def

    def on_keypress (self, event=None, *args, **kw):
        _key = event.keysym
        if _key == "Escape":
            self.hide()
        elif _key in ("Up", "Down"):
            self.on_key_arrows(_key)
        elif _key in ("Tab", "Return"):
            self.on_insert_selection()
        else:
            return None
        # end if
        return "break"
    # end def

    def on_listbox_double_click (self, *args, **kw):
        self.on_insert_selection()
    # end def

    def on_listbox_select (self, *args, **kw):
        self.current_index = self.w_listbox.curselection()[0]
    # end def

    def route_events (self, *args, **kw):
        self.w_text.bindtags((str(self),) + self.w_text.bindtags())
    # end def

    def show (self, *args, **kw):
        # ensure no tkinter.messagebox is up there
        try:
            if self.grab_current():
                return self.hide()
            # end if
        except:
            return self.hide()
        # end try
        if self.w_text.winfo_viewable():
            self.fill_listbox(kw.get("choices"))
            _start_index = kw.get("start_index") or "insert"
            try:
                _x, _y, _w, _h = self.w_text.bbox(_start_index)
                _xi, _yi, _wi, _hi = self.w_text.bbox("insert")
            except:
                pass
            else:
                self.start_index = _start_index
                _x += self.w_text.winfo_rootx()
                _y = self.w_text.winfo_rooty() + _h + max(_y, _yi)
                self.geometry("+{}+{}".format(_x, _y))
                self.deiconify()
                self.route_events(*args, **kw)
                if self.win32:
                    self.lift(self.w_text)
                # end if
            # end try
        else:
            self.hide()
        # end if
    # end def

    def unroute_events (self, *args, **kw):
        try:
            _tags = list(self.w_text.bindtags())
            _tags.remove(str(self))
            self.w_text.bindtags(tuple(_tags))
        except:
            pass
        # end try
    # end def

# end class
