#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Library main package.

Current implementation does the followings when merely importing
package by name i.e. `import tkRAD`:

* set up i18n language translation support;
* make `RADApplication`, `RADMainWindow`, `RADXMLMainWindow`,
  `RADWindow`, `RADXMLWindow` and `RADXMLFrame` classes directly
  available from here.

Example 1::

    import tkRAD
    tkRAD.RADApplication().run()

Example 2::

    import tkRAD
    xml_source = \"""
        <tkwidget>
            <label
                text="Hello, good people!"
                layout="pack"
                resizable="yes"
            />
            <button
                text="Quit"
                command="@quit"
                layout="pack"
            />
        </tkwidget>
    \"""
    tkRAD.RADApplication().run(xml=xml_source)

Example 3::

    import tkRAD
    xml_source = \"""
        <tkwidget>
            <tkmenu>
                <menu label="_File">
                    <command
                        label="Quit"
                        accelerator="Ctrl+Q"
                        command="@quit"
                    />
                </menu>
                <menu label="_View">
                    <checkbutton
                        label="Show/hide statusbar"
                        command="@ToggleStatusbar"
                        variable="show_statusbar"
                    />
                </menu>
            </tkmenu>
            <label
                text="Hello, good people!"
                layout="pack"
                resizable="yes"
            />
        </tkwidget>
    \"""
    tkRAD.RADApplication().run(xml=xml_source)

Example 4::

    import tkRAD
    xml_source = \"""
        <tkwidget>
            <tkmenu>
                <menu label="_File">
                    <command
                        label="Quit"
                        accelerator="Ctrl+Q"
                        command="@quit"
                    />
                </menu>
                <menu label="_View">
                    <checkbutton
                        label="Show/hide statusbar"
                        command="@ToggleStatusbar"
                        variable="show_statusbar"
                    />
                </menu>
            </tkmenu>
            <label
                text="Hello, good people!"
                layout="pack"
                resizable="yes"
            />
        </tkwidget>
    \"""
    w = tkRAD.RADXMLMainWindow()
    w.xml_build(xml_source)
    w.connect_statusbar("show_statusbar")
    w.run()

Example 5::

    import tkRAD
    xml_source = \"""
        <tkwidget>
            <tkmenu>
                <menu label="_File">
                    <command
                        label="Quit"
                        accelerator="Ctrl+Q"
                        command="@quit"
                    />
                </menu>
                <menu label="_View">
                    <checkbutton
                        label="Show/hide statusbar"
                        command="@ToggleStatusbar"
                        variable="show_statusbar"
                    />
                </menu>
            </tkmenu>
            <label
                text="Hello, good people!"
                layout="pack"
                resizable="yes"
            />
        </tkwidget>
    \"""
    w = tkRAD.RADMainWindow()
    w.mainframe = tkRAD.RADXMLFrame(w)
    w.xml_build(xml_source)
    w.connect_statusbar("show_statusbar")
    w.run()

Example 6::

    import tkRAD
    xml_source = \"""
        <tkwidget>
            <tkmenu>
                <menu label="_File">
                    <command
                        label="Quit"
                        accelerator="Ctrl+Q"
                        command="@quit"
                    />
                </menu>
                <menu label="_View">
                    <checkbutton
                        label="Show/hide statusbar"
                        command="@ToggleStatusbar"
                        variable="show_statusbar"
                    />
                </menu>
            </tkmenu>
            <label
                text="Hello, good people!"
                layout="pack"
                resizable="yes"
            />
        </tkwidget>
    \"""
    w = tkRAD.RADXMLWindow(hide_tk=True)
    w.xml_build(xml_source)
    w.connect_statusbar("show_statusbar")
    w.run()

Example 7::

    import tkRAD
    xml_source = \"""
        <tkwidget>
            <tkmenu>
                <menu label="_File">
                    <command
                        label="Quit"
                        accelerator="Ctrl+Q"
                        command="@quit"
                    />
                </menu>
                <menu label="_View">
                    <checkbutton
                        label="Show/hide statusbar"
                        command="@ToggleStatusbar"
                        variable="show_statusbar"
                    />
                </menu>
            </tkmenu>
            <label
                text="Hello, good people!"
                layout="pack"
                resizable="yes"
            />
        </tkwidget>
    \"""
    w = tkRAD.RADWindow(hide_tk=True)
    w.mainframe = tkRAD.RADXMLFrame(w)
    w.xml_build(xml_source)
    w.connect_statusbar("show_statusbar")
    w.run()

For more information about what's on / what's off with this package,
simply type in a Python interactive console::

    import tkRAD
    print(dir(tkRAD)) # show all accessable defs

*NEW* since **tkRAD v1.7**:

Many portions of code had to be entirely rewritten in many modules of
all library packages.

Please, feel free to post an issue ticket on GitHub's project
bugtracker if you face any troubleshooting with this new version of
tkRAD's library.

GitHub's project bugtracker: https://github.com/tarball69/tkRAD/issues
(requires a GitHub member account to proceed).
"""

"""
    tkRAD - tkinter Rapid Application Development library

    (c) 2013+ Raphaël SEBAN <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public
    License along with this program.

    If not, see: http://www.gnu.org/licenses/
"""

import warnings
import functools

warnings.simplefilter("default")


def _deprecated (msg):
    """Application-wide @deprecated function decorator.

    This is a decorator which can be used to mark functions as
    deprecated. It will result in a warning being emitted when the
    function is used.

    Requires additional 'msg' message for coders, e.g.::

        @deprecated("Please, use 'attr' class attribute instead.")
        def old_function_to_mark_as_deprecated (...):
    """
    def wrapper (func):
        @functools.wraps(func)
        def new_func(*args, **kw):
            warnings.warn_explicit(
                "Call to deprecated attribute or function '{}'. "
                "May be removed on later versions of tkRAD library. {}"
                .format(func.__name__, msg),
                category=DeprecationWarning,
                filename=func.__code__.co_filename,
                lineno=func.__code__.co_firstlineno + 1
            )
            return func(*args, **kw)
        # end def
        return new_func
    # end def
    return wrapper
# end def


# set @deprecated("message") decorator to be app-wide
try:
    __builtins__["deprecated"] = _deprecated
except:
    __builtins__.deprecated = _deprecated
# end try


# set i18n support by default
from .core import i18n

i18n.install()

# restrict to daily use classes
from .widgets.rad_application import RADApplication
from .widgets.rad_mainwindow import RADMainWindow
from .xml.rad_xml_mainwindow import RADXMLMainWindow
from .widgets.rad_window import RADWindow
from .xml.rad_xml_window import RADXMLWindow
from .xml.rad_xml_frame import RADXMLFrame
