#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Asynchronous task module.

Implements:

* module function get_async_manager -- retrieve an app-wide async
  manager (unique instance).
* class AsyncTask -- asynchronous task management class for tkinter GUI
  lib.
* class AsyncTaskError -- error handler for AsyncTask class.

**Caution:** tkRAD's asynchronous task maechanism relies on tkinter's
asynchronous task maechanism i.e. `w.after()`, `w.after_idle()` methods
and should *NOT* be confused with Python's **threading.Thread**
maechanism.

The goal of this module is to make tkinter's async tasks **safer** by
using a locker maechanism that ensures a same function/method could
*NOT* be called several times while still executing at the first call.

If you need real Thread objects, please, refer to Python's official
documentation:

https://docs.python.org/3/library/threading.html

Example::

    import tkinter as tk
    from tkRAD.core.async import get_async_manager

    def parse_text ():
        # put some time-consuming task here
        print("Parsing text...")
    # end def

    def on_keypress (event=None):
        # we want this to happen only once user has stopped typing
        async.run_after(500, parse_text)
    # end def

    root = tk.Tk()
    # put this only once a Tk() is created
    async = get_async_manager()
    text = tk.Text(root, width=20, height=5)
    text.bind("<Key>", on_keypress)
    text.pack()
    root.mainloop()

*NEW* since **tkRAD v1.7**:

* 'TK' module alias name has been renamed to 'tk' to best comply with
  Python's community conventions;
* since now, may raise **AsyncTaskError** if an object instance of
  *AsyncTask* is initialized *BEFORE* any existing *Tk()* object
  instance.

Please, feel free to post an issue ticket on GitHub's project
bugtracker if you face any troubleshooting with this new version of
tkRAD's library.

GitHub's project bugtracker: https://github.com/tarball69/tkRAD/issues
(requires a GitHub member account to proceed).
"""

"""
    tkRAD - tkinter Rapid Application Development library

    (c) 2013+ Raphaël SEBAN <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see: http://www.gnu.org/licenses/
"""

try:
    import tkinter as tk
except:
    import Tkinter as tk
# end try


__async = None


def get_async_manager ():
    """Retrieve an app-wide async manager (unique instance).

    Use this function whenever you need only one unique async manager
    object instance over all your app code.

    For several private-shared managers, merely instantiate AsyncTask
    class instead.
    """
    global __async
    if not isinstance(__async, AsyncTask):
        __async = AsyncTask()
    # end if
    return __async
# end def


class AsyncTask:
    """Asynchronous task management class for tkinter GUI lib."""

    # locker states
    STATE_LOCKED = 1
    STATE_ATOMIC = 2

    def __init__ (self):
        """Class constructor."""
        self.tid = dict()
        self.lockers = dict()
        self.root = tk._default_root
        if not self.root:
            raise AsyncTaskError(
                "Cannot initialize an AsyncTask object instance "
                "*BEFORE* any existing Tk() object instance."
            )
        # end if
    # end def

    def _atomic (self, callback, *args):
        """Atomic call for callback.

        Parameters:
            callback -- callback to run in atomic mode.

            args -- ordered list of arguments to pass to callback on
            atomic call.

        Ensures 'callback' runs in atomic mode i.e. is neither
        interrupted nor called several times.
        """
        if not self.lockers.get(callback):
            self.lockers[callback] = self.STATE_ATOMIC
            callback(*args)
            if self.lockers.get(callback) == self.STATE_ATOMIC:
                self.lockers.pop(callback, None)
            # end if
        # end if
    # end def

    def clear_all (self, *args, **kw):
        """Clear all registered tasks.

        Event handler signature.

        Stops all scheduled tasks and clears all registered lockers.

        This method has the same effect than self.release_all().
        """
        self.release_all(*args, **kw)
    # end def

    def lock (self, *callbacks):
        """Lock listed callbacks.

        Parameter:
            callbacks -- ordered list of callbacks to lock in.

        Stops each callback in 'callbacks', if scheduled.

        Locks each callback in 'callbacks' to prevent from further
        calls, even if not previously scheduled.
        """
        for _cb in callbacks:
            self.stop(_cb)
            self.lockers[_cb] = self.STATE_LOCKED
        # end for
    # end def

    def lock_all (self, *args, **kw):
        """Stop and lock all scheduled tasks.

        Event handler signature.
        """
        self.lock(*self.tid.keys())
    # end def

    def release (self, *callbacks):
        """Release lockers.

        Parameter:
            callbacks -- ordered list of callbacks to release.

        Releases locker for each callback in 'callbacks'.

        This also frees up callback reference for each and every
        callback in 'callbacks'.
        """
        for _cb in callbacks:
            self.stop(_cb)
            self.lockers.pop(_cb, None)
        # end for
    # end def

    def release_all (self, *args, **kw):
        """Release all lockers for all tasks.

        Event handler signature.

        Stops all scheduled tasks and clears all registered lockers.

        This method has the same effect than self.clear_all().
        """
        self.stop_all(*args, **kw)
        self.lockers.clear()
    # end def

    def run_after (self, delay, callback, *args):
        """Schedule a new task.

        Parameters:
            delay -- time delay in milliseconds (integer).

            callback -- function or method to call back in time.

            args -- ordered list of arguments to pass to callback on
            future call.

        Schedules a new task to fire up after 'delay' milliseconds,
        calling then 'callback' with 'args' arguments list.

        A repeated call to this method re-schedule previous scheduled
        identical task that has not yet been fired up, making a call to
        this task a unique call, even if this call might somewhat be
        delayed in time.
        """
        delay = max(1, int(delay))
        self.stop(callback)
        self.tid[callback] = self.root.after(
            delay, self._atomic, callback, *args
        )
    # end def

    def run_after_idle (self, callback, *args):
        """Schedule a new task after mainloop enters in idle mode.

        Parameters:
            callback -- function or method to call back in time.

            args -- ordered list of arguments to pass to callback on
            future call.

        Schedules a new task to fire up after tkinter's mainloop enters
        in idle mode i.e. when all tkinter events are done, calling
        then 'callback' with 'args' arguments list.

        A repeated call to this method re-schedule previous scheduled
        identical task that has not yet been fired up, making a call to
        this task a unique call, even if this call might somewhat be
        delayed in time.
        """
        self.stop(callback)
        self.tid[callback] = self.root.after_idle(
            self._atomic, callback, *args
        )
    # end def

    def stop (self, *callbacks):
        """Stop scheduled tasks.

        Parameter:
            callbacks -- ordered list of callbacks to stop.

        Stops scheduled task for each callback in 'callbacks' and then
        clears up callback reference from scheduled tasks.
        """
        for _cb in callbacks:
            self.root.after_cancel(self.tid.get(_cb) or 0)
            self.tid.pop(_cb, None)
        # end for
    # end def

    def stop_all (self, *args, **kw):
        """Stop all scheduled tasks.

        Event handler signature.

        Stops and then clears up all scheduled tasks and all callback
        references in scheduled tasks.
        """
        for _tid in self.tid.values():
            self.root.after_cancel(_tid)
        # end for
        self.tid.clear()
    # end def

# end class


class AsyncTaskError (Exception):
    """Error handler for AsyncTask class."""
    pass
# end class
