#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Application checkups module.

Implements:

* module function check_directories -- check for application required
  directories.
* module function parse_version -- parse a version string of chars.
* module function python_require -- verify application user's Python
  interpreter version.

You may find an example into the source code of
*tkRAD.widgets.rad_application* module, at
*RADApplication.initialize()* method beginning.

Short sample::

        import inspect
        import os
        from ..core import checkups
        from ..core import options
        from ..core import path
        from ..core import services
        # check required Python version
        _python = {
            "version": tools.choose_str(
                kw.get("python_version"),
                self.PYTHON.get("version"),
                "3.2",
            ),
            "strict": tools.choose(
                kw.get("python_strict"),
                self.PYTHON.get("strict"),
                False,
            ),
        }
        checkups.python_require(**_python)
        # set application's root directory
        path.set_app_root_dir(
            tools.choose_str(
                kw.get("app_root_dir"),
                inspect.stack()[-1][1],
                ".",
            )
        )
        _cwd = path.get_app_root_dir()
        os.chdir(_cwd)
        # check required directories
        if not kw.get("no_dependencies"):
            checkups.check_directories(
                _cwd, *(kw.get("check_dirs") or self.DIRECTORIES or ())
            )
        # end if

*NEW* since **tkRAD v1.7**:

Please, feel free to post an issue ticket on GitHub's project
bugtracker if you face any troubleshooting with this new version of
tkRAD's library.

GitHub's project bugtracker: https://github.com/tarball69/tkRAD/issues
(requires a GitHub member account to proceed).
"""

"""
    tkRAD - tkinter Rapid Application Development library

    (c) 2013+ Raphaël SEBAN <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see: http://www.gnu.org/licenses/
"""


def check_directories (root_dir, *dir_list):
    """Check for application required directories.

    Parameters:
        root_dir -- application's root directory path.

        dir_list -- ordered list of relative paths to application
        required directories.

    Verifies if a given list of application required directories do
    exist.

    May raise **FileNotFoundError** on first missing required
    directory.

    Returns True on success, False otherwise.
    """
    import os.path
    root_dir = os.path.abspath(
        os.path.realpath(os.path.expanduser(root_dir))
    )
    for _dir in set(dir_list):
        _dir = os.path.join(root_dir, _dir)
        if not os.path.isdir(_dir):
            raise FileNotFoundError(
                "Required directory '{reqdir}' could not be found."
                .format(reqdir=_dir)
            )
            return False
        # end if
    # end for
    return True
# end def


def parse_version (version, get_string=False):
    """Parse a version string of chars.

    Parameters:
        version -- version string of chars to parse.

        get_string -- flag for return value (default: False).

    Tries to parse a version string and put version numbers into a
    hashable Python tuple() object for best comparison process.

    This object is limited to *at most* (major, minor, release) version
    numbers.

    If 'get_string' flag is omitted or False, function will return the
    hashable tuple() object.

    If 'get_string' flag is True, function will return a cleaned-up
    string version of 'version' string of chars.
    """
    import re
    _version = re.sub(r"\.+", ".", str(version))
    _version = re.sub(r"[^0-9\.].*$", "", _version).strip(".")
    if get_string:
        return _version
    # end if
    return tuple(map(int, ("0" + _version).split(".")))[:3]
# end def


def python_require (version, strict=False):
    """Verify application user's Python interpreter version.

    Parameters:
        version -- application required Python interpreter version
        (string of chars).

        strict -- boolean flag for strict version comparison (default:
        False). If omitted or False, user's Python interpreter version
        will be compared as a loose version. Strict comparison
        otherwise.

    Verifies if user's Python interpreter version matches at least
    required 'version' when 'strict' is False (loose version) or
    compares strictly otherwise.

    May raise **SystemError** on version mismatch and application will
    exit with status 1 i.e. will stop for fatal error, in this case.
    """
    import platform
    _req_version = str(version)
    _cur_version = str(platform.python_version())
    # compare strict version
    if strict and (_req_version != _cur_version):
        raise SystemError(
            "This program expects *STRICT* Python "
            "version '{vreq}' while current version is '{vcur}'."
            .format(vreq=_req_version, vcur=_cur_version)
        )
        exit(1)
    # compare loose version
    elif parse_version(_req_version) > parse_version(_cur_version):
        raise SystemError(
            "This program expects *AT LEAST* Python "
            "version '{vreq}' while current version is '{vcur}'."
            .format(
                vreq=parse_version(_req_version, get_string=True),
                vcur=_cur_version
            )
        )
        exit(1)
    # end if
# end def
