#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""SQLite3 database management module.

Implements:

* module function get_database -- retrieve an app-wide database manager
  (unique instance).
* class Database -- SQLite3 database management class.
* class DatabaseError -- error handler for Database class.

Example 1::

    from tkRAD.core.database import get_database
    db = get_database(db_path=":memory:")
    db.sql_script(\"""
        CREATE TABLE 'Books' (
            'ID' INTEGER PRIMARY KEY,
            'Title' TEXT,
            'Author' TEXT,
            'ISBN' TEXT,
            'Pages' INTEGER
        );
        INSERT INTO 'Books' VALUES
            (NULL,
                'Tkinter GUI Application Development Blueprints',
                'Bhaskar Chaudhary', '9781785889738', '340'),
            (NULL,
                'Python Data Science Cookbook',
                'Gopi Subramanian', '9781784396404', '438'),
            (NULL,
                'Python Machine Learning',
                'Sebastian Raschka', '9781783555130', '454')
        ;
        VACUUM;
    \""")
    db.dump_tables("Books")

Example 2::

    from tkRAD.core.database import get_database
    db = get_database(db_path=":memory:")
    db.sql_script(\"""
        CREATE TABLE 'Books' (
            'ID' INTEGER PRIMARY KEY,
            'Title' TEXT,
            'Author' TEXT,
            'ISBN' TEXT,
            'Pages' INTEGER
        );
        INSERT INTO 'Books' VALUES
            (NULL,
                'Tkinter GUI Application Development Blueprints',
                'Bhaskar Chaudhary', '9781785889738', '340'),
            (NULL,
                'Python Data Science Cookbook',
                'Gopi Subramanian', '9781784396404', '438'),
            (NULL,
                'Python Machine Learning',
                'Sebastian Raschka', '9781783555130', '454')
        ;
        VACUUM;
    \""")
    with open("dump_tables.log", "w") as output_file:
        db.dump_tables("Books", output=output_file)
    # end with

*NEW* since **tkRAD v1.7**:

* added new *_SPECIAL_PATH* protected class member attribute;
* added new *normalize_path()* method;
* added support for SQLite3's special ":memory:" database path into
  memory instead of a disk file.

Please, feel free to post an issue ticket on GitHub's project
bugtracker if you face any troubleshooting with this new version of
tkRAD's library.

GitHub's project bugtracker: https://github.com/tarball69/tkRAD/issues
(requires a GitHub member account to proceed).
"""

"""
    tkRAD - tkinter Rapid Application Development library

    (c) 2013+ Raphaël SEBAN <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see: http://www.gnu.org/licenses/
"""

import sqlite3

try:
    from . import path
except:
    import path
# end try


__database = None


def get_database (**kw):
    """Retrieve an app-wide database manager (unique instance).

    Parameter:
        kw -- keyword arguments.

    Supported keyword arguments:
        auto_commit -- boolean flag (default: True). If omitted or
        True, let the database manager decide on transaction commits
        automatically (recommended). Disable feature otherwise.

        db_path -- database file path (default: Database.DEFAULT_PATH).

    Use this function whenever you need only one unique database
    manager object instance over all your app code.

    For several database managers i.e. one for each database used by
    application, merely instantiate Database class instead.
    """
    global __database
    if not isinstance(__database, Database):
        __database = Database(**kw)
    # end if
    return __database
# end def


class Database:
    """SQLite3 database management class."""

    ALL = "all"
    DEFAULT_PATH = "^/data/sqlite3/database.db"
    _SPECIAL_PATH = ":memory:"

    def __init__ (self, *args, **kw):
        """Class constructor.

        Parameter:
            args -- variable list of arguments.

            kw -- keyword arguments.

        Supported keyword arguments:
            auto_commit -- boolean flag (default: True). If omitted or
            True, let the database manager decide on transaction
            commits automatically (recommended). Disable feature
            otherwise.

            db_path -- database file path (default: DEFAULT_PATH).
        """
        self._SPECIAL_PATH = str(self._SPECIAL_PATH).lower()
        self.auto_commit = bool(kw.get("auto_commit", True))
        self.db_path = kw.get("db_path") or self.DEFAULT_PATH
        self.connection = None
        self.cursor = None
        self.initialize(*args, **kw)
    # end def

    def close_database (self, *args, **kw):
        """Close current database connection.

        Event handler signature.

        May raise **DatabaseError** when called with no available
        connection.
        """
        if self.connection:
            if self.auto_commit:
                self.commit()
            # end if
            self.connection.close()
            self.connection = None
            self.cursor = None
        else:
            raise DatabaseError(
                "Could not close database: "
                "no available connection by now."
            )
        # end if
    # end def

    def commit (self, *args, **kw):
        """Commit transaction in database.

        Event handler signature.

        May raise **DatabaseError** when called with no available
        connection.
        """
        if self.connection:
            self.connection.commit()
        else:
            raise DatabaseError(
                "Could not commit current transaction in database: "
                "no available connection by now."
            )
        # end if
    # end def

    @property
    def db_path (self):
        """Database file path - read-write attribute.

        Database's file path may be normalized along with tkRAD's hat
        '^' notation i.e. replace starting hat '^' in file path by
        application's root directory.

        *NEW* since **tkRAD v1.7**:

        * added support for ":memory:" special database path;
        * removed exception error: now *db_path* class attribute can be
          set to empty string at coder's choice (None will be converted
          to an empty string implicitly).
        """
        return self.__db_path
    # end def

    @db_path.setter
    def db_path (self, value):
        self.__db_path = self.normalize_path(value)
    # end def

    def dump_data (self, table_or_view, limit=100, formatter=None,
                                                        output=None):
        """Dump extracted data into output file object.

        Parameters:
            table_or_view -- database table name or view name (string
            of chars).

            limit -- integer number of rows to dump out (default: 100).
            Useful for fixing a limited number of rows when dumping
            large tables or views.

            formatter -- callback data formatter for extracted rows
            (default: None). If omitted or None, will be replaced by
            'self.dump_data_formatter'. This callback should expect
            'table_or_view' and 'rows' arguments.

            output -- output file to write dumped data in (default:
            None). If omitted or None, will be replaced by 'sys.stdout'
            (console output).

        Dumps all data extracted from 'table_or_view', up to 'limit'
        rows, formatted by 'formatter' callback and written to 'output'
        file.

        This method is merely a debug utility procedure.
        """
        if not callable(formatter):
            formatter = self.dump_data_formatter
        # end if
        print(
            "\nDumping data from '{tov}' limited to max. {lim} rows\n"
            .format(tov=table_or_view, lim=limit),
            file=output,
        )
        _rows = self.get_all(table_or_view, limit)
        if _rows:
            print(formatter(table_or_view, _rows), file=output)
        else:
            print("This table/view is *EMPTY*.", file=output)
        # end if
    # end def

    def dump_data_formatter (self, table_or_view, rows):
        """Generic data formatter.

        Parameters:
            table_or_view -- database table name or view name (string
            of chars).

            rows -- list of data rows. Must be consistently extracted
            from 'table_or_view'.

        Tries to generate a fancy table representation of extracted
        data with decorated columns and rows.

        Returns result as a unique formatted string of chars for all
        data extracted from 'rows' list.
        """
        _line_sep = lambda width: "+-{}-+\n".format("-" * width)
        _line_start = "| "
        _line_end = " |"
        _get_line = lambda data, align="": (
            "{s}{d}{e}\n".format(
                s=_line_start,
                d=_field_sep.join(
                    [
                        "{{v:{a}{w}}}"
                        .format(a=align, w=_widths[_i])
                        .format(v=_v)
                        for _i, _v in enumerate(data)
                    ]
                ),
                e=_line_end
            )
        )
        _field_sep = " | "
        _fields = self.get_column_names()
        _widths = [len(_f) for _f in _fields]
        for _row in rows:
            for _index, _field in enumerate(_row):
                _widths[_index] = (
                    max(_widths[_index], len(str(_field)))
                )
            # end for
        # end for
        _line_width = int(
            sum(_widths) + len(_field_sep) * (len(_fields) - 1)
        )
        _fstring = _line_sep(_line_width)
        _fstring += (
            "{s}{{t:^{w}}}{e}\n"
            .format(s=_line_start, w=_line_width, e=_line_end)
            .format(t=table_or_view)
        )
        _fstring += _line_sep(_line_width)
        _fstring += _get_line(_fields, align="^")
        _fstring += _line_sep(_line_width)
        for _row in rows:
            _fstring += _get_line(_row)
            _fstring += _line_sep(_line_width)
        # end for
        return _fstring
    # end def

    def dump_tables (self, *args, limit=100, formatter=None,
                                                        output=None):
        """Dump listed tables or views into an output file.

        Parameters:
            args -- ordered list of database table names and/or view
            names (list of strings of chars).

            limit -- integer number of rows to dump out (default: 100).
            Useful for fixing a limited number of rows when dumping
            large tables or views.

            formatter -- callback data formatter for extracted rows
            (default: None). If omitted or None, will be replaced by
            'self.dump_data_formatter'. This callback should expect
            'table_or_view' and 'rows' arguments.

            output -- output file object to write dumped data in
            (default: None). If omitted or None, will be replaced by
            'sys.stdout' (console output).

        Dumps an ordered list of database table names and/or view
        names, up to 'limit' number of rows, formatted by 'formatter'
        callback and written to 'output' file object.

        This method is merely a debug utility procedure.
        """
        for _table_or_view in args:
            self.dump_data(_table_or_view, limit, formatter, output)
        # end for
    # end def

    def fetch (self, qty=1, default=[]):
        """Fetch rows from a resulting query.

        Parameters
            qty -- number of rows to fetch out from query result
            (default: 1). This parameter can get the following values:

            * "all" -- fetches all available rows;
            * None -- fetches as many remaining rows as possible;
            * integer < 2 -- fetches only one row, if available;
            * integer >= 2 -- fetches *at most* 'integer' available
              rows.

            default -- default value to return instead of empty result
            or when there are no more available rows (default: []).

        Fetches at most 'qty' available rows from a resulting query and
        returns a list of these rows or 'default' value when no more
        rows are available at the moment of call of this method.

        May raise **DatabaseError** when called with no available
        database cursor e.g. closed database.
        """
        if self.cursor:
            if qty is None:
                return self.cursor.fetchmany() or default
            elif str(qty).lower() == self.ALL:
                return self.cursor.fetchall() or default
            elif qty < 2:
                return self.cursor.fetchone() or default or None
            else:
                return self.cursor.fetchmany(qty) or default
            # end if
        else:
            raise DatabaseError(
                "Could not fetch data: no available cursor by now."
            )
        # end if
    # end def

    def finalize (self, *args, **kw):
        """Finalize database.

        Parameter:
            args -- variable list of arguments.

            kw -- keyword arguments.

        Supported keyword arguments:
            -- none at this stage --

        Call this hook method on closing application.

        Current implementation merely calls *close_database()* method.

        Feel free to extend or to override this hook method in your own
        subclasses to best meet your needs.

        *NEW* since **tkRAD v1.7**.
        """
        self.close_database(*args, **kw)
    # end def

    def get_all (self, table_name, limit=0):
        """Fetch all rows from table.

        Parameters:
            table_name -- database table name (string of chars).

            limit -- integer number of rows to fetch (default: 0). If
            omitted or less than 1, stands for fetching **all** rows,
            unlimited number.

        Returns a list of all rows fetched from 'table_name', either up
        to 'limit' number of rows (when 'limit' >= 1) or up to an
        unlimited number of rows (when 'limit' < 1).
        """
        if limit < 1:
            self.sql_query(
                "SELECT * FROM '{}'".format(table_name)
            )
        else:
            self.sql_query(
                "SELECT * FROM '{}' LIMIT {}".format(table_name, limit)
            )
        # end if
        return self.fetch(self.ALL)
    # end def

    def get_column_names (self):
        """Get field names of the last SQL query, if any.

        Returns a tuple list of all field (column) names extracted from
        the last SQL query, if a resulting query.

        May raise **DatabaseError** when called with no available
        database cursor e.g. closed database.
        """
        if self.cursor:
            return tuple(c[0] for c in self.cursor.description or "")
        else:
            raise DatabaseError(
                "Could not retrieve column names: "
                "no available cursor by now."
            )
        # end if
    # end def

    def get_record (self, table_name, row_id):
        """Get a unique record from table.

        Parameters:
            table_name -- database table name (string of chars).

            row_id -- row identifier, as stated for SQLite3 'ROWID'
            field. See:
            https://www.sqlite.org/lang_createtable.html#rowid

        Retrieves from 'table_name' the unique row that matches
        'row_id'.

        Returns that row, if available, returns None otherwise.
        """
        self.sql_query(
            "SELECT * FROM '{}' WHERE ROWID = ? LIMIT 1"
            .format(table_name),
            row_id
        )
        return self.fetch(1)
    # end def

    @deprecated("Please, use 'initialize()' method instead.")
    def init_database (self, **kw):
        """Hook method.

        Parameter:
            kw -- keyword arguments.

        Supported keyword arguments:
            auto_commit -- boolean flag (default: True). If omitted or
            True, let the database manager decide on transaction
            commits automatically (recommended). Disable feature
            otherwise.

            db_path -- database file path (default: DEFAULT_PATH).

        This hook method is called after *open_database()* method.

        Current implementation does nothing.

        *DEPRECATED* since **tkRAD v1.7**.

        Please, use *initialize()* hook method instead.
        """
        pass
    # end def

    @deprecated("Please, use 'initialize()' method instead.")
    def init_members (self, **kw):
        """Hook method.

        Parameter:
            kw -- keyword arguments.

        Supported keyword arguments:
            auto_commit -- boolean flag (default: True). If omitted or
            True, let the database manager decide on transaction
            commits automatically (recommended). Disable feature
            otherwise.

            db_path -- database file path (default: DEFAULT_PATH).

        This hook method is called before *open_database()* method.

        Current implementation does nothing.

        *DEPRECATED* since **tkRAD v1.7**.

        Please, use *initialize()* hook method instead.
        """
        pass
    # end def

    def initialize (self, *args, **kw):
        """Initialize database.

        Parameter:
            args -- variable list of arguments.

            kw -- keyword arguments.

        Supported keyword arguments:
            auto_commit -- boolean flag (default: True). If omitted or
            True, let the database manager decide on transaction
            commits automatically (recommended). Disable feature
            otherwise.

            db_path -- database file path (default: DEFAULT_PATH).

        This hook method is called from *__init__()* method with 'args'
        and 'kw' initialized from there.

        Current implementation calls deprecated *init_members()* and
        *init_database()* hook methods, but should merely call only
        *open_database()* method in later versions of tkRAD library.

        Feel free to extend or to override this hook method in your own
        subclasses to best meet your needs.

        *NEW* since **tkRAD v1.7**.
        """
        self.init_members(**kw)
        self.open_database(*args, **kw)
        self.init_database(**kw)
    # end def

    def is_open (self):
        """Indicate if database is currently opened or not."""
        return bool(self.connection and self.cursor)
    # end def

    @property
    def last_row_id (self):
        """Row id of the last modified row - read-only attribute.

        Returns the row id - as defined by SQLite3's documentation - of
        the last modified row by an SQL INSERT statement.

        Returns None when not relevant.

        May raise **DatabaseError** when called with no available
        database cursor e.g. closed database.

        Python documentation page:
        https://docs.python.org/3/library/sqlite3.html#sqlite3.Cursor.lastrowid

        SQLite3 documentation page:
        https://www.sqlite.org/lang_createtable.html#rowid
        """
        if self.cursor:
            return self.cursor.lastrowid
        else:
            raise DatabaseError(
                "Could not retrieve last row id: "
                "no available cursor by now."
            )
        # end if
    # end def

    def normalize_path (self, db_path):
        """Normalize path to database.

        Parameter:
            db_path -- database file path (string of chars). Supports
            SQLite3's specific ":memory:" database path into memory
            (instead of a disk file).

        Normalizes 'db_path' database file path with special support
        for SQLite3's ":memory:" database path into memory instead of a
        disk file.

        *NEW* since **tkRAD v1.7**.
        """
        if str(db_path).lower() == self._SPECIAL_PATH:
            return self._SPECIAL_PATH
        else:
            return path.normalize(db_path)
        # end if
    # end def

    def open_database (self, *args, **kw):
        """Open database.

        Parameters:
            args -- variable list of arguments.

            kw -- keyword arguments.

        Supported keyword arguments:
            auto_commit -- boolean flag (default: True). If omitted or
            True, let the database manager decide on transaction
            commits automatically (recommended). Disable feature
            otherwise.

            db_path -- database file path (default: DEFAULT_PATH).

        Opens database using self.db_path attribute, 'db_path' keyword
        argument or self.DEFAULT_PATH as fallback values, otherwise.

        Sets self.row_factory attribute to sqlite3.Row recommended
        class type.

        Python documentation:
        https://docs.python.org/3/library/sqlite3.html#sqlite3.Row

        Initializes database's cursor into *self.cursor* class member
        attribute on-the-fly.
        """
        self.connection = sqlite3.connect(
            self.normalize_path(
                self.db_path or kw.get("db_path") or self.DEFAULT_PATH
            )
        )
        self.row_factory = sqlite3.Row
        self.cursor = self.connection.cursor()
    # end def

    def rollback (self, *args, **kw):
        """Cancel current transaction in database.

        Event handler signature.

        May raise **DatabaseError** when called with no available
        connection.
        """
        if self.connection:
            self.connection.rollback()
        else:
            raise DatabaseError(
                "Could not rollback current transaction in database: "
                "no available connection by now."
            )
        # end if
    # end def

    @property
    def row_factory (self):
        """Database connection's row factory - read-write attribute.

        Use this attribute to get or set database connection's row
        factory type.

        Recommended factory type: sqlite3.Row

        Python documentation:
        https://docs.python.org/3/library/sqlite3.html#sqlite3.Row

        May raise **DatabaseError** when called with no available
        connection.
        """
        if self.connection:
            return self.connection.row_factory
        else:
            raise DatabaseError(
                "Could not retrieve current row factory: "
                "no available connection by now."
            )
        # end if
    # end def

    @row_factory.setter
    def row_factory (self, value):
        if self.connection and callable(value):
            self.connection.row_factory = value
        else:
            raise DatabaseError(
                "Could not set up row factory for database: "
                "value is not a callable or "
                "no available connection by now."
            )
        # end if
    # end def

    def sql_query (self, query, *args, **kw):
        """Execute a unique SQL statement from a query.

        Parameters:
            query -- SQL statement query (string of chars).

            args -- ordered list of question mark arguments, as stated
            in Python documentation.

            kw -- keyword arguments for named placeholders, as stated
            in Python documentation.

        Python documentation:
        https://docs.python.org/3/library/sqlite3.html#sqlite3.Cursor.execute

        Notice: it is not possible to use both qmarks and named
        placeholders into a same SQL statement query.

        For executing multiple SQL statements in a same query script,
        please use self.sql_script() method.

        May raise **DatabaseError** when called with no available
        database cursor e.g. closed database.
        """
        if self.cursor:
            self.cursor.execute(query, args or kw)
            if self.auto_commit:
                self.commit()
            # end if
        else:
            raise DatabaseError(
                "Could not execute SQL statement: "
                "no available cursor by now."
            )
        # end if
    # end def

    def sql_script (self, script):
        """Execute multiple SQL statements from a script.

        Parameter:
            script -- multiple SQL statement script (string of chars).

        Executes a script containing multiple SQL statements, each of
        them separated by a semi-colon ';' character.

        For executing a unique SQL statement query with parameters,
        please use self.sql_query() method.

        May raise **DatabaseError** when called with no available
        database cursor e.g. closed database.
        """
        if self.cursor:
            self.cursor.executescript(script)
            if self.auto_commit:
                self.commit()
            # end if
        else:
            raise DatabaseError(
                "Could not execute SQL script: "
                "no available cursor by now."
            )
        # end if
    # end def

# end class


class DatabaseError (Exception):
    """Error handler for Database class."""
    pass
# end class
