#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Deferred actions queue module.

Implements:

* module function get_defer_queue -- retrieve an app-wide deferred
  actions queue (unique instance).
* class DeferQueue -- queue class for deferred actions.
* class DeferQueueError -- error handler for DeferQueue class.
* class QueueItem -- subcomponent class used by DeferQueue.

This module is mainly used by tkRAD's XML parsers, but can also be used
whenever you need to defer some tasks in time.

*NEW* since **tkRAD v1.7**:

Please, feel free to post an issue ticket on GitHub's project
bugtracker if you face any troubleshooting with this new version of
tkRAD's library.

GitHub's project bugtracker: https://github.com/tarball69/tkRAD/issues
(requires a GitHub member account to proceed).
"""

"""
    tkRAD - tkinter Rapid Application Development library

    (c) 2013+ Raphaël SEBAN <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see: http://www.gnu.org/licenses/
"""

__queue = None


def get_defer_queue ():
    """Retrieve an app-wide deferred actions queue (unique instance).

    Use this function whenever you need only one unique object instance
    over all your app code.

    For several deferred action queues, merely instantiate DeferQueue
    class instead.
    """
    global __queue
    if not isinstance(__queue, DeferQueue):
        __queue = DeferQueue()
    # end if
    return __queue
# end def


class DeferQueue:
    """Queue class for deferred actions."""

    def __init__ (self):
        """Class constructor."""
        self._queue = dict()
    # end def

    def clear (self, section=None):
        """Clear deferred actions for a section.

        Parameter:
            section -- section name, string of chars (default: None).
            If omitted or None, will clear all registered actions in
            queue without calling them first.

        Clears all deferred actions registered under a same given
        'section' name.

        Clears up all queue at once, if no 'section' name is given e.g.
        `queue.clear()`.

        A clearing procedure does *NOT* run deferred actions before
        clearing them up, as it might be done by other methods.
        """
        if section:
            self._queue.pop(section, None)
        else:
            self._queue.clear()
        # end if
    # end def

    def defer (self, section, callback, *args, **kw):
        """Register action to be deferred.

        Parameters:
            section -- section name (string of chars).

            callback -- function or method to be called in a later time
            (deferred action).

            args -- optional ordered list of arguments to pass to
            'callback' at call time.

            kw -- optional keyword arguments to pass to 'callback' at
            call time.

        Registers a callable 'callback' with its 'args' and 'kw'
        optional arguments under a given 'section' name, in order to be
        called later on as a deferred action.

        Deferred actions should always be grouped by section names for
        clearer organization.

        Notice: a same 'callback' action can be deferred as often as
        needed.
        """
        _buffer = self._queue.setdefault(section, list())
        _buffer.append(self.queue_item_type(callback, *args, **kw))
        self._queue[section] = _buffer
    # end def

    def flush (self, section, *args, **kw):
        """Call deferred actions and then clear section.

        Parameters:
            section -- section name (string of chars).

            args -- optional ordered list of arguments to pass to a
            callable.

            kw -- optional keyword arguments to pass to a callable.

        Calls each callable registered under 'section' name with common
        optional 'args' and 'kw' arguments.

        If a callable was registered with its own arguments, new 'args'
        and 'kw' arguments come **in addition** to former arguments.

        Notice: new common **keyword** arguments might override former
        private keyword arguments, if they share the same keyword name.

        Example::

            queue.defer("group1", my_callable, kw1="do_this", kw2="hello")
            # ...some piece of code...
            queue.flush("group1", kw1="do_that")
            # my_callable will be called with overridden kw1
            # i.e. my_callable(kw1="do_that", kw2="hello")

        Clears up 'section' name once all deferred actions are done.
        """
        _buffer = self._queue.setdefault(section, list())
        for _item in _buffer:
            _item.call(*args, **kw)
        # end for
        self.clear(section)
    # end def

    def flush_all (self, *args, **kw):
        """Flush all registered section names.

        Parameters:
            args -- optional ordered list of arguments to pass to a
            callable.

            kw -- optional keyword arguments to pass to a callable.

        Calls each callable registered under each registered section
        name with common optional 'args' and 'kw' arguments.

        If a callable was registered with its own arguments, new 'args'
        and 'kw' arguments come **in addition** to former arguments.

        Notice: new common **keyword** arguments might override former
        private keyword arguments, if they share the same keyword name.

        Example::

            queue.defer("group1", my_callable, kw1="do_this", kw2="hello")
            # ...some piece of code...
            queue.flush_all(kw1="do_that")
            # my_callable will be called with overridden kw1
            # i.e. my_callable(kw1="do_that", kw2="hello")

        Clears up all queue once all deferred actions are done.
        """
        for _section in self._queue:
            self.flush(_section, *args, **kw)
        # end for
        self.clear()
    # end def

    def get_queue (self):
        """Return queue's registry object.

        Returns a strong reference to a Python dict() object, so
        please, use with caution.
        """
        return self._queue
    # end def

    @property
    def queue_item_type (self):
        """Preferred queue item type - read-only attribute.

        Current implementation returns QueueItem class type.

        Override this attribute whenever you need another class type on
        registering deferred actions.

        Other class types may at least implement a self.call() method
        (see QueueItem implementation for more detail).

        Best practice still remains to subclass QueueItem to meet your
        own needs.
        """
        return QueueItem
    # end def

# end class


class DeferQueueError (Exception):
    """Error handler for DeferQueue class."""
    pass
# end class


class QueueItem:
    """Subcomponent class used by DeferQueue."""

    def __init__ (self, callback, *args, **kw):
        """Class constructor."""
        self.callback = callback
        self.arguments = args
        self.keywords = kw
    # end def

    def call (self, *args, **kw):
        """Call registered callable.

        Parameters:
            args -- optional ordered list of arguments to pass to
            registered callable.

            kw -- optional keyword arguments to pass to registered
            callable.

        Calls registered callable with optional new 'args' and 'kw'
        arguments.

        If the callable was registered with its own arguments, new
        'args' and 'kw' arguments come **in addition** to former
        arguments.

        Notice: new common **keyword** arguments might override former
        private keyword arguments, if they share the same keyword name.

        Example::

            _item = QueueItem(my_callable, kw1="do_this", kw2="hello")
            # ...some piece of code...
            _item.call(kw1="do_that")
            # my_callable will be called with overridden kw1
            # i.e. my_callable(kw1="do_that", kw2="hello")

        Returns callable's return value if callable is actually a
        callable object, returns None otherwise.
        """
        if callable(self.callback):
            _args = self.arguments + args
            _kw = self.keywords.copy()
            _kw.update(kw)
            return self.callback(*_args, **_kw)
        # end if
        return None
    # end def

# end class
