#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Event management module.

Implements:

* module function get_event_manager -- retrieve an app-wide event
  manager (unique instance).
* class EventManager -- event management class.
* class EventManagerError -- error handler for EventManager class.

This module uses simplified application-related event management
maechanism that do *NOT* rely on tkinter's widget event maechanism.

As EventManager class works with a Python dict() object for registering
callbacks, one may pay attention to object's reference held indirectly
by callback methods and which may lead to traps on garbage collection
(some objects may never be destroyed because of these strong references
in yet registered callbacks).

A good practice still remains to always disconnect() callbacks that
have been connect()-ed previously in app code, especially when
callback's holder is an object to be destroyed during normal program
flow.

Notice: tkRAD's RADxxxxxxx objects already manage automatic
disconnection of their related callbacks when w.destroy() method is
called, either programmatically or by implied tkinter parent widget
calls.

*NEW* since **tkRAD v1.7.6**:

* renamed 'event' --> 'event_name' to avoid naming conflicts with
  keyword arguments.

*NEW* since **tkRAD v1.7**:

Please, feel free to post an issue ticket on GitHub's project
bugtracker if you face any troubleshooting with this new version of
tkRAD's library.

GitHub's project bugtracker: https://github.com/tarball69/tkRAD/issues
(requires a GitHub member account to proceed).
"""

"""
    tkRAD - tkinter Rapid Application Development library

    (c) 2013+ Raphaël SEBAN <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see: http://www.gnu.org/licenses/
"""

__event_manager = None


def get_event_manager ():
    """Retrieve an app-wide event manager (unique instance).

    Use this function whenever you need only one unique event manager
    object instance over all your app code.

    For several private-shared managers, merely instantiate
    EventManager class instead.
    """
    global __event_manager
    if not isinstance(__event_manager, EventManager):
        __event_manager = EventManager()
    # end if
    return __event_manager
# end def


class EventManager:
    """Event management class."""

    def __init__ (self):
        """Class constructor."""
        self._connections = dict()
    # end def

    def _on_widget_destroy (self, widget):
        for _event, _callbacks in self._connections.copy().items():
            _dcallbacks = set()
            for _callback in _callbacks:
                if _callback is widget or \
                        (hasattr(_callback, "__self__") and \
                        _callback.__self__ is widget):
                    _dcallbacks.add(_callback)
                # end if
            # end for
            if _dcallbacks:
                self.disconnect(_event, *_dcallbacks)
            # end if
        # end for
    # end def

    def connect (self, event_name, *callbacks):
        """Associate callbacks to event name.

        Parameters:
            event_name -- event arbitrary name (string of chars).
            Caution: event names are case-sensitive.

            callbacks -- variable list of functions and/or methods to
            call back when given 'event_name' fires up at some point in
            program flow.

        Registers a variable list of 'callbacks' to call when given
        'event_name' fires up at some point in program flow.

        Caution: callbacks are stored in Python set() objects, which
        means they will be **called in random order** when given
        'event_name' fires up.

        Only callable objects will be actually registered in there.
        """
        _callbacks = self._connections.setdefault(event_name, set())
        if isinstance(_callbacks, set):
            _callbacks.update(callbacks)
            self._connections[event_name] = set(
                filter(callable, _callbacks)
            )
            return True
        # end if
        return False
    # end def

    def connect_dict (self, events_dict):
        """Make connections for an event mapping dict() object.

        Parameter:
            events_dict -- mapping dict() object of key/value pairs
            where key = event name and value = variable list of
            callable objects (callbacks).

        Browse a mapping dict() object of key/value pairs and connects
        each event name (key in pair) to its related list of callbacks
        (value in pair).

        Callbacks may either be a single callable object or any of
        tuple, list, set iterable list of callable objects to connect
        at once.

        Example::

            self.events.connect_dict(
                # Python dict() mapping type
                {
                    # unique callback
                    "App:Init": self.do_app_inits,

                    # tuple list of callbacks
                    "Project:New": (self.new_tab,
                        self.project.reset_to_new, self.options.reset),

                    # mutable list of callbacks
                    "Project:Modified": [self.on_modified,
                        self.project.modified, self.database.update],

                    # set list of callbacks
                    "Project:Preview": {self.do_this,
                        self.project.preview, self.do_that},
                }
            )

        May raise **EventManagerError** if 'events_dict' is an empty
        dict() object or not of this type at all.
        """
        if events_dict and isinstance(events_dict, dict):
            for (_event, _callbacks) in events_dict.items():
                if isinstance(_callbacks, (tuple, list, set)):
                    self.connect(_event, *_callbacks)
                else:
                    self.connect(_event, _callbacks)
                # end if
            # end for
            return True
        else:
            raise EventManagerError(
                "Expected non-empty Python dict() object type."
            )
        # end if
        return False
    # end def

    def disconnect (self, event_name, *callbacks):
        """Disconnect list of callbacks from event name.

        Parameters:
            event_name -- event arbitrary name (string of chars).
            Caution: event names are case-sensitive.

            callbacks -- variable list of functions and/or methods that
            were previously connect()-ed to 'event_name'.

        Unregisters any callback in 'callbacks' that has been
        previously registered through connect() or connect_dict()
        methods to 'event_name'.

        Unrelated callbacks in 'callbacks' do *NOT* raise any error on
        disconnection: they are merely ignored.

        Only callable objects will be kept after this operation in
        'event_name' group of callbacks.
        """
        _callbacks = self._connections.get(event_name)
        if _callbacks and isinstance(_callbacks, set):
            _callbacks.difference_update(callbacks)
            self._connections[event_name] = set(
                filter(callable, _callbacks)
            )
            return True
        # end if
        return False
    # end def

    def disconnect_all (self, *events):
        """Disconnect all listed event names.

        Parameters:
            events -- variable list of event arbitrary names (list of
            strings of chars). Caution: event names are case-sensitive.

        Disconnects all callbacks registered in each event name listed
        in 'events'.

        Caution: a call to this method with *NO* given event name
        merely **clears all connections everywhere**.

        Examples::

            # disconnects all callbacks for given event names
            self.events.disconnect_all("Project:New", "Project:Save")

            # disconnects absolutely everything... app-wide!
            # use this with much much caution
            self.events.disconnect_all()

        Unregistered or unknown listed event names are merely ignored
        and might *NOT* raise any error.
        """
        if not events:
            self._connections.clear()
        else:
            for _event in set(events):
                self._connections.pop(_event, None)
            # end for
        # end if
    # end def

    def raise_event (self, event_name, *args, **kw):
        """Fire up an event name.

        Parameters:
            event_name -- event arbitrary name (string of chars).
            Caution: event names are case-sensitive.

            args -- optional variable list of arguments to pass to a
            callable.

            kw -- optional keyword arguments to pass to a callable.

        Calls each callable object (callback) registered under
        'event_name' and passes optional 'args' arguments and 'kw'
        keyword arguments to the callback.

        Example::

            # first connect callbacks
            self.events.connect("Project:New", self.on_new_project,
                self.project.reset_to_new)

            # ...some piece of code...

            # fire up event with a keyword argument
            self.events.raise_event("Project:New", project_name="Untitled")
            # at this point, callbacks will be called as follows:
            self.on_new_project(project_name="Untitled")
            self.project.reset_to_new(project_name="Untitled")

        Notice: only callable objects will be actually called at this
        time, other objects are merely ignored.
        """
        _callbacks = self._connections.get(event_name)
        if _callbacks and isinstance(_callbacks, set):
            _callbacks = set(filter(callable, _callbacks))
            for _callback in _callbacks:
                _callback(*args, **kw)
            # end for
            return True
        # end if
        return False
    # end def

# end class


class EventManagerError (Exception):
    """Error handler for EventManager class."""
    pass
# end class
