#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Internationalization (i18n) module.

Implements:

* module function _ -- translate given text.
* module function dump_pot_file -- dump all registered entries into a
  POT file.
* module function get_lc_lang -- retrieve system's current LANG value.
* module function get_translations_dir -- get locale translations
  directory path.
* module function get_translations_lang -- get locale translations
  language.
* module function get_translations_table -- get locale translations
  mapping table.
* module function install -- install translations table.
* module function load_translations_table -- load translations mapping
  table.
* module function set_translations_dir -- set up locale translations
  directory path.
* module function set_translations_lang -- set up locale translations
  language.
* module function set_translations_table -- set up locale translations
  mapping table.
* module function switch_off -- switch i18n support OFF.
* module function switch_on -- switch i18n support ON.

Example 1::

    # This installs user's locale language by default.
    import tkRAD
    # Automatic translations support: this is the default usage.
    tkRAD.RADApplication().run()

There is nothing special to do else than providing the right
translations file corresponding to user's locale language e.g. for
french users, you should provide some *^/locale/fr_FR.po* translation
file filled out with all necessary translations.

Notice: tkRAD's hat '^' notation merely indicates **application's root
directory**.

Example 2::

    # This installs user's locale language by default.
    import tkRAD
    # Run your application.
    tkRAD.RADApplication().run()
    # Then try to get a template.pot translation file template
    # for *.po translation files e.g. fr_FR.po translations file.
    # This requires to have at least an actual translation
    # directory path to future template file e.g. ^/locale directory.
    tkRAD.i18n.dump_pot_file()

Example 3::

    # This installs user's locale language by default.
    import tkRAD
    # But you can force to a given static language,
    # if corresponding translation file does exist, of course,
    tkRAD.i18n.install(lc_lang="nl")
    # and then run your application.
    tkRAD.RADApplication().run()

Example 4::

    # This installs user's locale language by default.
    import tkRAD
    # The i18n support is switched ON by default.
    # Switch OFF i18n support for a debugging session,
    tkRAD.i18n.switch_off()
    # and then run your application.
    tkRAD.RADApplication().run()

*NEW* since **tkRAD v1.7**:

Please, feel free to post an issue ticket on GitHub's project
bugtracker if you face any troubleshooting with this new version of
tkRAD's library.

GitHub's project bugtracker: https://github.com/tarball69/tkRAD/issues
(requires a GitHub member account to proceed).
"""

"""
    tkRAD - tkinter Rapid Application Development library

    (c) 2013+ Raphaël SEBAN <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see: http://www.gnu.org/licenses/
"""


import re
import locale
import os.path

try:
    from . import path
    from . import tools
except:
    import path
    import tools
# end try


# Caution: global inits at bottom script.


def _ (text):
    """Translate given text.

    Parameter:
        text -- text to translate (string of chars).

    If i18n support is switch_off(), merely returns 'text' itself.

    If i18n support is switch_on() (the default, on a simple module
    import), tries to find a translated entry into translations table
    and returns it - if found - or merely returns genuine 'text'
    otherwise.
    """
    if __switch_off:
        return text
    # end if
    __pot_entries.add(text)
    return tools.choose_str(__translations_table.get(text), text)
# end def


try:
    __builtins__["_"] = _
except:
    __builtins__._ = _
# end try


def dump_pot_file (filename="template.pot"):
    """Dump all registered entries into a POT file.

    Parameter:
        filename -- POT file filename (default: "template.pot").

    POT file is located at '${lc_dir}/${filename}' i.e. most of time
    located at '^/locale/template.pot', tkRAD's hat notation '^'
    merely indicating *application's root directory*.

    *NEW* since **tkRAD v1.7**: added 'filename' parameter.
    """
    _fpath = path.normalize(os.path.join(__translations_dir, filename))
    _tpl = 'msgid ""\n"{msgid}"\nmsgstr ""\n""\n'
    with open(_fpath, "w") as _pot_file:
        _pot_file.write(
            "\n".join(
                map(
                    lambda s:_tpl.format(
                        msgid=s.replace("\n", '\\n"\n"')
                    ),
                    sorted(__pot_entries)
                )
            )
        )
    # end with
# end def


def get_lc_lang ():
    """Retrieve system's current LANG value.

    Merely returns *locale.getdefaultlocale()[0]* string of chars.

    See: https://docs.python.org/3/library/locale.html#locale.getdefaultlocale

    *NEW* since **tkRAD v1.7**.
    """
    return tools.choose_str(locale.getdefaultlocale()[0])
# end def


def get_translations_dir ():
    """Get locale translations directory path.

    This directory should contain all ${lc_lang}.po translation files.
    """
    return __translations_dir
# end def


def get_translations_lang ():
    """Get locale translations language.

    This language value should comply with IETF's RFC 1766 country code
    tags e.g. fr, fr_FR, de, it, en, en_US and so on.

    See: https://tools.ietf.org/html/rfc1766
    """
    return __translations_lang
# end def


def get_translations_table ():
    """Get locale translations mapping table."""
    return __translations_table
# end def


def install (lc_dir=None, lc_lang=None):
    """Install translations table.

    Parameters:
        lc_dir -- locale translations directory path (default: None).
        If omitted or None, will be replaced by current
        get_translations_dir() path.

        lc_lang -- locale translations language (default: None). If
        omitted or None, will be replaced by current
        get_translations_lang() string of chars. This parameter should
        comply with IETF's RFC 1766 country code tags
        (https://tools.ietf.org/html/rfc1766 ).

    Resets current translations directory and language to respectively
    'lc_dir' and 'lc_lang', tries to load new table of translations and
    then install translations to work app-wide.
    """
    set_translations_dir(lc_dir)
    set_translations_lang(lc_lang)
    try:
        load_translations_table()
    except:
        set_translations_table(dict())
    # end try
# end def


def load_translations_table (lc_dir=None, lc_lang=None):
    """Load translations mapping table.

    Parameters:
        lc_dir -- locale translations directory path (default: None).
        If omitted or None, will be replaced by current
        get_translations_dir() path.

        lc_lang -- locale translations language (default: None). If
        omitted or None, will be replaced by current
        get_translations_lang() string of chars. This parameter should
        comply with IETF's RFC 1766 country code tags
        (https://tools.ietf.org/html/rfc1766 ).

    Opens a "${lc_dir}/${lc_lang}.po" translations file and converts it
    to a Python mapping dict() translations table object.
    """
    global __translations_table
    lc_dir = tools.choose_str(lc_dir, __translations_dir)
    lc_lang = tools.choose_str(lc_lang, __translations_lang)
    _path = path.normalize(os.path.join(lc_dir, lc_lang + ".po"))
    with open(_path, "r", encoding="UTF-8") as _file:
        _data = _file.read()
    # end with
    _data = re.sub(r"(?m)^#.*$", r"", _data)
    _data = re.sub(r"(?i)msgid", r"msgid", _data)
    _data = _data.split("msgid")
    del _data[0]
    _data = ",".join(_data)
    _data = re.sub(r"(?i)msgstr", r":", _data)
    __translations_table = eval("{" + _data + "}")
# end def


def set_translations_dir (arg):
    """Set up locale translations directory path.

    This directory should contain all ${lc_lang}.po translation files.
    """
    global __translations_dir
    __translations_dir = tools.choose_str(arg, __translations_dir)
# end def


def set_translations_lang (arg):
    """Set up locale translations language.

    This language value should comply with IETF's RFC 1766 country code
    tags e.g. fr, fr_FR, de, it, en, en_US and so on.

    See: https://tools.ietf.org/html/rfc1766
    """
    global __translations_lang
    __translations_lang = tools.choose_str(arg, __translations_lang)
# end def


def set_translations_table (arg):
    """Set up locale translations mapping table."""
    global __translations_table
    __translations_table = dict(arg)
# end def


def switch_off ():
    """Switch i18n support OFF."""
    global __switch_off
    __switch_off = True
# end def


def switch_on ():
    """Switch i18n support ON."""
    global __switch_off
    __switch_off = False
# end def


__pot_entries = set()
__switch_off = False
__translations_dir = "^/locale"
__translations_lang = get_lc_lang()
__translations_table = dict()
