#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""RC options file management module.

Implements:

* module function get_option_manager -- retrieve an app-wide option
  manager (unique instance).
* class OptionManager -- RC options file management class.
* class OptionManagerError -- error handler for OptionManager class.

**Notice:** there are many definitions for 'RC' mention e.g. runcom,
run commands, resource control, run control, runtime configuration,
resource configuration, etc.

In this module terminology, we'll stand 'RC' file for 'Remote
Configuration' file -- in opposition to *software internal
configuration* -- which seems to be the best definition regarding to
its actual feature.

Example 1::

    # using RC options out from usual tkRAD context
    import tkinter as tk
    from tkRAD.core.options import get_option_manager

    def save_to_rc_file (event=None):
        options["my_section"]["my_option_name"] = entry_cvar.get()
        options.save()
        root.destroy()
    # end def

    root = tk.Tk()
    tk.Label(root, text="Please, insert some text:").pack()
    entry_cvar = tk.StringVar()
    tk.Entry(root, textvariable=entry_cvar).pack()
    tk.Button(
        root, text="Save to RC file", command=save_to_rc_file
    ).pack()
    options = get_option_manager(rc_dir="etc", rc_file="app.rc")
    options.set_sections("my_section")
    options.load()
    entry_cvar.set(
        options.get("my_section", "my_option_name", fallback="")
    )
    root.mainloop()

Example 2::

    # same thing in a tkRAD context
    import tkRAD

    class MyApp (tkRAD.RADXMLMainWindow):
        def run (self):
            xml_source = \"""
                <tkwidget>
                    <label
                        text="Please, insert some text:"
                        layout="pack"
                    />
                    <entry
                        textvariable="entry_cvar"
                        layout="pack"
                    />
                    <button
                        text="Save to RC file"
                        command=".save_to_rc_file"
                        layout="pack"
                    />
                </tkwidget>
            \"""
            self.xml_build(xml_source)
            self.entry_cvar = (
                self.mainframe.get_stringvar("entry_cvar")
            )
            self.entry_cvar.set(
                self.options.get(
                    "my_section", "my_option_name", fallback="")
            )
            self.mainloop()
        # end def

        def save_to_rc_file (self, *args, **kw):
            self.options.set_sections("my_section")
            self.options["my_section"]["my_option_name"] = (
                self.entry_cvar.get()
            )
            self.events.raise_event("quit")
        # end def
    # end class

    MyApp(rc_dir="etc", rc_file="app.rc").run()

*NEW* since **tkRAD v1.7**:

Please, feel free to post an issue ticket on GitHub's project
bugtracker if you face any troubleshooting with this new version of
tkRAD's library.

GitHub's project bugtracker: https://github.com/tarball69/tkRAD/issues
(requires a GitHub member account to proceed).
"""

"""
    tkRAD - tkinter Rapid Application Development library

    (c) 2013+ Raphaël SEBAN <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see: http://www.gnu.org/licenses/
"""

import configparser
import os
import re

try:
    from . import path as P
    from . import tools
except:
    import path as P
    import tools
# end try


__option_manager = None


def get_option_manager (**kw):
    """Retrieve an app-wide option manager (unique instance).

    Parameter:
        kw -- keyword arguments.

    Supported keyword arguments:
        rc_dir -- directory path to RC file, string of chars (default:
        None). If omitted or None, will be replaced by class attribute
        OptionManager.CONFIG["dir"].

        rc_file -- RC file filename, string of chars (default: None).
        If omitted or None, will be replaced by class attribute
        OptionManager.CONFIG["file"].

    Use this function whenever you need only one unique option manager
    object instance over all your app code.

    For several private-shared managers, merely instantiate
    OptionManager class instead.
    """
    global __option_manager
    if not isinstance(__option_manager, OptionManager):
        __option_manager = OptionManager(**kw)
    # end if
    return __option_manager
# end def


class OptionManager (configparser.ConfigParser):
    """RC options file management class."""

    CONFIG = {
        "dir": "~/.config/tkRAD_application",
        "file": "application.rc",
    }

    SECTIONS = (
        "dirs",  "files",  "geometry",  "gui",  "xml",
    )

    def __init__ (self, **kw):
        """Class constructor.

        Parameter:
            kw -- keyword arguments.

        Supported keyword arguments:
            rc_dir -- directory path to RC file, string of chars
            (default: None). If omitted or None, will be replaced by
            CONFIG["dir"].

            rc_file -- RC file filename, string of chars (default:
            None). If omitted or None, will be replaced by
            CONFIG["file"].
        """
        super().__init__()
        self.rc_dir = kw.get("rc_dir")
        self.rc_file = kw.get("rc_file")
        self.set_sections("DEFAULT", *self.SECTIONS)
        self._reset_load()
    # end def

    def _reset_load (self):
        self.__loaded = False
    # end def

    @deprecated("Please, use 'rc_dir' class attribute instead.")
    def get_config_dir (self):
        """Get directory path to RC file.

        *DEPRECATED* since **tkRAD v1.7**.

        Please, use 'rc_dir' attribute instead.
        """
        return self.rc_dir
    # end def

    @deprecated("Please, use 'rc_file' class attribute instead.")
    def get_config_file (self):
        """Get RC file filename.

        *DEPRECATED* since **tkRAD v1.7**.

        Please, use 'rc_file' attribute instead.
        """
        return self.rc_file
    # end def

    def load (self):
        """Load RC file.

        Loads RC file located at 'rc_path' attribute the first time and
        then locks.

        Once locked, each further call to this method will do nothing.

        This is made for large apps using 'self.options.load()' into
        their own component inits, allowing this way to declare it in a
        blind manner and thus not to bother with 'who did what?'
        issues.

        Please, use 'reload()' method whenever you actually need to
        force reloading of RC file options.

        Returns a list of read file(s) on success, None otherwise.

        *NEW* since **tkRAD v1.7**: remove irrelevant implicit fallback
        values.
        """
        _success = None
        if not self.__loaded:
            _success = self.read(self.rc_path)
            self.__loaded = tools.is_plist(_success)
        # end if
        return _success
    # end def

    @property
    def rc_dir (self):
        """Directory path to RC file - read-write attribute.

        *NEW* since **tkRAD v1.7**.
        """
        return self.__rc_dir
    # end def

    @rc_dir.setter
    def rc_dir (self, value):
        self.__rc_dir = P.normalize(
            value or self.CONFIG["dir"], raise_error=True
        )
        self._reset_load()
    # end def

    @property
    def rc_file (self):
        """RC file filename - read-write attribute.

        *NEW* since **tkRAD v1.7**.
        """
        return self.__rc_file
    # end def

    @rc_file.setter
    def rc_file (self, value):
        self.__rc_file = os.path.basename(value or self.CONFIG["file"])
        self._reset_load()
    # end def

    @property
    def rc_path (self):
        """Full RC file path - read-only attribute.

        Returns full RC file's path built along with 'rc_dir' and
        'rc_file' attributes.

        *NEW* since **tkRAD v1.7**.
        """
        return os.path.join(self.rc_dir, self.rc_file)
    # end def

    def reload (self):
        """Force reload of RC file options.

        Please, see 'load()' method documentation for more detail.

        Returns a list of read file(s) on success, None otherwise.
        """
        self._reset_load()
        return self.load()
    # end def

    def save (self):
        """Save options to RC file.

        Saves modified options into 'rc_path' RC file.

        May make 'rc_dir' directory on-the-fly, if missing on system.

        *NEW* since **tkRAD v1.7**: remove useless exception catchups
        and let any exception arise normally for finer error control.
        """
        os.makedirs(self.rc_dir, exist_ok=True)
        with open(self.rc_path, "w") as _file:
            self.write(_file)
        # end with
    # end def

    @deprecated("Please, use 'rc_dir' class attribute instead.")
    def set_config_dir (self, value):
        """Set up directory path to RC file.

        *DEPRECATED* since **tkRAD v1.7**.

        Please, use 'rc_dir' attribute instead.
        """
        self.rc_dir = value
    # end def

    @deprecated("Please, use 'rc_file' class attribute instead.")
    def set_config_file (self, value):
        """Set up RC file filename.

        *DEPRECATED* since **tkRAD v1.7**.

        Please, use 'rc_file' attribute instead.
        """
        self.rc_file = value
    # end def

    def set_defaults (self, **kw):
        """Set up 'DEFAULT' section.

        Parameter:
            kw -- keyword arguments.

        Sets up 'DEFAULT' section with new 'kw' keyword arguments.

        Already existing keyword names will be overridden by same
        keyword names' new values.

        New keyword names will merely be added to section.
        """
        self["DEFAULT"].update(kw)
    # end def

    def set_sections (self, *names):
        """Set up new sections.

        Parameter:
            names -- variable list of section names (list of strings of
            chars).

        Adds new section names to option manager.

        Already existing listed names will be merely ignored.
        """
        for _section in set(names):
            self.setdefault(_section, dict())
        # end for
    # end def

# end class


class OptionManagerError (Exception):
    """Error handler for OptionManager class."""
    pass
# end class
