#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""File path management module.

Implements:

* module function get_app_root_dir -- get application's root directory.
* module function normalize -- normalize file path.
* module function set_app_root_dir -- set up application's root
  directory.
* module function shorten_path -- shorten file path to a limited number
  of columns.

This module implements tkRAD's specific hat notation '^' one may set at
the beginning of file paths, which stands for **application's root
directory**.

*NEW* since **tkRAD v1.7**:

Please, feel free to post an issue ticket on GitHub's project
bugtracker if you face any troubleshooting with this new version of
tkRAD's library.

GitHub's project bugtracker: https://github.com/tarball69/tkRAD/issues
(requires a GitHub member account to proceed).
"""

"""
    tkRAD - tkinter Rapid Application Development library

    (c) 2013+ Raphaël SEBAN <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see: http://www.gnu.org/licenses/
"""

import os.path


__app_root_dir = os.path.realpath(".")


def get_app_root_dir ():
    """Get application's root directory."""
    return __app_root_dir
# end def


def normalize (path, raise_error=False):
    """Normalize file path.

    Parameters:
        path -- file path to normalize (string of chars).

        raise_error -- boolean flag (default: False). If True, may
        raise **TypeError** if 'path' is an empty string or not of
        string type at all. Remains silent otherwise.

    Normalizes a file path applying the following features:

    * os.path.normpath(path);
    * os.path.normcase(path);
    * tkRAD's hat notation: replaces a heading '^' character by
      application's root directory in file path;
    * os.path.expanduser(path);
    * os.path.realpath(path);
    * and finally, os.path.abspath(path).

    Returns normalized file path on success, empty string otherwise.
    """
    if path and isinstance(path, str):
        path = os.path.normcase(os.path.normpath(path))
        if path.startswith("^"):
            path = os.path.join(
                __app_root_dir, path.lstrip("^" + os.path.sep)
            )
        # end if
        return os.path.abspath(
            os.path.realpath(os.path.expanduser(path))
        )
    elif raise_error:
        raise TypeError("expected plain string of chars.")
    # end if
    return ""
# end def


def set_app_root_dir (dirpath):
    """Set up application's root directory."""
    global __app_root_dir
    __app_root_dir = normalize(dirpath)
    if not os.path.isdir(__app_root_dir):
        raise NotADirectoryError(
            "Application's root directory path is *NOT* "
            "an actual directory."
        )
    # end if
# end def


def shorten_path (path, width=40, placeholder="..."):
    """Shorten file path to a limited number of columns.

    Parameters:
        path -- file path string of chars to shorten.

        width -- max. number of columns for resulting string of chars
        (default: 40).

        placeholder -- string of chars used to indicate path string has
        been actually shortened (default: '...'). Will raise a
        **ValueError** if an empty string of chars. Even when not used
        under certain circumstances, this parameter MUST be a plain
        string of chars.

    Returns a given file 'path' shortened to a string of chars of
    'width' maximum number of chars, including 'placeholder' chars,
    whenever necessary.
    """
    if not placeholder:
        raise ValueError("A placeholder string of chars is required.")
    else:
        width = max(0, width)
        if isinstance(path, str) and len(path) > width:
            (_head, _tail) = os.path.split(path)
            placeholder = placeholder[:width]
            _swidth = width - len(placeholder)
            if _swidth > 0:
                _tail = _tail[-_swidth:]
                _swidth = max(0, _swidth - len(_tail))
                _head = _head[:_swidth]
            else:
                _head = _tail = ""
            # end if
            path = "{}{}{}".format(_head, placeholder, _tail)
        # end if
    # end if
    return path
# end def
