#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Service object management module.

Implements:

* module function ask_for -- ask for a given service object by name.
* module function get_service_manager -- retrieve an app-wide service
  manager (unique instance).
* module function register_service -- register a new service object by
  name.
* class ServiceManager -- service object management class.
* class ServiceManagerError -- error handler for ServiceManager class.

This module uses a simplified application-related service management.

A service object is an object that other app components may ask for
some given tasks. For example, *RADApplication* class registers itself
on instance init as an app-wide service for any other app components.
That way, anyone may query current running *RADApplication* object for
a given method, member or attribute somewhat considered as a provided
service.

Notice: as *ServiceManager* class works with a Python dict() object for
registering service objects, one may pay attention to object' strong
reference which may lead to traps on garbage collection (some objects
may never be destroyed because of these strong references in yet
registered services).

A good practice still remains to always unregister service objects that
have been registered previously in app code, especially when service is
an object to be destroyed during normal program flow.

Notice: tkRAD's RADxxxxxxx objects already manage automatic
unregistering of their related service objects when w.destroy() method
is called, either programmatically or by implied tkinter parent widget
calls.

*NEW* since **tkRAD v1.7**:

Please, feel free to post an issue ticket on GitHub's project
bugtracker if you face any troubleshooting with this new version of
tkRAD's library.

GitHub's project bugtracker: https://github.com/tarball69/tkRAD/issues
(requires a GitHub member account to proceed).
"""

"""
    tkRAD - tkinter Rapid Application Development library

    (c) 2013+ Raphaël SEBAN <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see: http://www.gnu.org/licenses/
"""

__service_manager = None


def ask_for (service_name, **kw):
    """Ask for a given service object by name.

    Parameters:
        service_name -- name of service object (string of chars).

        kw -- keyword arguments.

    Supported keyword arguments:
        silent_mode -- boolean flag (default: False). If omitted or
        False, may raise **ServiceManagerError** on unregistered
        'service_name'.

    Returns the service object registered under 'service_name'.

    Might return None on unregistered 'service_name' with 'silent_mode'
    set to True.
    """
    return get_service_manager().get_service(service_name, **kw)
# end def


def get_service_manager ():
    """Retrieve an app-wide service manager (unique instance).

    Use this function whenever you need only one unique service manager
    object instance over all your app code.

    For several private-shared managers, merely instantiate
    ServiceManager class instead.
    """
    global __service_manager
    if not isinstance(__service_manager, ServiceManager):
        __service_manager = ServiceManager()
    # end if
    return __service_manager
# end def


def register_service (service_name, service_object, **kw):
    """Register a new service object by name.

    Parameters:
        service_name -- service name (string of chars).

        service_object -- actual service object.

        kw -- keyword arguments.

    Supported keyword arguments:
        silent_mode -- boolean flag (default: False). If omitted or
        False, may raise **ServiceManagerError** on already registered
        service name.

    Registers a new 'service_object' under a given 'service_name'.

    May raise **ServiceManagerError** if given 'service_name' has
    already been previously registered and 'silent_mode' is False.

    Returns True on success, False otherwise.
    """
    return get_service_manager().register_service(
        service_name, service_object, **kw
    )
# end def


class ServiceManager:
    """Service object management class."""

    def __init__ (self):
        """Class constructor."""
        self.services = dict()
    # end def

    def _on_widget_destroy (self, widget):
        for _name, _service in self.services.copy().items():
            if _service is widget or \
                    (hasattr(_service, "__self__") and \
                    _service.__self__ is widget):
                self.services.pop(_name, None)
            # end if
        # end for
    # end def

    def clear_all (self):
        """Clear all registered service objects."""
        self.services.clear()
    # end def

    def delete_service (self, service_name):
        """Remove service object.

        Parameter:
            service_name -- service name (string of chars).

        Silently removes service object that has been registered under
        'service_name'.

        Unregistered 'service_name' is merely ignored.
        """
        self.services.pop(str(service_name), None)
    # end def

    def get_service (self, service_name, **kw):
        """Get service object by name.

        Parameters:
            service_name -- name of service object (string of chars).

            kw -- keyword arguments.

        Supported keyword arguments:
            silent_mode -- boolean flag (default: False). If omitted or
            False, may raise **ServiceManagerError** on unregistered
            'service_name'.

        Returns the service object registered under 'service_name'.

        Might return None on unregistered 'service_name' with
        'silent_mode' set to True.
        """
        service_name = str(service_name)
        if service_name in self.services:
            return self.services.get(service_name)
        elif not kw.get("silent_mode"):
            raise ServiceManagerError(
                "Service '{name}' is *NOT* registered."
                .format(name=service_name)
            )
        # end if
        return None
    # end def

    def register_service (self, service_name, service_object, **kw):
        """Register a new service object by name.

        Parameters:
            service_name -- service name (string of chars).

            service_object -- actual service object.

            kw -- keyword arguments.

        Supported keyword arguments:
            silent_mode -- boolean flag (default: False). If omitted or
            False, may raise **ServiceManagerError** on already
            registered service name.

        Registers a new 'service_object' under a given 'service_name'.

        May raise **ServiceManagerError** if given 'service_name' has
        already been previously registered and 'silent_mode' is False.

        Returns True on success, False otherwise.
        """
        service_name = str(service_name)
        if service_name not in self.services:
            self.services[service_name] = service_object
            return True
        elif not kw.get("silent_mode"):
            raise ServiceManagerError(
                "Service '{name}' is already registered."
                .format(name=service_name)
            )
        # end if
        return False
    # end def

    def replace_service (self, service_name, service_object, **kw):
        """Replace a registered service object by a new one.

        Parameters:
            service_name -- service name (string of chars).

            service_object -- actual service object.

            kw -- keyword arguments.

        Supported keyword arguments:
            silent_mode -- boolean flag (default: False). If omitted or
            False, may raise **ServiceManagerError** on unregistered
            'service_name'.

        Replaces the service object registered under 'service_name' by
        a new 'service_object'.

        May raise **ServiceManagerError** if given 'service_name' has
        *NOT* already been previously registered and 'silent_mode' is
        False.

        Returns True on success, False otherwise.
        """
        service_name = str(service_name)
        if service_name in self.services:
            self.services[service_name] = service_object
            return True
        elif not kw.get("silent_mode"):
            raise ServiceManagerError(
                "Service '{name}' is *NOT* already registered. "
                "Must exist before being replaced."
                .format(name=service_name)
            )
        # end if
        return False
    # end def

# end class


class ServiceManagerError (Exception):
    """Error handler for ServiceManager class."""
    pass
# end class
