#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Extended dictionary module.

Implements:

* class StructDict -- extended dictionary for transparent class
  structure value access.
* class StructDictError -- error handler for StructDict class.

A StructDict() object instance may be used to manage transparent value
getting/setting of a given class structure.

In fact, the StructDict() object holds key/value pairs like a Python
dict() object, but the 'value' part of the pair is itself an object of
a given type from which the StructDict() will get or set values.

In order to work properly, object of given type must implement a value
getter method and a value setter method, no matter what the names of
these two methods are.

Let's take a look to an example::

    #!/usr/bin/env python3
    # -*- coding: utf-8 -*-

    import tkRAD.core.struct_dict as SD

    class MyStructure:

        def __init__ (self, value):
            self.my_value = value
        # end def

        def my_value_getter (self):
            return self.my_value
        # end def

        def my_value_setter (self, new_value):
            self.my_value = new_value
        # end def

        # put here any other needed implementations

    # end class

    _sdict = SD.StructDict()
    _sdict.item_type = MyStructure
    _sdict.item_value_getter = "my_value_getter"
    _sdict.item_value_setter = "my_value_setter"

    _sdict.update(
        key1=MyStructure(10), key2=MyStructure(20), key3=MyStructure(30)
    )

    # show actual contents
    print("StructDict() object contents:")
    print(_sdict)

    # show actual item
    print('_sdict.get_item("key1"):', _sdict.get_item("key1"))

    # transparent value getting
    print('_sdict.get("key1"):', _sdict.get("key1"))
    print('_sdict["key1"]:', _sdict["key1"])

    # transparent value setting
    print('now setting _sdict["key1"] = 12')
    _sdict["key1"] = 12

    # show actual item
    print('_sdict.get_item("key1"):', _sdict.get_item("key1"))

    # transparent value getting
    print('_sdict.get("key1"):', _sdict.get("key1"))
    print('_sdict["key1"]:', _sdict["key1"])

    # changing item itself
    print("now changing item to an unexpected type:")
    _sdict.set_item("key1", value="hello")
    print('_sdict.set_item("key1", value="hello")')

    # show actual item
    print('_sdict.get_item("key1"):', _sdict.get_item("key1"))

    # transparent value getting
    print('_sdict.get("key1"):', _sdict.get("key1"))
    print('_sdict["key1"]:', _sdict["key1"])

You may get a console output similar to this::

    StructDict() object contents:
    {'key1': <__main__.MyStructure object at 0x7ff308454eb8>, 'key3':
    <__main__.MyStructure object at 0x7ff308454fd0>, 'key2':
    <__main__.MyStructure object at 0x7ff308454f28>}
    _sdict.get_item("key1"): <__main__.MyStructure object at 0x7ff308454eb8>
    _sdict.get("key1"): 10
    _sdict["key1"]: 10
    now setting _sdict["key1"] = 12
    _sdict.get_item("key1"): <__main__.MyStructure object at 0x7ff308454eb8>
    _sdict.get("key1"): 12
    _sdict["key1"]: 12
    now changing item to an unexpected type:
    _sdict.set_item("key1", value="hello")
    _sdict.get_item("key1"): hello
    _sdict.get("key1"): hello
    _sdict["key1"]: hello

As you can see, a StructDict() object can be set up to any key/value
pair you like, but when you use an expected class structure type, it
behaves as if objects of this type were transparent to user and gets or
sets values through them, directly.

For example giving in an actual context, this class is extended by
**tkRAD.xml.RADXMLAttributesDict** class to help tkRAD lib components
manage XML attributes transparently.

*NEW* since **tkRAD v1.7**:

* *StructDict.item_value_getter* and *StructDict.item_value_setter*
  class member attributes **cannot be deleted** any more.

Please, feel free to post an issue ticket on GitHub's project
bugtracker if you face any troubleshooting with this new version of
tkRAD's library.

GitHub's project bugtracker: https://github.com/tarball69/tkRAD/issues
(requires a GitHub member account to proceed).
"""

"""
    tkRAD - tkinter Rapid Application Development library

    (c) 2013+ Raphaël SEBAN <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see: http://www.gnu.org/licenses/
"""


class StructDict (dict):
    """Extended dictionary for transparent class structure value
    access.

    Please, notice the following features are also implemented here::

        _sdict = StructDict()

        # direct setting (see also set() method documentation)
        _sdict["key1"] = "hello"

        # direct getting (same behaviour than a Python's dict() object
        # all but the pretty difference mentioned in get() method
        # documentation)
        print(_sdict["key1"])

    """

    def __getitem__ (self, key):
        _item = super().__getitem__(key)
        if isinstance(_item, self.item_type):
            return getattr(_item, self.item_value_getter)()
        else:
            return _item
        # end if
    # end def

    def __init__ (self, *args, **kw):
        """Class constructor."""
        super().__init__(*args, **kw)
        self.item_type = None
        self.item_value_getter = "get_value"
        self.item_value_setter = "set_value"
    # end def

    def __setitem__ (self, key, value):
        _item = super().get(key)
        if isinstance(_item, self.item_type):
            getattr(_item, self.item_value_setter)(value)
        else:
            super().__setitem__(key, value)
        # end if
    # end def

    def flatten (self):
        """Return flatten values dict() object.

        Returns a new Python dict() object of key/value pairs where all
        expected structure type items are replaced by their actual
        values and all other unexpected types are kept unchanged.

        This is done to get a flat dictionary of key/value pairs as if
        expected structure type had never existed.

        Genuine 'self' dictionary is left *UNTOUCHED* after this
        operation.
        """
        return {_key: self[_key] for _key in self}
    # end def

    def get (self, key, default=None):
        """Transparent value getter.

        Parameters:
            key -- dict key index.

            default -- default value to provide if 'key' is *NOT* in
            dict (default: None).

        Works in a similar way than Python's dict.get() method, at the
        pretty difference that when an expected type item is met, it is
        replaced by its actual value.
        """
        _item = super().get(key, default)
        if isinstance(_item, self.item_type):
            return getattr(_item, self.item_value_getter)()
        else:
            return _item
        # end if
    # end def

    def get_item (self, key, default=None):
        """Raw value getter.

        Parameters:
            key -- dict key index.

            default -- default value to provide if 'key' is *NOT* in
            dict (default: None).

        Returns the real 'value' item in dict's 'key/value' pair
        regardless to its expected type or not.
        """
        return super().get(key, default)
    # end def

    def get_value (self, key, default=None):
        """Transparent value getter (alias).

        Parameters:
            key -- dict key index.

            default -- default value to provide if 'key' is *NOT* in
            dict (default: None).

        Works in a similar way than Python's dict.get() method, at the
        pretty difference that when an expected type item is met, it is
        replaced by its actual value.

        This method is merely an alias for self.get() method provided
        for comfort, readability and easiness.
        """
        return self.get(key, default)
    # end def

    @property
    def item_value_getter (self):
        """Expected type item value getter - read-write attribute.

        This is the name of the method retrieving the actual value from
        expected type item.

        This method *MUST* be implemented into expected class structure
        type.

        May raise **StructDictError** if attribute is *NOT* a plain
        string of chars.
        """
        return self.__item_value_getter
    # end def

    @item_value_getter.setter
    def item_value_getter (self, getter):
        if getter and isinstance(getter, str):
            self.__item_value_getter = getter
        else:
            raise StructDictError(
                "Item value getter must be a plain string of chars."
            )
        # end if
    # end def

    @property
    def item_value_setter (self):
        """Expected type item value setter - read-write attribute.

        This is the name of the method setting the actual value into
        expected type item.

        This method *MUST* be implemented into expected class structure
        type.

        May raise **StructDictError** if attribute is *NOT* a plain
        string of chars.
        """
        return self.__item_value_setter
    # end def

    @item_value_setter.setter
    def item_value_setter (self, setter):
        if setter and isinstance(setter, str):
            self.__item_value_setter = setter
        else:
            raise StructDictError(
                "Item value setter must be a plain string of chars."
            )
        # end if
    # end def

    def set (self, key, value):
        """Transparent value setter.

        Parameters:
            key -- dict key index.

            value -- value to set for 'key' index.

        Works in a similar way than Python's dict[key] = value, at the
        pretty difference that when an expected type item is met, it is
        its actual value that is set to 'value'.

        The expected type item is *NOT* replaced by 'value' in this
        case.

        Notice, you can merely use Python's self[key] = value notation
        instead of this method, it will work too.
        """
        self.__setitem__(key, value)
    # end def

    def set_item (self, key, value):
        """Raw value setter.

        Parameters:
            key -- dict key index.

            value -- value to set for 'key' index.

        Merely sets key/value pair indexed by 'key' to 'value',
        regardless to its expected type or not.

        The expected type item *IS* replaced by 'value' in this case.
        """
        super().__setitem__(key, value)
    # end def

    def set_value (self, key, value):
        """Transparent value setter (alias).

        Parameters:
            key -- dict key index.

            value -- value to set for 'key' index.

        Works in a similar way than Python's dict[key] = value, at the
        pretty difference that when an expected type item is met, it is
        its actual value that is set to 'value'.

        The expected type item is *NOT* replaced by 'value' in this
        case.

        Notice, you can merely use Python's self[key] = value notation
        instead of this method, it will work too.

        This method is merely an alias for self.set() method provided
        for comfort, readability and easiness.
        """
        self.set(key, value)
    # end def

# end class


class StructDictError (Exception):
    """Error handler for StructDict class."""
    pass
# end class
