#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Extra tools module.

Implements:

* module function choose -- return the first True-tested argument.
* module function choose_if -- return the first True callback-tested
  argument.
* module function choose_num -- return the first True callback-tested
  numeric argument.
* module function choose_str -- return the first plain string of chars
  argument.
* module function choose_type -- return the first class type matching
  argument.
* module function dict_delete_items -- return new dict() object with
  stripped listed keys.
* module function dict_only_keys -- return new dict() object with only
  listed keys.
* module function ensure_float -- ensure argument is actually a
  floating point number.
* module function ensure_int -- ensure argument is actually an integer
  number.
* module function is_num -- indicate if argument is numeric (int or
  float).
* module function is_pdict -- indicate if argument is a plain Python
  dict() object.
* module function is_plist -- indicate if argument is a plain Python
  list() object.
* module function is_pstr -- indicate if argument is a plain string of
  chars.
* module function is_ptuple -- indicate if argument is a plain Python
  tuple() object.
* module function normalize_id -- normalize to Python's identifier
  syntax.
* module function normalize_import -- normalize to Python's module
  importation syntax.
* module function normalize_relative_module -- normalize to Python's
  relative module importation syntax.
* module function str_complete -- string completion.

*NEW* since **tkRAD v1.7**:

Please, feel free to post an issue ticket on GitHub's project
bugtracker if you face any troubleshooting with this new version of
tkRAD's library.

GitHub's project bugtracker: https://github.com/tarball69/tkRAD/issues
(requires a GitHub member account to proceed).
"""

"""
    tkRAD - tkinter Rapid Application Development library

    (c) 2013+ Raphaël SEBAN <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see: http://www.gnu.org/licenses/
"""

import re


def choose (*args):
    """Return the first True-tested argument.

    Parameter:
        args -- variable list of arguments.

    Returns the first argument from 'args' which has been tested True
    i.e. non-zero value or non-empty object.

    Returns None if no argument succeeds.
    """
    for _arg in args:
        if _arg:
            return _arg
        # end if
    # end for
    return None
# end def


def choose_if (callback, *args):
    """Return the first True callback-tested argument.

    Parameter:
        callback -- function or method to use as a filtering feature.

        args -- variable list of arguments.

    Returns the first argument from 'args' which has been tested True
    through 'callback' filtering feature.

    The 'callback' feature must expect at least one required argument:
    the argument to be tested.

    Returns None if no argument succeeds.
    """
    for _arg in args:
        if callback(_arg):
            return _arg
        # end if
    # end for
    return None
# end def


def choose_num (callback, *args):
    """Return the first True callback-tested numeric argument.

    Parameter:
        callback -- function or method to use as a filtering feature.

        args -- variable list of numeric arguments (int or float).

    Returns the first argument from 'args' which is a number (int or
    float) and which has also been tested True through 'callback'
    filtering feature.

    The 'callback' feature must expect at least one required argument:
    the argument to be tested.

    Returns 0 if no argument succeeds.
    """
    if not callable(callback):
        callback = lambda n: True
    # end if
    for _arg in args:
        if is_num(_arg) and callback(_arg):
            return _arg
        # end if
    # end for
    return 0
# end def


def choose_str (*args):
    """Return the first plain string of chars argument.

    Parameter:
        args -- variable list of arguments.

    Returns the first argument from 'args' which is a non-empty (plain)
    string of chars.

    Returns empty string if no argument succeeds.
    """
    for _arg in args:
        if is_pstr(_arg):
            return _arg
        # end if
    # end for
    return ""
# end def


def choose_type (class_type, *args):
    """Return the first class type matching argument.

    Parameter:
        class_type -- class type to match (subclasses accepted).

        args -- variable list of arguments.

    Returns the first argument from 'args' that is an instance of
    'class_type' class type or at least an instance of one of its
    subclasses, if any.

    Returns None if no argument succeeds.

    *NEW* since **tkRAD v1.7**: renamed 'object_type' parameter to
    'class_type'.
    """
    for _arg in args:
        if isinstance(_arg, class_type):
            return _arg
        # end if
    # end for
    return None
# end def


def dict_delete_items (dict_object, *keys):
    """Return new dict() object with stripped listed keys.

    Parameters:
        dict_object -- Python dict() object.

        keys -- variable list of 'dict_object' keys.

    Copies 'dict_object' into a new dict() object and then removes each
    key listed in 'keys' from copied object.

    Will return a new dict() object whatever 'dict_object' parameter
    actually is.

    The eventual dict() object in 'dict_object' parameter is left
    *UNTOUCHED* after this operation.

    *NEW* since **tkRAD v1.7**: renamed 'args' parameter to 'keys'.

    *NEW* since **tkRAD v1.2**.
    """
    _dict = dict()
    if is_pdict(dict_object):
        _dict = dict_object.copy()
        _keys = set(keys).intersection(_dict.keys())
        for _key in _keys:
            _dict.pop(_key, None)
        # end for
    # end if
    return _dict
# end def


def dict_only_keys (dict_object, *keys):
    """Return new dict() object with only listed keys.

    Parameters:
        dict_object -- Python dict() object.

        keys -- variable list of 'dict_object' keys.

    Creates a new dict() object and then copies only listed 'keys'
    items from 'dict_object' into it.

    Will return a new dict() object whatever 'dict_object' parameter
    actually is.

    The eventual dict() object in 'dict_object' parameter is left
    *UNTOUCHED* after this operation.

    *NEW* since **tkRAD v1.7**: renamed 'args' parameter to 'keys'.

    *NEW* since **tkRAD v1.2**.
    """
    _dict = dict()
    if is_pdict(dict_object):
        _keys = set(keys).intersection(dict_object.keys())
        for _key in _keys:
            _dict[_key] = dict_object[_key]
        # end for
    # end if
    return _dict
# end def


def ensure_float (arg):
    """Ensure argument is actually a floating point number.

    Parameter:
        arg -- argument to operate.

    Returns 'arg' converted to an actual float() object on success or
    0.0 otherwise.
    """
    try:
        return float(eval(str(arg)))
    except:
        return 0.0
    # end try
# end def


def ensure_int (arg):
    """Ensure argument is actually an integer number.

    Parameter:
        arg -- argument to operate.

    Returns 'arg' converted to an actual int() object on success or 0
    otherwise.
    """
    try:
        return int(eval(str(arg)))
    except:
        return 0
    # end try
# end def


def is_num (arg):
    """Indicate if argument is numeric (int or float)."""
    return isinstance(arg, (int, float))
# end def


def is_pdict (arg):
    """Indicate if argument is a plain Python dict() object."""
    return bool(arg and isinstance(arg, dict))
# end def


def is_plist (arg):
    """Indicate if argument is a plain Python list() object."""
    return bool(arg and isinstance(arg, list))
# end def


def is_pstr (arg):
    """Indicate if argument is a plain string of chars."""
    return bool(arg and isinstance(arg, str))
# end def


def is_ptuple (arg):
    """Indicate if argument is a plain Python tuple() object."""
    return bool(arg and isinstance(arg, tuple))
# end def


def normalize_id (arg, raise_error=False):
    """Normalize to Python's identifier syntax.

    Parameters:
        arg -- argument to operate.

        raise_error -- boolean flag (default: False). If True, may
        raise **TypeError** if 'arg' is *NOT* a plain string of chars.
        Remains silent otherwise.

    Filters 'arg' to match with Python's identifier syntax i.e. only
    [a-zA-Z0-9\_] characters are allowed.

    Returns formatted string on success, empty string otherwise.
    """
    if is_pstr(arg):
        return re.sub(r"\W+", "", arg)
    elif raise_error:
        raise TypeError("Expected plain string of chars.")
    # end if
    return ""
# end def


def normalize_import (arg, raise_error=False):
    """Normalize to Python's module importation syntax.

    Parameters:
        arg -- argument to operate.

        raise_error -- boolean flag (default: False). If True, may
        raise **TypeError** if 'arg' is *NOT* a plain string of chars.
        Remains silent otherwise.

    Filters 'arg' to match with Python's module importation syntax e.g.
    'from sequence import **identifier_or_star**'.

    Returns formatted string on success, empty string otherwise.

    *NEW* since **tkRAD v1.7**: fixed some little quirks in syntax
    conformation.
    """
    if "*" in str(arg):
        return "*"
    else:
        return normalize_relative_module(arg, raise_error).strip(".")
    # end if
# end def


def normalize_relative_module (arg, raise_error=False):
    """Normalize to Python's relative module importation syntax.

    Parameters:
        arg -- argument to operate.

        raise_error -- boolean flag (default: False). If True, may
        raise **TypeError** if 'arg' is *NOT* a plain string of chars.
        Remains silent otherwise.

    Filters 'arg' to match with Python's relative module importation
    syntax e.g. 'from **..pkg.subpkg.submodule** import identifier'.

    Returns formatted string on success, empty string otherwise.

    *NEW* since **tkRAD v1.7**: fixed some little quirks in syntax
    conformation.
    """
    if is_pstr(arg):
        arg = re.sub(r"[^\.\w]+", "", arg)
        arg = re.sub(r"\b\.+", ".", arg)
        arg = re.sub(r"^\.{3,}", "..", arg).rstrip(".")
        return arg
    elif raise_error:
        raise TypeError("Expected plain string of chars.")
    # end if
    return ""
# end def


def str_complete (str_format, str_value, default=""):
    """String completion.

    Parameters:
        str_format -- string of chars containing str.format() syntax
        elements for string substitution / completion.

        str_value -- 'str_format' substitution value (string of chars).

        default -- default value to provide on failure (default: "").

    Tries to make string substitution / completion of 'str_format'
    along with 'str_value' substitution value.

    Returns formatted string on success, 'default' value otherwise.

    Please, notice 'str_value' parameter must be a non-empty (plain)
    string of chars to be considered under substitution process.

    This is made to ensure 'str_format' is actually completed with
    plain values (non-zero, non-empty values).

    Also notice if 'str_format' does *NOT* contain **only one unnamed**
    field to substitute, it may raise an exception error on
    str.format() operation.

    Example::

        # good
        _result = tools.str_complete(
            "use this only if '{}' exists", my_value, "no way!"
        )

        # IndexError
        _result = tools.str_complete(
            "use this only if '{}{}' exists", my_value, "no way!"
        )

        # KeyError
        _result = tools.str_complete(
            "use this only if '{value}' exists", my_value, "no way!"
        )

    """
    if is_pstr(str_format) and is_pstr(str_value):
        return str_format.format(str_value)
    else:
        return default
    # end if
# end def
