#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Simple XML widget factory module.

Implements:

* module function build -- build XML source into a tkinter parent
  widget.
* module function is_pstr -- indicate if argument is a plain string of
  chars or not.
* module function normalize_id -- normalize string of chars.
* class Builder -- simple XML widget factory class.
* class BuilderError -- error handler for Builder class.

Example 1::

    from tkRAD.easy import builder
    xml_source = \"""
        <root>
            <label text="Hello good people!" />
            <button text="Quit" command="self.quit" />
        </root>
    \"""
    builder.build(xml_source)

Example 2::

    import tkinter as tk
    from tkRAD.easy.builder import Builder
    xml_source = \"""
        <root>
            <label text="Please, choose your meal:"/>
            <checkbutton text="Pizza" anchor="w"/>
            <checkbutton text="Hotdog" anchor="w"/>
            <checkbutton text="Hamburger" anchor="w"/>
        </root>
    \"""
    def show_contents ():
        button.configure(text="Quit", command=root.destroy)
        frame.build(xml_source)
    # end def
    root = tk.Tk()
    frame = Builder(root)
    frame.pack(padx=5, pady=5)
    button = tk.Button(root, text="Click here", command=show_contents)
    button.pack(padx=5, pady=5)
    root.mainloop()

*NEW* since **tkRAD v1.7**:

* by now, this module can be copied elsewhere as a standalone script
  and run either with a Python2 interpreter or with a Python3
  interpreter;
* code has been rewritten to best comply with good programming
  practices i.e. PEP8, PEP0257 and PEP0263.

Please, feel free to post an issue ticket on GitHub's project
bugtracker if you face any troubleshooting with this new version of
tkRAD's library.

GitHub's project bugtracker: https://github.com/tarball69/tkRAD/issues
(requires a GitHub member account to proceed).
"""

"""
    tkRAD - tkinter Rapid Application Development library

    (c) 2013+ Raphaël SEBAN <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see: http://www.gnu.org/licenses/
"""

import re
import xml.etree.ElementTree as etree

try:
    import tkinter as tk
    from tkinter import messagebox as msgbox
except:
    import Tkinter as tk
    import tkMessageBox as msgbox
# end try


def build (xml, master=None, **kw):
    """Build XML source into a tkinter parent widget.

    Parameters:
        xml -- XML source (string of chars). Can either be a direct XML
        script or a file path to an XML script.

        master -- tkinter widgets container (default: None). If omitted
        or None, will automagically be replaced by a 'Tk()' Toplevel
        window instance.

        kw -- keyword arguments.

    Supported keyword arguments:
        layout_options -- w.pack() layout options. see documentation:
        http://effbot.org/tkinterbook/pack.htm

    Builds 'xml' source into 'master' tkinter parent widget.

    Returns a Builder() object instance i.e. a subclass of
    tkinter.Frame widgets container.
    """
    return Builder(master).build(xml, **kw)
# end def


def is_pstr (arg):
    """Indicate if argument is a plain string of chars or not."""
    return arg and isinstance(arg, str)
# end def


def normalize_id (arg):
    """Normalize string of chars.

    Parameter:
        arg -- string of chars.

    Cleans up 'arg' in order to match only [a-zA-Z0-9\_] characters.

    Any other character will be removed from 'arg'.

    Returns normalized string of chars or empty string if 'arg' was of
    unexpected type.
    """
    if is_pstr(arg):
        return re.sub(r"\W+", "", arg)
    # end if
    return ""
# end def


class Builder (tk.Frame):
    """Simple XML widget factory class."""

    ATTRIBUTE_PARSER = "_parse_attr_{attr}"

    OI_COUNT = 1

    PACK_OPTIONS = {
        "side": tk.TOP,
        "expand": 1,
        "fill": tk.BOTH,
        "padx": 0,
        "pady": 0,
    }

    TK_CLASSES = {
        "button":           "tk.Button",
        "canvas":           "tk.Canvas",
        "checkbutton":      "tk.Checkbutton",
        "entry":            "tk.Entry",
        "frame":            "tk.Frame",
        "label":            "tk.Label",
        "labelframe":       "tk.LabelFrame",
        "listbox":          "tk.Listbox",
        "menu":             "tk.Menu",
        "menubutton":       "tk.Menubutton",
        "message":          "tk.Message",
        "optionmenu":       "tk.OptionMenu",
        "panedwindow":      "tk.PanedWindow",
        "radiobutton":      "tk.Radiobutton",
        "scale":            "tk.Scale",
        "scrollbar":        "tk.Scrollbar",
        "spinbox":          "tk.Spinbox",
        "text":             "tk.Text",
        "toplevel":         "tk.Toplevel",
    }

    def __init__ (self, master=None, **tk_options):
        """Class constructor."""
        tk.Frame.__init__(self, master, **tk_options)
        self.__autorun = not master
        self.xml_tree = None
        self.objects = dict()
    # end def

    def _build_element (self, xml_element, tk_parent):
        _tag = xml_element.tag
        _tagl = _tag.lower()
        if _tagl in ("root", "tkwidget"):
            _widget = tk_parent
        else:
            self._parse_xml_attributes(xml_element, tk_parent)
            _id = xml_element.attrib.pop("id", None)
            _classtype = eval(self.TK_CLASSES.get(_tagl, _tag))
            _widget = _classtype(tk_parent, **xml_element.attrib)
            self._register_object_by_id(_widget, _id)
            _widget.pack(**self.layout_options)
        # end if
        for _child in xml_element:
            self._build_element(_child, _widget)
        # end for
    # end def

    def _get_correct_id (self, attr_id):
        return normalize_id(attr_id) or self._get_unique_id("object")
    # end def

    def _get_unique_id (self, radix):
        if is_pstr(radix):
            while self.OI_COUNT:
                _uid = radix + str(self.OI_COUNT)
                self.OI_COUNT += 1
                if _uid not in self.objects:
                    return _uid
                # end if
            # end while
        # end if
        return None
    # end def

    def _init_xml_tree (self, xml_source):
        self.xml_tree = None
        self.OI_COUNT = 1
        if is_pstr(xml_source):
            if re.search("<.*?>", xml_source):
                self.xml_tree = etree.ElementTree(
                    element=etree.fromstring(xml_source)
                )
            else:
                self.xml_tree = etree.parse(xml_source)
            # end if
        else:
            raise BuilderError(
                "XML source must be a plain string of chars."
            )
        # end if
    # end def

    def _parse_attr_command (self, value, attrs, **kw):
        attrs["command"] = eval(value)
    # end def

    def _parse_attr_id (self, value, attrs, **kw):
        attrs["id"] = self._get_correct_id(value)
    # end def

    def _parse_xml_attributes (self, xml_element, tk_parent):
        _kw = {"xml_element": xml_element, "tk_parent": tk_parent}
        _attrs = xml_element.attrib
        for (_attr, _value) in _attrs.items():
            _parser = str(self.ATTRIBUTE_PARSER).format(
                attr=_attr.lower()
            )
            try:
                getattr(self, _parser)(_value, _attrs, **_kw)
            except:
                pass
            # end try
        # end for
    # end def

    def _register_object_by_id (self, built_object, attr_id):
        self.objects[self._get_correct_id(attr_id)] = built_object
    # end def

    def build (self, xml_source, **kw):
        """Build widgets from XML source.

        Parameters:
            xml_source -- XML source (string of chars). Can either be a
            direct XML script or a file path to an XML script.

            kw -- keyword arguments.

        Supported keyword arguments:
            layout_options -- w.pack() layout options. see
            documentation: http://effbot.org/tkinterbook/pack.htm

        Builds widgets from 'xml_source'.

        Returns 'self' as a subclass of tkinter.Frame widgets
        container.
        """
        try:
            self._init_xml_tree(xml_source)
            self.layout_options = kw.get(
                "layout_options", self.PACK_OPTIONS
            )
            self._build_element(self.xml_tree.getroot(), self)
            if self.__autorun:
                self.pack(**self.PACK_OPTIONS)
                self.mainloop()
                try:
                    self.master.destroy()
                except:
                    pass
                # end try
                return None
            # end if
            return self
        except Exception as err:
            msgbox.showerror(
                title="Error",
                message="An error has occurred:\n\n{}".format(err),
                parent=self,
            )
            raise
        # end try
        return None
    # end def

    def get_object_by_id (self, attr_id):
        """Get tkinter widget by its XML id attribute.

        Parameter:
            attr_id -- XML 'id' attribute (string of chars).

        Returns the tkinter widget corresponding to 'attr_id' XML id
        attribute on success, None otherwise.
        """
        return self.objects.get(normalize_id(attr_id))
    # end def

    def hide_widgets (self, *args, **kw):
        """Hide all child widgets.

        Event handler signature.
        """
        for _widget in self.winfo_children():
            _widget.pack_forget()
        # end for
    # end def

    def set_pack_options (self, **options):
        """Reset default w.pack() layout options.

        Parameter:
            options -- w.pack() layout options (keyword arguments). See
            documentation: http://effbot.org/tkinterbook/pack.htm

        Merely sets 'PACK_OPTIONS' class member attribute to a shallow
        copy of 'options' mapping dict().
        """
        self.PACK_OPTIONS = options.copy()
    # end def

# end class


class BuilderError (Exception):
    """Error handler for Builder class."""
    pass
# end class


if __name__ == "__main__":
    xml = """
        <root>
            <label text="hello good people!"/>
            <button text="OK" command="self.quit"/>
        </root>
    """
    build(xml)
# end if
