#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Application module.

Implements:

* class RADApplication -- tkRAD generic application framework class.

Example::

    import tkRAD
    xml_source = \"""
        <tkwidget>
            <label
                text="Hello good people!"
                layout="pack"
                resizable="yes"
            />
            <button
                text="Quit"
                command="@quit"
                layout="pack"
            />
        </tkwidget>
    \"""
    tkRAD.RADApplication().run(xml=xml_source)

*NEW* since **tkRAD v1.7**:

* class implementation has entirely been rewritten -- less code, more
  freedom;
* class attribute *RC_OPTIONS* now uses "rc_dir" and "rc_file" key
  entries instead of "user_dir", "user_file", "app_dir" and "app_file"
  key entries;
* joined all RC configuration files into one unique application's
  global configuration file, all options from now accessable through
  *tkRAD.core.options.get_option_manager()* or the *self.options* class
  member attribute;
* please, notice *RC_OPTIONS["rc_dir"]* entry will replace
  *"tkRAD_application"* mention by **current class name**. Subclassing
  *RADApplication* with your own application name e.g. let's say
  *MyCoolApp* will make RC_OPTIONS["rc_dir"] become something like
  *"~/.config/MyCoolApp"*, unless you redefine all *RC_OPTIONS* class
  attribute itself, of course;
* removed GUI/CLI running mode switcher: this should be author's coding
  choice;
* new features taken apart, backward compatibility with older versions
  of tkRAD has been respected as much as possible.

Please, feel free to post an issue ticket on GitHub's project
bugtracker if you face any troubleshooting with this new version of
tkRAD's library.

GitHub's project bugtracker: https://github.com/tarball69/tkRAD/issues
(requires a GitHub member account to proceed).
"""

"""
    tkRAD - tkinter Rapid Application Development library

    (c) 2013+ Raphaël SEBAN <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see: http://www.gnu.org/licenses/
"""

from ..core import i18n
from ..core import tools


class RADApplication:
    """tkRAD generic application framework class.

    This class has been designed for comfort and easiness but is *NOT*
    required to make a tkRAD application run properly.

    Feel free to extend or to override any provided hook methods to
    best meet your own needs.
    """

    APP = {
        "name": _("application name"),
        "version": _("0.1a"),
        "description": _("application description"),
        "title": _("My application is wonderful!"),
        "author": _("author name <email@domain.org>"),
        "copyright": _("(c) YEAR author name."),
        "license": _("Licensed under GNU General Public License v3."),
        "license_url": _("http://www.gnu.org/licenses/"),
    }

    DIRECTORIES = (
        #~ "locale", "src", "xml",
    )

    PYTHON = {
        "version": "3.2",
        "strict": False,
    }

    RC_OPTIONS = {
        "rc_dir": "~/.config/tkRAD_application",
        "rc_file": "application.rc",
    }

    def __init__ (self, *args, **kw):
        """Class constructor."""
        self.initialize(*args, **kw)
    # end def

    def finalize (self, *args, **kw):
        """Finalize application - hook method.

        Parameters:
            args -- variable list of arguments.

            kw -- keyword arguments.

        Supported keyword arguments:
            -- none at this stage --

        Finalizes application after having ran it.

        Feel free to extend or to override this hook method to best
        meet your own needs.

        *NEW* since **tkRAD v1.7**.
        """
        self.options.save()
        try:
            i18n.dump_pot_file()
        except:
            pass
        # end try
    # end def

    def initialize (self, *args, **kw):
        """Initialize application - hook method.

        Parameters:
            args -- variable list of arguments.

            kw -- keyword arguments.

        Supported keyword arguments:
            app_root_dir -- application's root directory, string of
            chars (default: None). If omitted or None, will be replaced
            by path.get_app_root_dir() default value.

            check_dirs -- list of directory names to check up as
            application required directories (default: None). If
            omitted or None, will be replaced by 'DIRECTORIES' class
            attribute value.

            lc_dir -- locale translations directory path (default:
            None). If omitted or None, will be replaced by current
            i18n.get_translations_dir() path.

            lc_lang -- locale translations language (default: None). If
            omitted or None, will be replaced by current
            i18n.get_translations_lang() string of chars. This option
            should comply with IETF's RFC 1766 country code tags e.g.
            fr, fr_FR, de, it, en, en_US and so on. See:
            https://tools.ietf.org/html/rfc1766

            no_dependencies -- boolean flag (default: None). If True,
            disables application required directories checking. This
            option has priority over 'check_dirs'.

            python_version -- required Python version to run this
            application, string of chars (default: None). If omitted or
            None, will be replaced by 'PYTHON["version"]' class
            attribute value.

            python_strict -- boolean flag for strict mode version
            checking (default: None). If omitted or None, will be
            replaced by 'PYTHON["strict"]' class attribute value.

            rc_dir -- directory path to configuration file, string of
            chars (default: None). If omitted or None, will be replaced
            by 'RC_OPTIONS["rc_dir"]' class attribute value.

            rc_file -- configuration file's filename, string of chars
            (default: None). If omitted or None, will be replaced by
            'RC_OPTIONS["rc_file"]' class attribute value.

        Initializes application before actually running it.

        Feel free to extend or to override this hook method to best
        meet your own needs.

        *NEW* since **tkRAD v1.7**.
        """
        import inspect
        import os
        from ..core import checkups
        from ..core import options
        from ..core import path
        from ..core import services
        # check required Python version
        _python = {
            "version": tools.choose_str(
                kw.get("python_version"),
                self.PYTHON.get("version"),
                "3.2",
            ),
            "strict": tools.choose(
                kw.get("python_strict"),
                self.PYTHON.get("strict"),
                False,
            ),
        }
        checkups.python_require(**_python)
        # set application's root directory
        path.set_app_root_dir(
            tools.choose_str(
                kw.get("app_root_dir"),
                os.path.dirname(
                    os.path.realpath(inspect.stack()[-1][1])),
                ".",
            )
        )
        _cwd = path.get_app_root_dir()
        os.chdir(_cwd)
        # check required directories
        if not kw.get("no_dependencies"):
            checkups.check_directories(
                _cwd, *(kw.get("check_dirs") or self.DIRECTORIES or ())
            )
        # end if
        # i18n support
        i18n.install(kw.get("lc_dir"), kw.get("lc_lang"))
        # load application's options
        try:
            self.RC_OPTIONS["rc_dir"] = (
                self.RC_OPTIONS["rc_dir"].replace(
                    "tkRAD_application", self.__class__.__name__
                )
            )
        except:
            pass
        # end try
        self.options = options.get_option_manager(
            rc_dir=tools.choose_str(
                kw.get("rc_dir"),
                self.RC_OPTIONS.get("rc_dir"),
                "~/.config/tkRAD_application",
            ),
            rc_file=tools.choose_str(
                kw.get("rc_file"),
                self.RC_OPTIONS.get("rc_file"),
                "application.rc",
            )
        )
        self.options.load()
        # register application as a service
        self.services = services.get_service_manager()
        services.register_service("app", self)
        services.register_service("application", self)
    # end def

    def main (self, *args, **kw):
        """Application's main procedure - hook method.

        Parameters:
            args -- variable list of arguments.

            kw -- keyword arguments.

        Supported keyword arguments:
            xml -- XML source for application's mainwindow contents,
            string of chars (default: None). This can either be an
            actual XML script or a file path to an XML script. If
            omitted or None, will be replaced by some welcome XML code.

            xml_menu -- XML source for application's toplevel menu,
            string of chars (default: None). This can either be an
            actual XML script or a file path to an XML script. If
            omitted or None, will *NOT* build any toplevel menu.

        This is the heart of application running. Should be called
        between 'initialize()' and 'finalize()' hook methods.

        Feel free to extend or to override this hook method to best
        meet your own needs.

        *NEW* since **tkRAD v1.7**.
        """
        from ..xml import rad_xml_mainwindow
        self.mainwindow = rad_xml_mainwindow.RADXMLMainWindow()
        try:
            if kw.get("xml_menu"):
                self.mainwindow.topmenu.xml_build(kw.get("xml_menu"))
            # end if
            self.mainwindow.mainframe.xml_build(kw.get("xml"))
            self.mainwindow.connect_statusbar("show_statusbar")
            self.mainwindow.run()
        except FileNotFoundError:
            if self.mainwindow.mainframe.winfo_children():
                raise
            else:
                self.mainwindow.mainframe.xml_build(
                    r"""
                    <tkwidget>
                        <frame
                            layout="pack"
                            resizable="yes"
                        />
                        <label
                            text="this should better work with:"
                            layout="pack"
                        />
                        <label
                            text="{path}"
                            fg="red"
                            layout="pack"
                        />
                        <label
                            text="/!\ don't forget to create missing directories /!\"
                            layout="pack"
                        />
                        <button
                            text="Quit"
                            command="@quit"
                            layout="pack"
                        />
                        <frame
                            layout="pack"
                            resizable="yes"
                        />
                    </tkwidget>
                    """
                    .format(
                        path=self.mainwindow.mainframe.get_xml_path()
                    )
                )
                self.mainwindow.run()
            # end if
        # end try
    # end def

    def run (self, *args, **kw):
        """Run application - hook method.

        Parameters:
            args -- variable list of arguments.

            kw -- keyword arguments.

        Supported keyword arguments:
            xml -- XML source for application's mainwindow contents,
            string of chars (default: None). This can either be an
            actual XML script or a file path to an XML script. If
            omitted or None, will be replaced by some welcome XML code.

            xml_menu -- XML source for application's toplevel menu,
            string of chars (default: None). This can either be an
            actual XML script or a file path to an XML script. If
            omitted or None, will *NOT* build any toplevel menu.

        Application's actual entry point.

        Feel free to extend or to override this hook method to best
        meet your own needs.

        *NEW* since **tkRAD v1.7**: entirely reimplemented this class.
        """
        self.main(*args, **kw)
        self.finalize(*args, **kw)
    # end def

# end class
