#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""tkRAD Dialog widget module.

Implements:

* class RADDialog -- tkRAD base dialog widget class.

Example 1::

    import tkinter as tk
    from tkRAD.widgets.rad_dialog import RADDialog
    root = tk.Tk()
    root.withdraw()
    RADDialog().show()

Example 2::

    import tkinter as tk
    from tkRAD.widgets.rad_dialog import RADDialog

    def show_dialog ():
        RADDialog(root).show()
    # end def

    root = tk.Tk()
    root.geometry("320x240+400+200")
    tk.Button(
        root, text="Show dialog", command=show_dialog
    ).pack(pady=5)
    tk.Button(root, text="Quit", command=root.destroy).pack(pady=5)
    root.mainloop()

*NEW* since **tkRAD v1.7**:

* entirely rewritten all code in this module;
* removed now useless *RADButtonsDialog* class, as *RADDialog* suffices
  for all user interaction dialog purpose;
* new code is likely to break backward compatibility with older
  versions of tkRAD;
* it is *NOT* recommended to use this module with old apps, as it may
  lead to a full code migration process.

Please, feel free to post an issue ticket on GitHub's project
bugtracker if you face any troubleshooting with this new version of
tkRAD's library.

GitHub's project bugtracker: https://github.com/tarball69/tkRAD/issues
(requires a GitHub member account to proceed).
"""

"""
    tkRAD - tkinter Rapid Application Development library

    (c) 2013+ Raphaël SEBAN <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see: http://www.gnu.org/licenses/
"""

import tkinter as tk
from tkinter import ttk

from . import rad_widget_base as rwb


class RADDialog (rwb.RADWidgetBase, tk.Toplevel):
    """tkRAD base dialog widget class.

    Dialog boxes are roughly very simple user interfaces (GUI) aimed to
    act as helpers in user's interaction with the application.

    Dialog boxes are transient whenever possible and modal i.e. they
    grab all events from caller.

    A call to *dialog.show()* method will block caller's thread of
    control until dialog terminates and returns its internal return
    value.

    If you need more sophisticated windows, please, take a look at
    *tkRAD.widgets.RADWindow* or *tkRAD.xml.RADXMLWindow* classes.
    """

    CONFIG = dict()

    def __init__ (self, master=None, **kw):
        self.CONFIG = self.CONFIG.copy()
        self.CONFIG.update(kw)
        tk.Toplevel.__init__(self, master)
        self.configure(**self._only_tk(self.CONFIG))
        rwb.RADWidgetBase.__init__(self, **kw)
        self.initialize(**kw)
    # end def

    def cancel (self, *args, **kw):
        """Cancel dialog window - hook method.

        Event handler signature.

        Cancels current dialog window.

        Current implementation merely destroys dialog window.

        Feel free to extend or to override this hook method to best
        meet your own needs.

        *NEW* since **tkRAD v1.7**.
        """
        self.destroy()
    # end def

    def center_dialog (self, *args, **kw):
        """Center dialog on parent or screen.

        Event handler signature.

        Centers dialog on parent window if it is visible, on all screen
        otherwise.
        """
        self.update_idletasks()
        if self.master.winfo_viewable():
            _root_width = self.master.winfo_width()
            _root_height = self.master.winfo_height()
            _root_x = self.master.winfo_rootx()
            _root_y = self.master.winfo_rooty()
            (_rel_x, _rel_y) = (0.5, 0.3)
        else:
            _root_width = self.winfo_screenwidth()
            _root_height = self.winfo_screenheight()
            _root_x = _root_y = 0
            _rel_x = _rel_y = 0.5
        # end if
        _dlg_width = self.winfo_reqwidth()
        _dlg_height = self.winfo_reqheight()
        _left = int(_root_x + (_root_width - _dlg_width) * _rel_x)
        _top = int(_root_y + (_root_height - _dlg_height) * _rel_y)
        self.geometry("+{x}+{y}".format(x=_left, y=_top))
    # end def

    def finalize (self, *args, **kw):
        """Finalize dialog window - hook method.

        Event handler signature.

        Finalizes current dialog window before actually destroying it.

        Current implementation does nothing.

        Feel free to extend or to override this hook method to best
        meet your own needs.

        *NEW* since **tkRAD v1.7**.
        """
        pass
    # end def

    def hide (self, *args, **kw):
        """Hide current window.

        Event handler signature.
        """
        self.withdraw()
    # end def

    def initialize (self, *args, **kw):
        """Initialize current widget - hook method.

        Parameters:
            args -- variable list of arguments.

            kw -- keyword arguments.

        Feel free to extend or to override this hook method to best
        meet your own needs.

        *NEW* since **tkRAD v1.7**.
        """
        self.hide()
        self.protocol("WM_DELETE_WINDOW", self.cancel)
        self.resizable(width=False, height=False)
        try:
            self.title(
                kw.get("title") or self.master.winfo_toplevel().title()
            )
        except:
            self.title("Dialog")
        # end try
        self.return_value = None
        self.setup_gui(*args, **kw)
    # end def

    @property
    def return_value (self):
        """Dialog's return value - read-write attribute.

        Set this attribute to whatever you need e.g. user's dialog
        response.

        Attribute's value will be returned to caller thread of control
        *after* the dialog window has been actually destroyed.

        Notice: this class attribute cannot be deleted.

        *NEW* since **tkRAD v1.7**.
        """
        return self.__return_value
    # end def

    @return_value.setter
    def return_value (self, value):
        self.__return_value = value
    # end def

    def set_transient (self, parent=None):
        """Set current dialog window a transient window for parent
        window.

        Parameter:
            parent -- parent window (default: None). If omitted or
            None, will be replaced by dialog's current parent widget
            i.e. *self.master* class member attribute.

        Sets current dialog window a transient window for 'parent'
        window i.e. if 'parent' window is visible, makes current dialog
        window always visible over 'parent' window.

        *NEW* since **tkRAD v1.7**.
        """
        parent = parent or self.master
        if parent.winfo_viewable():
            self.transient(parent)
        # end if
    # end def

    def setup_gui (self, *args, **kw):
        """Set up dialog window's GUI contents.

        Parameters:
            args -- variable list of arguments.

            kw -- keyword arguments.

        Supported keyword arguments:
            -- none at this stage --

        Current implementation merely sets up some welcome GUI contents
        for demonstration purpose.

        Feel free to extend or to override this hook method to best
        meet your own needs.

        *NEW* since **tkRAD v1.7**.
        """
        # set up body
        _body = ttk.Frame(self, padding=10)
        ttk.Label(
            _body, text=_("Put here whatever you need.")
        ).pack(padx=20, pady=20)
        # set up button box
        _buttonbox = ttk.Frame(self, padding=10)
        ttk.Button(
            _buttonbox, text=_("OK"), command=self.validate,
        ).pack(side=tk.LEFT, padx=10)
        ttk.Button(
            _buttonbox, text=_("Cancel"), command=self.cancel,
        ).pack(side=tk.RIGHT, padx=10)
        # layout management
        _body.pack(expand=1, fill=tk.BOTH)
        _buttonbox.pack(expand=1, fill=tk.X)
    # end def

    def show (self, *args, **kw):
        """Show current dialog window.

        Event handler signature.

        Shows current dialog window centered on parent widget or
        screen, blocks caller thread of control until dialog window is
        destroyed, returns *return_value* class attribute as user's
        interaction response.

        *NEW* since **tkRAD v1.7**: code has been entirely rewritten in
        this method.
        """
        self.center_dialog()
        self.set_transient()
        self.deiconify()
        self.wait_visibility(self)
        self.grab_set()
        self.wait_window(self)
        return self.return_value
    # end def

    def validate (self, *args, **kw):
        """Validate dialog window - hook method.

        Event handler signature.

        Validates current dialog window e.g. user clicked on dialog's
        'OK' button.

        Current implementation merely calls *cancel()* method by
        default.

        Feel free to extend or to override this hook method to best
        meet your own needs.

        *NEW* since **tkRAD v1.7**.
        """
        self.cancel()
    # end def

# end class
