#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Statusbar widget module.

Implements:

* class RADStatusBar -- tkRAD default statusbar widget.
* class RADStatusBarError -- error handler for RADStatusBar class.

Example::

    import tkinter as tk
    from tkRAD.widgets.rad_statusbar import RADStatusBar
    root = tk.Tk()
    # you can mix tkinter.Frame init options with tkRAD init options
    # unsupported options will be merely ignored
    my_statusbar = RADStatusBar(
        root,
        background="white",
        borderwidth=3,
        rc_dir="~/.config/testing",
        rc_file="statusbar.rc",
        relief=tk.SUNKEN,
    )
    my_statusbar.pack(
        anchor=tk.S, expand=1, fill=tk.X, padx=5, pady=5
    )
    root.mainloop()

*NEW* since **tkRAD v1.7**:

* error messages are *NOT* translated any more;
* added new *RADStatusBarError* error handler class;
* reviewed (and mostly rewritten) all code.

Please, feel free to post an issue ticket on GitHub's project
bugtracker if you face any troubleshooting with this new version of
tkRAD's library.

GitHub's project bugtracker: https://github.com/tarball69/tkRAD/issues
(requires a GitHub member account to proceed).
"""

"""
    tkRAD - tkinter Rapid Application Development library

    (c) 2013+ Raphaël SEBAN <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see: http://www.gnu.org/licenses/
"""

import tkinter as tk
from tkinter import ttk

from ..core import tools
from .rad_frame import RADFrame


class RADStatusBar (RADFrame):
    """tkRAD default statusbar widget."""

    MINIMUM_CONSISTENT_DELAY = 0.5  # seconds
    NOTIFICATION_DELAY = 5          # seconds

    def _get_state (self, value):
        return int(tools.ensure_int(value) != 0)
    # end def

    def _stop_notification (self):
        self.after_cancel(self.__notify_tid)
        self.__notify_tid = 0
    # end def

    def _toggle_var_state (self):
        return self._get_state(self.toggle_var.get())
    # end def

    def apply_options (self, *args, **kw):
        """Apply RC options to display state value.

        Event handler signature.

        Applies configuration file options to current statusbar display
        state (show/hide).

        Shows statusbar by default, if no RC option is given.

        *NEW* since **tkRAD v1.7**.

        *NEW* since **tkRAD v1.7.5**:

        * statusbar is now init'ed by its owner's RC section name;
        * section name falls back to 'gui' if owner hasn't a
          *get_rc_section()* method.
        """
        try:
            _section = self.tk_owner.get_rc_section()
        except:
            _section = "gui"
        # end try
        self.toggle_var_set(
            self.options.get(_section, "show_statusbar", fallback=1)
        )
    # end def

    @property
    def delay (self):
        """Time delay for notification display - read-write attribute.

        This class attribute is expressed in *seconds* (float).
        """
        return self.__delay
    # end def

    @delay.setter
    def delay (self, value):
        self.__delay = tools.ensure_float(
            max(value, self.MINIMUM_CONSISTENT_DELAY)
        )
    # end def

    @delay.deleter
    def delay (self):
        del self.__delay
    # end def

    def finalize (self, *args, **kw):
        """Finalize widget - hook method.

        Event handler signature.

        Finalizes current widget before actually destroying it.

        Feel free to extend or to override this hook method to best
        meet your own needs.

        *NEW* since **tkRAD v1.7**.

        *NEW* since **tkRAD v1.7.5**:

        * statusbar saves its status into owner's RC section name;
        * section name falls back to 'gui' if owner hasn't a
          *get_rc_section()* method.
        """
        try:
            _section = self.tk_owner.get_rc_section()
        except:
            _section = "gui"
        # end try
        self.options[_section]["show_statusbar"] = str(
            self._toggle_var_state()
        )
        self.options.save()
    # end def

    def get_correct_delay (self, delay=None):
        """Try to get the best notification delay (in seconds).

        Parameter:
            delay -- time delay in seconds (default: None).

        Looks through 'delay' and other different values and tries to
        get the best value for a consistent time delay for displaying a
        notification into the statusbar.
        """
        return tools.ensure_float(
            max(
                abs(self.MINIMUM_CONSISTENT_DELAY),
                tools.choose_num(
                    lambda x: x > 0,
                    delay,
                    self.options["gui"].get(
                        "statusbar_notification_delay"),
                    self.delay,
                    self.NOTIFICATION_DELAY,
                    5 # last but not least
                )
            )
        )
    # end def

    def hide (self, *args, **kw):
        """Hide current statusbar.

        Event handler signature.
        """
        self.toggle_var_set(0)
    # end def

    def info (self, text=None):
        """Display a static information text.

        Parameter:
            text -- text to display, string of chars (default: None).
            If omitted or None, will be replaced by default values e.g.
            "Ready." mention.

        Displays a static 'text' into statusbar.

        Notice: static texts are *NOT* related to any time delay.
        """
        self._stop_notification()
        self.__static_text = tools.choose_str(
            text, self.__static_text, _("Ready."), "Ready."
        )
        self.message_cvar.set(self.__static_text)
        self.update_idletasks()
    # end def

    def initialize (self, *args, **kw):
        """Initialize current widget - hook method.

        This method replaces deprecated *init_widget()* hook method.

        Feel free to extend or to override this hook method to best
        meet your own needs.

        *NEW* since **tkRAD v1.7**.
        """
        self.__notify_tid = 0
        self.__previous_state = 0
        self.__static_text = None
        self.delay = self.NOTIFICATION_DELAY
        self.options.set_sections("gui")
        self.options.load()
        self.toggle_var = tk.StringVar()
        self.apply_options()
        ttk.Sizegrip(self).pack(padx=2, pady=2, side=tk.RIGHT)
        self.message_cvar = tk.StringVar()
        self.label = ttk.Label(
            self,
            textvariable=self.message_cvar,
            anchor=tk.W,
            justify=tk.LEFT,
            relief=tk.SUNKEN,
        )
        self.label.pack(padx=2, pady=2, **self.PACK_OPTIONS)
        self.info()
    # end def

    def notify (self, text, delay=None):
        """Display a notification text for a time delay (in seconds).

        Parameter:
            text -- text to display (string of chars).

            delay -- time delay in seconds, floating point value
            (default: None). If omitted or None, will be replaced by a
            value extracted from *get_correct_delay()* method.

        Displays a notification 'text' into statusbar during 'delay'
        delay of time, expressed in seconds (float value).

        After time delay has expired, the previous static text set by
        *info()* method is displayed again.

        Notice: **no text** will be displayed if 'text' is *NOT* a
        plain string of chars i.e. is an empty string of chars or not
        of string type at all.
        """
        if tools.is_pstr(text):
            self._stop_notification()
            self.message_cvar.set(text)
            self.update_idletasks()
            self.__notify_tid = self.after(
                tools.ensure_int(
                    self.get_correct_delay(delay) * 1000.0
                ),
                self.info
            )
        # end if
    # end def

    def show (self, *args, **kw):
        """Show current statusbar.

        Event handler signature.
        """
        self.toggle_var_set(1)
    # end def

    def toggle (self, *args, **kw):
        """Toggle display state of current statusbar.

        Event handler signature.

        Shows statusbar if it were hidden, hides otherwise.
        """
        self.toggle_var_set(1 - self.__previous_state)
    # end def

    @property
    def toggle_var (self):
        """Tkinter control variable for statusbar's displaying state -
        read-write attribute.

        May raise **RADStatusBarError** if set to a different type than
        *tkinter.StringVar* control variable.
        """
        return self.__toggle_var
    # end def

    @toggle_var.setter
    def toggle_var (self, value):
        if isinstance(value, tk.StringVar):
            self.__toggle_var = value
        else:
            raise RADStatusBarError(
                "Statusbar toggle variable must be of "
                "'tkinter.StringVar' type."
            )
            self.__toggle_var = tk.StringVar()
        # end if
    # end def

    @toggle_var.deleter
    def toggle_var (self):
        del self.__toggle_var
    # end def

    def toggle_var_set (self, value):
        """Switch toggle control variable to value.

        Parameter:
            value -- should be 0 or 1 and nothing else.

        Sets toggle control variable to 'value' and then updates
        display state of current statusbar.
        """
        self.__previous_state = self._get_state(value)
        self.toggle_var.set(self.__previous_state)
        self.update_display()
    # end def

    def update_display (self, *args, **kw):
        """Update widget display state (show/hide).

        Event handler signature.

        Shows or hides current statusbar widget along with toggle
        control variable values.

        Raises 'StatusbarShow' event with *widget=self* keyword
        argument after having shown up statusbar.

        Raises 'StatusbarHide' event with *widget=self* keyword
        argument after having hidden statusbar.

        *NEW* since **tkRAD v1.7**.
        """
        if self._toggle_var_state():
            self.grid()
            self.events.raise_event("StatusbarShow", widget=self)
        else:
            self.grid_remove()
            self.events.raise_event("StatusbarHide", widget=self)
        # end if
    # end def

# end class


class RADStatusBarError (Exception):
    """Error handler for RADStatusBar class."""
    pass
# end class
