#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Base widget module.

Implements:

* class RADWidgetBase -- mixin base class for all RADxxxxxxx family
  widgets.
* class RADWidgetBaseError -- error handler for RADWidgetBase class.

This module is intended to be used internally with all RADxxxxxxx
family widgets.

*NEW* since **tkRAD v1.7**:

* improved many code portions;
* 'TK' module alias name has been renamed to 'tk' to best comply with
  Python's community conventions;
* by now, error handling is made through *RADWidgetBaseError* class and
  i18n support is abandoned for error messages.

Please, feel free to post an issue ticket on GitHub's project
bugtracker if you face any troubleshooting with this new version of
tkRAD's library.

GitHub's project bugtracker: https://github.com/tarball69/tkRAD/issues
(requires a GitHub member account to proceed).
"""

"""
    tkRAD - tkinter Rapid Application Development library

    (c) 2013+ Raphaël SEBAN <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see: http://www.gnu.org/licenses/
"""

import tkinter as tk


class RADWidgetBase:
    """Mixin base class for all RADxxxxxxx family widgets."""

    STICKY_ALL = tk.NW + tk.SE

    GRID_OPTIONS = dict(
        padx=0, pady=0, ipadx=0, ipady=0, sticky=STICKY_ALL
    )

    PACK_OPTIONS = dict(
        side=tk.TOP, fill=tk.BOTH, expand=1
    )

    RC_OPTIONS = dict()

    def __del__ (self):
        self.events._on_widget_destroy(self)
        self.services._on_widget_destroy(self)
    # end def

    def __init__ (self, tk_owner=None, **kw):
        """Class constructor.

        *NEW* since **tkRAD v1.7**:

        * method signature reduced to only mentioned keyword arguments;
        * renamed 'h_owner' parameter to 'h_owner' i.e. meaning
          'handlers_owner';
        * keyword argument 'subclassed' is useless by now and no more
          supported;
        * removed call to 'init_widget()' method: not a mixin feature,
          must be coder's choice in subclasses.
        """
        from ..core import events
        from ..core import options
        from ..core import services
        for _name in dir(self):
            _value = getattr(self, _name, None)
            if _name.isupper() and isinstance(_value, dict):
                setattr(self, _name, _value.copy())
            # end if
        # end for
        self.tk_owner = tk_owner or self.master
        self.h_owner = kw.get("h_owner") or self.tk_owner
        self.app = services.ask_for("app", silent_mode=True)
        self.events = events.get_event_manager()
        self.services = services.get_service_manager()
        self.options = options.get_option_manager(**kw)
        self.options.set_sections(
            self.RC_OPTIONS.setdefault(
                "section", self.get_rc_section()
            )
        )
    # end def

    def _only_tk (self, kw):
        _dict = dict()
        if hasattr(self, "tk") and hasattr(self, "configure"):
            _attrs = set(self.configure().keys()) & set(kw.keys())
            for _key in _attrs:
                _dict[_key] = kw.get(_key)
            # end for
        # end if
        return _dict
    # end def

    def cast_parent (self, widget):
        """Cast if widget is of tkinter parent widget type or not.

        Parameter:
            widget -- object instance.

        Returns True if 'widget' is actually one of (or a subclass of)
        tkinter.Widget, tkinter.BaseWidget or tkinter.Tk type.

        May raise **RADWidgetBaseError** otherwise and after then
        return False.
        """
        if self.is_tk_parent(widget):
            return True
        else:
            raise RADWidgetBaseError(
                "Object must be one of (or a subclass of) tkinter "
                "Widget, BaseWidget or Tk type, not '{}'."
                .format(widget.__class__.__name__)
            )
            return False
        # end if
    # end def

    def cast_widget (self, widget):
        """Cast if widget is of tkinter.Widget type or not.

        Parameter:
            widget -- object instance.

        Returns True if 'widget' is actually one of (or a subclass of)
        tkinter.Widget type.

        May raise **RADWidgetBaseError** otherwise and after then
        return False.
        """
        if self.is_tk_widget(widget):
            return True
        else:
            raise RADWidgetBaseError(
                "Object must be of (or a subclass of) "
                "tkinter.Widget type, not '{}'."
                .format(widget.__class__.__name__)
            )
            return False
        # end if
    # end def

    def classname (self):
        """Return class name.

        Returns the class name string of chars of current object
        instance.
        """
        return self.__class__.__name__
    # end def

    def destroy (self):
        """Widget destroy.

        Should override tkinter's w.destroy() universal method in
        derived subclass.

        Calls *finalize()* hook method then makes automatic
        unregistering of related methods from app-wide event and
        service managers, releasing as much as possible any external
        strong references that might block a proper garbage collection.
        """
        self.finalize()
        self.__del__()
        super().destroy()
    # end def

    def enable (self, state=True):
        """Enable or disable current widget.

        Parameter:
            state -- boolean flag (default: True).

        May enable or disable current widget along with 'state' boolean
        value i.e. True or False.
        """
        self.enable_widget(self, state)
    # end def

    def enable_widget (self, tkwidget, state=True):
        """Enable or disable a given tkinter widget.

        Parameters:
            tkwidget -- expected tkinter widget.

            state -- boolean flag (default: True).

        May enable or disable given 'tkwidget' tkinter widget along
        with 'state' boolean value i.e. True or False.

        If 'state' is None (which is different from False), will pass
        through and will **do nothing**.
        """
        if state is not None:
            tkwidget.configure(
                state=tk.NORMAL if state else tk.DISABLED
            )
        # end if
    # end def

    def enabled (self):
        """Indicate if current widget is enabled or not."""
        return self.widget_enabled(self)
    # end def

    def finalize (self, *args, **kw):
        """Finalize widget - hook method.

        Event handler signature.

        Finalizes current widget before actually destroying it.

        Current implementation does nothing.

        Feel free to extend or to override this hook method to best
        meet your own needs.

        *NEW* since **tkRAD v1.7**.
        """
        pass
    # end def

    def get_rc_section (self):
        """Return an RC configuration file section name.

        Current implementation merely returns current widget's class
        name as an RC configuration file section name for this widget.
        """
        return self.classname()
    # end def

    @deprecated("Please, use initialize() method instead.")
    def init_widget (self, **kw):
        """Initialize current widget - hook method.

        Parameter:
            kw -- keyword arguments.

        *NEW* since **tkRAD v1.7**:

        This method is now *DEPRECATED*.

        Please, better use 'initialize()' method instead.
        """
        pass
    # end def

    def initialize (self, *args, **kw):
        """Initialize current widget - hook method.

        Parameters:
            args -- variable list of arguments.

            kw -- keyword arguments.

        This method is a coding helper and should be called immediately
        after '__init__()' method.

        Current implementation merely calls deprecated 'init_widget()'
        hook method as a helper indirection for old code migration.

        Feel free to extend or to override this hook method to best
        meet your own needs.

        *NEW* since **tkRAD v1.7**.
        """
        self.init_widget(**kw)
    # end def

    def is_tk_parent (self, widget):
        """Indicate if widget is a tkinter parent widget.

        Parameter:
            widget -- object instance.

        Returns True if 'widget' is actually one of (or a subclass of)
        tkinter.Widget, tkinter.BaseWidget or tkinter.Tk type.

        Returns False otherwise.
        """
        return isinstance(widget, (tk.Widget, tk.BaseWidget, tk.Tk))
    # end def

    def is_tk_widget (self, widget):
        """Indicate if widget is a tkinter widget.

        Parameter:
            widget -- object instance.

        Returns True if 'widget' is actually one of (or a subclass of)
        tkinter.Widget type.

        Returns False otherwise.
        """
        return isinstance(widget, tk.Widget)
    # end def

    def widget_enabled (self, tkwidget):
        """Indicate if given widget is enabled or not."""
        return bool(tkwidget.cget("state") != tk.DISABLED)
    # end def

# end class


class RADWidgetBaseError (Exception):
    """Error handler for RADWidgetBase class."""
    pass
# end class
