#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""tkRAD mixin base window module.

Implements:

* class RADWindowBase -- mixin base class for tkRAD window widgets.
* class RADWindowBaseError -- error handler for RADWindowBase class.

This module is intended to be used internally with RADxxxxxxx window
widgets family.

*NEW* since **tkRAD v1.7**:

* from now, each subclass of mixin RADWidgetBase *MUST* call
  self.initialize() hook method by itself;
* as a mixin base class, *RADWindowBase* does *NOT* call *initialize()*
  hook method by itself. It either delegates responsibility to real
  widget subclasses;
* error messages are *NOT* translated any more;
* renamed erroneous 'pending' mention to 'running' everywhere in this
  module;
* replaced erroneous 'slot' mention by event handlers' more fitting
  names e.g. 'on_mouse_clicked';
* removed a bunch of protected method defs;
* moved 'WM_DELETE_WINDOW' protocol to new *self.on_close_window()*
  event handler as it is more relevant than specific
  *self.on_quit_app()* event handler i.e. app secundary Toplevel()
  windows do *NOT* have to quit application, but merely close window;
* from now, window saves its geometry and window state ('normal',
  'maximized') at the very end, just before closing;
* tried to keep as much as possible backward compatibility with older
  versions of tkRAD.

Please, feel free to post an issue ticket on GitHub's project
bugtracker if you face any troubleshooting with this new version of
tkRAD's library.

GitHub's project bugtracker: https://github.com/tarball69/tkRAD/issues
(requires a GitHub member account to proceed).
"""

"""
    tkRAD - tkinter Rapid Application Development library

    (c) 2013+ Raphaël SEBAN <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see: http://www.gnu.org/licenses/
"""

import warnings

import tkinter as tk
import tkinter.messagebox as mb

from .rad_widget_base import RADWidgetBase
from .rad_statusbar import RADStatusBar

from ..xml import rad_xml_menu as rxm
from ..core import tools


class RADWindowBase (RADWidgetBase):
    """Mixin base class for tkRAD window widgets."""

    WINDOW_ID = "window"

    def __init__ (self, tk_owner=None, **kw):
        """Class constructor."""
        super().__init__(tk_owner, **kw)
        self.protocol("WM_DELETE_WINDOW", self.on_close_window)
        self.WINDOW_ID_STATE = "{}_state".format(self.WINDOW_ID)
        self.STATES = {
            "hidden": self.hide,
            "minimized": self.minimize,
            "maximized": self.maximize,
            "normal": self.show,
        }
        self.__running_task = False
    # end def

    def confirm_quit (self, *args, **kw):
        """Ask user to confirm before quitting application.

        Event handler signature.

        Pops up a dialog box and asks user to confirm before actually
        quitting application.
        """
        return mb.askokcancel(
            title=_("Quit app?"),
            message=_(
                "Are you sure you want to quit this application?"
            ),
            parent=self,
        )
    # end def

    def connect_statusbar (self, stringvarname):
        """Try to connect window's statusbar display state to a tkinter
        control variable.

        Parameter:
            stringvarname -- name of the tkinter.StringVar() control
            variable (string of chars).

        Tries to connect current window's statusbar widget to a given
        'stringvarname' tkinter control variable.

        The tkinter.StringVar() control variable should then manage
        statusbar's displaying state (shown/hidden) along with its own
        internal values.

        Configuration file options are applied on connection success.
        """
        self.statusbar.toggle_var = (
            self.topmenu.get_stringvar(stringvarname)
            or self.mainframe.get_stringvar(stringvarname)
            or tk.StringVar()
        )
        self.statusbar.apply_options()
    # end def

    def finalize (self, *args, **kw):
        """Finalize window - hook method.

        Event handler signature.

        Finalizes current window before actually destroying it.

        Feel free to extend or to override this hook method to best
        meet your own needs.

        *NEW* since **tkRAD v1.7**.
        """
        _state = "normal"
        try:
            if int(self.wm_attributes("-zoomed")):
                _state = "maximized"
            # end if
        except:
            pass
        # end try
        self.options["geometry"][self.WINDOW_ID_STATE] = _state
        if _state == "normal":
            self.options["geometry"][self.WINDOW_ID] = str(
                self.geometry()
            )
        # end if
        self.options.save()
    # end def

    @deprecated("Please, use 'running_task' class attribute instead.")
    def get_pending_task (self):
        """Get important running task boolean state.

        Notice: 'pending' mention is erroneous. Task is running, not
        pending.

        *DEPRECATED* since **tkRAD v1.7**.

        Please, use *running_task* class attribute instead.
        """
        return self.running_task
    # end def

    @deprecated("Please, use 'window_state' class attribute instead.")
    def get_window_state (self):
        """Get current window state.

        *DEPRECATED* since **tkRAD v1.7**.

        Please, use *window_state* class attribute instead.
        """
        return self.window_state
    # end def

    def hide (self, *args, **kw):
        """Hide current window.

        Event handler signature.
        """
        self.withdraw()
    # end def

    def initialize (self, *args, **kw):
        """Initialize current widget - hook method.

        Parameters:
            args -- variable list of arguments.

            kw -- keyword arguments.

        This method is a coding helper and should be called immediately
        after '__init__()' method.

        Feel free to extend or to override this hook method to best
        meet your own needs.

        *NEW* since **tkRAD v1.7**.

        *NEW* since **tkRAD v1.7.5**:

        * moved mandatory inits to class' constructor;
        * removed support for window settings i.e. resizable, minsize,
          geometry and window_state;
        * less code, more freedom.
        """
        self.options.set_sections("geometry", "gui", "topmenu")
        self.options.set_defaults(
            **tools.choose(
                kw.get("rc_defaults"),
                {
                    "maximized": "0",
                    self.WINDOW_ID: "640x480+20+20",
                    self.WINDOW_ID_STATE: "normal",
                },
            )
        )
        self.options.load()
        try:
            _app_title = self.app.APP["title"]
        except:
            _app_title = None
        # end try
        self.title(
            tools.choose_str(
                kw.get("title"),
                _app_title,
                _(self.WINDOW_ID),
                "Window",
            )
        )
        self.topmenu = kw.get("topmenu") or rxm.RADXMLMenu(self)
        self.topmenu.xml_filename = (
            kw.get("topmenu_xml_filename") or "topmenu"
        )
        self.mainframe = kw.get("mainframe") or tk.Frame(self)
        self.statusbar = kw.get("statusbar") or RADStatusBar(self)
        self.rowconfigure(0, weight=1)
        self.rowconfigure(1, weight=0)
        self.columnconfigure(0, weight=1)
        self.events.connect_dict(
            {
                "close": self.on_close_window,
                "Close": self.on_close_window,
                "closewindow": self.on_close_window,
                "CloseWindow": self.on_close_window,
                "quit": self.on_quit_app,
                "Quit": self.on_quit_app,
                "quitapp": self.on_quit_app,
                "QuitApp": self.on_quit_app,
                "runningtaskon": self.on_running_task_on,
                "RunningTaskOn": self.on_running_task_on,
                "runningtaskoff": self.on_running_task_off,
                "RunningTaskOff": self.on_running_task_off,
                "togglestatusbar": self.statusbar.toggle,
                "ToggleStatusbar": self.statusbar.toggle,
            }
        )
    # end def

    @property
    def mainframe (self):
        """Main widgets container frame - read-write attribute.

        This should be set to *at least* a tkinter.Widget object
        instance.

        Recommended widgets containers are of (or subclass of) tk.Frame
        and ttk.Frame types.

        Grids automatically new widgets container to current window.

        May raise **RADWidgetBaseError** on unsupported type.
        """
        return self.__mainframe
    # end def

    @mainframe.setter
    def mainframe (self, widget):
        if self.cast_widget(widget):
            self.__mainframe = widget
            widget.grid(
                in_=self, row=0, column=0, **self.GRID_OPTIONS
            )
        # end if
    # end def

    @mainframe.deleter
    def mainframe (self):
        del self.__mainframe
    # end def

    def maximize (self, *args, **kw):
        """Try to maximize current window.

        Event handler signature.

        This feature is *NOT* guaranteed to work on any platform.

        May print a warning message in console output if feature is not
        supported on a given platform.
        """
        try:
            self.deiconify()
            self.attributes("-zoomed", "1")
            self.update()
        except:
            warnings.warn("Could *NOT* maximize current window.")
        # end try
    # end def

    def minimize (self, *args, **kw):
        """Minimize current window.

        Event handler signature.

        Reduces current window frame to platform's task bar (iconify).
        """
        self.iconify()
    # end def

    def on_close_window (self, *args, **kw):
        """Before closing window - event handler.

        Event handler signature.

        Called before actually closing current window. May decide if
        window has to be closed or not.

        This event handler is useful for app secundary windows which
        may *NOT* ask to quit application on closing e.g. tool windows.

        Current implementation merely calls *on_quit_app()* event
        handler method by default.

        Feel free to override this hook method to best meet your own
        needs.

        *NEW* since **tkRAD v1.7**.
        """
        self.on_quit_app(*args, **kw)
    # end def

    def on_quit_app (self, *args, **kw):
        """Quit application - event handler.

        Event handler signature.

        Verifies if some important task is not running, asks user for
        confirmation and then actually destroys current window, which
        implies for main windows to terminate application itself.

        Replaces former misspelled *_slot_quit_app()* method.

        *NEW* since **tkRAD v1.7**.
        """
        if self.running_task:
            mb.showwarning(
                title=_("Running operation"),
                message=_(
                    "Some very important task is running by now. "
                    "Please wait for completion and then retry."
                ),
                parent=self,
            )
        elif self.confirm_quit(*args, **kw):
            self.quit()
            self.destroy()
        # end if
    # end def

    def on_running_task_off (self, *args, **kw):
        """Important running task has ended - event handler.

        Event handler signature.

        Replaces former misspelled *_slot_pending_task_off()* method.

        *NEW* since **tkRAD v1.7**.
        """
        self.running_task = False
        self.statusbar.notify(
            tools.choose_str(
                kw.get("message"),
                kw.get("info"),
                _("An important task has ended."),
            )
        )
    # end def

    def on_running_task_on (self, *args, **kw):
        """Important running task has started - event handler.

        Event handler signature.

        Replaces former misspelled *_slot_pending_task_on()* method.

        *NEW* since **tkRAD v1.7**.
        """
        self.running_task = True
        self.statusbar.notify(
            tools.choose_str(
                kw.get("message"),
                kw.get("info"),
                _("An important task has started."),
            )
        )
    # end def

    def run (self):
        """Enter tkinter's event management mainloop.

        Current implementation merely calls *self.mainloop()*.
        """
        self.mainloop()
    # end def

    @property
    def running_task (self):
        """Flag for important running task - read-write attribute.

        Switch this flag to True (task ON) whenever you need to inform
        application an important task is running.

        Switch this flag to False (task OFF) whenever you need to
        inform application an important task has ended.

        Please, better use this new class attribute instead of
        *get_pending_task()* which is now *DEPRECATED*.

        *NEW* since **tkRAD v1.7**.
        """
        return self.__running_task
    # end def

    @running_task.setter
    def running_task (self, value):
        self.__running_task = bool(value)
    # end def

    @running_task.deleter
    def running_task (self):
        del self.__running_task
    # end def

    @deprecated("Please, use 'window_state' class attribute instead.")
    def set_window_state (self, state):
        """Set current window state.

        *DEPRECATED* since **tkRAD v1.7**.

        Please, use *window_state* class attribute instead.
        """
        self.window_state = state
    # end def

    def show (self, *args, **kw):
        """Show current window.

        Event handler signature.

        Deiconifies current window if it were iconified.
        """
        self.deiconify()
    # end def

    @property
    def statusbar (self):
        """Current window statusbar widget - read-write attribute.

        Grids automatically new widget to current window.

        May raise **RADWidgetBaseError** on unsupported type.
        """
        return self.__statusbar
    # end def

    @statusbar.setter
    def statusbar (self, widget):
        if self.cast_widget(widget):
            self.__statusbar = widget
            widget.grid(
                in_=self, row=1, column=0, **self.GRID_OPTIONS
            )
        # end if
    # end def

    @statusbar.deleter
    def statusbar (self):
        del self.__statusbar
    # end def

    @property
    def topmenu (self):
        """Current window topmenu menubar - read-write attribute.

        May raise **RADWindowBaseError** if not set to at least an
        object instance of (or subclass of) tkinter.Menu or RADXMLMenu
        type.
        """
        return self.__topmenu
    # end def

    @topmenu.setter
    def topmenu (self, widget):
        if isinstance(widget, (tk.Menu, rxm.RADXMLMenu)):
            self.__topmenu = widget
        else:
            raise RADWindowBaseError(
                "Topmenu menubar must be one of 'tkinter.Menu' or "
                "'tkRAD.xml.RADXMLMenu' type."
            )
        # end if
    # end def

    @topmenu.deleter
    def topmenu (self):
        del self.__topmenu
    # end def

    @property
    def window_state (self):
        """Current window state - read-write attribute.

        This attribute relies on *STATES* class attribute items.

        Supported values should always be at least 'normal', 'hidden',
        'minimized' and 'maximized'.

        Gets current window state or sets actual window state to one of
        supported states.

        May raise **RADWindowBaseError** if set to an unsupported state
        value.

        *NEW* since **tkRAD v1.7**.
        """
        return self.__window_state
    # end def

    @window_state.setter
    def window_state (self, state):
        state = str(state).lower()
        try:
            self.STATES[state]()
        except:
            raise RADWindowBaseError(
                "Unsupported window state '{}'.".format(state)
            ) from None
        else:
            self.__window_state = state
        # end try
    # end def

    def xml_build (self, xml_source=None, **kw):
        """Build XML source.

        Parameters:
            xml_source -- string of chars, can either be an actual XML
            script or a file path to an XML script (default: None). If
            omitted or None, will be replaced by *mainframe*'s XML
            filename and looked up into e.g. default "^/xml/widget"
            directory.

            kw -- keyword arguments.

        Supported keyword arguments:
            -- none at this stage --

        Builds an XML source into *mainframe* widgets container.

        May raise errors if *mainframe* is *NOT* at least an object
        instance of (or subclass of) *RADXMLFrame* type.

        *NEW* since **tkRAD v1.7**:

        * renamed erroneous 'filename' parameter to 'xml_source';
        * removed deprecated 'silent_mode' parameter -- no more
          supported by now.
        """
        return self.mainframe.xml_build(xml_source)
    # end def

# end class


class RADWindowBaseError (Exception):
    """Error handler for RADWindowBase class."""
    pass
# end class
