#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""tkRAD XML attributes module.

Implements:

* class RADXMLAttribute -- XML attribute structure for tkRAD parser.
* class RADXMLAttributeError -- error handler for RADXMLAttribute
  class.

This module is intended to be used internally by tkRAD parsers only.

*NEW* since **tkRAD v1.7**:

* renamed module alias 'ET' to 'etree';
* added new *RADXMLAttributeError* error handler class for
  *RADXMLAttribute* class;
* error messages are *NOT* translated any more;
* *RADXMLAttribute* class member attributes cannot be deleted any more.

Please, feel free to post an issue ticket on GitHub's project
bugtracker if you face any troubleshooting with this new version of
tkRAD's library.

GitHub's project bugtracker: https://github.com/tarball69/tkRAD/issues
(requires a GitHub member account to proceed).
"""

"""
    tkRAD - tkinter Rapid Application Development library

    (c) 2013+ Raphaël SEBAN <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see: http://www.gnu.org/licenses/
"""

import xml.etree.ElementTree as etree


def reset_attributes (xml_attributes, xml_element=None):
    """Copy a flat dict() of XML attributes into a new dict() of
    RADXMLAttribute() XML attribute structures.

    Parameters:
        xml_attributes -- mapping dict() of flat XML attributes or
        xml.etree.ElementTree parsed XML element.

        xml_element -- xml.etree.ElementTree parsed XML element
        (default: None). If omitted or None, will be replaced by
        'xml_attributes' parameter.

    If 'xml_attributes' is a parsed XML element, it takes precedence
    over 'xml_element' parameter and after then is changed to a flat
    mapping dict() of XML attributes.

    Copies a flat mapping dict() of XML attributes into a new mapping
    dict() object, where values are all instances of RADXMLAttribute
    class.

    Returns the new copied mapping dict() object.

    Original mapping dict() is left *UNTOUCHED* after this operation.

    May raise **RADXMLAttributeError** on unsupported parameter types.
    """
    if etree.iselement(xml_attributes):
        if not etree.iselement(xml_element):
            xml_element = xml_attributes
        # end if
        xml_attributes = xml_attributes.attrib
    # end if
    if isinstance(xml_attributes, dict):
        _attrs = xml_attributes.copy()
        for (_name, _value) in _attrs.items():
            if not isinstance(_value, RADXMLAttribute):
                _attrs[_name] = (
                    RADXMLAttribute(xml_element, _name, _value)
                )
            # end if
        # end for
        return _attrs
    else:
        raise RADXMLAttributeError(
            "Parameter 'xml_attributes' must be of 'dict' type."
        )
        return None
    # end if
# end def


class RADXMLAttribute:
    """XML attribute structure for tkRAD parser."""

    def __init__ (self, xml_element, attr_name, attr_value, parsed=0):
        """Class constructor."""
        self.xml_element = xml_element
        self.name = attr_name
        self.value = attr_value
        self.parsed = parsed
    # end def

    def get_value (self):
        """Value getter for tkRAD.xml.RADXMLAttributesDict class."""
        return self.value
    # end def

    @property
    def name (self):
        """XML attribute name - read-write attribute.

        This class attribute *MUST* be a plain string of chars or it
        will raise **RADXMLAttributeError**.
        """
        return self.__attr_name
    # end def

    @name.setter
    def name (self, value):
        if value and isinstance(value, str):
            self.__attr_name = value
        else:
            raise RADXMLAttributeError(
                "XML attribute name must be a plain string of chars."
            )
        # end if
    # end def

    @property
    def parsed (self):
        """Counter for parsed XML attribute - read-write attribute.

        Increments an internal counter each time this class attribute
        is set to the boolean *True* value.

        If set to *False* or zero, resets counter to zero.

        If set to a positive integer value (n > 0), resets counter to
        that value.

        Negative integer values will reset counter to 1.
        """
        return self.__attr_parsed
    # end def

    @parsed.setter
    def parsed (self, value):
        if not value:
            self.__attr_parsed = 0
        elif isinstance(value, int):
            self.__attr_parsed = min(1, value)
        else:
            try:
                self.__attr_parsed += 1
            except:
                self.__attr_parsed = 1
            # end try
        # end if
    # end def

    def set_value (self, value):
        """Value setter for tkRAD.xml.RADXMLAttributesDict class."""
        self.value = value
    # end def

    def update_xml_element (self, new_name=None):
        """Update internal XML element's 'name' XML attribute.

        Parameter:
            new_name -- string of chars, new name for internal XML
            element's 'name' XML attribute (default: None). If omitted
            or None, will be replaced by *self.value* class member
            attribute.

        Updates internal XML element's 'name' XML attribute to
        'new_name', if given, or to *self.value* otherwise.

        Please, notice it is possible to set 'name' XML attribute to an
        empty string of chars i.e. *update_xml_element("")*.

        Returns True on success, False otherwise.

        May raise **RADXMLAttributeError** if internal XML element is
        *NOT* of *xml.etree.Element* type.

        *NEW* since **tkRAD v1.7**: renamed 'value' parameter to more
        convenient 'new_name' parameter.
        """
        if new_name is None:
            new_name = self.value
        # end if
        if etree.iselement(self.xml_element):
            self.xml_element.set(self.name, new_name)
            return True
        else:
            raise RADXMLAttributeError(
                "XML element must be of "
                "'xml.etree.ElementTree.Element' type."
            )
        # end if
        return False
    # end def

# end class


class RADXMLAttributeError (Exception):
    """Error handler for RADXMLAttribute class."""
    pass
# end class
