#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""XML widget factory mixin base class module.

Implements:

* class RADXMLBase -- mixin base class for XML widget factory feature.
* class RADXMLBaseError -- error handler for RADXMLBase class.

Notice: all XML parsing features in this module rely on Python's
official *xml.etree.ElementTree* package modules.

Please, visit official documentation for more detail:
https://docs.python.org/3/library/xml.etree.elementtree.html

*NEW* since **tkRAD v1.7**:

* from now, each subclass of mixin RADWidgetBase *MUST* call
  *initialize()* hook method by itself;
* as a mixin base class, *RADXMLBase* does *NOT* call *initialize()*
  hook method by itself. It either delegates responsibility to real
  widget subclasses;
* error messages are *NOT* translated any more;
* added new *RADXMLBaseError* error handler class for *RADXMLBase*
  class;
* added new *_BITMAPS* protected class member attribute;
* added new *_CVARTYPES* protected class member attribute;
* added support for *tkinter.BooleanVar()* control variables;
* tried to keep as much as possible backward compatibility with older
  versions of tkRAD.

Please, feel free to post an issue ticket on GitHub's project
bugtracker if you face any troubleshooting with this new version of
tkRAD's library.

GitHub's project bugtracker: https://github.com/tarball69/tkRAD/issues
(requires a GitHub member account to proceed).
"""

"""
    tkRAD - tkinter Rapid Application Development library

    (c) 2013+ Raphaël SEBAN <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see: http://www.gnu.org/licenses/
"""

import re
import os.path
import warnings
import xml.etree.ElementTree as etree

import tkinter as tk

from ..core import path
from ..core import tools

from ..widgets.rad_widget_base import RADWidgetBase
from .rad_xml_attributes_dict import RADXMLAttributesDict
from . import rad_xml_attribute as rxa


class RADXMLBase (RADWidgetBase):
    """Mixin base class for XML widget factory feature.

    This class implements the core maechanism of tkRAD's XML widget
    factory.
    """

    ATTRIBUTE_PARSER = "_parse_{xml_element}_attr_{xml_attribute}"

    DOCTYPE = "tkbase"

    ELEMENT_BUILDER = "_build_element_{xml_element}"

    RC_OPTIONS = {
        "dir": "default_xml_dir",
        "filename": "default_xml_filename",
        "file_ext": "default_xml_file_ext",
    }

    XML_RC = {
        "dir": "^/xml",
        "filename": "lowercased_classname",
        "file_ext": ".xml",
    }

    _BITMAPS = (
        "error", "gray75", "gray50", "gray25", "gray12", "hourglass",
        "info", "questhead", "question", "warning"
    )

    _CVARTYPES = {
        "booleanvar": tk.BooleanVar,
        "doublevar": tk.DoubleVar,
        "intvar": tk.IntVar,
        "stringvar": tk.StringVar,
    }

    __OI_COUNT = 1

    def __init__ (self, tk_owner=None, **kw):
        """Class constructor."""
        super().__init__(tk_owner, **kw)
        from ..core import defer
        self.xml_dir = kw.get("xml_dir")
        self.xml_file_ext = kw.get("xml_file_ext")
        self.xml_filename = kw.get("xml_filename")
        self.xml_tree = None
        self.XML_RC.setdefault("filename", self.classname().lower())
        self.DOCTYPE = str(self.DOCTYPE).lower()
        self._queue = defer.DeferQueue()
        self.__objects = dict()
        self.__images = dict()
        self.__tk_variables = {
            vtype.lower(): dict() for vtype in self._CVARTYPES
        }
    # end def

    def _before_building_element(self, **kw):
        pass
    # end def

    def _before_parsing_attribute(self, **kw):
        pass
    # end def

    def _build_element (self, xml_element, tk_parent):
        if self.cast_element(xml_element):
            _tag = self.normalize_tag(xml_element)
            _elt_builder = tools.normalize_id(
                str(self.ELEMENT_BUILDER).format(xml_element=_tag)
            )
            _elt_builder = getattr(self, _elt_builder, None)
            if callable(_elt_builder):
                self._before_building_element(
                    xml_tag=_tag,
                    xml_element=xml_element,
                    tk_parent=tk_parent,
                )
                return _elt_builder(_tag, xml_element, tk_parent)
            else:
                raise RADXMLBaseError(
                    "Unsupported XML tag <{}>.".format(_tag)
                )
            # end if
        # end if
        return False
    # end def

    def _cast_root_element (self, xml_element):
        if self.cast_element(xml_element):
            if self.normalize_tag(xml_element) in self.DOCTYPE:
                return True
            else:
                raise RADXMLBaseError(
                    "XML root element *MUST* be one of {} "
                    "markup tag(s)."
                    .format(self.DOCTYPE)
                )
        # end if
        return False
    # end def

    def _get_object_id (self, built_object, attr_id=None):
        attr_id = tools.normalize_id(attr_id)
        if not tools.is_pstr(attr_id):
            attr_id = self._get_unique_id(
                built_object.__class__.__name__
            )
        # end if
        return attr_id.lower()
    # end def

    def _get_oi_count_str (self):
        _str = str(self.__OI_COUNT)
        self.__OI_COUNT += 1
        return _str
    # end def

    def _get_unique_id (self, radix):
        if tools.is_pstr(radix):
            while self.__OI_COUNT:
                _uid = radix + self._get_oi_count_str()
                if _uid not in self.__objects:
                    return _uid
                # end if
            # end while
        # end if
        return None
    # end def

    def _loop_on_children (self, xml_element, tk_parent, accept=None):
        if self.cast_element(xml_element):
            _ptag = self.normalize_tag(xml_element)
            _ret = True
            for _xml_child in xml_element:
                _ctag = self.normalize_tag(_xml_child)
                if not accept or _ctag in accept:
                    _ret = (
                        self._build_element(_xml_child, tk_parent)
                        and _ret
                    )
                else:
                    raise RADXMLBaseError(
                        "XML child element <{}> is *NOT* "
                        "allowed inside <{}> element."
                        .format(_ctag, _ptag)
                    )
                    return False
                # end if
            # end for
            return _ret
        # end if
        return False
    # end def

    def _parse_xml_attributes (self, xml_element, tk_parent, **kw):
        if self.cast_element(xml_element):
            _tag = self.normalize_tag(xml_element)
            _attrs = RADXMLAttributesDict(
                rxa.reset_attributes(
                    kw.get("xml_attrs", xml_element.attrib),
                    xml_element
                )
            )
            for (_attr_name, _attr_object) in _attrs.items():
                _attr_name = str(_attr_name).lower()
                _parser = tools.normalize_id(
                    str(self.ATTRIBUTE_PARSER)
                    .format(
                        xml_element=_tag,
                        xml_attribute=_attr_name,
                    )
                )
                _parse_attr = getattr(self, _parser, None)
                if callable(_parse_attr):
                    kw.update(
                        attribute=_attr_object,
                        attrs=_attrs,
                        tk_parent=tk_parent,
                        xml_element=xml_element,
                        xml_tag=_tag,
                        xml_attr=_attr_name,
                    )
                    self._before_parsing_attribute(**kw)
                    _parse_attr(**kw)
                else:
                    warnings.warn(
                        "Optional parser '{}()' is *NOT* implemented."
                        .format(_parser)
                    )
                # end if
            # end for
            """
                $ 2013-12-18 RS $
                New support: using RADXMLAttributesDict by now.
                Notice: flatten() provides a 'flat' dict() object
                containing simple (key, value) pairs.
                All RADXMLAttribute extra data are lost at this point.

                $ 2014-03-10 RS $
                since v1.4: deferred tasks
                do *NOT* flatten() dict any more(!)
            """
            return _attrs
        # end if
        return None
    # end def

    def _register_object_by_id (self, built_object, attr_id):
        _id = self._get_object_id(built_object, attr_id)
        if _id not in self.__objects:
            self.__objects[_id] = built_object
        else:
            raise RADXMLBaseError(
                "Object with id '{}' already exists."
                .format(_id)
            )
        # end if
    # end def

    def _reset_oi_count (self, value=1):
        self.__OI_COUNT = max(1, tools.ensure_int(value))
    # end def

    def _set_class_member (self, name, widget):
        if tools.is_pstr(name):
            if hasattr(self.tk_owner, name):
                raise RADXMLBaseError(
                    "Class member '{}' already exists in '{}'."
                    .format(name, self.tk_owner.__class__.__name__)
                )
            else:
                setattr(self.tk_owner, name, widget)
            # end if
        # end if
    # end def

    def cast_element (self, xml_element):
        """Cast if given XML element is of
        'xml.etree.ElementTree.Element' type.

        Parameter:
            xml_element -- xml.etree.ElementTree parsed XML element.

        Returns True if 'xml_element' is of supported type.

        May raise **RADXMLBaseError** otherwise and then return False.
        """
        if self.is_element(xml_element):
            return True
        else:
            raise RADXMLBaseError(
                "XML element must be of "
                "'xml.etree.ElementTree.Element' type."
            )
            return False
        # end if
    # end def

    def cast_tree (self, tree_object):
        """Cast if given XML tree data structure is of
        'xml.etree.ElementTree.ElementTree' type.

        Parameter:
            tree_object -- XML tree data structure.

        Returns True if 'tree_object' is of supported type.

        May raise **RADXMLBaseError** otherwise and then return False.
        """
        if self.is_tree(tree_object):
            return True
        else:
            raise RADXMLBaseError(
                "XML tree must be of "
                "'xml.etree.ElementTree.ElementTree' type."
            )
            return False
        # end if
    # end def

    @deprecated(
        "Please, use tkRAD.core.tools.dict_delete_items() instead.")
    def delete_dict_items (self, dict_object, *args):
        """Return new dict() object with stripped listed keys.

        *DEPRECATED* since **tkRAD v1.2**.

        Please, use *tkRAD.core.tools.dict_delete_items()* instead.
        """
        return tools.dict_delete_items(dict_object, *args)
    # end def

    def element_get_id (self, xml_element):
        """Get an ID value for XML element's 'id' XML attribute.

        Parameter:
            xml_element -- xml.etree.ElementTree parsed XML element.

        Tries to get a unique ID value for 'id' XML attribute of parsed
        'xml_element'.

        Sets 'id' XML attribute of 'xml_element' and then returns this
        unique ID value on success, returns None otherwise.
        """
        if self.cast_element(xml_element):
            _id = self.get_correct_id(xml_element.get("id"))
            xml_element.set("id", _id)
            return _id
        # end if
        return None
    # end def

    def get_bitmap_path (self, value):
        """Get formatted filepath for 'bitmap' option.

        Parameter:
            value -- bitmap name or filepath (string of chars).

        Formats 'value' to match tkinter's 'bitmap' configuration
        option requirements.

        Example::

            import tkinter as tk
            root = tk.Tk()
            root.geometry("320x240+400+200")
            # supported bitmap name
            tk.Label(root, bitmap="questhead").pack(expand=1, fill=tk.BOTH)
            try:
                # requires an actual bitmap resource file
                tk.Label(root, bitmap="@/path/to/bitmap.xbm").pack()
            except Exception as err:
                print(err)
            # end try
            root.mainloop()

        See:
        http://infohost.nmt.edu/tcc/help/pubs/tkinter/web/bitmaps.html

        Returns formatted value.

        *NEW* since **tkRAD v1.7**:

        * renamed 'path' parameter to more convenient 'value'
          parameter.
        """
        if tools.is_pstr(value) and value not in self._BITMAPS:
            value = "@" + path.normalize(value.lstrip("@"))
        # end if
        return value
    # end def

    def get_booleanvar (self, varname):
        """Get tkinter.BooleanVar control variable of given name.

        Parameter:
            varname -- name of tkinter control variable (string of
            chars).

        Returns the control variable corresponding to 'varname' name or
        None if it does not exist.
        """
        return self.__tk_variables["booleanvar"].get(varname)
    # end def

    def get_booleanvars (self):
        """Get all registered control variables of tkinter.BooleanVar
        type (mapping dict() object).
        """
        return self.__tk_variables["booleanvar"]
    # end def

    def get_correct_id(self, value):
        """Return a normalized ID value.

        Parameter:
            value -- suggested ID value (string of chars).

        Tries to normalize given ID 'value' i.e. keeps only matching
        [a-zA-Z0-9\_] characters in 'value'.

        If 'value' fails to empty string, it is then replaced by a
        unique "objectN" ID value ('N' is an integer index value) e.g.
        "object1", "object20", "object300" and so on.

        Returns a formatted string of chars intended for 'id' XML
        attribute, which should always be at least a plain string of
        chars.

        Notice: if suggested 'value' is correctly formatted on input,
        it will then return *AS IS*, no matter if it is a unique ID
        value or *NOT*.
        """
        value = tools.normalize_id(value)
        if not tools.is_pstr(value):
            value = self._get_unique_id("object")
        # end if
        return value
    # end def

    def get_cvar (self, vartype, varname):
        """Return tkinter control variable.

        Parameters:
            vartype -- string of chars, tkinter control variable type
            i.e. one of 'BooleanVar', 'DoubleVar', 'IntVar' or
            'StringVar' type (case insensitive, here).

            varname -- name of requested control variable (string of
            chars).

        Returns the tkinter control variable matching 'vartype' and
        'varname'.

        Returns None if 'varname' does not exist for requested
        'vartype' type.

        May raise **RADXMLBaseError** on unsupported control variable
        types.
        """
        vartype = str(vartype).lower()
        if vartype in self.__tk_variables:
            return self.__tk_variables[vartype].get(varname)
        else:
            raise RADXMLBaseError(
                "Tkinter control variable must be one of "
                "{} type (case insensitive, here)."
                .format(tuple(self._CVARTYPES))
            )
            return None
        # end if
    # end def

    def get_cvars (self):
        """Return all registered tkinter control variables (a mapping
        dict() of dict() objects).

        Returns a strong reference to internal mapping dict()
        (registry) of all registered tkinter control variables for any
        supported control variable types.

        Please, use with caution.
        """
        return self.__tk_variables
    # end def

    def get_doublevar (self, varname):
        """Get tkinter.DoubleVar control variable of given name.

        Parameter:
            varname -- name of tkinter control variable (string of
            chars).

        Returns the control variable corresponding to 'varname' name or
        None if it does not exist.
        """
        return self.__tk_variables["doublevar"].get(varname)
    # end def

    def get_doublevars (self):
        """Get all registered control variables of tkinter.DoubleVar
        type (mapping dict() object).
        """
        return self.__tk_variables["doublevar"]
    # end def

    def get_element_by_id (self, attr_id, xml_tree=None):
        """Retrieve an XML element by its 'id' XML attribute value.

        Parameters:
            attr_id -- 'id' XML attribute value (string of chars).
            Should always be a unique ID value through all XML tree
            data structure.

            xml_tree -- XML tree data structure (default: None). If
            omitted or None, will be replaced by current internal XML
            tree data structure i.e. *self.xml_tree* class member
            attribute. This parameter should always be at least of
            'xml.etree.ElementTree.ElementTree' type.

        Searches for the XML element matching 'attr_id' unique ID value
        into 'xml_tree' XML tree data structure.

        Returns the corresponding XML element if found, None otherwise.

        Notice: returned XML element should always be at least of
        'xml.etree.ElementTree.Element' type.
        """
        attr_id = tools.normalize_id(attr_id)
        if tools.is_pstr(attr_id):
            if not self.is_tree(xml_tree):
                xml_tree = self.xml_tree
            # end if
            return xml_tree.find(".//*[@id='{}']".format(attr_id))
        # end if
        return None
    # end def

    def get_image (self, filepath):
        """Get registered tk.PhotoImage() by its file path.

        Parameter:
            filepath -- path to actual image resource file (string of
            chars). Current supported formats are 'GIF' and 'PNG'.

        Returns the registered tk.PhotoImage() object corresponding to
        'filepath' image resource path or None, if it has not already
        been registered.

        *NEW* since **tkRAD v1.7**:

        * renamed 'path' parameter to more convenient 'filepath'
          parameter.
        """
        return self.__images.get(path.normalize(filepath))
    # end def

    def get_intvar (self, varname):
        """Get tkinter.IntVar control variable of given name.

        Parameter:
            varname -- name of tkinter control variable (string of
            chars).

        Returns the control variable corresponding to 'varname' name or
        None if it does not exist.
        """
        return self.__tk_variables["intvar"].get(varname)
    # end def

    def get_intvars (self):
        """Get all registered control variables of tkinter.IntVar type
        (mapping dict() object).
        """
        return self.__tk_variables["intvar"]
    # end def

    def get_object_by_id (self, attr_id, default=None):
        """Get XML-generated widget by its 'id' XML attribute value.

        Parameters:
            attr_id -- 'id' XML attribute value (string of chars).
            Should always be a unique ID value through all XML tree
            data structure.

            default -- default value to return on failure (default:
            None).

        Tries to retrieve the tkinter XML-generated widget built upon
        XML element of 'attr_id' unique 'id' XML attribute value.

        Example::

            import tkRAD
            xml_source = \"""
                <tkwidget>
                    <label
                        id="label000"
                        text="hello good people!"
                        layout="pack"
                    />
                </tkwidget>
            \"""
            frame = tkRAD.RADXMLFrame()
            frame.xml_build(xml_source)
            frame.pack(padx=20, pady=20)
            # ask for widget created by XML element of id 'label000'
            widget = frame.get_object_by_id("label000")
            print("Widget info:", repr(widget))
            frame.mainloop()

        Returns this widget on success or 'default' default value
        otherwise.
        """
        return self.__objects.get(
            tools.normalize_id(attr_id).lower(), default
        )
    # end def

    def get_objects (self):
        """Get shallow copy of XML-generated objects registry (mapping
        dict() object).
        """
        return self.__objects.copy()
    # end def

    def get_stringvar (self, varname):
        """Get tkinter.StringVar control variable of given name.

        Parameter:
            varname -- name of tkinter control variable (string of
            chars).

        Returns the control variable corresponding to 'varname' name or
        None if it does not exist.
        """
        return self.__tk_variables["stringvar"].get(varname)
    # end def

    def get_stringvars (self):
        """Get all registered control variables of tkinter.StringVar
        type (mapping dict() object).
        """
        return self.__tk_variables["stringvar"]
    # end def

    @deprecated("Please, use 'xml_dir' class attribute instead.")
    def get_xml_dir (self):
        """Get directory path to XML file.

        *DEPRECATED* since **tkRAD v1.7**.

        Please, use 'xml_dir' class attribute instead.
        """
        return self.xml_dir
    # end def

    @deprecated("Please, use 'xml_file_ext' class attribute instead.")
    def get_xml_file_ext (self):
        """Get file extension part of XML filepath.

        *DEPRECATED* since **tkRAD v1.7**.

        Please, use 'xml_file_ext' class attribute instead.
        """
        return self.xml_file_ext
    # end def

    @deprecated("Please, use 'xml_filename' class attribute instead.")
    def get_xml_filename (self):
        """Get filename part of XML filepath.

        *DEPRECATED* since **tkRAD v1.7**.

        Please, use 'xml_filename' class attribute instead.
        """
        return self.xml_filename
    # end def

    def get_xml_path (self, xml_source=None):
        """Try to build a consistent path to XML file.

        Parameter:
            xml_source -- can either be a file object open in read mode
            or a string of chars i.e. a filename part of an XML
            filepath or a complete path to an existing XML file
            (default: None). If omitted or None, will be replaced by
            one of seeked default values.

        If 'xml_source' is a file object open in read mode, merely
        returns that file object as is.

        Otherwise, uses 'xml_source' to build up a consistent path to
        an XML file.

        Many default values are looked up to define the whole path to
        XML file.

        Please, take a look at source code for more detail.

        Returns the most relevant path from all searched values.

        May raise **RADXMLBaseError** if unable to determine a valid
        XML filename.

        *NEW* since **tkRAD v1.7**:

        * renamed 'filename' parameter to more convenient 'xml_source'
          parameter;
        * added support for file object open in read mode as 'source'
          parameter.
        """
        if hasattr(xml_source, "read"):
            return xml_source
        elif tools.is_pstr(xml_source):
            _filepath = path.normalize(xml_source)
            if os.path.isfile(_filepath) or os.path.sep in xml_source:
                return _filepath
            # end if
        # end if
        filename = tools.choose_str(
            xml_source,
            self.xml_filename,
            self.options[self.RC_OPTIONS["section"]]
                .get(self.RC_OPTIONS["filename"]),
            self.XML_RC.get("filename"),
            self.classname(),
            "component",
        )
        filename = re.sub(r"\..*$", "", filename.strip("."))
        if not filename:
            raise RADXMLBaseError(
                "Unable to determine a valid XML filename."
            )
            return None
        # end if
        _ext = tools.choose_str(
            self.xml_file_ext,
            self.options[self.RC_OPTIONS["section"]]
                .get(self.RC_OPTIONS["file_ext"]),
            self.XML_RC.get("file_ext"),
            ".xml",
        ).strip(".")
        if _ext:
            filename += "." + _ext
        # end if
        _dir = tools.choose_str(
            self.xml_dir,
            self.options[self.RC_OPTIONS["section"]]
                .get(self.RC_OPTIONS["dir"]),
            self.XML_RC.get("dir"),
            "^/xml",
        )
        return os.path.join(path.normalize(_dir), filename)
    # end def

    @deprecated("Please, use 'xml_tree' class attribute instead.")
    def get_xml_tree (self):
        """Get current internal XML tree data structure.

        *DEPRECATED* since **tkRAD v1.7**.

        Please, use 'xml_tree' class attribute instead.
        """
        return self.xml_tree
    # end def

    def is_element (self, xml_element):
        """Indicate if given XML element is of
        'xml.etree.ElementTree.Element' type or not.
        """
        return isinstance(xml_element, etree.Element)
    # end def

    def is_tree (self, tree_object):
        """Indicate if given tree object is of
        'xml.etree.ElementTree.ElementTree' type or not.
        """
        return isinstance(tree_object, etree.ElementTree)
    # end def

    def is_xml (self, content):
        """Indicate if given content may be an XML script or not."""
        return tools.is_pstr(content) and re.search("<.+?>", content)
    # end def

    def normalize_tag (self, xml_element):
        """Normalize XML tag.

        Parameter:
            xml_element -- xml.etree.ElementTree parsed XML element.

        Returns a lowercased string of chars of 'xml_element' tag.
        """
        return str(xml_element.tag).lower()
    # end def

    def set_booleanvar (self, varname):
        """Create a tkinter.BooleanVar control variable.

        Parameter:
            varname -- name of requested control variable (string of
            chars).

        Creates a tkinter.BooleanVar control variable and registers it
        under 'varname' control variable name.

        Does nothing if control variable already exists.

        Returns control variable object instance corresponding to
        requested parameter value, newly created or not.
        """
        return self.set_cvar("booleanvar", varname)
    # end def

    def set_cvar (self, vartype, varname):
        """Create a new tkinter control variable.

        Parameters:
            vartype -- string of chars, tkinter control variable type
            i.e. one of 'BooleanVar', 'DoubleVar', 'IntVar' or
            'StringVar' type (case-insensitive, here).

            varname -- name of requested control variable (string of
            chars).

        Creates a tkinter control variable of 'vartype' control
        variable type and registers it under 'varname' control variable
        name.

        Does nothing if control variable already exists for requested
        parameter values.

        Returns control variable object instance corresponding to
        requested parameter values, newly created or not.

        May raise **RADXMLBaseError** on unsupported control variable
        types.
        """
        vartype = str(vartype).lower()
        if vartype in self.__tk_variables:
            if varname not in self.__tk_variables[vartype]:
                _cvar = self._CVARTYPES[vartype]
                self.__tk_variables[vartype][varname] = _cvar()
            # end if
            return self.__tk_variables[vartype].get(varname)
        else:
            raise RADXMLBaseError(
                "Tkinter control variable must be one of "
                "{} type (case insensitive, here)."
                .format(tuple(self._CVARTYPES))
            )
            return None
        # end if
    # end def

    def set_doublevar (self, varname):
        """Create a tkinter.DoubleVar control variable.

        Parameter:
            varname -- name of requested control variable (string of
            chars).

        Creates a tkinter.DoubleVar control variable and registers it
        under 'varname' control variable name.

        Does nothing if control variable already exists.

        Returns control variable object instance corresponding to
        requested parameter value, newly created or not.
        """
        return self.set_cvar("doublevar", varname)
    # end def

    def set_image (self, filepath):
        """Register a new tk.PhotoImage() object.

        Parameter:
            filepath -- path to actual image resource file (string of
            chars). Current supported formats are 'GIF' and 'PNG'.

        Tries to create a new tk.PhotoImage() object along with
        'filepath' image resource path if not already exists.

        Returns the 'filepath' corresponding tk.PhotoImage() object,
        newly created or not.

        *NEW* since **tkRAD v1.7**:

        * renamed 'path' parameter to more convenient 'filepath'
          parameter.
        """
        filepath = path.normalize(filepath)
        if filepath and filepath not in self.__images:
            self.__images[filepath] = tk.PhotoImage(file=filepath)
        # end if
        return self.__images.get(filepath)
    # end def

    def set_intvar (self, varname):
        """Create a tkinter.IntVar control variable.

        Parameter:
            varname -- name of requested control variable (string of
            chars).

        Creates a tkinter.IntVar control variable and registers it
        under 'varname' control variable name.

        Does nothing if control variable already exists.

        Returns control variable object instance corresponding to
        requested parameter value, newly created or not.
        """
        return self.set_cvar("intvar", varname)
    # end def

    def set_stringvar (self, varname):
        """Create a tkinter.StringVar control variable.

        Parameter:
            varname -- name of requested control variable (string of
            chars).

        Creates a tkinter.StringVar control variable and registers it
        under 'varname' control variable name.

        Does nothing if control variable already exists.

        Returns control variable object instance corresponding to
        requested parameter value, newly created or not.
        """
        return self.set_cvar("stringvar", varname)
    # end def

    @deprecated("Please, use 'xml_dir' class attribute instead.")
    def set_xml_dir (self, value):
        """Set directory path to XML file.

        *DEPRECATED* since **tkRAD v1.7**.

        Please, use 'xml_dir' class attribute instead.
        """
        self.xml_dir = value
    # end def

    @deprecated("Please, use 'xml_file_ext' class attribute instead.")
    def set_xml_file_ext (self, value):
        """Set file extension part of XML filepath.

        *DEPRECATED* since **tkRAD v1.7**.

        Please, use 'xml_file_ext' class attribute instead.
        """
        self.xml_file_ext = value
    # end def

    @deprecated("Please, use 'xml_filename' class attribute instead.")
    def set_xml_filename (self, value):
        """Set filename part of XML filepath.

        *DEPRECATED* since **tkRAD v1.7**.

        Please, use 'xml_filename' class attribute instead.
        """
        self.xml_filename = value
    # end def

    def set_xml_tree (self, **kw):
        """Set a new internal XML tree data structure.

        Parameter:
            kw -- keyword arguments.

        Supported keyword arguments:
            element -- XML tree data structure's root XML element
            object of 'xml.etree.ElementTree.Element' type (default:
            None).

            file -- XML file (default: None). Can either be a file
            object open in read mode or a path to an existing XML file
            (string of chars). This parameter takes precedence over
            'element' parameter if both are given.

        Sets up internal XML tree data structure with a new object
        instance of 'xml.etree.ElementTree.ElementTree' type.
        """
        self.xml_tree = etree.ElementTree(**kw)
    # end def

    def xml_build (self, xml_source=None, **kw):
        """Build XML source into widgets container.

        Parameter:
            xml_source -- can either be a file object open in read mode
            or a string of chars i.e. a filename part of an XML
            filepath or a complete path to an existing XML file
            (default: None). If omitted or None, will be replaced by
            default values and looked up into e.g.
            "^/xml/lowercased_classname.xml".

        Builds widgets from 'xml_source' into current widgets container
        i.e. into tkinter widget referred to by *self.tk_owner* class
        member attribute.

        Returns True on build success, False otherwise.

        *NEW* since **tkRAD v1.7**:

        * renamed 'filename' parameter to more convenient 'xml_source'
          parameter;
        * removed 'silent_mode' parameter: not supported any more;
        * may raise exception errors by now;
        * arranged signature to keep backward compatibility.
        """
        if tools.is_pstr(xml_source) or not self.xml_tree:
            self.xml_load(xml_source)
        # end if
        _root = self.xml_tree.getroot()
        if self._cast_root_element(_root):
            _build_ok = self._build_element(_root, self.tk_owner)
            self._queue.flush_all()
            return _build_ok
        else:
            return False
        # end if
    # end def

    @property
    def xml_dir (self):
        """Directory path to XML file - read-write attribute.

        This class attribute cannot be deleted.

        *NEW* since **tkRAD v1.7**.
        """
        return self.__xml_dir
    # end def

    @xml_dir.setter
    def xml_dir (self, value):
        self.__xml_dir = value
    # end def

    @property
    def xml_file_ext (self):
        """File extension part of XML filepath - read-write attribute.

        This class attribute cannot be deleted.

        *NEW* since **tkRAD v1.7**.
        """
        return self.__xml_file_ext
    # end def

    @xml_file_ext.setter
    def xml_file_ext (self, value):
        self.__xml_file_ext = value
    # end def

    @property
    def xml_filename (self):
        """Filename part of XML filepath - read-write attribute.

        This class attribute cannot be deleted.

        *NEW* since **tkRAD v1.7**.
        """
        return self.__xml_filename
    # end def

    @xml_filename.setter
    def xml_filename (self, value):
        self.__xml_filename = value
    # end def

    def xml_load (self, xml_source=None):
        """Load and parse XML source.

        Parameter:
            xml_source -- can either be a file object open in read mode
            or a string of chars i.e. a filename part of an XML
            filepath or a complete path to an existing XML file
            (default: None). If omitted or None, will be replaced by
            default values and looked up into e.g.
            "^/xml/lowercased_classname.xml".

        Load and parse data coming from 'xml_source'.

        May raise **xml.etree.ElementTree.ParseError** while parsing
        data.

        *NEW* since **tkRAD v1.7**:

        * renamed 'filename' parameter to more convenient 'xml_source'
          parameter;
        * may raise exception errors by now.
        """
        if self.is_xml(xml_source):
            self.set_xml_tree(element=etree.fromstring(xml_source))
        else:
            self.set_xml_tree(file=self.get_xml_path(xml_source))
        # end if
    # end def

    def xml_save (self, to_file=None):
        """Write internal XML tree data structure into a file.

        Parameter:
            to_file -- can either be a file object open in write mode
            or a string of chars i.e. a filename part of an XML
            filepath or a complete path to an existing XML file
            (default: None). If omitted or None, will be entirely
            replaced by default values.

        Writes all internal XML tree data structure into 'to_file' file
        object or filepath.

        *NEW* since **tkRAD v1.7**:

        * renamed 'filename' parameter to more convenient 'to_file'
          parameter;
        * added support for file object open in write mode as 'to_file'
          parameter.
        """
        self.xml_tree.write(
            self.get_xml_path(to_file),
            encoding="UTF-8",
            xml_declaration=True,
            method="xml"
        )
    # end def

    @property
    def xml_tree (self):
        """Internal XML tree data structure - read-write attribute.

        This class attribute cannot be deleted and cannot be set to
        something else than None or an instance of
        xml.etree.ElementTree class.

        May raise **RADXMLBaseError** otherwise.

        *NEW* since **tkRAD v1.7**.
        """
        return self.__xml_tree
    # end def

    @xml_tree.setter
    def xml_tree (self, value):
        if not value:
            self.__xml_tree = None
        elif isinstance(value, etree.ElementTree):
            self.__xml_tree = value
        else:
            raise RADXMLBaseError(
                "XML tree must be of "
                "'xml.etree.ElementTree.ElementTree' type."
            )
        # end if
    # end def

# end class


class RADXMLBaseError (Exception):
    """Error handler for RADXMLBase class."""
    pass
# end class
