#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""XML menu factory module.

Implements:

* class RADXMLMenu -- XML menu builder class.
* class RADXMLMenu -- error handler for RADXMLMenu class.

The *RADXMLMenu* class is intended to be used internally by tkRAD
RADXMLxxxxxxx family widgets.

Example of XML menu generation using *RADXMLMenu*::

    import tkRAD
    xml_source = \"""
        <tkwidget>
            <tkmenu>
                <menu label="_File">
                    <command
                        label="_Quit"
                        accelerator="Ctrl+Q"
                        command="@quit"
                    />
                </menu>
                <menu label="_Testings">
                    <menu label="_Checkbuttons">
                        <checkbutton
                            label="option 1"
                            checked="checked"
                            onvalue="1"
                            variable="chkoption1"
                        />
                        <checkbutton
                            label="option 2"
                            checked="no"
                            onvalue="2"
                            variable="chkoption1"
                        />
                        <checkbutton
                            label="option 3"
                            onvalue="3"
                            variable="chkoption3"
                        />
                    </menu>
                    <menu label="_Radiobuttons">
                        <radiobutton
                            label="option 1"
                            selected="selected"
                            value="1"
                            variable="group1"
                        />
                        <radiobutton
                            label="option 2"
                            value="2"
                            variable="group1"
                        />
                        <radiobutton
                            label="option 3"
                            value="3"
                            variable="group1"
                        />
                    </menu>
                    <separator/>
                    <command label="Sim_ple command"/>
                </menu>
            </tkmenu>
            <menubutton text="tkinter.menubutton" layout="pack">
                <menu>
                    <command label="_Command"/>
                    <separator/>
                    <menu label="_Submenu">
                        <command label="Another _command"/>
                    </menu>
                </menu>
            </menubutton>
            <ttkmenubutton text="ttk.menubutton" layout="pack">
                <menu>
                    <command label="_Command"/>
                    <separator/>
                    <menu label="_Submenu">
                        <command label="Another _command"/>
                    </menu>
                </menu>
            </ttkmenubutton>
            <label
                text="Hello good people!"
                layout="pack"
                resizable="width"
            />
        </tkwidget>
    \"""
    tkRAD.RADApplication().run(xml=xml_source)

*NEW* since **tkRAD v1.7**:

* fixed many buggy behaviours in menu widget building, especially when
  XML menu script is embedded into a more global XML widget script.

Please, feel free to post an issue ticket on GitHub's project
bugtracker if you face any troubleshooting with this new version of
tkRAD's library.

GitHub's project bugtracker: https://github.com/tarball69/tkRAD/issues
(requires a GitHub member account to proceed).
"""

"""
    tkRAD - tkinter Rapid Application Development library

    (c) 2013+ Raphaël SEBAN <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see: http://www.gnu.org/licenses/
"""

import re

import tkinter as tk
from tkinter import ttk

from ..core import tools
from .rad_xml_widget_base import RADXMLWidgetBase


class RADXMLMenu (RADXMLWidgetBase):
    """XML menu builder class."""

    ATTRS = {
        "generic": {
            "id": None,
        },
        "checkable": {
        },
        "menu": {
            "tearoff": 0,
        },
        "child": {
            "label": None,
            "underline": None,
        },
    }

    DOCTYPE = ("tkmenu", "menu")

    DTD = {
        "menu": (
            "menu", "command", "checkbutton", "radiobutton",
            "separator",
        ),
        "tkmenu": (
            "menu",
        ),
    }

    KEYS = {
        "generic": ("id",),
        "checkable": ("selected", "checked"),
        "menu": ("activebackground", "activeborderwidth",
        "activeforeground", "background", "bd", "bg", "borderwidth",
        "cursor", "disabledforeground", "fg", "font", "foreground",
        "postcommand", "relief", "selectcolor", "tearoff",
        "tearoffcommand", "title"),
        "child": ("accelerator", "activebackground",
        "activeforeground", "background", "bitmap", "columnbreak",
        "command", "compound", "font", "foreground", "hidemargin",
        "image", "label", "menu", "offvalue", "onvalue",
        "selectcolor", "selectimage", "state", "underline", "value",
        "variable"),
    }

    SYMBOLS = (
        (re.compile(r"\^+|C-|(?i)co?n?tro?l"), "Control-"),
        (re.compile(r"M-|(?i)meta|(?i)alt"), "Alt-"),
        (re.compile(r"(?i)shi?ft"), "Shift-"),
        (re.compile(r"\+$"), "plus"),
        (re.compile(r"\-$"), "minus"),
        (re.compile(r"\*$"), "asterisk"),
        (re.compile(r"\/$"), "slash"),
        (re.compile(r"\.$"), "period"),
        (re.compile(r"\,$"), "comma"),
        (re.compile(r"\:$"), "colon"),
        (re.compile(r"\;$"), "semicolon"),
        (re.compile(r"\?$"), "question"),
        (re.compile(r"\!$"), "exclam"),
        (re.compile(r"\$$"), "dollar"),
        (re.compile(r"\%$"), "percent"),
        (re.compile(r"\@$"), "at"),
        (re.compile(r"\&$"), "ampersand"),
        (re.compile(r"\#$"), "numbersign"),
        (re.compile(r"\_$"), "underscore"),
        (re.compile(r"(?i)less|(?i)\blt\b"), "less"),
        (re.compile(r"(?i)greater|(?i)\bgt\b"), "greater"),
        (re.compile(r"(?i)spa?ce?"), "space"),
        (re.compile(r"(?i)ba?ckspa?ce?"), "BackSpace"),
        (re.compile(r"(?i)del(?:ete)?\b"), "Delete"),
        (re.compile(r"(?i)bre?a?k|(?i)ca?nce?l"), "Cancel"),
        (re.compile(r"(?i)esc(?:ape)?\b"), "Escape"),
        (re.compile(r"(?i)tab(?:ulate)?"), "Tab"),
        (re.compile(r"(?i)ho?me?"), "Home"),
        (re.compile(r"(?i)end"), "End"),
        (re.compile(r"(?i)page[\s\+\-]*?up"), "Prior"),
        (re.compile(r"(?i)page[\s\+\-]*?do?w?n"), "Next"),
        (re.compile(r"(?i)(?:arrow)?[\s\+\-]*?up"), "Up"),
        (re.compile(r"(?i)(?:arrow)?[\s\+\-]*?do?w?n"), "Down"),
        (re.compile(r"(?i)(?:arrow)?[\s\+\-]*?left"), "Left"),
        (re.compile(r"(?i)(?:arrow)?[\s\+\-]*?right"), "Right"),
        (re.compile(r"(?i)f(\d+)$"), r"F\1"),
        (re.compile(r"[<>]+"), ""),
        (re.compile(r"^\W+|\W+$"), ""),
        (re.compile(r"\W+"), "-"),
    )

    XML_RC = {
        "dir": "^/xml/menu",
        "file_ext": ".xml",
    }

    def _build_element_checkbutton (self, xml_tag, xml_element,
                                                            tk_parent):
        _attrs = self._init_checkables(xml_element, tk_parent)
        _coptions = self._init_coptions(xml_element, tk_parent)
        _cvar = _coptions.get("variable")
        if _cvar and _attrs.get("checked"):
            _cvar.set(_coptions.get("onvalue") or "")
        # end if
        tk_parent.add_checkbutton(**_coptions)
        return True
    # end def

    def _build_element_command (self, xml_tag, xml_element, tk_parent):
        _coptions = self._init_coptions(xml_element, tk_parent)
        tk_parent.add_command(**_coptions)
        return True
    # end def

    def _build_element_menu (self, xml_tag, xml_element, tk_parent):
        if self.is_menu_handler(tk_parent):
            _new_menu = self._get_menu_object(xml_element, tk_parent)
            if self.is_menu(tk_parent):
                _coptions = self._init_coptions(xml_element, tk_parent)
                _coptions["menu"] = _new_menu
                tk_parent.add_cascade(**_coptions)
                del _coptions
            else:
                tk_parent["menu"] = _new_menu
            # end if
            return self._loop_on_children(
                xml_element, _new_menu, accept=self.DTD.get(xml_tag)
            )
        else:
            raise RADXMLMenuError(
                "Cannot build menu object into '{}'."
                .format(repr(tk_parent))
            )
            return False
        # end if
    # end def

    def _build_element_radiobutton (self, xml_tag, xml_element,
                                                            tk_parent):
        _attrs = self._init_checkables(xml_element, tk_parent)
        _coptions = self._init_coptions(xml_element, tk_parent)
        _cvar = _coptions.get("variable")
        if _cvar and _attrs.get("selected"):
            _cvar.set(_coptions.get("value") or "")
        # end if
        tk_parent.add_radiobutton(**_coptions)
        return True
    # end def

    def _build_element_separator (self, xml_tag, xml_element,
                                                            tk_parent):
        tk_parent.add_separator()
        return True
    # end def

    def _build_element_tkmenu (self, xml_tag, xml_element, tk_parent):
        if not self.is_menu_handler(tk_parent):
            tk_parent = self.tk_owner.winfo_toplevel()
        # end if
        _new_menu = self._get_menu_object(xml_element, tk_parent)
        tk_parent["menu"] = _new_menu
        if self.is_topmenu_handler(tk_parent):
            _dtd = "tkmenu"
        else:
            _dtd = "menu"
        # end if
        return self._loop_on_children(
            xml_element, _new_menu, accept=self.DTD.get(_dtd)
        )
    # end def

    def _get_menu_object (self, xml_element, tk_parent):
        _attrs = self._init_generics(xml_element, tk_parent)
        _moptions = self._init_moptions(xml_element, tk_parent)
        _new_menu = tk.Menu(tk_parent)
        self._register_object_by_id(_new_menu, _attrs.get("id"))
        self._queue.flush("widget", widget=_new_menu)
        _new_menu.configure(**_moptions.flatten())
        del _attrs, _moptions
        return _new_menu
    # end def

    def _init_checkables (self, xml_element, tk_parent):
        _attrs = self.ATTRS["checkable"].copy()
        _attrs.update(
            tools.dict_only_keys(
                xml_element.attrib, *self.KEYS["checkable"]
            )
        )
        return self._parse_xml_attributes(
            xml_element, tk_parent, xml_attrs=_attrs
        ).flatten()
    # end def

    def _init_coptions (self, xml_element, tk_parent):
        _coptions = self.ATTRS["child"].copy()
        _coptions.update(
            tools.dict_only_keys(
                xml_element.attrib, *self.KEYS["child"]
            )
        )
        _coptions = self._parse_xml_attributes(
            xml_element, tk_parent, xml_attrs=_coptions
        )
        self._queue.flush("widget")
        _coptions = _coptions.flatten()
        _acc = self.TK_ACCEL
        _cmd = _coptions.get("command")
        if tools.is_pstr(_acc) and callable(_cmd):
            self.tk_owner.bind_all(_acc, _cmd)
        # end if
        return _coptions
    # end def

    def _init_generics (self, xml_element, tk_parent):
        return dict(id=self.element_get_id(xml_element))
    # end def

    def _init_moptions (self, xml_element, tk_parent):
        _moptions = self.ATTRS["menu"].copy()
        _moptions.update(
            tools.dict_only_keys(
                xml_element.attrib, *self.KEYS["menu"]
            )
        )
        return self._parse_xml_attributes(
            xml_element, tk_parent, xml_attrs=_moptions
        )
    # end def

    def _parse_attr_accelerator (self, attribute, **kw):
        if self._is_new(attribute):
            _acc = attribute.value
            for (_search, _replace) in self.SYMBOLS:
                _acc = _search.sub(_replace, _acc)
            # end for
            _chunks = _acc.split("-")
            _detail = _chunks[-1]
            if len(_detail) == 1:
                if "shift" in _acc.lower():
                    _detail = _detail.upper()
                else:
                    _detail = _detail.lower()
                # end if
                _chunks[-1] = _detail
                _acc = "-".join(_chunks)
            # end if
            self.TK_ACCEL = "<" + _acc + ">"
            attribute.parsed = True
        # end if
    # end def

    def _parse_attr_activeborderwidth (self, attribute, **kw):
        self._tkRAD_dimension_support(attribute, **kw)
    # end def

    def _parse_attr_columnbreak (self, attribute, **kw):
        self._tkRAD_boolean_support(attribute, **kw)
    # end def

    def _parse_attr_hidemargin (self, attribute, **kw):
        self._tkRAD_boolean_support(attribute, **kw)
    # end def

    def _parse_attr_label (self, attribute, **kw):
        self._tkRAD_label_support(attribute, **kw)
    # end def

    def _parse_attr_postcommand (self, attribute, **kw):
        self._tkRAD_command_support(attribute, **kw)
    # end def

    def _parse_attr_tearoff (self, attribute, **kw):
        self._tkRAD_boolean_support(attribute, **kw)
    # end def

    def _parse_attr_tearoffcommand (self, attribute, **kw):
        self._tkRAD_command_support(attribute, **kw)
    # end def

    def _parse_attr_title (self, attribute, **kw):
        self._tkRAD_label_support(attribute, **kw)
    # end def

    def get_menu (self, attr_id):
        """Get menu object by its 'id' XML attribute value.

        Parameter:
            attr_id -- 'id' XML attribute value (string of chars).
            Should always be a unique ID value through all XML tree
            data structure.

        Returns the menu object registered under 'id' XML attribute
        value or None if not exists.
        """
        return self.get_object_by_id(attr_id)
    # end def

    def is_menu (self, widget):
        """Indicate if given widget is of (or subclass of)
        'tkinter.Menu' type or not.
        """
        return isinstance(widget, tk.Menu)
    # end def

    def is_menu_handler (self, widget):
        """Indicate if given widget can handle a menu object i.e.
        widget is one of (or subclass of) tkinter.Menu,
        tkinter.Menubutton, ttk.Menubutton, tkinter.Tk or not.
        """
        return isinstance(
            widget, (tk.Menu, tk.Menubutton, ttk.Menubutton, tk.Tk)
        )
    # end def

    def is_topmenu_handler (self, widget):
        """Indicate if given widget can handle a top menu object i.e.
        widget is of (or subclass of) 'tkinter.Tk' type or not.
        """
        return isinstance(widget, tk.Tk)
    # end def

# end class


class RADXMLMenuError (Exception):
    """Error handler for RADXMLMenu class."""
    pass
# end class
