#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""XML widget factory module.

Implements:

* class RADXMLWidget -- XML widget builder class.
* class RADXMLWidgetError -- error handler for RADXMLWidget class.

The *RADXMLWidget* XML widget builder factory is intended to be used as
a mixin class with tkinter widgets container such as *tkinter.Frame* or
*ttk.Frame*.

Please, take a look at *RADXMLFrame* documentation page for more
detail.

Example of XML widget generation using *RADXMLWidget*::

    import tkRAD
    xml_source = \"""
        <tkwidget>
            <tkmenu>
                <menu label="_File">
                    <command
                        label="_Quit"
                        accelerator="Ctrl+Q"
                        command="@quit"
                    />
                </menu>
                <menu label="_Testings">
                    <menu label="_Checkbuttons">
                        <checkbutton
                            label="option 1"
                            checked="checked"
                            onvalue="1"
                            variable="chkoption1"
                        />
                        <checkbutton
                            label="option 2"
                            checked="no"
                            onvalue="2"
                            variable="chkoption1"
                        />
                        <checkbutton
                            label="option 3"
                            onvalue="3"
                            variable="chkoption3"
                        />
                    </menu>
                    <menu label="_Radiobuttons">
                        <radiobutton
                            label="option 1"
                            selected="selected"
                            value="1"
                            variable="group1"
                        />
                        <radiobutton
                            label="option 2"
                            value="2"
                            variable="group1"
                        />
                        <radiobutton
                            label="option 3"
                            value="3"
                            variable="group1"
                        />
                    </menu>
                    <separator/>
                    <command label="Sim_ple command"/>
                </menu>
            </tkmenu>
            <menubutton text="tkinter.menubutton" layout="pack">
                <menu>
                    <command label="_Command"/>
                    <separator/>
                    <menu label="_Submenu">
                        <command label="Another _command"/>
                    </menu>
                </menu>
            </menubutton>
            <ttkmenubutton text="ttk.menubutton" layout="pack">
                <menu>
                    <command label="_Command"/>
                    <separator/>
                    <menu label="_Submenu">
                        <command label="Another _command"/>
                    </menu>
                </menu>
            </ttkmenubutton>
            <label
                text="Hello good people!"
                layout="pack"
                resizable="width"
            />
            <frame
                layout="pack"
                resizable="yes"
            >
                <text
                    id="mytext"
                    height="5"
                    width="20"
                    layout="grid"
                    layout_options="row=0, column=0"
                    resizable="yes"
                />
                <ttkscrollbar
                    orient="vertical"
                    connect="mytext"
                    layout="grid"
                    layout_options="row=0, column=1"
                    resizable="height"
                />
                <ttkscrollbar
                    orient="horizontal"
                    connect="mytext"
                    layout="grid"
                    layout_options="row=1, column=0"
                    resizable="width"
                />
            </frame>
        </tkwidget>
    \"""
    tkRAD.RADApplication().run(xml=xml_source)

*NEW* since **tkRAD v1.7**:

* added new *RADXMLWidgetError* error handler class for *RADXMLWidget*
  class;
* added new '_LEGAL_ANCHORS' protected class member attribute;
* renamed all erroneous 'signal' mention to 'event' everywhere in this
  module;
* renamed all erroneous 'slot' mention to 'handler' everywhere in this
  module;
* error messages are *NOT* translated any more;
* optimized some portions of code.

Please, feel free to post an issue ticket on GitHub's project
bugtracker if you face any troubleshooting with this new version of
tkRAD's library.

GitHub's project bugtracker: https://github.com/tarball69/tkRAD/issues
(requires a GitHub member account to proceed).
"""

"""
    tkRAD - tkinter Rapid Application Development library

    (c) 2013+ Raphaël SEBAN <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see: http://www.gnu.org/licenses/
"""

import re
import warnings

import tkinter as tk
from tkinter import ttk

from ..core import tools
from ..core import path

from .rad_xml_widget_base import RADXMLWidgetBase


class RADXMLWidget (RADXMLWidgetBase):
    """XML widget builder class."""

    ANCHORS = (
        (re.compile(r"(?i)north|top|up"), tk.N),
        (re.compile(r"(?i)south|bottom|down"), tk.S),
        (re.compile(r"(?i)east|right"), tk.E),
        (re.compile(r"(?i)west|left"), tk.W),
        (re.compile(r"\W+"), r""),
        (re.compile(tk.N + "+"), tk.N),
        (re.compile(tk.S + "+"), tk.S),
        (re.compile(tk.E + "+"), tk.E),
        (re.compile(tk.W + "+"), tk.W),
        (re.compile(tk.W + tk.N), tk.NW),
        (re.compile(tk.E + tk.N), tk.NE),
        (re.compile(tk.W + tk.S), tk.SW),
        (re.compile(tk.E + tk.S), tk.SE),
    )

    ATTRS = {
        "common": {
            "id": None,
        },
        "button": {
            "underline": None,
        },
        "checkbutton": {
            "underline": None,
        },
        "label": {
            "underline": None,
        },
        "listbox": {
            #~ "name": None,
            "class": None,
            "args": None,
            "module": None,
            "choices": None,
            "start": None,
            "layout": None,
            "layout_options": None,
            "resizable": "no",
        },
        "menubutton": {
            "underline": None,
        },
        "optionmenu": {
            #~ "name": None,
            "listvariable": None,
            "variable": None,
            "choices": None,
            "start": None,
            "layout": None,
            "layout_options": None,
            "resizable": "no",
        },
        "radiobutton": {
            "underline": None,
        },
        "tkwidget": {
        },
        "ttkbutton": {
            "underline": None,
        },
        "ttkcheckbutton": {
            "underline": None,
        },
        "ttklabel": {
            "underline": None,
        },
        "ttkmenubutton": {
            "underline": None,
        },
        "ttkradiobutton": {
            "underline": None,
        },
        "ttktab": {
            "sticky": "all",
            "underline": -1,
        },
        "widget": {
            #~ "name": None,
            "class": None,
            "args": None,
            "module": "",
            "layout": None,
            "layout_options": None,
            "resizable": "no",
        },
        "include": {
            #~ "name": None,
            "src": None,
            "xml_dir": None,
            "xml_filename": None,
            "xml_file_ext": None,
        },
        "module": {
            "from": None,
            "import": None,
            "as": None,
        },
        "configure": {
            "widget": None,
        },
        "layout": {
            "widget": None,
            "layout": "pack",
            "layout_options": None,
            "resizable": "no",
        },
        "event": {
            "event": None,
            "handler": None,
        },
        "tkevent": {
            "widget": None,
            "bind": "bind",
            "class": None,
            "seq": None,
            "handler": None,
            "add": None,
        },
    }

    CLASSES = {
        "button":           "tk.Button",
        "canvas":           "tk.Canvas",
        "checkbutton":      "tk.Checkbutton",
        "entry":            "tk.Entry",
        "frame":            "tk.Frame",
        "label":            "tk.Label",
        "labelframe":       "tk.LabelFrame",
        "listbox":          "tk.Listbox",
        "menu":             "tk.Menu",
        "menubutton":       "tk.Menubutton",
        "message":          "tk.Message",
        "optionmenu":       "tk.OptionMenu",
        "panedwindow":      "tk.PanedWindow",
        "radiobutton":      "tk.Radiobutton",
        "scale":            "tk.Scale",
        "scrollbar":        "tk.Scrollbar",
        "spinbox":          "tk.Spinbox",
        "text":             "tk.Text",
        "toplevel":         "tk.Toplevel",
        "ttkbutton":        "ttk.Button",
        "ttkcheckbutton":   "ttk.Checkbutton",
        "ttkcombobox":      "ttk.Combobox",
        "ttkentry":         "ttk.Entry",
        "ttkframe":         "ttk.Frame",
        "ttklabel":         "ttk.Label",
        "ttklabelframe":    "ttk.LabelFrame",
        "ttkmenubutton":    "ttk.Menubutton",
        "ttknotebook":      "ttk.Notebook",
        "ttkpanedwindow":   "ttk.PanedWindow",
        "ttkprogressbar":   "ttk.Progressbar",
        "ttkradiobutton":   "ttk.Radiobutton",
        "ttkscale":         "ttk.Scale",
        "ttkscrollbar":     "ttk.Scrollbar",
        "ttkseparator":     "ttk.Separator",
        "ttksizegrip":      "ttk.Sizegrip",
        "ttktab":           "ttk.Frame",
        "ttktreeview":      "ttk.Treeview",
    }

    DOCTYPE = "tkwidget"

    DTD = {
        "menubutton": ("menu", "tkmenu"),
        "ttkmenubutton": ("menu", "tkmenu"),
        "ttknotebook": ("ttktab", ),
        "widget": (
            "configure", "event", "include", "layout", "module",
            "style", "tkevent", "tkmenu", "ttkstyle", "ttktheme",
            "widget",
        ) + tuple(CLASSES.keys()),
    }

    XML_RC = {
        "dir": "^/xml/widget",
        "file_ext": ".xml",
    }

    _LEGAL_ANCHORS = (
        tk.N, tk.S, tk.E, tk.W, tk.NW, tk.NE, tk.SW, tk.SE
    )

    def _build_element_button (self, xml_tag, xml_element, tk_parent):
        return self._build_tk_native(xml_tag, xml_element, tk_parent)
    # end def

    def _build_element_canvas (self, xml_tag, xml_element, tk_parent):
        return self._build_tk_native(xml_tag, xml_element, tk_parent)
    # end def

    def _build_element_checkbutton (self, xml_tag, xml_element,
                                                            tk_parent):
        _ok = self._build_tk_native(xml_tag, xml_element, tk_parent)
        if xml_element.get("checked"):
            self.WIDGET.select()
        else:
            self.WIDGET.deselect()
        # end if
        return _ok
    # end def

    def _build_element_configure (self, xml_tag, xml_element,
                                                            tk_parent):
        if self.cast_element(xml_element):
            _attributes = self._init_attributes(
                xml_tag, xml_element, tk_parent
            )
            _widget = _attributes.get("widget", tk_parent)
            return self._set_widget_config(_widget, self.TK_CONFIG)
        # end if
        return False
    # end def

    def _build_element_entry (self, xml_tag, xml_element, tk_parent):
        return self._build_tk_native(xml_tag, xml_element, tk_parent)
    # end def

    def _build_element_event (self, xml_tag, xml_element, tk_parent):
        if self.cast_element(xml_element):
            _attributes = self._init_attributes(
                xml_tag, xml_element, tk_parent
            )
            return self.events.connect(
                _attributes.get("event"), _attributes.get("handler")
            )
        # end if
        return False
    # end def

    def _build_element_frame (self, xml_tag, xml_element, tk_parent):
        return self._build_tk_native(xml_tag, xml_element, tk_parent)
    # end def

    def _build_element_include (self, xml_tag, xml_element, tk_parent):
        if self.cast_element(xml_element):
            _attributes = self._init_attributes(
                xml_tag, xml_element, tk_parent
            )
            _widget = RADXMLWidget(
                tk_owner=tk_parent, h_owner=self.h_owner, **_attributes
            )
            _widget.XML_RC.clear()
            _widget.xml_load(_attributes.get("src"))
            xml_element = _widget.xml_tree.getroot()
            del _attributes, _widget
            return self._loop_on_children(
                xml_element, tk_parent, accept=self.DTD.get("widget")
            )
        # end if
        return False
    # end def

    def _build_element_label (self, xml_tag, xml_element, tk_parent):
        return self._build_tk_native(xml_tag, xml_element, tk_parent)
    # end def

    def _build_element_labelframe (self, xml_tag, xml_element,
                                                            tk_parent):
        return self._build_tk_native(xml_tag, xml_element, tk_parent)
    # end def

    def _build_element_layout (self, xml_tag, xml_element, tk_parent):
        if self.cast_element(xml_element):
            _attributes = self._init_attributes(
                xml_tag, xml_element, tk_parent
            )
            _widget = _attributes.get("widget", tk_parent)
            self._set_layout(_widget, _attributes, tk_parent)
            return True
            # end if
        # end if
        return False
    # end def

    def _build_element_listbox (self, xml_tag, xml_element, tk_parent):
        if self.is_tk_parent(tk_parent):
            _attributes = self._init_deferred_attributes(
                xml_tag, xml_element, tk_parent
            )
            _args = _attributes.get("args") or ""
            if "tk_parent" not in _args:
                _args = "tk_parent, " + _args
            # end if
            _widget = eval("tk.Listbox({})".format(_args))
            self._queue.flush("widget", widget=_widget)
            _attributes = _attributes.flatten()
            self._register_object_by_id(_widget, _attributes.get("id"))
            self._set_class_member(_attributes.get("name"), _widget)
            _widget.delete(0, tk.END)
            _choices = _attributes.get("choices")
            if _choices:
                _widget.insert(0, *_choices)
                _start = _attributes.get("start")
                if tools.is_num(_start):
                    if _start > 0:
                        _start = min(_start, len(_choices) - 1)
                    elif _start < 0:
                        _start = max(0, len(_choices) + _start)
                    # end if
                elif _start in _choices:
                    _start = _choices.index(_start)
                else:
                    _start = -1
                # end if
                _widget.selection_anchor(_start)
                _widget.selection_set(_start)
                _widget.activate(_start)
                _widget.see(_start)
            # end if
            self._set_widget_config(_widget, self.TK_CONFIG)
            self._set_layout(_widget, _attributes, tk_parent)
            return True
        else:
            raise RADXMLWidgetError(
                "Cannot build '{}' object into '{}' object."
                .format(
                    xml_element.get("class") or self.WIDGET_CLASS,
                    repr(tk_parent)
                )
            )
            return False
        # end if
    # end def

    def _build_element_menu (self, xml_tag, xml_element, tk_parent):
        return self._build_element_tkmenu(
            xml_tag, xml_element, tk_parent
            )
    # end def

    def _build_element_menubutton (self, xml_tag, xml_element,
                                                            tk_parent):
        return self._build_tk_native(xml_tag, xml_element, tk_parent)
    # end def

    def _build_element_message (self, xml_tag, xml_element, tk_parent):
        return self._build_tk_native(xml_tag, xml_element, tk_parent)
    # end def

    def _build_element_module (self, xml_tag, xml_element, tk_parent):
        if self.cast_element(xml_element):
            _attributes = self._init_attributes(
                xml_tag, xml_element, tk_parent
            )
            _attrs = {
                "from": tools.str_complete(
                    "from {} ", _attributes.get("from")
                ),
                "import": tools.str_complete(
                    "import {}", _attributes.get("import")
                ),
                "as": tools.str_complete(
                    " as {}", _attributes.get("as")
                ),
            }
            exec("{from}{import}{as}".format(**_attrs), globals())
            return True
        # end if
        return False
    # end def

    def _build_element_optionmenu (self, xml_tag, xml_element,
                                                            tk_parent):
        if self.is_tk_parent(tk_parent):
            _attributes = self._init_deferred_attributes(
                xml_tag, xml_element, tk_parent
            )
            _cvar = tools.choose(
                _attributes.get("listvariable"),
                _attributes.get("variable"),
                tk.StringVar(),
            )
            _choices = tools.choose(
                _attributes.get("choices"),
                [_("<empty>")],
                ["<empty>"],
            )
            _widget = tk.OptionMenu(tk_parent, _cvar, *_choices)
            self._queue.flush("widget", widget=_widget)
            _attributes = _attributes.flatten()
            self._register_object_by_id(_widget, _attributes.get("id"))
            self._set_class_member(_attributes.get("name"), _widget)
            _start = _attributes.get("start")
            if tools.is_num(_start):
                if _start > 0:
                    _start = min(_start, len(_choices) - 1)
                elif _start < 0:
                    _start = max(0, len(_choices) + _start)
                # end if
                _start = _choices[_start]
            elif _start not in _choices:
                _start = _choices[0]
            # end if
            _cvar.set(str(_start))
            self._set_layout(_widget, _attributes, tk_parent)
            return True
        else:
            raise RADXMLWidgetError(
                "Cannot build '{}' object into '{}' object."
                .format(
                    xml_element.get("class") or self.WIDGET_CLASS,
                    repr(tk_parent)
                )
            )
            return False
        # end if
    # end def

    def _build_element_panedwindow (self, xml_tag, xml_element,
                                                            tk_parent):
        return self._build_tk_native(xml_tag, xml_element, tk_parent)
    # end def

    def _build_element_radiobutton (self, xml_tag, xml_element,
                                                            tk_parent):
        _ok = self._build_tk_native(xml_tag, xml_element, tk_parent)
        if xml_element.get("selected"):
            self.WIDGET.select()
        else:
            self.WIDGET.deselect()
        # end if
        return _ok
    # end def

    def _build_element_scale (self, xml_tag, xml_element, tk_parent):
        return self._build_tk_native(xml_tag, xml_element, tk_parent)
    # end def

    def _build_element_scrollbar (self, xml_tag, xml_element,
                                                            tk_parent):
        _ok = self._build_tk_native(xml_tag, xml_element, tk_parent)
        _scrollbar = self.WIDGET
        _target = self.TK_CHILD_CONFIG.get("connect")
        if _target and _scrollbar:
            try:
                if str(_scrollbar.cget("orient")) == tk.VERTICAL:
                    _target.configure(yscrollcommand=_scrollbar.set)
                    _scrollbar.configure(command=_target.yview)
                else:
                    _target.configure(xscrollcommand=_scrollbar.set)
                    _scrollbar.configure(command=_target.xview)
                # end if
            except:
                raise RADXMLWidgetError(
                    "Cannot connect widget '{}' of id '{}' "
                    "to scrollbar."
                    .format(
                        _target.__class__.__name__,
                        xml_element.get("connect")
                    )
                ) from None
            # end try
        # end if
        return _ok
    # end def

    def _build_element_spinbox (self, xml_tag, xml_element, tk_parent):
        return self._build_tk_native(xml_tag, xml_element, tk_parent)
    # end def

    def _build_element_style (self, xml_tag, xml_element, tk_parent):
        if self.cast_element(xml_element):
            _attributes = self._init_attributes(
                xml_tag, xml_element, tk_parent
            )
            _id = _attributes.pop("id", None)
            self._register_object_by_id(_attributes, _id)
            return True
        # end if
        return False
    # end def

    def _build_element_text (self, xml_tag, xml_element, tk_parent):
        return self._build_tk_native(xml_tag, xml_element, tk_parent)
    # end def

    def _build_element_tkevent (self, xml_tag, xml_element, tk_parent):
        if self.cast_element(xml_element):
            _attributes = self._init_attributes(
                xml_tag, xml_element, tk_parent
            )
            _widget = _attributes.get("widget") or tk_parent
            _bind = _attributes.get("bind") or ""
            _seq = _attributes.get("seq") or ""
            _add = _attributes.get("add") or ""
            _handler = _attributes.get("handler")
            _method = getattr(_widget, _bind)
            if _bind == "bind_class":
                _class = _attributes.get("class") or ""
                _method(_class, _seq, _handler, _add)
            else:
                _method(_seq, _handler, _add)
            # end if
            return True
        # end if
        return False
    # end def

    def _build_element_tkmenu (self, xml_tag, xml_element, tk_parent):
        if self.cast_element(xml_element):
            from .rad_xml_menu import RADXMLMenu
            _attributes = self._init_deferred_attributes(
                xml_tag, xml_element, tk_parent
            )
            _widget = RADXMLMenu(
                tk_owner=tk_parent, h_owner=self.h_owner
            )
            self._queue.flush("widget", widget=_widget)
            _attributes = _attributes.flatten()
            self._register_object_by_id(_widget, _attributes.get("id"))
            self._set_class_member(_attributes.get("name"), _widget)
            _widget.set_xml_tree(element=xml_element)
            _ok = _widget.xml_build()
            self.get_stringvars().update(_widget.get_stringvars())
            return _ok
        # end if
        return False
    # end def

    def _build_element_tkwidget (self, xml_tag, xml_element,
                                                            tk_parent):
        if self.is_tk_parent(tk_parent):
            return self._loop_on_children(
                xml_element, tk_parent, accept=self.DTD.get("widget"),
            )
        else:
            raise RADXMLWidgetError(
                "Cannot build root widget into '{}' object."
                .format(repr(tk_parent))
            )
            return False
        # end if
    # end def

    def _build_element_toplevel (self, xml_tag, xml_element,
                                                            tk_parent):
        return self._build_tk_native(xml_tag, xml_element, tk_parent)
    # end def

    def _build_element_ttkbutton (self, xml_tag, xml_element,
                                                            tk_parent):
        return self._build_tk_native(xml_tag, xml_element, tk_parent)
    # end def

    def _build_element_ttkcheckbutton (self, xml_tag, xml_element,
                                                            tk_parent):
        _ok = self._build_tk_native(xml_tag, xml_element, tk_parent)
        if xml_element.get("checked"):
            self.WIDGET.invoke()
        # end if
        return _ok
    # end def

    def _build_element_ttkcombobox (self, xml_tag, xml_element,
                                                            tk_parent):
        return self._build_tk_native(xml_tag, xml_element, tk_parent)
    # end def

    def _build_element_ttkentry (self, xml_tag, xml_element,
                                                            tk_parent):
        return self._build_tk_native(xml_tag, xml_element, tk_parent)
    # end def

    def _build_element_ttkframe (self, xml_tag, xml_element,
                                                            tk_parent):
        return self._build_tk_native(xml_tag, xml_element, tk_parent)
    # end def

    def _build_element_ttklabel (self, xml_tag, xml_element,
                                                            tk_parent):
        return self._build_tk_native(xml_tag, xml_element, tk_parent)
    # end def

    def _build_element_ttklabelframe (self, xml_tag, xml_element,
                                                            tk_parent):
        return self._build_tk_native(xml_tag, xml_element, tk_parent)
    # end def

    def _build_element_ttkmenubutton (self, xml_tag, xml_element,
                                                            tk_parent):
        return self._build_tk_native(xml_tag, xml_element, tk_parent)
    # end def

    def _build_element_ttknotebook (self, xml_tag, xml_element,
                                                            tk_parent):
        return self._build_tk_native(xml_tag, xml_element, tk_parent)
    # end def

    def _build_element_ttkpanedwindow (self, xml_tag, xml_element,
                                                            tk_parent):
        return self._build_tk_native(xml_tag, xml_element, tk_parent)
    # end def

    def _build_element_ttkprogressbar (self, xml_tag, xml_element,
                                                            tk_parent):
        return self._build_tk_native(xml_tag, xml_element, tk_parent)
    # end def

    def _build_element_ttkradiobutton (self, xml_tag, xml_element,
                                                            tk_parent):
        _ok = self._build_tk_native(xml_tag, xml_element, tk_parent)
        if xml_element.get("selected"):
            self.WIDGET.invoke()
        # end if
        return _ok
    # end def

    def _build_element_ttkscale (self, xml_tag, xml_element,
                                                            tk_parent):
        return self._build_tk_native(xml_tag, xml_element, tk_parent)
    # end def

    def _build_element_ttkscrollbar (self, xml_tag, xml_element,
                                                            tk_parent):
        return self._build_element_scrollbar(
            xml_tag, xml_element, tk_parent
        )
    # end def

    def _build_element_ttkseparator (self, xml_tag, xml_element,
                                                            tk_parent):
        return self._build_tk_native(xml_tag, xml_element, tk_parent)
    # end def

    def _build_element_ttksizegrip (self, xml_tag, xml_element,
                                                            tk_parent):
        return self._build_tk_native(xml_tag, xml_element, tk_parent)
    # end def

    def _build_element_ttkstyle (self, xml_tag, xml_element,
                                                            tk_parent):
        if self.cast_element(xml_element):
            _attributes = self._init_attributes(
                xml_tag, xml_element, tk_parent
            )
            _attributes.pop("id", None)
            _apply = _attributes.pop("apply", ".")
            _style = ttk.Style()
            _style.configure(_apply, **_attributes)
            if tools.is_pstr(xml_element.text):
                _cdata = (
                    xml_element.text
                        .strip("\n\t ;")
                        .replace("\n", "")
                        .replace('"', "'")
                )
                _cdata = re.sub(r"/\*.*?\*/", "", _cdata)
                _cdata = _cdata.replace("*", ".")
                _cdata = list(
                    filter(
                        None, set(re.split(r"(.*?\{.*?\})", _cdata))
                    )
                )
                for _def in _cdata:
                    _elements, _attrs = _def.split("{")
                    _elements = (
                        re.sub(r"[^\w,.!:]+", "", _elements).split(",")
                    )
                    _attrs = self._parse_xml_attributes(
                        xml_element,
                        tk_parent,
                        xml_attrs=eval(
                            "{{{}}}"
                            .format(
                                re.sub(
                                    r"\s*(\w+)\s*:\s*(.*?)\s*;",
                                    r'"\1":"\2",',
                                    _attrs.strip("{ };") + ";"
                                ).strip(";")
                            )
                        ),
                    ).flatten()
                    if tools.is_pdict(_attrs):
                        for _element in _elements:
                            _element = _element.split(":")
                            if len(_element) > 1:
                                _states = _element[1:]
                                _element = _element[0]
                                _mattrs = _attrs.copy()
                                for (_key, _value) in _mattrs.items():
                                    _map = _style.map(_element, _key)
                                    _map.insert(
                                        0, tuple(_states + [_value])
                                    )
                                    _mattrs[_key] = _map
                                # end for
                                _style.map(_element, **_mattrs)
                            else:
                                _style.configure(_element[0], **_attrs)
                            # end if - mapping
                        # end for - elements
                    # end if - attrs
                # end for - cdata
            # end if - CSS-like syntax
            return True
        # end if
        return False
    # end def

    def _build_element_ttktab (self, xml_tag, xml_element, tk_parent):
        return self._build_tk_native(xml_tag, xml_element, tk_parent)
    # end def

    def _build_element_ttktheme (self, xml_tag, xml_element,
                                                            tk_parent):
        if self.cast_element(xml_element):
            _attributes = self._init_attributes(
                xml_tag, xml_element, tk_parent
            )
            ttk.Style().theme_use(_attributes.get("use"))
            return True
        # end if
        return False
    # end def

    def _build_element_ttktreeview (self, xml_tag, xml_element,
                                                            tk_parent):
        return self._build_tk_native(xml_tag, xml_element, tk_parent)
    # end def

    def _build_element_widget (self, xml_tag, xml_element, tk_parent,
                                                                **kw):
        if self.is_tk_parent(tk_parent):
            kw.update(addon_attrs=self.ATTRS.get("widget"))
            _attributes = self._init_deferred_attributes(
                xml_tag, xml_element, tk_parent, **kw
            )
            _class = eval("{module}{class}".format(**_attributes))
            _args = _attributes.get("args") or ""
            if issubclass(_class, tk.BaseWidget) \
                                        and "tk_parent" not in _args:
                _args = "master=tk_parent, " + _args
            # end if
            _widget = eval("_class({})".format(_args))
            self._queue.flush("widget", widget=_widget)
            _attributes = _attributes.flatten()
            self._register_object_by_id(_widget, _attributes.get("id"))
            self.WIDGET = _widget
            self._set_class_member(_attributes.get("name"), _widget)
            self._set_widget_config(_widget, self.TK_CONFIG)
            self._set_layout(_widget, _attributes, tk_parent)
            del _class, _args, self.TK_CONFIG
            _build_ok = self._loop_on_children(
                xml_element,
                _widget,
                accept=self.DTD.get(xml_tag) or self.DTD.get("widget"),
            )
            _init = _attributes.get("init")
            if callable(_init):
                kw.update(
                    widget=_widget,
                    parent=tk_parent,
                    xml_attributes=_attributes,
                )
                _init(**kw)
            # end if
            return _build_ok
        else:
            raise RADXMLWidgetError(
                "Cannot build '{}' object into '{}' object."
                .format(
                    xml_element.get("class") or self.WIDGET_CLASS,
                    repr(tk_parent)
                )
            )
            return False
        # end if
    # end def

    def _build_tk_native (self, xml_tag, xml_element, tk_parent):
        if self.cast_element(xml_element):
            _cname = tools.choose_str(
                self.CLASSES.get(xml_tag), xml_tag, "Frame",
            )
            _cname = _cname.strip(".").split(".")
            _module = xml_element.get("module") or ""
            if len(_cname) > 1:
                _module = _cname[0] + "."
            # end if
            xml_element.set("class", tools.normalize_id(_cname[-1]))
            xml_element.set("module", _module)
            return self._build_element_widget(
                xml_tag,  xml_element,  tk_parent,
            )
        # end if
        return False
    # end def

    def _ensure_string_value (self, attribute, **kw):
        if self._is_unparsed(attribute):
            attribute.value = tools.choose_str(
                attribute.value,
                kw.get("default"),
            )
            self._tk_config(attribute, **kw)
        # end if
    # end def

    def _init_attributes (self, xml_tag, xml_element, tk_parent, **kw):
        _attributes = self._init_deferred_attributes(
            xml_tag, xml_element, tk_parent, **kw
        )
        self._queue.flush("widget", **kw)
        return _attributes.flatten()
    # end def

    @deprecated("Please, use '_init_attributes()' method instead.")
    def _init_attributes_flat (self, xml_tag, xml_element, tk_parent,
                                                                **kw):
        return self._init_attributes(
            xml_tag, xml_element, tk_parent, **kw
        )
    # end def

    def _init_deferred_attributes (self, xml_tag, xml_element,
                                                    tk_parent, **kw):
        _dicts = (
            self.ATTRS.get("common"),
            self.ATTRS.get(xml_tag),
            kw.get("addon_attrs"),
            xml_element.attrib,
        )
        _attributes = dict()
        for _dict in _dicts:
            if tools.is_pdict(_dict):
                _attributes.update(_dict)
            # end if
        # end for
        kw["xml_attrs"] = _attributes
        return self._parse_xml_attributes(xml_element, tk_parent, **kw)
    # end def

    def _layout_toplevel (self, widget, attrs, tk_parent):
        _wstate = attrs.get("visibility") or ""
        _resizable = attrs.get("resizable") or ""
        widget.transient(attrs.get("transient"))
        widget.title(attrs.get("title"))
        widget.minsize(
            width=attrs.get("minwidth"), height=attrs.get("minheight")
        )
        _maxwidth = attrs.get("maxwidth")
        _maxheight = attrs.get("maxheight")
        widget.maxsize(width=_maxwidth, height=_maxheight)
        widget.resizable(
            width=(_resizable in ("yes", "width")),
            height=(_resizable in ("yes", "height")),
        )
        if _wstate == "maximized" and _resizable == "yes" \
                                    and not (_maxwidth or _maxheight):
            widget.deiconify()
            try:
                widget.wm_attributes("-zoomed", "1")
            except:
                warnings.warn("Unable to maximize window.")
            # end try
        elif _wstate == "minimized":
            widget.iconify()
        elif _wstate == "hidden":
            widget.withdraw()
        else:
            widget.deiconify()
        # end if
    # end def

    def _parse_attr__after (self, attribute, **kw):
        kw.update(tk_child_config=True)
        self._tkRAD_widget_support(attribute, **kw)
    # end def

    def _parse_attr__before (self, attribute, **kw):
        kw.update(tk_child_config=True)
        self._tkRAD_widget_support(attribute, **kw)
    # end def

    def _parse_attr__height (self, attribute, **kw):
        kw.update(tk_child_config=True)
        self._tkRAD_dimension_support(attribute, **kw)
    # end def

    def _parse_attr__minsize (self, attribute, **kw):
        kw.update(tk_child_config=True)
        self._tkRAD_dimension_support(attribute, **kw)
    # end def

    def _parse_attr__padx (self, attribute, **kw):
        kw.update(tk_child_config=True)
        self._tkRAD_dimension_support(attribute, **kw)
    # end def

    def _parse_attr__pady (self, attribute, **kw):
        kw.update(tk_child_config=True)
        self._tkRAD_dimension_support(attribute, **kw)
    # end def

    def _parse_attr__sticky (self, attribute, **kw):
        if self._is_new(attribute):
            _sticky = attribute.value.lower()
            if not set(_sticky).issubset(self.STICKY_ALL):
                _sticky = self.STICKY_ALL
            # end if
            attribute.value = _sticky
            kw.update(tk_child_config=True)
            self._tk_config(attribute, **kw)
        # end if
    # end def

    def _parse_attr__width (self, attribute, **kw):
        kw.update(tk_child_config=True)
        self._tkRAD_dimension_support(attribute, **kw)
    # end def

    def _parse_attr_activerelief (self, attribute, **kw):
        self._tkRAD_relief_support(attribute, **kw)
    # end def

    def _parse_attr_activestyle (self, attribute, **kw):
        kw.update(
            default="dotbox", values=("underline", "none"),
        )
        self._fix_values(attribute, **kw)
    # end def

    def _parse_attr_add (self, attribute, **kw):
        attribute.value = "+" * tools.is_pstr(attribute.value)
    # end def

    def _parse_attr_after (self, attribute, **kw):
        self._parse_attr__after(attribute, **kw)
    # end def

    def _parse_attr_anchor (self, attribute, **kw):
        if self._is_new(attribute):
            _anchor = attribute.value
            for (_search, _replace) in self.ANCHORS:
                _anchor = _search.sub(_replace, _anchor)
            # end for
            if _anchor not in self._LEGAL_ANCHORS:
                _anchor = tk.CENTER
            # end if
            attribute.value = _anchor
            self._tk_config(attribute, **kw)
        # end if
    # end def

    def _parse_attr_apply (self, attribute, **kw):
        if self._is_new(attribute):
            attribute.value = re.sub(r"[^\w\.]+", "", attribute.value)
            attribute.parsed = True
        # end if
    # end def

    def _parse_attr_args (self, attribute, **kw):
        if tools.is_pstr(attribute.value):
            _args = re.sub(
                r"(?i)\b(?:self|parent)\b",
                "tk_parent",
                attribute.value
            )
            _args = self._replace_alias(_args, **kw)
        else:
            _args = ""
        # end if
        attribute.value = _args
    # end def

    def _parse_attr_as (self, attribute, **kw):
        attribute.value = tools.normalize_id(attribute.value)
    # end def

    def _parse_attr_aspect (self, attribute, **kw):
        self._tkRAD_integer_support(attribute, **kw)
    # end def

    def _parse_attr_autoseparators (self, attribute, **kw):
        self._tkRAD_boolean_support(attribute, **kw)
    # end def

    def _parse_attr_before (self, attribute, **kw):
        self._parse_attr__before(attribute, **kw)
    # end def

    def _parse_attr_bind (self, attribute, **kw):
        kw.update(
            default="bind",
            values=("bind_class", "bind_all"),
            no_tk_config=True,
        )
        self._fix_values(attribute, **kw)
    # end def

    def _parse_attr_buttonbackground (self, attribute, **kw):
        self._tkRAD_color_support(attribute, **kw)
    # end def

    def _parse_attr_buttoncursor (self, attribute, **kw):
        self._tkRAD_cursor_support(attribute, **kw)
    # end def

    def _parse_attr_buttondownrelief (self, attribute, **kw):
        self._tkRAD_relief_support(attribute, **kw)
    # end def

    def _parse_attr_buttonup (self, attribute, **kw):
        self._tkRAD_relief_support(attribute, **kw)
    # end def

    def _parse_attr_choices (self, attribute, **kw):
        if self._is_new(attribute):
            attribute.value = list(map(str,
                eval("[{}]".format(attribute.value.strip("()[]{}")))
            ))
            attribute.parsed = True
        # end if
    # end def

    def _parse_attr_class (self, attribute, **kw):
        if self._is_unparsed(attribute):
            attribute.value = tools.choose_str(
                tools.normalize_id(attribute.value),
                self.WIDGET_CLASS,
                "Frame",
            )
            attribute.parsed = True
        # end if
    # end def

    def _parse_attr_class_ (self, attribute, **kw):
        self._tkRAD_any_value_support(attribute, **kw)
    # end def

    def _parse_attr_closeenough (self, attribute, **kw):
        self._tkRAD_float_support(attribute, **kw)
    # end def

    def _parse_attr_columns (self, attribute, **kw):
        self._parse_attr_values(attribute, **kw)
    # end def

    def _parse_attr_confine (self, attribute, **kw):
        self._tkRAD_boolean_support(attribute, **kw)
    # end def

    def _parse_attr_connect (self, attribute, **kw):
        kw.update(tk_child_config=True)
        self._tkRAD_widget_support(attribute, **kw)
    # end def

    def _parse_attr_default (self, attribute, **kw):
        self._tkRAD_state_support(attribute, **kw)
    # end def

    def _parse_attr_digits (self, attribute, **kw):
        self._tkRAD_integer_support(attribute, **kw)
    # end def

    def _parse_attr_direction (self, attribute, **kw):
        kw.update(
            default="below",
            values=("above", "flush", "left", "right"),
        )
        self._fix_values(attribute, **kw)
    # end def

    def _parse_attr_disabledbackground (self, attribute, **kw):
        self._tkRAD_color_support(attribute, **kw)
    # end def

    def _parse_attr_displaycolumns (self, attribute, **kw):
        self._parse_attr_values(attribute, **kw)
    # end def

    def _parse_attr_elementborderwidth (self, attribute, **kw):
        self._tkRAD_dimension_support(attribute, **kw)
    # end def

    def _parse_attr_event (self, attribute, **kw):
        kw.update(no_tk_config=True)
        self._ensure_string_value(attribute, **kw)
    # end def

    def _parse_attr_exportselection (self, attribute, **kw):
        self._tkRAD_boolean_support(attribute, **kw)
    # end def

    def _parse_attr_fieldbackground (self, attribute, **kw):
        self._tkRAD_color_support(attribute, **kw)
    # end def

    def _parse_attr_format (self, attribute, **kw):
        if self._is_new(attribute):
            _fmt = re.search(
                r"\D*(\d*\.\d+)|\D*(\d*)", attribute.value
            )
            if _fmt:
                _fmt = tools.str_complete(
                    "%{}f",
                    "".join(filter(None, _fmt.groups()))
                )
            # end if
            attribute.value = _fmt
            self._tk_config(attribute, **kw)
        # end if
    # end def

    def _parse_attr_from (self, attribute, **kw):
        if self._is_unparsed(attribute):
            attribute.value = (
                tools.normalize_relative_module(attribute.value)
            )
            attribute.parsed = True
        # end if
    # end def

    def _parse_attr_from_ (self, attribute, **kw):
        attribute.name = "from"
        self._tkRAD_float_support(attribute, **kw)
    # end def

    def _parse_attr_handlepad (self, attribute, **kw):
        self._tkRAD_dimension_support(attribute, **kw)
    # end def

    def _parse_attr_handler (self, attribute, **kw):
        self._tkRAD_command_support(attribute, **kw)
    # end def

    def _parse_attr_handlesize (self, attribute, **kw):
        self._tkRAD_dimension_support(attribute, **kw)
    # end def

    def _parse_attr_height (self, attribute, xml_tag, **kw):
        if xml_tag in ("button", "checkbutton", "label", "listbox",
        "menubutton", "radiobutton", "text"):
            self._tkRAD_integer_support(attribute, **kw)
        else:
            self._tkRAD_dimension_support(attribute, **kw)
        # end if
    # end def

    def _parse_attr_highlightbackground (self, attribute, **kw):
        self._tkRAD_color_support(attribute, **kw)
    # end def

    def _parse_attr_highlightcolor (self, attribute, **kw):
        self._tkRAD_color_support(attribute, **kw)
    # end def

    def _parse_attr_highlightthickness (self, attribute, **kw):
        self._tkRAD_dimension_support(attribute, **kw)
    # end def

    def _parse_attr_import (self, attribute, **kw):
        if self._is_unparsed(attribute):
            attribute.value = tools.normalize_import(attribute.value)
            attribute.parsed = True
        # end if
    # end def

    def _parse_attr_increment (self, attribute, **kw):
        self._tkRAD_float_support(attribute, **kw)
    # end def

    def _parse_attr_indicatorcolor (self, attribute, **kw):
        self._tkRAD_color_support(attribute, **kw)
    # end def

    def _parse_attr_indicatoron (self, attribute, **kw):
        self._tkRAD_boolean_support(attribute, **kw)
    # end def

    def _parse_attr_init (self, attribute, **kw):
        kw.update(no_tk_config=True)
        self._tkRAD_command_support(attribute, **kw)
    # end def

    def _parse_attr_insertbackground (self, attribute, **kw):
        self._tkRAD_color_support(attribute, **kw)
    # end def

    def _parse_attr_insertborderwidth (self, attribute, **kw):
        self._tkRAD_dimension_support(attribute, **kw)
    # end def

    def _parse_attr_insertofftime (self, attribute, **kw):
        self._tkRAD_integer_support(attribute, **kw)
    # end def

    def _parse_attr_insertontime (self, attribute, **kw):
        self._tkRAD_integer_support(attribute, **kw)
    # end def

    def _parse_attr_insertwidth (self, attribute, **kw):
        self._tkRAD_dimension_support(attribute, **kw)
    # end def

    def _parse_attr_invalidcommand (self, attribute, **kw):
        self._tkRAD_command_support(attribute, **kw)
    # end def

    def _parse_attr_jump (self, attribute, **kw):
        self._tkRAD_boolean_support(attribute, **kw)
    # end def

    def _parse_attr_justify (self, attribute, **kw):
        kw.update(
            default=tk.CENTER, values=(tk.LEFT, tk.RIGHT),
        )
        self._fix_values(attribute, **kw)
    # end def

    def _parse_attr_label (self, attribute, **kw):
        self._tkRAD_label_support(attribute, **kw)
    # end def

    def _parse_attr_labelanchor (self, attribute, **kw):
        self._parse_attr_anchor(attribute, **kw)
    # end def

    def _parse_attr_labelwidget (self, attribute, **kw):
        self._tkRAD_widget_support(attribute, **kw)
    # end def

    def _parse_attr_layout (self, attribute, **kw):
        kw.update(
            default="none",
            values=("pack", "grid", "place"),
            no_tk_config=True,
        )
        self._fix_values(attribute, **kw)
    # end def

    def _parse_attr_layout_options (self, attribute, **kw):
        if self._is_unparsed(attribute):
            _lopts = attribute.value
            if tools.is_pstr(_lopts):
                _lopts = self._replace_alias(_lopts, **kw)
                _lopts = (
                    eval("dict({})".format(_lopts.strip("()[]{}")))
                )
            elif not tools.is_pdict(_lopts):
                _lopts = dict()
            # end if
            attribute.value = _lopts
            attribute.parsed = True
        # end if
    # end def

    def _parse_attr_length (self, attribute, **kw):
        self._tkRAD_dimension_support(attribute, **kw)
    # end def

    def _parse_attr_listvariable (self, attribute, **kw):
        self._tkRAD_cvar_support(attribute, **kw)
    # end def

    def _parse_attr_maxheight (self, attribute, **kw):
        kw.update(no_tk_config=True)
        self._tkRAD_integer_support(attribute, **kw)
    # end def

    def _parse_attr_maximum (self, attribute, **kw):
        self._tkRAD_integer_support(attribute, **kw)
    # end def

    def _parse_attr_maxundo (self, attribute, **kw):
        self._tkRAD_integer_support(attribute, **kw)
    # end def

    def _parse_attr_maxwidth (self, attribute, **kw):
        kw.update(no_tk_config=True)
        self._tkRAD_integer_support(attribute, **kw)
    # end def

    def _parse_attr_minheight (self, attribute, **kw):
        kw.update(no_tk_config=True)
        self._tkRAD_integer_support(attribute, **kw)
    # end def

    def _parse_attr_minsize (self, attribute, **kw):
        self._parse_attr__minsize(attribute, **kw)
    # end def

    def _parse_attr_minwidth (self, attribute, **kw):
        kw.update(no_tk_config=True)
        self._tkRAD_integer_support(attribute, **kw)
    # end def

    def _parse_attr_mode (self, attribute, **kw):
        kw.update(
            default="determinate", values=("indeterminate", ),
        )
        self._fix_values(attribute, **kw)
    # end def

    def _parse_attr_module (self, attribute, **kw):
        if self._is_new(attribute):
            _module = attribute.value.lstrip(".")
            if _module.endswith("."):
                _name = _module
            else:
                _name = ""
                _module = self.get_element_by_id(_module)
                if self.is_element(_module):
                    _import = (
                        tools.normalize_import(_module.get("import"))
                    )
                    if _import != "*":
                        _name = tools.choose_str(
                            tools.normalize_id(_module.get("as")),
                            _import,
                        ) + "."
                    # end if
                else:
                    raise RADXMLWidgetError(
                        "Module of id '{}' has *NOT* been found."
                        .format(attribute.value)
                    )
                # end if
            # end if
            attribute.value = _name.lstrip(".")
            attribute.parsed = True
        # end if
    # end def

    def _parse_attr_offrelief (self, attribute, **kw):
        self._tkRAD_relief_support(attribute, **kw)
    # end def

    def _parse_attr_opaqueresize (self, attribute, **kw):
        self._tkRAD_boolean_support(attribute, **kw)
    # end def

    def _parse_attr_orient (self, attribute, attrs, xml_tag, **kw):
        kw.update(
            default=tk.VERTICAL, values=(tk.HORIZONTAL, ),
        )
        self._fix_values(attribute, **kw)
        if xml_tag == "ttkpanedwindow" and "args" in attrs:
            _args = tools.choose_str(attrs["args"]).split(",")
            _args.append("orient='{}'".format(attribute.value))
            attrs["args"] = ",".join(filter(None, _args))
            self.TK_CONFIG.pop("orient", None)
        # end if
    # end def

    def _parse_attr_overrelief (self, attribute, **kw):
        self._tkRAD_relief_support(attribute, **kw)
    # end def

    def _parse_attr_padding (self, attribute, **kw):
        self._tkRAD_dimension_support(attribute, **kw)
    # end def

    def _parse_attr_padx (self, attribute, **kw):
        self._tkRAD_dimension_support(attribute, **kw)
    # end def

    def _parse_attr_pady (self, attribute, **kw):
        self._tkRAD_dimension_support(attribute, **kw)
    # end def

    def _parse_attr_readonlybackground (self, attribute, **kw):
        self._tkRAD_color_support(attribute, **kw)
    # end def

    def _parse_attr_repeatdelay (self, attribute, **kw):
        self._tkRAD_integer_support(attribute, **kw)
    # end def

    def _parse_attr_repeatinterval (self, attribute, **kw):
        self._tkRAD_integer_support(attribute, **kw)
    # end def

    def _parse_attr_resizable (self, attribute, **kw):
        kw.update(
            default="no",
            values=("yes", "width", "height"),
            no_tk_config=True,
        )
        self._fix_values(attribute, **kw)
    # end def

    def _parse_attr_resolution (self, attribute, **kw):
        self._tkRAD_float_support(attribute, **kw)
    # end def

    def _parse_attr_sashpad (self, attribute, **kw):
        self._tkRAD_dimension_support(attribute, **kw)
    # end def

    def _parse_attr_sashrelief (self, attribute, **kw):
        self._tkRAD_relief_support(attribute, **kw)
    # end def

    def _parse_attr_sashwidth (self, attribute, **kw):
        self._tkRAD_dimension_support(attribute, **kw)
    # end def

    def _parse_attr_scrollregion (self, attribute, **kw):
        if self._is_new(attribute):
            attribute.value = tuple(map(tools.ensure_int,
                eval("[{}]".format(attribute.value.strip("(){}[]")))
            ))
            self._tk_config(attribute, **kw)
        # end if
    # end def

    def _parse_attr_selectbackground (self, attribute, **kw):
        self._tkRAD_color_support(attribute, **kw)
    # end def

    def _parse_attr_selectborderwidth (self, attribute, **kw):
        self._tkRAD_dimension_support(attribute, **kw)
    # end def

    def _parse_attr_selectforeground (self, attribute, **kw):
        self._tkRAD_color_support(attribute, **kw)
    # end def

    def _parse_attr_selectmode (self, attribute, xml_tag, **kw):
        if xml_tag == "ttktreeview":
            kw.update(
                default="extended", values=("none", "browse"),
            )
        else:
            kw.update(
                default="browse",
                values=("single", "multiple", "extended"),
            )
        # end if
        self._fix_values(attribute, **kw)
    # end def

    def _parse_attr_seq (self, attribute, **kw):
        if self._is_new(attribute):
            _seq = attribute.value.replace("<", "", 1)
            _seq = _seq.replace(">", "", 1)
            attribute.value = "<" + _seq + ">"
            attribute.parsed = True
        # end if
    # end def

    def _parse_attr_show (self, attribute, xml_tag, **kw):
        if xml_tag == "ttktreeview":
            kw.update(
                default="headings", values=("tree", ),
            )
            self._fix_values(attribute, **kw)
        else:
            self._tkRAD_any_value_support(attribute, **kw)
        # end if
    # end def

    def _parse_attr_showhandle (self, attribute, **kw):
        self._tkRAD_boolean_support(attribute, **kw)
    # end def

    def _parse_attr_showvalue (self, attribute, **kw):
        self._tkRAD_boolean_support(attribute, **kw)
    # end def

    def _parse_attr_sliderlength (self, attribute, **kw):
        self._tkRAD_dimension_support(attribute, **kw)
    # end def

    def _parse_attr_sliderrelief (self, attribute, **kw):
        self._tkRAD_relief_support(attribute, **kw)
    # end def

    def _parse_attr_spacing1 (self, attribute, **kw):
        self._tkRAD_dimension_support(attribute, **kw)
    # end def

    def _parse_attr_spacing2 (self, attribute, **kw):
        self._tkRAD_dimension_support(attribute, **kw)
    # end def

    def _parse_attr_spacing3 (self, attribute, **kw):
        self._tkRAD_dimension_support(attribute, **kw)
    # end def

    def _parse_attr_src (self, attribute, **kw):
        if self._is_new(attribute):
            attribute.value = path.normalize(attribute.value)
            attribute.parsed = True
        # end if
    # end def

    def _parse_attr_start (self, attribute, **kw):
        if self._is_new(attribute):
            _start = attribute.value
            if _start.startswith("@"):
                _start = tools.ensure_int(_start.lstrip("@"))
            else:
                _start = re.sub(r"\\([@'])", r"\1", _start)
            # end if
            attribute.value = _start
            attribute.parsed = True
        # end if
    # end def

    def _parse_attr_state (self, attribute, xml_tag, **kw):
        if xml_tag in ("entry", "spinbox"):
            _values = ("disabled", "readonly", )
        else:
            _values = ("disabled", )
        # end if
        kw.update(default="normal", values=_values)
        self._fix_values(attribute, **kw)
    # end def

    def _parse_attr_sticky (self, attribute, **kw):
        self._parse_attr__sticky(attribute, **kw)
    # end def

    def _parse_attr_style (self, attribute, **kw):
        if self._is_new(attribute):
            _style = attribute.value
            if "." not in _style:
                _style = self.get_object_by_id(_style, _style)
            # end if
            attribute.value = _style
            self._tk_config(attribute)
        # end if
    # end def

    def _parse_attr_tabs (self, attribute, **kw):
        self._parse_attr_choices(attribute, **kw)
        self._tk_config(attribute, **kw)
    # end def

    def _parse_attr_takefocus (self, attribute, **kw):
        self._tkRAD_boolean_support(attribute, **kw)
    # end def

    def _parse_attr_text (self, attribute, **kw):
        self._tkRAD_label_support(attribute, **kw)
    # end def

    def _parse_attr_textvariable (self, attribute, **kw):
        self._tkRAD_cvar_support(attribute, **kw)
    # end def

    def _parse_attr_tickinterval (self, attribute, **kw):
        self._tkRAD_float_support(attribute, **kw)
    # end def

    def _parse_attr_title (self, attribute, **kw):
        kw.update(no_tk_config=True)
        self._tkRAD_label_support(attribute, **kw)
    # end def

    def _parse_attr_to (self, attribute, **kw):
        self._tkRAD_float_support(attribute, **kw)
    # end def

    def _parse_attr_transient (self, attribute, **kw):
        kw.update(no_tk_config=True)
        self._tkRAD_widget_support(attribute, **kw)
    # end def

    def _parse_attr_troughcolor (self, attribute, **kw):
        self._tkRAD_color_support(attribute, **kw)
    # end def

    def _parse_attr_undo (self, attribute, **kw):
        self._tkRAD_boolean_support(attribute, **kw)
    # end def

    def _parse_attr_use (self, attribute, **kw):
        kw.update(
            default="default", values=ttk.Style().theme_names(),
        )
        self._fix_values(attribute, **kw)
    # end def

    def _parse_attr_validate (self, attribute, **kw):
        kw.update(
            default="none",
            values=("focus", "focusin", "focusout", "key", "all"),
        )
        self._fix_values(attribute, **kw)
    # end def

    def _parse_attr_validatecommand (self, attribute, **kw):
        self._tkRAD_command_support(attribute, **kw)
    # end def

    def _parse_attr_value (self, attribute, xml_tag, **kw):
        if xml_tag == "ttkscale":
            self._tkRAD_float_support(attribute, **kw)
        else:
            self._tkRAD_any_value_support(attribute, **kw)
        # end if
    # end def

    def _parse_attr_values (self, attribute, **kw):
        self._parse_attr_choices(attribute, **kw)
        self._tk_config(attribute, **kw)
    # end def

    def _parse_attr_visibility (self, attribute, **kw):
        kw.update(
            no_tk_config=True,
            default="normal",
            values=("maximized", "minimized", "hidden"),
        )
        self._fix_values(attribute, **kw)
    # end def

    def _parse_attr_weight (self, attribute, **kw):
        kw.update(tk_child_config=True)
        self._tkRAD_integer_support(attribute, **kw)
    # end def

    def _parse_attr_width (self, attribute, xml_tag, **kw):
        if xml_tag in ("button", "checkbutton", "entry", "label",
                "listbox", "menubutton", "radiobutton", "spinbox",
                "text"):
            self._tkRAD_integer_support(attribute, **kw)
        else:
            self._tkRAD_dimension_support(attribute, **kw)
        # end if
    # end def

    def _parse_attr_wrap (self, attribute, xml_tag, **kw):
        if xml_tag == "text":
            kw.update(
                default="none", values=("char", "word"),
            )
            self._fix_values(attribute, **kw)
        else:
            self._tkRAD_boolean_support(attribute, **kw)
        # end if
    # end def

    def _parse_attr_wraplength (self, attribute, **kw):
        self._tkRAD_dimension_support(attribute, **kw)
    # end def

    def _parse_attr_xml_dir (self, attribute, **kw):
        kw.update(
            no_tk_config=True,
            default=self.XML_RC.get("dir")
        )
        self._ensure_string_value(attribute, **kw)
    # end def

    def _parse_attr_xml_file_ext (self, attribute, **kw):
        kw.update(
            no_tk_config=True,
            default=self.XML_RC.get("file_ext")
        )
        self._ensure_string_value(attribute, **kw)
    # end def

    def _parse_attr_xml_filename (self, attribute, **kw):
        kw.update(no_tk_config=True)
        self._ensure_string_value(attribute, **kw)
    # end def

    def _parse_attr_xscrollcommand (self, attribute, **kw):
        self._tkRAD_command_support(attribute, **kw)
    # end def

    def _parse_attr_xscrollincrement (self, attribute, **kw):
        self._tkRAD_dimension_support(attribute, **kw)
    # end def

    def _parse_attr_yscrollcommand (self, attribute, **kw):
        self._tkRAD_command_support(attribute, **kw)
    # end def

    def _parse_attr_yscrollincrement (self, attribute, **kw):
        self._tkRAD_dimension_support(attribute, **kw)
    # end def

    def _replace_alias (self, str_value, attrs, **kw):
        if tools.is_pstr(str_value):
            self._parse_attr_module(attrs.get_item("module"), **kw)
            return str_value.replace("@", attrs.get("module") or "")
        # end if
        return ""
    # end def

    def _set_layout (self, widget, attrs, tk_parent):
        if isinstance(widget, tk.Toplevel):
            self._layout_toplevel(widget, attrs, tk_parent)
        elif isinstance(tk_parent, ttk.Notebook):
            tk_parent.add(
                widget,
                **tools.dict_only_keys(
                    attrs, "compound", "image", "padding", "sticky",
                    "text", "underline",
                )
            )
        elif isinstance(tk_parent, ttk.PanedWindow):
            tk_parent.add(
                widget, weight=self.TK_CHILD_CONFIG.get("weight")
            )
        elif isinstance(tk_parent, tk.PanedWindow):
            if attrs.get("layout") in ("pack", "grid", "place"):
                _sticky = {
                    "width": tk.E + tk.W,
                    "height": tk.N + tk.S,
                    "yes": tk.NW + tk.SE,
                }.get(attrs.get("resizable"))
                self.TK_CHILD_CONFIG.setdefault("sticky", _sticky)
                tk_parent.add(widget, **self.TK_CHILD_CONFIG)
            # end if
        elif hasattr(widget, str(attrs.get("layout"))):
            self._set_resizable(widget, attrs, tk_parent)
            exec("widget.{layout}(**{layout_options})".format(**attrs))
        # end if
    # end def

    def _set_resizable (self, widget, attrs, tk_parent):
        _resizable = (attrs.get("resizable") or "no").lower()
        _layout = attrs.get("layout")
        _lopts = attrs.get("layout_options") or dict()
        if _layout and _resizable in ("yes", "width", "height"):
            if _layout == "pack":
                _lopts.update(
                    {
                        "yes": dict(expand=1, fill=tk.BOTH),
                        "width": dict(expand=0, fill=tk.X),
                        "height": dict(expand=1, fill=tk.Y),
                    }.get(_resizable)
                )
            elif _layout == "grid":
                _lopts["sticky"] = {
                    "width": tk.W + tk.E,
                    "height": tk.N + tk.S,
                }.get(_resizable, self.STICKY_ALL)
                if _resizable in ("yes", "width"):
                    tk_parent.columnconfigure(
                        tools.ensure_int(_lopts.get("column", 0)),
                        weight=1
                    )
                # end if
                if _resizable in ("yes", "height"):
                    tk_parent.rowconfigure(
                        tools.ensure_int(_lopts.get("row", 0)),
                        weight=1
                    )
                # end if
            elif _layout == "place":
                _lopts.update(
                    {
                        "yes": dict(relwidth=1, relheight=1),
                        "width": dict(relwidth=1, relheight=None),
                        "height": dict(relwidth=None, relheight=1),
                    }.get(_resizable)
                )
            # end if
        # end if
        attrs["layout_options"] = _lopts
    # end def

    def _set_widget_config (self, widget, config):
        if hasattr(widget, "configure") and isinstance(config, dict):
            _attrs = config.get("style")
            if not tools.is_pdict(_attrs):
                _attrs = dict()
            # end if
            _attrs.update(config)
            if tools.is_pdict(widget.configure()):
                _attrs = tools.dict_only_keys(
                    _attrs, *widget.configure().keys()
                )
            # end if
            widget.configure(**_attrs)
            return True
        # end if
        return False
    # end def

# end class


class RADXMLWidgetError (Exception):
    """Error handler for RADXMLWidget class."""
    pass
# end class
