#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Intermediary mixin base class module.

Implements:

* class RADXMLWidgetBase -- mixin base class for *RADXMLMenu* and
  *RADXMLWidget* classes.
* class RADXMLWidgetBaseError -- error handler for RADXMLWidgetBase
  class.

This module is mainly intended for internal use.

Intermediary mixin base class for *RADXMLMenu* and *RADXMLWidget*
classes.

*NEW* since **tkRAD v1.7**:

Please, feel free to post an issue ticket on GitHub's project
bugtracker if you face any troubleshooting with this new version of
tkRAD's library.

GitHub's project bugtracker: https://github.com/tarball69/tkRAD/issues
(requires a GitHub member account to proceed).
"""

"""
    tkRAD - tkinter Rapid Application Development library

    (c) 2013+ Raphaël SEBAN <motus@laposte.net>

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.

    If not, see: http://www.gnu.org/licenses/
"""

import re
import tkinter as tk
from ..core import tools
from .rad_xml_base import RADXMLBase


class RADXMLWidgetBase (RADXMLBase):
    """Mixin base class for *RADXMLMenu* and *RADXMLWidget* classes."""

    ATTRIBUTE_PARSER = "_parse_attr_{xml_attribute}"

    ATTRS = {
        "common": {
            "id": None,
            "name": None,
            "widget": None,
        },
    }

    def _before_building_element (self, **kw):
        self.WIDGET_CLASS = "Frame"
        self.WIDGET = None
        self.TK_ACCEL = ""
        self.TK_CONFIG = dict()
        self.TK_CHILD_CONFIG = dict()
    # end def

    def _fix_values (self, attribute, default, values, **kw):
        if values and self._is_new(attribute):
            _value = attribute.value.lower()
            if _value not in values:
                _value = default
            # end if
            attribute.value = _value
            attribute.update_xml_element()
            self._tk_config(attribute, **kw)
        # end if
    # end def

    def _is_new (self, attribute):
        return (
            attribute and not attribute.parsed and
            tools.is_pstr(attribute.value)
        )
    # end def

    def _is_unparsed (self, attribute):
        return attribute and not attribute.parsed
    # end def

    def _parse_attr_activebackground (self, attribute, **kw):
        self._tkRAD_color_support(attribute, **kw)
    # end def

    def _parse_attr_activeforeground (self, attribute, **kw):
        self._tkRAD_color_support(attribute, **kw)
    # end def

    def _parse_attr_background (self, attribute, **kw):
        self._tkRAD_color_support(attribute, **kw)
    # end def

    def _parse_attr_bd (self, attribute, **kw):
        self._tkRAD_dimension_support(attribute, **kw)
    # end def

    def _parse_attr_bg (self, attribute, **kw):
        self._tkRAD_color_support(attribute, **kw)
    # end def

    def _parse_attr_bitmap (self, attribute, **kw):
        self._tkRAD_bitmap_support(attribute, **kw)
    # end def

    def _parse_attr_borderwidth (self, attribute, **kw):
        self._tkRAD_dimension_support(attribute, **kw)
    # end def

    def _parse_attr_checked (self, attribute, **kw):
        kw.update(no_tk_config=True)
        self._tkRAD_boolean_support(attribute, **kw)
    # end def

    def _parse_attr_command (self, attribute, **kw):
        self._tkRAD_command_support(attribute, **kw)
    # end def

    def _parse_attr_compound (self, attribute, xml_tag, **kw):
        _values = ("top", "bottom", "left", "right", "center")
        if tools.is_pstr(xml_tag) and xml_tag.startswith("ttk"):
            _values += ("image", "text")
        # end if
        kw.update(default="none", values=_values)
        self._fix_values(attribute, **kw)
    # end def

    def _parse_attr_cursor (self, attribute, **kw):
        self._tkRAD_cursor_support(attribute, **kw)
    # end def

    def _parse_attr_disabledforeground (self, attribute, **kw):
        self._tkRAD_color_support(attribute, **kw)
    # end def

    def _parse_attr_fg (self, attribute, **kw):
        self._tkRAD_color_support(attribute, **kw)
    # end def

    def _parse_attr_font (self, attribute, **kw):
        self._tkRAD_font_support(attribute, **kw)
    # end def

    def _parse_attr_foreground (self, attribute, **kw):
        self._tkRAD_color_support(attribute, **kw)
    # end def

    def _parse_attr_id (self, attribute, **kw):
        if self._is_unparsed(attribute):
            attribute.value = (
                self.element_get_id(attribute.xml_element)
            )
            attribute.parsed = True
        # end if
    # end def

    def _parse_attr_image (self, attribute, **kw):
        self._tkRAD_image_support(attribute, **kw)
    # end def

    def _parse_attr_menu (self, attribute, **kw):
        if self._is_new(attribute):
            attribute.value = None
            self._tk_config(attribute, **kw)
        # end if
    # end def

    def _parse_attr_name (self, attribute, **kw):
        if self._is_unparsed(attribute):
            _name = tools.choose_str(
                tools.normalize_id(attribute.value),
                self.element_get_id(attribute.xml_element),
            )
            attribute.value = _name
            attribute.parsed = True
        # end if
    # end def

    def _parse_attr_offvalue (self, attribute, **kw):
        self._tkRAD_any_value_support(attribute, **kw)
    # end def

    def _parse_attr_onvalue (self, attribute, **kw):
        self._tkRAD_any_value_support(attribute, **kw)
    # end def

    def _parse_attr_postcommand (self, attribute, **kw):
        self._tkRAD_command_support(attribute, **kw)
    # end def

    def _parse_attr_relief (self, attribute, **kw):
        self._tkRAD_relief_support(attribute, **kw)
    # end def

    def _parse_attr_selectcolor (self, attribute, **kw):
        self._tkRAD_color_support(attribute, **kw)
    # end def

    def _parse_attr_selected (self, attribute, **kw):
        kw.update(no_tk_config=True)
        self._tkRAD_boolean_support(attribute, **kw)
    # end def

    def _parse_attr_selectimage (self, attribute, **kw):
        self._tkRAD_image_support(attribute, **kw)
    # end def

    def _parse_attr_state (self, attribute, **kw):
        self._tkRAD_state_support(attribute, **kw)
    # end def

    def _parse_attr_underline (self, attribute, **kw):
        self._tkRAD_integer_support(attribute, **kw)
    # end def

    def _parse_attr_value (self, attribute, **kw):
        self._tkRAD_any_value_support(attribute, **kw)
    # end def

    def _parse_attr_variable (self, attribute, **kw):
        self._tkRAD_cvar_support(attribute, **kw)
    # end def

    def _parse_attr_widget (self, attribute, **kw):
        kw.update(no_tk_config=True)
        self._tkRAD_widget_support(attribute, **kw)
    # end def

    def _tkRAD_any_value_support (self, attribute, **kw):
        if self._is_new(attribute):
            self._tk_config(attribute, **kw)
        # end if
    # end def

    def _tkRAD_bitmap_support (self, attribute, **kw):
        if self._is_new(attribute):
            attribute.value = self.get_bitmap_path(attribute.value)
            self._tk_config(attribute)
        # end if
    # end def

    def _tkRAD_boolean_support (self, attribute, **kw):
        if self._is_new(attribute):
            attribute.value = bool(
                attribute.value.lower()
                in ("1", "yes", "true", attribute.name)
            )
            attribute.update_xml_element()
            self._tk_config(attribute, **kw)
        # end if
    # end def

    def _tkRAD_color_support (self, attribute, **kw):
        self._tkRAD_any_value_support(attribute, **kw)
    # end def

    def _tkRAD_command_support (self, attribute, **kw):
        self._queue.defer(
            "widget", self._tkRAD_deferred_command_support, attribute,
            **kw
        )
    # end def

    def _tkRAD_cursor_support (self, attribute, **kw):
        self._tkRAD_any_value_support(attribute, **kw)
    # end def

    def _tkRAD_cvar_support (self, attribute, **kw):
        if self._is_new(attribute):
            attribute.value = self.set_stringvar(attribute.value)
            self._tk_config(attribute, **kw)
        # end if
    # end def

    def _tkRAD_deferred_command_support (self, attribute, *args, **kw):
        if self._is_new(attribute):
            _cmd = re.sub(r"\(.*\)", "", attribute.value)
            if _cmd.startswith("@"):
                _cmd = (
                    lambda *args, _e=_cmd[1:], _s=self.events, kw=kw:
                        _s.raise_event(_e, *args, **kw)
                )
            elif _cmd.startswith("^") and hasattr(self, "app"):
                _cmd = _cmd.lstrip("^.@")
                try:
                    _cb = eval("self.app.{}".format(_cmd))
                except:
                    _cb = getattr(self.app, _cmd, None)
                # end try
                if callable(_cb):
                    _cmd = (
                        lambda *args, cb=_cb, kw=kw: cb(*args, **kw)
                    )
                else:
                    raise RADXMLWidgetBaseError(
                        "In widget '{widget}': cannot link command "
                        "'{cmd}' to '{app}' (self.app) - bad command "
                        "or invalid self.app class member."
                        .format(
                            widget=repr(self),
                            cmd=_cmd,
                            app=repr(self.app),
                        )
                    )
                    _cmd = None
                # end if
            elif _cmd.startswith(".") and hasattr(self, "h_owner"):
                _cmd = _cmd.lstrip(".^@")
                try:
                    _cb = eval("self.h_owner.{}".format(_cmd))
                except:
                    _cb = getattr(self.h_owner, _cmd, None)
                # end try
                if callable(_cb):
                    _cmd = (
                        lambda *args, cb=_cb, kw=kw: cb(*args, **kw)
                    )
                else:
                    raise RADXMLWidgetBaseError(
                        "In widget '{widget}': cannot link command "
                        "'{cmd}' to '{owner}' (self.h_owner) - bad "
                        "command or invalid self.h_owner class member."
                        .format(
                            widget=repr(self),
                            cmd=_cmd,
                            owner=repr(self.h_owner),
                        )
                    )
                    _cmd = None
                # end if
            else:
                _cmd = (
                    lambda *args, _cb=eval(_cmd), **kw:
                        _cb(*args, **kw)
                )
            # end if
            attribute.value = _cmd
            self._tk_config(attribute, **kw)
        # end if
    # end def

    def _tkRAD_dimension_support (self, attribute, **kw):
        self._tkRAD_any_value_support(attribute, **kw)
    # end def

    def _tkRAD_float_support (self, attribute, **kw):
        if self._is_new(attribute):
            attribute.value = tools.ensure_float(attribute.value)
            self._tk_config(attribute, **kw)
        # end if
    # end def

    def _tkRAD_font_support (self, attribute, **kw):
        if self._is_new(attribute):
            _font = attribute.value
            _font = re.sub(r"'(.*?)'", r"{\1}", str(_font)).lower()
            attribute.value = _font
            self._tk_config(attribute, **kw)
        # end if
    # end def

    def _tkRAD_image_support (self, attribute, **kw):
        if self._is_new(attribute):
            attribute.value = self.set_image(attribute.value)
            self._tk_config(attribute)
        # end if
    # end def

    def _tkRAD_integer_support (self, attribute, **kw):
        if self._is_new(attribute):
            attribute.value = tools.ensure_int(attribute.value)
            self._tk_config(attribute, **kw)
        # end if
    # end def

    def _tkRAD_label_support (self, attribute, attrs, **kw):
        if self._is_new(attribute):
            _label = _(attribute.value)
            if "underline" in attrs:
                _attr_underline = attrs.get_item("underline")
                _attr_underline.value = -1
                _pos = _label.find("_")
                if _pos >= 0:
                    _attr_underline.value = _pos
                    _label = _label[:_pos] + _label[_pos+1:]
                # end if
                self._tk_config(_attr_underline, **kw)
            # end if
            attribute.value = _label
            self._tk_config(attribute, **kw)
        # end if
    # end def

    def _tkRAD_relief_support (self, attribute, **kw):
        kw.update(
            default="flat",
            values=("raised", "sunken", "groove", "ridge", "solid"),
        )
        self._fix_values(attribute, **kw)
    # end def

    def _tkRAD_state_support (self, attribute, **kw):
        kw.update(
            default="normal", values=("disabled", ),
        )
        self._fix_values(attribute, **kw)
    # end def

    def _tkRAD_widget_support (self, attribute, **kw):
        if self._is_new(attribute):
            _wid = attribute.value
            if _wid.startswith("@"):
                _wid = _wid.strip("@").lower()
                _widget = {
                    "top": self.tk_owner.winfo_toplevel(),
                    "parent": kw.get("tk_parent"),
                }.get(_wid)
            else:
                _widget = self.get_object_by_id(_wid)
            # end if
            if _widget:
                attribute.value = _widget
                self._tk_config(attribute, **kw)
            else:
                raise RADXMLWidgetBaseError(
                    "Widget of id '{}' does not exist or "
                    "has not been registered yet."
                    .format(attribute.value)
                )
            # end if
        # end if
    # end def

    def _tk_config (self, attribute, **kw):
        if not kw.get("no_tk_config"):
            _name = attribute.name.lstrip("_")
            if kw.get("tk_child_config"):
                self.TK_CHILD_CONFIG[_name] = attribute.value
            else:
                self.TK_CONFIG[_name] = attribute.value
            # end if
        # end if
        attribute.parsed = True
    # end def

# end class


class RADXMLWidgetBaseError (Exception):
    """Error handler for RADXMLWidgetBase class."""
    pass
# end class
